//===-- X86InstrFoldTables.cpp - X86 Instruction Folding Tables -----------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//
//
// This file contains the X86 memory folding tables.
//
//===----------------------------------------------------------------------===//

#include "X86InstrFoldTables.h"
#include "X86InstrInfo.h"
#include "llvm/ADT/STLExtras.h"
#include <vector>

using namespace llvm;

// These tables are sorted by their RegOp value allowing them to be binary
// searched at runtime without the need for additional storage. The enum values
// are currently emitted in X86GenInstrInfo.inc in alphabetical order. Which
// makes sorting these tables a simple matter of alphabetizing the table.
//
// We also have a tablegen emitter that tries to autogenerate these tables
// by comparing encoding information. This can be enabled by passing
// X86_GEN_FOLD_TABLES=ON to cmake which fill produce X86GenFoldTables.inc
// in the build area. There are currently some bugs in the autogenerated table
// that require a manual review to copy them from the autogenerated table into
// this table. It is unclear if we will ever be able to fully automate this
// because as new instruction are added into holes in the X86 opcode map they
// potentially pair up with old instructions and create new entries in the
// tables that would be incorrect. The manual review process allows us a chance
// to catch these before they become observable bugs.
static const X86MemoryFoldTableEntry MemoryFoldTable2Addr[] = {
  { X86::ADD16ri8_DB, X86::ADD16mi8,   TB_NO_REVERSE },
  { X86::ADD16ri_DB,  X86::ADD16mi,    TB_NO_REVERSE },
  { X86::ADD16rr_DB,  X86::ADD16mr,    TB_NO_REVERSE },
  { X86::ADD32ri8_DB, X86::ADD32mi8,   TB_NO_REVERSE },
  { X86::ADD32ri_DB,  X86::ADD32mi,    TB_NO_REVERSE },
  { X86::ADD32rr_DB,  X86::ADD32mr,    TB_NO_REVERSE },
  { X86::ADD64ri32_DB,X86::ADD64mi32,  TB_NO_REVERSE },
  { X86::ADD64ri8_DB, X86::ADD64mi8,   TB_NO_REVERSE },
  { X86::ADD64rr_DB,  X86::ADD64mr,    TB_NO_REVERSE },
  { X86::ADD8ri_DB,   X86::ADD8mi,     TB_NO_REVERSE },
  { X86::ADD8rr_DB,   X86::ADD8mr,     TB_NO_REVERSE },
  { X86::ADC16ri,     X86::ADC16mi,    0 },
  { X86::ADC16ri8,    X86::ADC16mi8,   0 },
  { X86::ADC16rr,     X86::ADC16mr,    0 },
  { X86::ADC32ri,     X86::ADC32mi,    0 },
  { X86::ADC32ri8,    X86::ADC32mi8,   0 },
  { X86::ADC32rr,     X86::ADC32mr,    0 },
  { X86::ADC64ri32,   X86::ADC64mi32,  0 },
  { X86::ADC64ri8,    X86::ADC64mi8,   0 },
  { X86::ADC64rr,     X86::ADC64mr,    0 },
  { X86::ADC8ri,      X86::ADC8mi,     0 },
  { X86::ADC8ri8,     X86::ADC8mi8,    0 },
  { X86::ADC8rr,      X86::ADC8mr,     0 },
  { X86::ADD16ri,     X86::ADD16mi,    0 },
  { X86::ADD16ri8,    X86::ADD16mi8,   0 },
  { X86::ADD16rr,     X86::ADD16mr,    0 },
  { X86::ADD32ri,     X86::ADD32mi,    0 },
  { X86::ADD32ri8,    X86::ADD32mi8,   0 },
  { X86::ADD32rr,     X86::ADD32mr,    0 },
  { X86::ADD64ri32,   X86::ADD64mi32,  0 },
  { X86::ADD64ri8,    X86::ADD64mi8,   0 },
  { X86::ADD64rr,     X86::ADD64mr,    0 },
  { X86::ADD8ri,      X86::ADD8mi,     0 },
  { X86::ADD8ri8,     X86::ADD8mi8,    0 },
  { X86::ADD8rr,      X86::ADD8mr,     0 },
  { X86::AND16ri,     X86::AND16mi,    0 },
  { X86::AND16ri8,    X86::AND16mi8,   0 },
  { X86::AND16rr,     X86::AND16mr,    0 },
  { X86::AND32ri,     X86::AND32mi,    0 },
  { X86::AND32ri8,    X86::AND32mi8,   0 },
  { X86::AND32rr,     X86::AND32mr,    0 },
  { X86::AND64ri32,   X86::AND64mi32,  0 },
  { X86::AND64ri8,    X86::AND64mi8,   0 },
  { X86::AND64rr,     X86::AND64mr,    0 },
  { X86::AND8ri,      X86::AND8mi,     0 },
  { X86::AND8ri8,     X86::AND8mi8,    0 },
  { X86::AND8rr,      X86::AND8mr,     0 },
  { X86::BTC16ri8,    X86::BTC16mi8,   0 },
  { X86::BTC32ri8,    X86::BTC32mi8,   0 },
  { X86::BTC64ri8,    X86::BTC64mi8,   0 },
  { X86::BTR16ri8,    X86::BTR16mi8,   0 },
  { X86::BTR32ri8,    X86::BTR32mi8,   0 },
  { X86::BTR64ri8,    X86::BTR64mi8,   0 },
  { X86::BTS16ri8,    X86::BTS16mi8,   0 },
  { X86::BTS32ri8,    X86::BTS32mi8,   0 },
  { X86::BTS64ri8,    X86::BTS64mi8,   0 },
  { X86::DEC16r,      X86::DEC16m,     0 },
  { X86::DEC32r,      X86::DEC32m,     0 },
  { X86::DEC64r,      X86::DEC64m,     0 },
  { X86::DEC8r,       X86::DEC8m,      0 },
  { X86::INC16r,      X86::INC16m,     0 },
  { X86::INC32r,      X86::INC32m,     0 },
  { X86::INC64r,      X86::INC64m,     0 },
  { X86::INC8r,       X86::INC8m,      0 },
  { X86::NEG16r,      X86::NEG16m,     0 },
  { X86::NEG32r,      X86::NEG32m,     0 },
  { X86::NEG64r,      X86::NEG64m,     0 },
  { X86::NEG8r,       X86::NEG8m,      0 },
  { X86::NOT16r,      X86::NOT16m,     0 },
  { X86::NOT32r,      X86::NOT32m,     0 },
  { X86::NOT64r,      X86::NOT64m,     0 },
  { X86::NOT8r,       X86::NOT8m,      0 },
  { X86::OR16ri,      X86::OR16mi,     0 },
  { X86::OR16ri8,     X86::OR16mi8,    0 },
  { X86::OR16rr,      X86::OR16mr,     0 },
  { X86::OR32ri,      X86::OR32mi,     0 },
  { X86::OR32ri8,     X86::OR32mi8,    0 },
  { X86::OR32rr,      X86::OR32mr,     0 },
  { X86::OR64ri32,    X86::OR64mi32,   0 },
  { X86::OR64ri8,     X86::OR64mi8,    0 },
  { X86::OR64rr,      X86::OR64mr,     0 },
  { X86::OR8ri,       X86::OR8mi,      0 },
  { X86::OR8ri8,      X86::OR8mi8,     0 },
  { X86::OR8rr,       X86::OR8mr,      0 },
  { X86::RCL16r1,     X86::RCL16m1,    0 },
  { X86::RCL16rCL,    X86::RCL16mCL,   0 },
  { X86::RCL16ri,     X86::RCL16mi,    0 },
  { X86::RCL32r1,     X86::RCL32m1,    0 },
  { X86::RCL32rCL,    X86::RCL32mCL,   0 },
  { X86::RCL32ri,     X86::RCL32mi,    0 },
  { X86::RCL64r1,     X86::RCL64m1,    0 },
  { X86::RCL64rCL,    X86::RCL64mCL,   0 },
  { X86::RCL64ri,     X86::RCL64mi,    0 },
  { X86::RCL8r1,      X86::RCL8m1,     0 },
  { X86::RCL8rCL,     X86::RCL8mCL,    0 },
  { X86::RCL8ri,      X86::RCL8mi,     0 },
  { X86::RCR16r1,     X86::RCR16m1,    0 },
  { X86::RCR16rCL,    X86::RCR16mCL,   0 },
  { X86::RCR16ri,     X86::RCR16mi,    0 },
  { X86::RCR32r1,     X86::RCR32m1,    0 },
  { X86::RCR32rCL,    X86::RCR32mCL,   0 },
  { X86::RCR32ri,     X86::RCR32mi,    0 },
  { X86::RCR64r1,     X86::RCR64m1,    0 },
  { X86::RCR64rCL,    X86::RCR64mCL,   0 },
  { X86::RCR64ri,     X86::RCR64mi,    0 },
  { X86::RCR8r1,      X86::RCR8m1,     0 },
  { X86::RCR8rCL,     X86::RCR8mCL,    0 },
  { X86::RCR8ri,      X86::RCR8mi,     0 },
  { X86::ROL16r1,     X86::ROL16m1,    0 },
  { X86::ROL16rCL,    X86::ROL16mCL,   0 },
  { X86::ROL16ri,     X86::ROL16mi,    0 },
  { X86::ROL32r1,     X86::ROL32m1,    0 },
  { X86::ROL32rCL,    X86::ROL32mCL,   0 },
  { X86::ROL32ri,     X86::ROL32mi,    0 },
  { X86::ROL64r1,     X86::ROL64m1,    0 },
  { X86::ROL64rCL,    X86::ROL64mCL,   0 },
  { X86::ROL64ri,     X86::ROL64mi,    0 },
  { X86::ROL8r1,      X86::ROL8m1,     0 },
  { X86::ROL8rCL,     X86::ROL8mCL,    0 },
  { X86::ROL8ri,      X86::ROL8mi,     0 },
  { X86::ROR16r1,     X86::ROR16m1,    0 },
  { X86::ROR16rCL,    X86::ROR16mCL,   0 },
  { X86::ROR16ri,     X86::ROR16mi,    0 },
  { X86::ROR32r1,     X86::ROR32m1,    0 },
  { X86::ROR32rCL,    X86::ROR32mCL,   0 },
  { X86::ROR32ri,     X86::ROR32mi,    0 },
  { X86::ROR64r1,     X86::ROR64m1,    0 },
  { X86::ROR64rCL,    X86::ROR64mCL,   0 },
  { X86::ROR64ri,     X86::ROR64mi,    0 },
  { X86::ROR8r1,      X86::ROR8m1,     0 },
  { X86::ROR8rCL,     X86::ROR8mCL,    0 },
  { X86::ROR8ri,      X86::ROR8mi,     0 },
  { X86::SAR16r1,     X86::SAR16m1,    0 },
  { X86::SAR16rCL,    X86::SAR16mCL,   0 },
  { X86::SAR16ri,     X86::SAR16mi,    0 },
  { X86::SAR32r1,     X86::SAR32m1,    0 },
  { X86::SAR32rCL,    X86::SAR32mCL,   0 },
  { X86::SAR32ri,     X86::SAR32mi,    0 },
  { X86::SAR64r1,     X86::SAR64m1,    0 },
  { X86::SAR64rCL,    X86::SAR64mCL,   0 },
  { X86::SAR64ri,     X86::SAR64mi,    0 },
  { X86::SAR8r1,      X86::SAR8m1,     0 },
  { X86::SAR8rCL,     X86::SAR8mCL,    0 },
  { X86::SAR8ri,      X86::SAR8mi,     0 },
  { X86::SBB16ri,     X86::SBB16mi,    0 },
  { X86::SBB16ri8,    X86::SBB16mi8,   0 },
  { X86::SBB16rr,     X86::SBB16mr,    0 },
  { X86::SBB32ri,     X86::SBB32mi,    0 },
  { X86::SBB32ri8,    X86::SBB32mi8,   0 },
  { X86::SBB32rr,     X86::SBB32mr,    0 },
  { X86::SBB64ri32,   X86::SBB64mi32,  0 },
  { X86::SBB64ri8,    X86::SBB64mi8,   0 },
  { X86::SBB64rr,     X86::SBB64mr,    0 },
  { X86::SBB8ri,      X86::SBB8mi,     0 },
  { X86::SBB8ri8,     X86::SBB8mi8,    0 },
  { X86::SBB8rr,      X86::SBB8mr,     0 },
  { X86::SHL16r1,     X86::SHL16m1,    0 },
  { X86::SHL16rCL,    X86::SHL16mCL,   0 },
  { X86::SHL16ri,     X86::SHL16mi,    0 },
  { X86::SHL32r1,     X86::SHL32m1,    0 },
  { X86::SHL32rCL,    X86::SHL32mCL,   0 },
  { X86::SHL32ri,     X86::SHL32mi,    0 },
  { X86::SHL64r1,     X86::SHL64m1,    0 },
  { X86::SHL64rCL,    X86::SHL64mCL,   0 },
  { X86::SHL64ri,     X86::SHL64mi,    0 },
  { X86::SHL8r1,      X86::SHL8m1,     0 },
  { X86::SHL8rCL,     X86::SHL8mCL,    0 },
  { X86::SHL8ri,      X86::SHL8mi,     0 },
  { X86::SHLD16rrCL,  X86::SHLD16mrCL, 0 },
  { X86::SHLD16rri8,  X86::SHLD16mri8, 0 },
  { X86::SHLD32rrCL,  X86::SHLD32mrCL, 0 },
  { X86::SHLD32rri8,  X86::SHLD32mri8, 0 },
  { X86::SHLD64rrCL,  X86::SHLD64mrCL, 0 },
  { X86::SHLD64rri8,  X86::SHLD64mri8, 0 },
  { X86::SHR16r1,     X86::SHR16m1,    0 },
  { X86::SHR16rCL,    X86::SHR16mCL,   0 },
  { X86::SHR16ri,     X86::SHR16mi,    0 },
  { X86::SHR32r1,     X86::SHR32m1,    0 },
  { X86::SHR32rCL,    X86::SHR32mCL,   0 },
  { X86::SHR32ri,     X86::SHR32mi,    0 },
  { X86::SHR64r1,     X86::SHR64m1,    0 },
  { X86::SHR64rCL,    X86::SHR64mCL,   0 },
  { X86::SHR64ri,     X86::SHR64mi,    0 },
  { X86::SHR8r1,      X86::SHR8m1,     0 },
  { X86::SHR8rCL,     X86::SHR8mCL,    0 },
  { X86::SHR8ri,      X86::SHR8mi,     0 },
  { X86::SHRD16rrCL,  X86::SHRD16mrCL, 0 },
  { X86::SHRD16rri8,  X86::SHRD16mri8, 0 },
  { X86::SHRD32rrCL,  X86::SHRD32mrCL, 0 },
  { X86::SHRD32rri8,  X86::SHRD32mri8, 0 },
  { X86::SHRD64rrCL,  X86::SHRD64mrCL, 0 },
  { X86::SHRD64rri8,  X86::SHRD64mri8, 0 },
  { X86::SUB16ri,     X86::SUB16mi,    0 },
  { X86::SUB16ri8,    X86::SUB16mi8,   0 },
  { X86::SUB16rr,     X86::SUB16mr,    0 },
  { X86::SUB32ri,     X86::SUB32mi,    0 },
  { X86::SUB32ri8,    X86::SUB32mi8,   0 },
  { X86::SUB32rr,     X86::SUB32mr,    0 },
  { X86::SUB64ri32,   X86::SUB64mi32,  0 },
  { X86::SUB64ri8,    X86::SUB64mi8,   0 },
  { X86::SUB64rr,     X86::SUB64mr,    0 },
  { X86::SUB8ri,      X86::SUB8mi,     0 },
  { X86::SUB8ri8,     X86::SUB8mi8,    0 },
  { X86::SUB8rr,      X86::SUB8mr,     0 },
  { X86::XOR16ri,     X86::XOR16mi,    0 },
  { X86::XOR16ri8,    X86::XOR16mi8,   0 },
  { X86::XOR16rr,     X86::XOR16mr,    0 },
  { X86::XOR32ri,     X86::XOR32mi,    0 },
  { X86::XOR32ri8,    X86::XOR32mi8,   0 },
  { X86::XOR32rr,     X86::XOR32mr,    0 },
  { X86::XOR64ri32,   X86::XOR64mi32,  0 },
  { X86::XOR64ri8,    X86::XOR64mi8,   0 },
  { X86::XOR64rr,     X86::XOR64mr,    0 },
  { X86::XOR8ri,      X86::XOR8mi,     0 },
  { X86::XOR8ri8,     X86::XOR8mi8,    0 },
  { X86::XOR8rr,      X86::XOR8mr,     0 },
};

static const X86MemoryFoldTableEntry MemoryFoldTable0[] = {
  { X86::BT16ri8,             X86::BT16mi8,             TB_FOLDED_LOAD },
  { X86::BT32ri8,             X86::BT32mi8,             TB_FOLDED_LOAD },
  { X86::BT64ri8,             X86::BT64mi8,             TB_FOLDED_LOAD },
  { X86::CALL16r,             X86::CALL16m,             TB_FOLDED_LOAD },
  { X86::CALL16r_NT,          X86::CALL16m_NT,          TB_FOLDED_LOAD },
  { X86::CALL32r,             X86::CALL32m,             TB_FOLDED_LOAD },
  { X86::CALL32r_NT,          X86::CALL32m_NT,          TB_FOLDED_LOAD },
  { X86::CALL64r,             X86::CALL64m,             TB_FOLDED_LOAD },
  { X86::CALL64r_NT,          X86::CALL64m_NT,          TB_FOLDED_LOAD },
  { X86::CMP16ri,             X86::CMP16mi,             TB_FOLDED_LOAD },
  { X86::CMP16ri8,            X86::CMP16mi8,            TB_FOLDED_LOAD },
  { X86::CMP16rr,             X86::CMP16mr,             TB_FOLDED_LOAD },
  { X86::CMP32ri,             X86::CMP32mi,             TB_FOLDED_LOAD },
  { X86::CMP32ri8,            X86::CMP32mi8,            TB_FOLDED_LOAD },
  { X86::CMP32rr,             X86::CMP32mr,             TB_FOLDED_LOAD },
  { X86::CMP64ri32,           X86::CMP64mi32,           TB_FOLDED_LOAD },
  { X86::CMP64ri8,            X86::CMP64mi8,            TB_FOLDED_LOAD },
  { X86::CMP64rr,             X86::CMP64mr,             TB_FOLDED_LOAD },
  { X86::CMP8ri,              X86::CMP8mi,              TB_FOLDED_LOAD },
  { X86::CMP8ri8,             X86::CMP8mi8,             TB_FOLDED_LOAD },
  { X86::CMP8rr,              X86::CMP8mr,              TB_FOLDED_LOAD },
  { X86::DIV16r,              X86::DIV16m,              TB_FOLDED_LOAD },
  { X86::DIV32r,              X86::DIV32m,              TB_FOLDED_LOAD },
  { X86::DIV64r,              X86::DIV64m,              TB_FOLDED_LOAD },
  { X86::DIV8r,               X86::DIV8m,               TB_FOLDED_LOAD },
  { X86::EXTRACTPSrr,         X86::EXTRACTPSmr,         TB_FOLDED_STORE },
  { X86::IDIV16r,             X86::IDIV16m,             TB_FOLDED_LOAD },
  { X86::IDIV32r,             X86::IDIV32m,             TB_FOLDED_LOAD },
  { X86::IDIV64r,             X86::IDIV64m,             TB_FOLDED_LOAD },
  { X86::IDIV8r,              X86::IDIV8m,              TB_FOLDED_LOAD },
  { X86::IMUL16r,             X86::IMUL16m,             TB_FOLDED_LOAD },
  { X86::IMUL32r,             X86::IMUL32m,             TB_FOLDED_LOAD },
  { X86::IMUL64r,             X86::IMUL64m,             TB_FOLDED_LOAD },
  { X86::IMUL8r,              X86::IMUL8m,              TB_FOLDED_LOAD },
  { X86::JMP16r,              X86::JMP16m,              TB_FOLDED_LOAD },
  { X86::JMP16r_NT,           X86::JMP16m_NT,           TB_FOLDED_LOAD },
  { X86::JMP32r,              X86::JMP32m,              TB_FOLDED_LOAD },
  { X86::JMP32r_NT,           X86::JMP32m_NT,           TB_FOLDED_LOAD },
  { X86::JMP64r,              X86::JMP64m,              TB_FOLDED_LOAD },
  { X86::JMP64r_NT,           X86::JMP64m_NT,           TB_FOLDED_LOAD },
  { X86::MOV16ri,             X86::MOV16mi,             TB_FOLDED_STORE },
  { X86::MOV16rr,             X86::MOV16mr,             TB_FOLDED_STORE },
  { X86::MOV32ri,             X86::MOV32mi,             TB_FOLDED_STORE },
  { X86::MOV32rr,             X86::MOV32mr,             TB_FOLDED_STORE },
  { X86::MOV64ri32,           X86::MOV64mi32,           TB_FOLDED_STORE },
  { X86::MOV64rr,             X86::MOV64mr,             TB_FOLDED_STORE },
  { X86::MOV8ri,              X86::MOV8mi,              TB_FOLDED_STORE },
  { X86::MOV8rr,              X86::MOV8mr,              TB_FOLDED_STORE },
  { X86::MOV8rr_NOREX,        X86::MOV8mr_NOREX,        TB_FOLDED_STORE },
  { X86::MOVAPDrr,            X86::MOVAPDmr,            TB_FOLDED_STORE | TB_ALIGN_16 },
  { X86::MOVAPSrr,            X86::MOVAPSmr,            TB_FOLDED_STORE | TB_ALIGN_16 },
  { X86::MOVDQArr,            X86::MOVDQAmr,            TB_FOLDED_STORE | TB_ALIGN_16 },
  { X86::MOVDQUrr,            X86::MOVDQUmr,            TB_FOLDED_STORE },
  { X86::MOVPDI2DIrr,         X86::MOVPDI2DImr,         TB_FOLDED_STORE },
  { X86::MOVPQIto64rr,        X86::MOVPQI2QImr,         TB_FOLDED_STORE | TB_NO_REVERSE },
  { X86::MOVSDto64rr,         X86::MOVSDmr,             TB_FOLDED_STORE | TB_NO_REVERSE },
  { X86::MOVSS2DIrr,          X86::MOVSSmr,             TB_FOLDED_STORE },
  { X86::MOVUPDrr,            X86::MOVUPDmr,            TB_FOLDED_STORE },
  { X86::MOVUPSrr,            X86::MOVUPSmr,            TB_FOLDED_STORE },
  { X86::MUL16r,              X86::MUL16m,              TB_FOLDED_LOAD },
  { X86::MUL32r,              X86::MUL32m,              TB_FOLDED_LOAD },
  { X86::MUL64r,              X86::MUL64m,              TB_FOLDED_LOAD },
  { X86::MUL8r,               X86::MUL8m,               TB_FOLDED_LOAD },
  { X86::PEXTRDrr,            X86::PEXTRDmr,            TB_FOLDED_STORE },
  { X86::PEXTRQrr,            X86::PEXTRQmr,            TB_FOLDED_STORE },
  { X86::PTWRITE64r,          X86::PTWRITE64m,          TB_FOLDED_LOAD },
  { X86::PTWRITEr,            X86::PTWRITEm,            TB_FOLDED_LOAD },
  { X86::PUSH16r,             X86::PUSH16rmm,           TB_FOLDED_LOAD },
  { X86::PUSH32r,             X86::PUSH32rmm,           TB_FOLDED_LOAD },
  { X86::PUSH64r,             X86::PUSH64rmm,           TB_FOLDED_LOAD },
  { X86::SETCCr,              X86::SETCCm,              TB_FOLDED_STORE },
  { X86::TAILJMPr,            X86::TAILJMPm,            TB_FOLDED_LOAD },
  { X86::TAILJMPr64,          X86::TAILJMPm64,          TB_FOLDED_LOAD },
  { X86::TAILJMPr64_REX,      X86::TAILJMPm64_REX,      TB_FOLDED_LOAD },
  { X86::TCRETURNri,          X86::TCRETURNmi,          TB_FOLDED_LOAD | TB_NO_FORWARD },
  { X86::TCRETURNri64,        X86::TCRETURNmi64,        TB_FOLDED_LOAD | TB_NO_FORWARD },
  { X86::TEST16ri,            X86::TEST16mi,            TB_FOLDED_LOAD },
  { X86::TEST16rr,            X86::TEST16mr,            TB_FOLDED_LOAD },
  { X86::TEST32ri,            X86::TEST32mi,            TB_FOLDED_LOAD },
  { X86::TEST32rr,            X86::TEST32mr,            TB_FOLDED_LOAD },
  { X86::TEST64ri32,          X86::TEST64mi32,          TB_FOLDED_LOAD },
  { X86::TEST64rr,            X86::TEST64mr,            TB_FOLDED_LOAD },
  { X86::TEST8ri,             X86::TEST8mi,             TB_FOLDED_LOAD },
  { X86::TEST8rr,             X86::TEST8mr,             TB_FOLDED_LOAD },
  { X86::VCVTPS2PHYrr,        X86::VCVTPS2PHYmr,        TB_FOLDED_STORE },
  { X86::VCVTPS2PHZ256rr,     X86::VCVTPS2PHZ256mr,     TB_FOLDED_STORE },
  { X86::VCVTPS2PHZrr,        X86::VCVTPS2PHZmr,        TB_FOLDED_STORE },
  { X86::VEXTRACTF128rr,      X86::VEXTRACTF128mr,      TB_FOLDED_STORE },
  { X86::VEXTRACTF32x4Z256rr, X86::VEXTRACTF32x4Z256mr, TB_FOLDED_STORE },
  { X86::VEXTRACTF32x4Zrr,    X86::VEXTRACTF32x4Zmr,    TB_FOLDED_STORE },
  { X86::VEXTRACTF32x8Zrr,    X86::VEXTRACTF32x8Zmr,    TB_FOLDED_STORE },
  { X86::VEXTRACTF64x2Z256rr, X86::VEXTRACTF64x2Z256mr, TB_FOLDED_STORE },
  { X86::VEXTRACTF64x2Zrr,    X86::VEXTRACTF64x2Zmr,    TB_FOLDED_STORE },
  { X86::VEXTRACTF64x4Zrr,    X86::VEXTRACTF64x4Zmr,    TB_FOLDED_STORE },
  { X86::VEXTRACTI128rr,      X86::VEXTRACTI128mr,      TB_FOLDED_STORE },
  { X86::VEXTRACTI32x4Z256rr, X86::VEXTRACTI32x4Z256mr, TB_FOLDED_STORE },
  { X86::VEXTRACTI32x4Zrr,    X86::VEXTRACTI32x4Zmr,    TB_FOLDED_STORE },
  { X86::VEXTRACTI32x8Zrr,    X86::VEXTRACTI32x8Zmr,    TB_FOLDED_STORE },
  { X86::VEXTRACTI64x2Z256rr, X86::VEXTRACTI64x2Z256mr, TB_FOLDED_STORE },
  { X86::VEXTRACTI64x2Zrr,    X86::VEXTRACTI64x2Zmr,    TB_FOLDED_STORE },
  { X86::VEXTRACTI64x4Zrr,    X86::VEXTRACTI64x4Zmr,    TB_FOLDED_STORE },
  { X86::VEXTRACTPSZrr,       X86::VEXTRACTPSZmr,       TB_FOLDED_STORE },
  { X86::VEXTRACTPSrr,        X86::VEXTRACTPSmr,        TB_FOLDED_STORE },
  { X86::VMOVAPDYrr,          X86::VMOVAPDYmr,          TB_FOLDED_STORE | TB_ALIGN_32 },
  { X86::VMOVAPDZ128rr,       X86::VMOVAPDZ128mr,       TB_FOLDED_STORE | TB_ALIGN_16 },
  { X86::VMOVAPDZ256rr,       X86::VMOVAPDZ256mr,       TB_FOLDED_STORE | TB_ALIGN_32 },
  { X86::VMOVAPDZrr,          X86::VMOVAPDZmr,          TB_FOLDED_STORE | TB_ALIGN_64 },
  { X86::VMOVAPDrr,           X86::VMOVAPDmr,           TB_FOLDED_STORE | TB_ALIGN_16 },
  { X86::VMOVAPSYrr,          X86::VMOVAPSYmr,          TB_FOLDED_STORE | TB_ALIGN_32 },
  { X86::VMOVAPSZ128rr,       X86::VMOVAPSZ128mr,       TB_FOLDED_STORE | TB_ALIGN_16 },
  { X86::VMOVAPSZ256rr,       X86::VMOVAPSZ256mr,       TB_FOLDED_STORE | TB_ALIGN_32 },
  { X86::VMOVAPSZrr,          X86::VMOVAPSZmr,          TB_FOLDED_STORE | TB_ALIGN_64 },
  { X86::VMOVAPSrr,           X86::VMOVAPSmr,           TB_FOLDED_STORE | TB_ALIGN_16 },
  { X86::VMOVDQA32Z128rr,     X86::VMOVDQA32Z128mr,     TB_FOLDED_STORE | TB_ALIGN_16 },
  { X86::VMOVDQA32Z256rr,     X86::VMOVDQA32Z256mr,     TB_FOLDED_STORE | TB_ALIGN_32 },
  { X86::VMOVDQA32Zrr,        X86::VMOVDQA32Zmr,        TB_FOLDED_STORE | TB_ALIGN_64 },
  { X86::VMOVDQA64Z128rr,     X86::VMOVDQA64Z128mr,     TB_FOLDED_STORE | TB_ALIGN_16 },
  { X86::VMOVDQA64Z256rr,     X86::VMOVDQA64Z256mr,     TB_FOLDED_STORE | TB_ALIGN_32 },
  { X86::VMOVDQA64Zrr,        X86::VMOVDQA64Zmr,        TB_FOLDED_STORE | TB_ALIGN_64 },
  { X86::VMOVDQAYrr,          X86::VMOVDQAYmr,          TB_FOLDED_STORE | TB_ALIGN_32 },
  { X86::VMOVDQArr,           X86::VMOVDQAmr,           TB_FOLDED_STORE | TB_ALIGN_16 },
  { X86::VMOVDQU16Z128rr,     X86::VMOVDQU16Z128mr,     TB_FOLDED_STORE },
  { X86::VMOVDQU16Z256rr,     X86::VMOVDQU16Z256mr,     TB_FOLDED_STORE },
  { X86::VMOVDQU16Zrr,        X86::VMOVDQU16Zmr,        TB_FOLDED_STORE },
  { X86::VMOVDQU32Z128rr,     X86::VMOVDQU32Z128mr,     TB_FOLDED_STORE },
  { X86::VMOVDQU32Z256rr,     X86::VMOVDQU32Z256mr,     TB_FOLDED_STORE },
  { X86::VMOVDQU32Zrr,        X86::VMOVDQU32Zmr,        TB_FOLDED_STORE },
  { X86::VMOVDQU64Z128rr,     X86::VMOVDQU64Z128mr,     TB_FOLDED_STORE },
  { X86::VMOVDQU64Z256rr,     X86::VMOVDQU64Z256mr,     TB_FOLDED_STORE },
  { X86::VMOVDQU64Zrr,        X86::VMOVDQU64Zmr,        TB_FOLDED_STORE },
  { X86::VMOVDQU8Z128rr,      X86::VMOVDQU8Z128mr,      TB_FOLDED_STORE },
  { X86::VMOVDQU8Z256rr,      X86::VMOVDQU8Z256mr,      TB_FOLDED_STORE },
  { X86::VMOVDQU8Zrr,         X86::VMOVDQU8Zmr,         TB_FOLDED_STORE },
  { X86::VMOVDQUYrr,          X86::VMOVDQUYmr,          TB_FOLDED_STORE },
  { X86::VMOVDQUrr,           X86::VMOVDQUmr,           TB_FOLDED_STORE },
  { X86::VMOVPDI2DIZrr,       X86::VMOVPDI2DIZmr,       TB_FOLDED_STORE },
  { X86::VMOVPDI2DIrr,        X86::VMOVPDI2DImr,        TB_FOLDED_STORE },
  { X86::VMOVPQIto64Zrr,      X86::VMOVPQI2QIZmr,       TB_FOLDED_STORE | TB_NO_REVERSE },
  { X86::VMOVPQIto64rr,       X86::VMOVPQI2QImr,        TB_FOLDED_STORE | TB_NO_REVERSE },
  { X86::VMOVSDto64Zrr,       X86::VMOVSDZmr,           TB_FOLDED_STORE | TB_NO_REVERSE },
  { X86::VMOVSDto64rr,        X86::VMOVSDmr,            TB_FOLDED_STORE | TB_NO_REVERSE },
  { X86::VMOVSS2DIZrr,        X86::VMOVSSZmr,           TB_FOLDED_STORE },
  { X86::VMOVSS2DIrr,         X86::VMOVSSmr,            TB_FOLDED_STORE },
  { X86::VMOVUPDYrr,          X86::VMOVUPDYmr,          TB_FOLDED_STORE },
  { X86::VMOVUPDZ128rr,       X86::VMOVUPDZ128mr,       TB_FOLDED_STORE },
  { X86::VMOVUPDZ256rr,       X86::VMOVUPDZ256mr,       TB_FOLDED_STORE },
  { X86::VMOVUPDZrr,          X86::VMOVUPDZmr,          TB_FOLDED_STORE },
  { X86::VMOVUPDrr,           X86::VMOVUPDmr,           TB_FOLDED_STORE },
  { X86::VMOVUPSYrr,          X86::VMOVUPSYmr,          TB_FOLDED_STORE },
  { X86::VMOVUPSZ128rr,       X86::VMOVUPSZ128mr,       TB_FOLDED_STORE },
  { X86::VMOVUPSZ256rr,       X86::VMOVUPSZ256mr,       TB_FOLDED_STORE },
  { X86::VMOVUPSZrr,          X86::VMOVUPSZmr,          TB_FOLDED_STORE },
  { X86::VMOVUPSrr,           X86::VMOVUPSmr,           TB_FOLDED_STORE },
  { X86::VPEXTRDZrr,          X86::VPEXTRDZmr,          TB_FOLDED_STORE },
  { X86::VPEXTRDrr,           X86::VPEXTRDmr,           TB_FOLDED_STORE },
  { X86::VPEXTRQZrr,          X86::VPEXTRQZmr,          TB_FOLDED_STORE },
  { X86::VPEXTRQrr,           X86::VPEXTRQmr,           TB_FOLDED_STORE },
  { X86::VPMOVDBZrr,          X86::VPMOVDBZmr,          TB_FOLDED_STORE },
  { X86::VPMOVDWZ256rr,       X86::VPMOVDWZ256mr,       TB_FOLDED_STORE },
  { X86::VPMOVDWZrr,          X86::VPMOVDWZmr,          TB_FOLDED_STORE },
  { X86::VPMOVQDZ256rr,       X86::VPMOVQDZ256mr,       TB_FOLDED_STORE },
  { X86::VPMOVQDZrr,          X86::VPMOVQDZmr,          TB_FOLDED_STORE },
  { X86::VPMOVQWZrr,          X86::VPMOVQWZmr,          TB_FOLDED_STORE },
  { X86::VPMOVSDBZrr,         X86::VPMOVSDBZmr,         TB_FOLDED_STORE },
  { X86::VPMOVSDWZ256rr,      X86::VPMOVSDWZ256mr,      TB_FOLDED_STORE },
  { X86::VPMOVSDWZrr,         X86::VPMOVSDWZmr,         TB_FOLDED_STORE },
  { X86::VPMOVSQDZ256rr,      X86::VPMOVSQDZ256mr,      TB_FOLDED_STORE },
  { X86::VPMOVSQDZrr,         X86::VPMOVSQDZmr,         TB_FOLDED_STORE },
  { X86::VPMOVSQWZrr,         X86::VPMOVSQWZmr,         TB_FOLDED_STORE },
  { X86::VPMOVSWBZ256rr,      X86::VPMOVSWBZ256mr,      TB_FOLDED_STORE },
  { X86::VPMOVSWBZrr,         X86::VPMOVSWBZmr,         TB_FOLDED_STORE },
  { X86::VPMOVUSDBZrr,        X86::VPMOVUSDBZmr,        TB_FOLDED_STORE },
  { X86::VPMOVUSDWZ256rr,     X86::VPMOVUSDWZ256mr,     TB_FOLDED_STORE },
  { X86::VPMOVUSDWZrr,        X86::VPMOVUSDWZmr,        TB_FOLDED_STORE },
  { X86::VPMOVUSQDZ256rr,     X86::VPMOVUSQDZ256mr,     TB_FOLDED_STORE },
  { X86::VPMOVUSQDZrr,        X86::VPMOVUSQDZmr,        TB_FOLDED_STORE },
  { X86::VPMOVUSQWZrr,        X86::VPMOVUSQWZmr,        TB_FOLDED_STORE },
  { X86::VPMOVUSWBZ256rr,     X86::VPMOVUSWBZ256mr,     TB_FOLDED_STORE },
  { X86::VPMOVUSWBZrr,        X86::VPMOVUSWBZmr,        TB_FOLDED_STORE },
  { X86::VPMOVWBZ256rr,       X86::VPMOVWBZ256mr,       TB_FOLDED_STORE },
  { X86::VPMOVWBZrr,          X86::VPMOVWBZmr,          TB_FOLDED_STORE },
};

static const X86MemoryFoldTableEntry MemoryFoldTable1[] = {
  { X86::AESIMCrr,             X86::AESIMCrm,             TB_ALIGN_16 },
  { X86::AESKEYGENASSIST128rr, X86::AESKEYGENASSIST128rm, TB_ALIGN_16 },
  { X86::BEXTR32rr,            X86::BEXTR32rm,            0 },
  { X86::BEXTR64rr,            X86::BEXTR64rm,            0 },
  { X86::BEXTRI32ri,           X86::BEXTRI32mi,           0 },
  { X86::BEXTRI64ri,           X86::BEXTRI64mi,           0 },
  { X86::BLCFILL32rr,          X86::BLCFILL32rm,          0 },
  { X86::BLCFILL64rr,          X86::BLCFILL64rm,          0 },
  { X86::BLCI32rr,             X86::BLCI32rm,             0 },
  { X86::BLCI64rr,             X86::BLCI64rm,             0 },
  { X86::BLCIC32rr,            X86::BLCIC32rm,            0 },
  { X86::BLCIC64rr,            X86::BLCIC64rm,            0 },
  { X86::BLCMSK32rr,           X86::BLCMSK32rm,           0 },
  { X86::BLCMSK64rr,           X86::BLCMSK64rm,           0 },
  { X86::BLCS32rr,             X86::BLCS32rm,             0 },
  { X86::BLCS64rr,             X86::BLCS64rm,             0 },
  { X86::BLSFILL32rr,          X86::BLSFILL32rm,          0 },
  { X86::BLSFILL64rr,          X86::BLSFILL64rm,          0 },
  { X86::BLSI32rr,             X86::BLSI32rm,             0 },
  { X86::BLSI64rr,             X86::BLSI64rm,             0 },
  { X86::BLSIC32rr,            X86::BLSIC32rm,            0 },
  { X86::BLSIC64rr,            X86::BLSIC64rm,            0 },
  { X86::BLSMSK32rr,           X86::BLSMSK32rm,           0 },
  { X86::BLSMSK64rr,           X86::BLSMSK64rm,           0 },
  { X86::BLSR32rr,             X86::BLSR32rm,             0 },
  { X86::BLSR64rr,             X86::BLSR64rm,             0 },
  { X86::BSF16rr,              X86::BSF16rm,              0 },
  { X86::BSF32rr,              X86::BSF32rm,              0 },
  { X86::BSF64rr,              X86::BSF64rm,              0 },
  { X86::BSR16rr,              X86::BSR16rm,              0 },
  { X86::BSR32rr,              X86::BSR32rm,              0 },
  { X86::BSR64rr,              X86::BSR64rm,              0 },
  { X86::BZHI32rr,             X86::BZHI32rm,             0 },
  { X86::BZHI64rr,             X86::BZHI64rm,             0 },
  { X86::CMP16rr,              X86::CMP16rm,              0 },
  { X86::CMP32rr,              X86::CMP32rm,              0 },
  { X86::CMP64rr,              X86::CMP64rm,              0 },
  { X86::CMP8rr,               X86::CMP8rm,               0 },
  { X86::COMISDrr,             X86::COMISDrm,             0 },
  { X86::COMISDrr_Int,         X86::COMISDrm_Int,         TB_NO_REVERSE },
  { X86::COMISSrr,             X86::COMISSrm,             0 },
  { X86::COMISSrr_Int,         X86::COMISSrm_Int,         TB_NO_REVERSE },
  { X86::CVTDQ2PDrr,           X86::CVTDQ2PDrm,           TB_NO_REVERSE },
  { X86::CVTDQ2PSrr,           X86::CVTDQ2PSrm,           TB_ALIGN_16 },
  { X86::CVTPD2DQrr,           X86::CVTPD2DQrm,           TB_ALIGN_16 },
  { X86::CVTPD2PSrr,           X86::CVTPD2PSrm,           TB_ALIGN_16 },
  { X86::CVTPS2DQrr,           X86::CVTPS2DQrm,           TB_ALIGN_16 },
  { X86::CVTPS2PDrr,           X86::CVTPS2PDrm,           TB_NO_REVERSE },
  { X86::CVTSD2SI64rr_Int,     X86::CVTSD2SI64rm_Int,     TB_NO_REVERSE },
  { X86::CVTSD2SIrr_Int,       X86::CVTSD2SIrm_Int,       TB_NO_REVERSE },
  { X86::CVTSD2SSrr,           X86::CVTSD2SSrm,           0 },
  { X86::CVTSI2SDrr,           X86::CVTSI2SDrm,           0 },
  { X86::CVTSI2SSrr,           X86::CVTSI2SSrm,           0 },
  { X86::CVTSI642SDrr,         X86::CVTSI642SDrm,         0 },
  { X86::CVTSI642SSrr,         X86::CVTSI642SSrm,         0 },
  { X86::CVTSS2SDrr,           X86::CVTSS2SDrm,           0 },
  { X86::CVTSS2SI64rr_Int,     X86::CVTSS2SI64rm_Int,     TB_NO_REVERSE },
  { X86::CVTSS2SIrr_Int,       X86::CVTSS2SIrm_Int,       TB_NO_REVERSE },
  { X86::CVTTPD2DQrr,          X86::CVTTPD2DQrm,          TB_ALIGN_16 },
  { X86::CVTTPS2DQrr,          X86::CVTTPS2DQrm,          TB_ALIGN_16 },
  { X86::CVTTSD2SI64rr,        X86::CVTTSD2SI64rm,        0 },
  { X86::CVTTSD2SI64rr_Int,    X86::CVTTSD2SI64rm_Int,    TB_NO_REVERSE },
  { X86::CVTTSD2SIrr,          X86::CVTTSD2SIrm,          0 },
  { X86::CVTTSD2SIrr_Int,      X86::CVTTSD2SIrm_Int,      TB_NO_REVERSE },
  { X86::CVTTSS2SI64rr,        X86::CVTTSS2SI64rm,        0 },
  { X86::CVTTSS2SI64rr_Int,    X86::CVTTSS2SI64rm_Int,    TB_NO_REVERSE },
  { X86::CVTTSS2SIrr,          X86::CVTTSS2SIrm,          0 },
  { X86::CVTTSS2SIrr_Int,      X86::CVTTSS2SIrm_Int,      TB_NO_REVERSE },
  { X86::IMUL16rri,            X86::IMUL16rmi,            0 },
  { X86::IMUL16rri8,           X86::IMUL16rmi8,           0 },
  { X86::IMUL32rri,            X86::IMUL32rmi,            0 },
  { X86::IMUL32rri8,           X86::IMUL32rmi8,           0 },
  { X86::IMUL64rri32,          X86::IMUL64rmi32,          0 },
  { X86::IMUL64rri8,           X86::IMUL64rmi8,           0 },
  { X86::LWPINS32rri,          X86::LWPINS32rmi,          0 },
  { X86::LWPINS64rri,          X86::LWPINS64rmi,          0 },
  { X86::LWPVAL32rri,          X86::LWPVAL32rmi,          0 },
  { X86::LWPVAL64rri,          X86::LWPVAL64rmi,          0 },
  { X86::LZCNT16rr,            X86::LZCNT16rm,            0 },
  { X86::LZCNT32rr,            X86::LZCNT32rm,            0 },
  { X86::LZCNT64rr,            X86::LZCNT64rm,            0 },
  { X86::MMX_CVTPD2PIirr,      X86::MMX_CVTPD2PIirm,      TB_ALIGN_16 },
  { X86::MMX_CVTPI2PDirr,      X86::MMX_CVTPI2PDirm,      0 },
  { X86::MMX_CVTPS2PIirr,      X86::MMX_CVTPS2PIirm,      TB_NO_REVERSE },
  { X86::MMX_CVTTPD2PIirr,     X86::MMX_CVTTPD2PIirm,     TB_ALIGN_16 },
  { X86::MMX_CVTTPS2PIirr,     X86::MMX_CVTTPS2PIirm,     TB_NO_REVERSE },
  { X86::MMX_MOVD64to64rr,     X86::MMX_MOVQ64rm,         0 },
  { X86::MMX_PABSBrr,          X86::MMX_PABSBrm,          0 },
  { X86::MMX_PABSDrr,          X86::MMX_PABSDrm,          0 },
  { X86::MMX_PABSWrr,          X86::MMX_PABSWrm,          0 },
  { X86::MMX_PSHUFWri,         X86::MMX_PSHUFWmi,         0 },
  { X86::MOV16rr,              X86::MOV16rm,              0 },
  { X86::MOV32rr,              X86::MOV32rm,              0 },
  { X86::MOV64rr,              X86::MOV64rm,              0 },
  { X86::MOV64toPQIrr,         X86::MOVQI2PQIrm,          TB_NO_REVERSE },
  { X86::MOV64toSDrr,          X86::MOVSDrm_alt,          TB_NO_REVERSE },
  { X86::MOV8rr,               X86::MOV8rm,               0 },
  { X86::MOVAPDrr,             X86::MOVAPDrm,             TB_ALIGN_16 },
  { X86::MOVAPSrr,             X86::MOVAPSrm,             TB_ALIGN_16 },
  { X86::MOVDDUPrr,            X86::MOVDDUPrm,            TB_NO_REVERSE },
  { X86::MOVDI2PDIrr,          X86::MOVDI2PDIrm,          0 },
  { X86::MOVDI2SSrr,           X86::MOVSSrm_alt,          0 },
  { X86::MOVDQArr,             X86::MOVDQArm,             TB_ALIGN_16 },
  { X86::MOVDQUrr,             X86::MOVDQUrm,             0 },
  { X86::MOVSHDUPrr,           X86::MOVSHDUPrm,           TB_ALIGN_16 },
  { X86::MOVSLDUPrr,           X86::MOVSLDUPrm,           TB_ALIGN_16 },
  { X86::MOVSX16rr8,           X86::MOVSX16rm8,           0 },
  { X86::MOVSX32rr16,          X86::MOVSX32rm16,          0 },
  { X86::MOVSX32rr8,           X86::MOVSX32rm8,           0 },
  { X86::MOVSX32rr8_NOREX,     X86::MOVSX32rm8_NOREX,     0 },
  { X86::MOVSX64rr16,          X86::MOVSX64rm16,          0 },
  { X86::MOVSX64rr32,          X86::MOVSX64rm32,          0 },
  { X86::MOVSX64rr8,           X86::MOVSX64rm8,           0 },
  { X86::MOVUPDrr,             X86::MOVUPDrm,             0 },
  { X86::MOVUPSrr,             X86::MOVUPSrm,             0 },
  { X86::MOVZPQILo2PQIrr,      X86::MOVQI2PQIrm,          TB_NO_REVERSE },
  { X86::MOVZX16rr8,           X86::MOVZX16rm8,           0 },
  { X86::MOVZX32rr16,          X86::MOVZX32rm16,          0 },
  { X86::MOVZX32rr8,           X86::MOVZX32rm8,           0 },
  { X86::MOVZX32rr8_NOREX,     X86::MOVZX32rm8_NOREX,     0 },
  { X86::MOVZX64rr16,          X86::MOVZX64rm16,          0 },
  { X86::MOVZX64rr8,           X86::MOVZX64rm8,           0 },
  { X86::PABSBrr,              X86::PABSBrm,              TB_ALIGN_16 },
  { X86::PABSDrr,              X86::PABSDrm,              TB_ALIGN_16 },
  { X86::PABSWrr,              X86::PABSWrm,              TB_ALIGN_16 },
  { X86::PCMPESTRIrr,          X86::PCMPESTRIrm,          0 },
  { X86::PCMPESTRMrr,          X86::PCMPESTRMrm,          0 },
  { X86::PCMPISTRIrr,          X86::PCMPISTRIrm,          0 },
  { X86::PCMPISTRMrr,          X86::PCMPISTRMrm,          0 },
  { X86::PF2IDrr,              X86::PF2IDrm,              0 },
  { X86::PF2IWrr,              X86::PF2IWrm,              0 },
  { X86::PFRCPrr,              X86::PFRCPrm,              0 },
  { X86::PFRSQRTrr,            X86::PFRSQRTrm,            0 },
  { X86::PHMINPOSUWrr,         X86::PHMINPOSUWrm,         TB_ALIGN_16 },
  { X86::PI2FDrr,              X86::PI2FDrm,              0 },
  { X86::PI2FWrr,              X86::PI2FWrm,              0 },
  { X86::PMOVSXBDrr,           X86::PMOVSXBDrm,           TB_NO_REVERSE },
  { X86::PMOVSXBQrr,           X86::PMOVSXBQrm,           TB_NO_REVERSE },
  { X86::PMOVSXBWrr,           X86::PMOVSXBWrm,           TB_NO_REVERSE },
  { X86::PMOVSXDQrr,           X86::PMOVSXDQrm,           TB_NO_REVERSE },
  { X86::PMOVSXWDrr,           X86::PMOVSXWDrm,           TB_NO_REVERSE },
  { X86::PMOVSXWQrr,           X86::PMOVSXWQrm,           TB_NO_REVERSE },
  { X86::PMOVZXBDrr,           X86::PMOVZXBDrm,           TB_NO_REVERSE },
  { X86::PMOVZXBQrr,           X86::PMOVZXBQrm,           TB_NO_REVERSE },
  { X86::PMOVZXBWrr,           X86::PMOVZXBWrm,           TB_NO_REVERSE },
  { X86::PMOVZXDQrr,           X86::PMOVZXDQrm,           TB_NO_REVERSE },
  { X86::PMOVZXWDrr,           X86::PMOVZXWDrm,           TB_NO_REVERSE },
  { X86::PMOVZXWQrr,           X86::PMOVZXWQrm,           TB_NO_REVERSE },
  { X86::POPCNT16rr,           X86::POPCNT16rm,           0 },
  { X86::POPCNT32rr,           X86::POPCNT32rm,           0 },
  { X86::POPCNT64rr,           X86::POPCNT64rm,           0 },
  { X86::PSHUFDri,             X86::PSHUFDmi,             TB_ALIGN_16 },
  { X86::PSHUFHWri,            X86::PSHUFHWmi,            TB_ALIGN_16 },
  { X86::PSHUFLWri,            X86::PSHUFLWmi,            TB_ALIGN_16 },
  { X86::PSWAPDrr,             X86::PSWAPDrm,             0 },
  { X86::PTESTrr,              X86::PTESTrm,              TB_ALIGN_16 },
  { X86::RCPPSr,               X86::RCPPSm,               TB_ALIGN_16 },
  { X86::RCPSSr,               X86::RCPSSm,               0 },
  { X86::RORX32ri,             X86::RORX32mi,             0 },
  { X86::RORX64ri,             X86::RORX64mi,             0 },
  { X86::ROUNDPDr,             X86::ROUNDPDm,             TB_ALIGN_16 },
  { X86::ROUNDPSr,             X86::ROUNDPSm,             TB_ALIGN_16 },
  { X86::ROUNDSDr,             X86::ROUNDSDm,             0 },
  { X86::ROUNDSSr,             X86::ROUNDSSm,             0 },
  { X86::RSQRTPSr,             X86::RSQRTPSm,             TB_ALIGN_16 },
  { X86::RSQRTSSr,             X86::RSQRTSSm,             0 },
  { X86::SARX32rr,             X86::SARX32rm,             0 },
  { X86::SARX64rr,             X86::SARX64rm,             0 },
  { X86::SHLX32rr,             X86::SHLX32rm,             0 },
  { X86::SHLX64rr,             X86::SHLX64rm,             0 },
  { X86::SHRX32rr,             X86::SHRX32rm,             0 },
  { X86::SHRX64rr,             X86::SHRX64rm,             0 },
  { X86::SQRTPDr,              X86::SQRTPDm,              TB_ALIGN_16 },
  { X86::SQRTPSr,              X86::SQRTPSm,              TB_ALIGN_16 },
  { X86::SQRTSDr,              X86::SQRTSDm,              0 },
  { X86::SQRTSSr,              X86::SQRTSSm,              0 },
  { X86::T1MSKC32rr,           X86::T1MSKC32rm,           0 },
  { X86::T1MSKC64rr,           X86::T1MSKC64rm,           0 },
  { X86::TZCNT16rr,            X86::TZCNT16rm,            0 },
  { X86::TZCNT32rr,            X86::TZCNT32rm,            0 },
  { X86::TZCNT64rr,            X86::TZCNT64rm,            0 },
  { X86::TZMSK32rr,            X86::TZMSK32rm,            0 },
  { X86::TZMSK64rr,            X86::TZMSK64rm,            0 },
  { X86::UCOMISDrr,            X86::UCOMISDrm,            0 },
  { X86::UCOMISDrr_Int,        X86::UCOMISDrm_Int,        TB_NO_REVERSE },
  { X86::UCOMISSrr,            X86::UCOMISSrm,            0 },
  { X86::UCOMISSrr_Int,        X86::UCOMISSrm_Int,        TB_NO_REVERSE },
  { X86::VAESIMCrr,            X86::VAESIMCrm,            0 },
  { X86::VAESKEYGENASSIST128rr,X86::VAESKEYGENASSIST128rm,0 },
  { X86::VBROADCASTF32X2Z256r, X86::VBROADCASTF32X2Z256m, TB_NO_REVERSE },
  { X86::VBROADCASTF32X2Zr,    X86::VBROADCASTF32X2Zm,    TB_NO_REVERSE },
  { X86::VBROADCASTI32X2Z128r, X86::VBROADCASTI32X2Z128m, TB_NO_REVERSE },
  { X86::VBROADCASTI32X2Z256r, X86::VBROADCASTI32X2Z256m, TB_NO_REVERSE },
  { X86::VBROADCASTI32X2Zr,    X86::VBROADCASTI32X2Zm,    TB_NO_REVERSE },
  { X86::VBROADCASTSDYrr,      X86::VBROADCASTSDYrm,      TB_NO_REVERSE },
  { X86::VBROADCASTSDZ256r,    X86::VBROADCASTSDZ256m,    TB_NO_REVERSE },
  { X86::VBROADCASTSDZr,       X86::VBROADCASTSDZm,       TB_NO_REVERSE },
  { X86::VBROADCASTSSYrr,      X86::VBROADCASTSSYrm,      TB_NO_REVERSE },
  { X86::VBROADCASTSSZ128r,    X86::VBROADCASTSSZ128m,    TB_NO_REVERSE },
  { X86::VBROADCASTSSZ256r,    X86::VBROADCASTSSZ256m,    TB_NO_REVERSE },
  { X86::VBROADCASTSSZr,       X86::VBROADCASTSSZm,       TB_NO_REVERSE },
  { X86::VBROADCASTSSrr,       X86::VBROADCASTSSrm,       TB_NO_REVERSE },
  { X86::VCOMISDZrr,           X86::VCOMISDZrm,           0 },
  { X86::VCOMISDZrr_Int,       X86::VCOMISDZrm_Int,       TB_NO_REVERSE },
  { X86::VCOMISDrr,            X86::VCOMISDrm,            0 },
  { X86::VCOMISDrr_Int,        X86::VCOMISDrm_Int,        TB_NO_REVERSE },
  { X86::VCOMISSZrr,           X86::VCOMISSZrm,           0 },
  { X86::VCOMISSZrr_Int,       X86::VCOMISSZrm_Int,       TB_NO_REVERSE },
  { X86::VCOMISSrr,            X86::VCOMISSrm,            0 },
  { X86::VCOMISSrr_Int,        X86::VCOMISSrm_Int,        TB_NO_REVERSE },
  { X86::VCVTDQ2PDYrr,         X86::VCVTDQ2PDYrm,         0 },
  { X86::VCVTDQ2PDZ128rr,      X86::VCVTDQ2PDZ128rm,      TB_NO_REVERSE },
  { X86::VCVTDQ2PDZ256rr,      X86::VCVTDQ2PDZ256rm,      0 },
  { X86::VCVTDQ2PDZrr,         X86::VCVTDQ2PDZrm,         0 },
  { X86::VCVTDQ2PDrr,          X86::VCVTDQ2PDrm,          TB_NO_REVERSE },
  { X86::VCVTDQ2PSYrr,         X86::VCVTDQ2PSYrm,         0 },
  { X86::VCVTDQ2PSZ128rr,      X86::VCVTDQ2PSZ128rm,      0 },
  { X86::VCVTDQ2PSZ256rr,      X86::VCVTDQ2PSZ256rm,      0 },
  { X86::VCVTDQ2PSZrr,         X86::VCVTDQ2PSZrm,         0 },
  { X86::VCVTDQ2PSrr,          X86::VCVTDQ2PSrm,          0 },
  { X86::VCVTNEPS2BF16Z128rr,  X86::VCVTNEPS2BF16Z128rm,  0 },
  { X86::VCVTNEPS2BF16Z256rr,  X86::VCVTNEPS2BF16Z256rm,  0 },
  { X86::VCVTNEPS2BF16Zrr,     X86::VCVTNEPS2BF16Zrm,     0 },
  { X86::VCVTPD2DQYrr,         X86::VCVTPD2DQYrm,         0 },
  { X86::VCVTPD2DQZ128rr,      X86::VCVTPD2DQZ128rm,      0 },
  { X86::VCVTPD2DQZ256rr,      X86::VCVTPD2DQZ256rm,      0 },
  { X86::VCVTPD2DQZrr,         X86::VCVTPD2DQZrm,         0 },
  { X86::VCVTPD2DQrr,          X86::VCVTPD2DQrm,          0 },
  { X86::VCVTPD2PSYrr,         X86::VCVTPD2PSYrm,         0 },
  { X86::VCVTPD2PSZ128rr,      X86::VCVTPD2PSZ128rm,      0 },
  { X86::VCVTPD2PSZ256rr,      X86::VCVTPD2PSZ256rm,      0 },
  { X86::VCVTPD2PSZrr,         X86::VCVTPD2PSZrm,         0 },
  { X86::VCVTPD2PSrr,          X86::VCVTPD2PSrm,          0 },
  { X86::VCVTPD2QQZ128rr,      X86::VCVTPD2QQZ128rm,      0 },
  { X86::VCVTPD2QQZ256rr,      X86::VCVTPD2QQZ256rm,      0 },
  { X86::VCVTPD2QQZrr,         X86::VCVTPD2QQZrm,         0 },
  { X86::VCVTPD2UDQZ128rr,     X86::VCVTPD2UDQZ128rm,     0 },
  { X86::VCVTPD2UDQZ256rr,     X86::VCVTPD2UDQZ256rm,     0 },
  { X86::VCVTPD2UDQZrr,        X86::VCVTPD2UDQZrm,        0 },
  { X86::VCVTPD2UQQZ128rr,     X86::VCVTPD2UQQZ128rm,     0 },
  { X86::VCVTPD2UQQZ256rr,     X86::VCVTPD2UQQZ256rm,     0 },
  { X86::VCVTPD2UQQZrr,        X86::VCVTPD2UQQZrm,        0 },
  { X86::VCVTPH2PSYrr,         X86::VCVTPH2PSYrm,         0 },
  { X86::VCVTPH2PSZ128rr,      X86::VCVTPH2PSZ128rm,      TB_NO_REVERSE },
  { X86::VCVTPH2PSZ256rr,      X86::VCVTPH2PSZ256rm,      0 },
  { X86::VCVTPH2PSZrr,         X86::VCVTPH2PSZrm,         0 },
  { X86::VCVTPH2PSrr,          X86::VCVTPH2PSrm,          TB_NO_REVERSE },
  { X86::VCVTPS2DQYrr,         X86::VCVTPS2DQYrm,         0 },
  { X86::VCVTPS2DQZ128rr,      X86::VCVTPS2DQZ128rm,      0 },
  { X86::VCVTPS2DQZ256rr,      X86::VCVTPS2DQZ256rm,      0 },
  { X86::VCVTPS2DQZrr,         X86::VCVTPS2DQZrm,         0 },
  { X86::VCVTPS2DQrr,          X86::VCVTPS2DQrm,          0 },
  { X86::VCVTPS2PDYrr,         X86::VCVTPS2PDYrm,         0 },
  { X86::VCVTPS2PDZ128rr,      X86::VCVTPS2PDZ128rm,      TB_NO_REVERSE },
  { X86::VCVTPS2PDZ256rr,      X86::VCVTPS2PDZ256rm,      0 },
  { X86::VCVTPS2PDZrr,         X86::VCVTPS2PDZrm,         0 },
  { X86::VCVTPS2PDrr,          X86::VCVTPS2PDrm,          TB_NO_REVERSE },
  { X86::VCVTPS2QQZ128rr,      X86::VCVTPS2QQZ128rm,      TB_NO_REVERSE },
  { X86::VCVTPS2QQZ256rr,      X86::VCVTPS2QQZ256rm,      0 },
  { X86::VCVTPS2QQZrr,         X86::VCVTPS2QQZrm,         0 },
  { X86::VCVTPS2UDQZ128rr,     X86::VCVTPS2UDQZ128rm,     0 },
  { X86::VCVTPS2UDQZ256rr,     X86::VCVTPS2UDQZ256rm,     0 },
  { X86::VCVTPS2UDQZrr,        X86::VCVTPS2UDQZrm,        0 },
  { X86::VCVTPS2UQQZ128rr,     X86::VCVTPS2UQQZ128rm,     TB_NO_REVERSE },
  { X86::VCVTPS2UQQZ256rr,     X86::VCVTPS2UQQZ256rm,     0 },
  { X86::VCVTPS2UQQZrr,        X86::VCVTPS2UQQZrm,        0 },
  { X86::VCVTQQ2PDZ128rr,      X86::VCVTQQ2PDZ128rm,      0 },
  { X86::VCVTQQ2PDZ256rr,      X86::VCVTQQ2PDZ256rm,      0 },
  { X86::VCVTQQ2PDZrr,         X86::VCVTQQ2PDZrm,         0 },
  { X86::VCVTQQ2PSZ128rr,      X86::VCVTQQ2PSZ128rm,      0 },
  { X86::VCVTQQ2PSZ256rr,      X86::VCVTQQ2PSZ256rm,      0 },
  { X86::VCVTQQ2PSZrr,         X86::VCVTQQ2PSZrm,         0 },
  { X86::VCVTSD2SI64Zrr_Int,   X86::VCVTSD2SI64Zrm_Int,   TB_NO_REVERSE },
  { X86::VCVTSD2SI64rr_Int,    X86::VCVTSD2SI64rm_Int,    TB_NO_REVERSE },
  { X86::VCVTSD2SIZrr_Int,     X86::VCVTSD2SIZrm_Int,     TB_NO_REVERSE },
  { X86::VCVTSD2SIrr_Int,      X86::VCVTSD2SIrm_Int,      TB_NO_REVERSE },
  { X86::VCVTSD2USI64Zrr_Int,  X86::VCVTSD2USI64Zrm_Int,  TB_NO_REVERSE },
  { X86::VCVTSD2USIZrr_Int,    X86::VCVTSD2USIZrm_Int,    TB_NO_REVERSE },
  { X86::VCVTSS2SI64Zrr_Int,   X86::VCVTSS2SI64Zrm_Int,   TB_NO_REVERSE },
  { X86::VCVTSS2SI64rr_Int,    X86::VCVTSS2SI64rm_Int,    TB_NO_REVERSE },
  { X86::VCVTSS2SIZrr_Int,     X86::VCVTSS2SIZrm_Int,     TB_NO_REVERSE },
  { X86::VCVTSS2SIrr_Int,      X86::VCVTSS2SIrm_Int,      TB_NO_REVERSE },
  { X86::VCVTSS2USI64Zrr_Int,  X86::VCVTSS2USI64Zrm_Int,  TB_NO_REVERSE },
  { X86::VCVTSS2USIZrr_Int,    X86::VCVTSS2USIZrm_Int,    TB_NO_REVERSE },
  { X86::VCVTTPD2DQYrr,        X86::VCVTTPD2DQYrm,        0 },
  { X86::VCVTTPD2DQZ128rr,     X86::VCVTTPD2DQZ128rm,     0 },
  { X86::VCVTTPD2DQZ256rr,     X86::VCVTTPD2DQZ256rm,     0 },
  { X86::VCVTTPD2DQZrr,        X86::VCVTTPD2DQZrm,        0 },
  { X86::VCVTTPD2DQrr,         X86::VCVTTPD2DQrm,         0 },
  { X86::VCVTTPD2QQZ128rr,     X86::VCVTTPD2QQZ128rm,     0 },
  { X86::VCVTTPD2QQZ256rr,     X86::VCVTTPD2QQZ256rm,     0 },
  { X86::VCVTTPD2QQZrr,        X86::VCVTTPD2QQZrm,        0 },
  { X86::VCVTTPD2UDQZ128rr,    X86::VCVTTPD2UDQZ128rm,    0 },
  { X86::VCVTTPD2UDQZ256rr,    X86::VCVTTPD2UDQZ256rm,    0 },
  { X86::VCVTTPD2UDQZrr,       X86::VCVTTPD2UDQZrm,       0 },
  { X86::VCVTTPD2UQQZ128rr,    X86::VCVTTPD2UQQZ128rm,    0 },
  { X86::VCVTTPD2UQQZ256rr,    X86::VCVTTPD2UQQZ256rm,    0 },
  { X86::VCVTTPD2UQQZrr,       X86::VCVTTPD2UQQZrm,       0 },
  { X86::VCVTTPS2DQYrr,        X86::VCVTTPS2DQYrm,        0 },
  { X86::VCVTTPS2DQZ128rr,     X86::VCVTTPS2DQZ128rm,     0 },
  { X86::VCVTTPS2DQZ256rr,     X86::VCVTTPS2DQZ256rm,     0 },
  { X86::VCVTTPS2DQZrr,        X86::VCVTTPS2DQZrm,        0 },
  { X86::VCVTTPS2DQrr,         X86::VCVTTPS2DQrm,         0 },
  { X86::VCVTTPS2QQZ128rr,     X86::VCVTTPS2QQZ128rm,     TB_NO_REVERSE },
  { X86::VCVTTPS2QQZ256rr,     X86::VCVTTPS2QQZ256rm,     0 },
  { X86::VCVTTPS2QQZrr,        X86::VCVTTPS2QQZrm,        0 },
  { X86::VCVTTPS2UDQZ128rr,    X86::VCVTTPS2UDQZ128rm,    0 },
  { X86::VCVTTPS2UDQZ256rr,    X86::VCVTTPS2UDQZ256rm,    0 },
  { X86::VCVTTPS2UDQZrr,       X86::VCVTTPS2UDQZrm,       0 },
  { X86::VCVTTPS2UQQZ128rr,    X86::VCVTTPS2UQQZ128rm,    TB_NO_REVERSE },
  { X86::VCVTTPS2UQQZ256rr,    X86::VCVTTPS2UQQZ256rm,    0 },
  { X86::VCVTTPS2UQQZrr,       X86::VCVTTPS2UQQZrm,       0 },
  { X86::VCVTTSD2SI64Zrr,      X86::VCVTTSD2SI64Zrm,      0 },
  { X86::VCVTTSD2SI64Zrr_Int,  X86::VCVTTSD2SI64Zrm_Int,  TB_NO_REVERSE },
  { X86::VCVTTSD2SI64rr,       X86::VCVTTSD2SI64rm,       0 },
  { X86::VCVTTSD2SI64rr_Int,   X86::VCVTTSD2SI64rm_Int,   TB_NO_REVERSE },
  { X86::VCVTTSD2SIZrr,        X86::VCVTTSD2SIZrm,        0 },
  { X86::VCVTTSD2SIZrr_Int,    X86::VCVTTSD2SIZrm_Int,    TB_NO_REVERSE },
  { X86::VCVTTSD2SIrr,         X86::VCVTTSD2SIrm,         0 },
  { X86::VCVTTSD2SIrr_Int,     X86::VCVTTSD2SIrm_Int,     TB_NO_REVERSE },
  { X86::VCVTTSD2USI64Zrr,     X86::VCVTTSD2USI64Zrm,     0 },
  { X86::VCVTTSD2USI64Zrr_Int, X86::VCVTTSD2USI64Zrm_Int, TB_NO_REVERSE },
  { X86::VCVTTSD2USIZrr,       X86::VCVTTSD2USIZrm,       0 },
  { X86::VCVTTSD2USIZrr_Int,   X86::VCVTTSD2USIZrm_Int,   TB_NO_REVERSE },
  { X86::VCVTTSS2SI64Zrr,      X86::VCVTTSS2SI64Zrm,      0 },
  { X86::VCVTTSS2SI64Zrr_Int,  X86::VCVTTSS2SI64Zrm_Int,  TB_NO_REVERSE },
  { X86::VCVTTSS2SI64rr,       X86::VCVTTSS2SI64rm,       0 },
  { X86::VCVTTSS2SI64rr_Int,   X86::VCVTTSS2SI64rm_Int,   TB_NO_REVERSE },
  { X86::VCVTTSS2SIZrr,        X86::VCVTTSS2SIZrm,        0 },
  { X86::VCVTTSS2SIZrr_Int,    X86::VCVTTSS2SIZrm_Int,    TB_NO_REVERSE },
  { X86::VCVTTSS2SIrr,         X86::VCVTTSS2SIrm,         0 },
  { X86::VCVTTSS2SIrr_Int,     X86::VCVTTSS2SIrm_Int,     TB_NO_REVERSE },
  { X86::VCVTTSS2USI64Zrr,     X86::VCVTTSS2USI64Zrm,     0 },
  { X86::VCVTTSS2USI64Zrr_Int, X86::VCVTTSS2USI64Zrm_Int, TB_NO_REVERSE },
  { X86::VCVTTSS2USIZrr,       X86::VCVTTSS2USIZrm,       0 },
  { X86::VCVTTSS2USIZrr_Int,   X86::VCVTTSS2USIZrm_Int,   TB_NO_REVERSE },
  { X86::VCVTUDQ2PDZ128rr,     X86::VCVTUDQ2PDZ128rm,     TB_NO_REVERSE },
  { X86::VCVTUDQ2PDZ256rr,     X86::VCVTUDQ2PDZ256rm,     0 },
  { X86::VCVTUDQ2PDZrr,        X86::VCVTUDQ2PDZrm,        0 },
  { X86::VCVTUDQ2PSZ128rr,     X86::VCVTUDQ2PSZ128rm,     0 },
  { X86::VCVTUDQ2PSZ256rr,     X86::VCVTUDQ2PSZ256rm,     0 },
  { X86::VCVTUDQ2PSZrr,        X86::VCVTUDQ2PSZrm,        0 },
  { X86::VCVTUQQ2PDZ128rr,     X86::VCVTUQQ2PDZ128rm,     0 },
  { X86::VCVTUQQ2PDZ256rr,     X86::VCVTUQQ2PDZ256rm,     0 },
  { X86::VCVTUQQ2PDZrr,        X86::VCVTUQQ2PDZrm,        0 },
  { X86::VCVTUQQ2PSZ128rr,     X86::VCVTUQQ2PSZ128rm,     0 },
  { X86::VCVTUQQ2PSZ256rr,     X86::VCVTUQQ2PSZ256rm,     0 },
  { X86::VCVTUQQ2PSZrr,        X86::VCVTUQQ2PSZrm,        0 },
  { X86::VEXP2PDZr,            X86::VEXP2PDZm,            0 },
  { X86::VEXP2PSZr,            X86::VEXP2PSZm,            0 },
  { X86::VEXPANDPDZ128rr,      X86::VEXPANDPDZ128rm,      TB_NO_REVERSE },
  { X86::VEXPANDPDZ256rr,      X86::VEXPANDPDZ256rm,      TB_NO_REVERSE },
  { X86::VEXPANDPDZrr,         X86::VEXPANDPDZrm,         TB_NO_REVERSE },
  { X86::VEXPANDPSZ128rr,      X86::VEXPANDPSZ128rm,      TB_NO_REVERSE },
  { X86::VEXPANDPSZ256rr,      X86::VEXPANDPSZ256rm,      TB_NO_REVERSE },
  { X86::VEXPANDPSZrr,         X86::VEXPANDPSZrm,         TB_NO_REVERSE },
  { X86::VFPCLASSPDZ128rr,     X86::VFPCLASSPDZ128rm,     0 },
  { X86::VFPCLASSPDZ256rr,     X86::VFPCLASSPDZ256rm,     0 },
  { X86::VFPCLASSPDZrr,        X86::VFPCLASSPDZrm,        0 },
  { X86::VFPCLASSPSZ128rr,     X86::VFPCLASSPSZ128rm,     0 },
  { X86::VFPCLASSPSZ256rr,     X86::VFPCLASSPSZ256rm,     0 },
  { X86::VFPCLASSPSZrr,        X86::VFPCLASSPSZrm,        0 },
  { X86::VFPCLASSSDZrr,        X86::VFPCLASSSDZrm,        TB_NO_REVERSE },
  { X86::VFPCLASSSSZrr,        X86::VFPCLASSSSZrm,        TB_NO_REVERSE },
  { X86::VFRCZPDYrr,           X86::VFRCZPDYrm,           0 },
  { X86::VFRCZPDrr,            X86::VFRCZPDrm,            0 },
  { X86::VFRCZPSYrr,           X86::VFRCZPSYrm,           0 },
  { X86::VFRCZPSrr,            X86::VFRCZPSrm,            0 },
  { X86::VFRCZSDrr,            X86::VFRCZSDrm,            TB_NO_REVERSE },
  { X86::VFRCZSSrr,            X86::VFRCZSSrm,            TB_NO_REVERSE },
  { X86::VGETEXPPDZ128r,       X86::VGETEXPPDZ128m,       0 },
  { X86::VGETEXPPDZ256r,       X86::VGETEXPPDZ256m,       0 },
  { X86::VGETEXPPDZr,          X86::VGETEXPPDZm,          0 },
  { X86::VGETEXPPSZ128r,       X86::VGETEXPPSZ128m,       0 },
  { X86::VGETEXPPSZ256r,       X86::VGETEXPPSZ256m,       0 },
  { X86::VGETEXPPSZr,          X86::VGETEXPPSZm,          0 },
  { X86::VGETMANTPDZ128rri,    X86::VGETMANTPDZ128rmi,    0 },
  { X86::VGETMANTPDZ256rri,    X86::VGETMANTPDZ256rmi,    0 },
  { X86::VGETMANTPDZrri,       X86::VGETMANTPDZrmi,       0 },
  { X86::VGETMANTPSZ128rri,    X86::VGETMANTPSZ128rmi,    0 },
  { X86::VGETMANTPSZ256rri,    X86::VGETMANTPSZ256rmi,    0 },
  { X86::VGETMANTPSZrri,       X86::VGETMANTPSZrmi,       0 },
  { X86::VMOV64toPQIZrr,       X86::VMOVQI2PQIZrm,        TB_NO_REVERSE },
  { X86::VMOV64toPQIrr,        X86::VMOVQI2PQIrm,         TB_NO_REVERSE },
  { X86::VMOV64toSDZrr,        X86::VMOVSDZrm_alt,        TB_NO_REVERSE },
  { X86::VMOV64toSDrr,         X86::VMOVSDrm_alt,         TB_NO_REVERSE },
  { X86::VMOVAPDYrr,           X86::VMOVAPDYrm,           TB_ALIGN_32 },
  { X86::VMOVAPDZ128rr,        X86::VMOVAPDZ128rm,        TB_ALIGN_16 },
  { X86::VMOVAPDZ256rr,        X86::VMOVAPDZ256rm,        TB_ALIGN_32 },
  { X86::VMOVAPDZrr,           X86::VMOVAPDZrm,           TB_ALIGN_64 },
  { X86::VMOVAPDrr,            X86::VMOVAPDrm,            TB_ALIGN_16 },
  { X86::VMOVAPSYrr,           X86::VMOVAPSYrm,           TB_ALIGN_32 },
  { X86::VMOVAPSZ128rr,        X86::VMOVAPSZ128rm,        TB_ALIGN_16 },
  { X86::VMOVAPSZ256rr,        X86::VMOVAPSZ256rm,        TB_ALIGN_32 },
  { X86::VMOVAPSZrr,           X86::VMOVAPSZrm,           TB_ALIGN_64 },
  { X86::VMOVAPSrr,            X86::VMOVAPSrm,            TB_ALIGN_16 },
  { X86::VMOVDDUPYrr,          X86::VMOVDDUPYrm,          0 },
  { X86::VMOVDDUPZ128rr,       X86::VMOVDDUPZ128rm,       TB_NO_REVERSE },
  { X86::VMOVDDUPZ256rr,       X86::VMOVDDUPZ256rm,       0 },
  { X86::VMOVDDUPZrr,          X86::VMOVDDUPZrm,          0 },
  { X86::VMOVDDUPrr,           X86::VMOVDDUPrm,           TB_NO_REVERSE },
  { X86::VMOVDI2PDIZrr,        X86::VMOVDI2PDIZrm,        0 },
  { X86::VMOVDI2PDIrr,         X86::VMOVDI2PDIrm,         0 },
  { X86::VMOVDI2SSZrr,         X86::VMOVSSZrm_alt,        0 },
  { X86::VMOVDI2SSrr,          X86::VMOVSSrm_alt,         0 },
  { X86::VMOVDQA32Z128rr,      X86::VMOVDQA32Z128rm,      TB_ALIGN_16 },
  { X86::VMOVDQA32Z256rr,      X86::VMOVDQA32Z256rm,      TB_ALIGN_32 },
  { X86::VMOVDQA32Zrr,         X86::VMOVDQA32Zrm,         TB_ALIGN_64 },
  { X86::VMOVDQA64Z128rr,      X86::VMOVDQA64Z128rm,      TB_ALIGN_16 },
  { X86::VMOVDQA64Z256rr,      X86::VMOVDQA64Z256rm,      TB_ALIGN_32 },
  { X86::VMOVDQA64Zrr,         X86::VMOVDQA64Zrm,         TB_ALIGN_64 },
  { X86::VMOVDQAYrr,           X86::VMOVDQAYrm,           TB_ALIGN_32 },
  { X86::VMOVDQArr,            X86::VMOVDQArm,            TB_ALIGN_16 },
  { X86::VMOVDQU16Z128rr,      X86::VMOVDQU16Z128rm,      0 },
  { X86::VMOVDQU16Z256rr,      X86::VMOVDQU16Z256rm,      0 },
  { X86::VMOVDQU16Zrr,         X86::VMOVDQU16Zrm,         0 },
  { X86::VMOVDQU32Z128rr,      X86::VMOVDQU32Z128rm,      0 },
  { X86::VMOVDQU32Z256rr,      X86::VMOVDQU32Z256rm,      0 },
  { X86::VMOVDQU32Zrr,         X86::VMOVDQU32Zrm,         0 },
  { X86::VMOVDQU64Z128rr,      X86::VMOVDQU64Z128rm,      0 },
  { X86::VMOVDQU64Z256rr,      X86::VMOVDQU64Z256rm,      0 },
  { X86::VMOVDQU64Zrr,         X86::VMOVDQU64Zrm,         0 },
  { X86::VMOVDQU8Z128rr,       X86::VMOVDQU8Z128rm,       0 },
  { X86::VMOVDQU8Z256rr,       X86::VMOVDQU8Z256rm,       0 },
  { X86::VMOVDQU8Zrr,          X86::VMOVDQU8Zrm,          0 },
  { X86::VMOVDQUYrr,           X86::VMOVDQUYrm,           0 },
  { X86::VMOVDQUrr,            X86::VMOVDQUrm,            0 },
  { X86::VMOVSHDUPYrr,         X86::VMOVSHDUPYrm,         0 },
  { X86::VMOVSHDUPZ128rr,      X86::VMOVSHDUPZ128rm,      0 },
  { X86::VMOVSHDUPZ256rr,      X86::VMOVSHDUPZ256rm,      0 },
  { X86::VMOVSHDUPZrr,         X86::VMOVSHDUPZrm,         0 },
  { X86::VMOVSHDUPrr,          X86::VMOVSHDUPrm,          0 },
  { X86::VMOVSLDUPYrr,         X86::VMOVSLDUPYrm,         0 },
  { X86::VMOVSLDUPZ128rr,      X86::VMOVSLDUPZ128rm,      0 },
  { X86::VMOVSLDUPZ256rr,      X86::VMOVSLDUPZ256rm,      0 },
  { X86::VMOVSLDUPZrr,         X86::VMOVSLDUPZrm,         0 },
  { X86::VMOVSLDUPrr,          X86::VMOVSLDUPrm,          0 },
  { X86::VMOVUPDYrr,           X86::VMOVUPDYrm,           0 },
  { X86::VMOVUPDZ128rr,        X86::VMOVUPDZ128rm,        0 },
  { X86::VMOVUPDZ256rr,        X86::VMOVUPDZ256rm,        0 },
  { X86::VMOVUPDZrr,           X86::VMOVUPDZrm,           0 },
  { X86::VMOVUPDrr,            X86::VMOVUPDrm,            0 },
  { X86::VMOVUPSYrr,           X86::VMOVUPSYrm,           0 },
  { X86::VMOVUPSZ128rr,        X86::VMOVUPSZ128rm,        0 },
  { X86::VMOVUPSZ256rr,        X86::VMOVUPSZ256rm,        0 },
  { X86::VMOVUPSZrr,           X86::VMOVUPSZrm,           0 },
  { X86::VMOVUPSrr,            X86::VMOVUPSrm,            0 },
  { X86::VMOVZPQILo2PQIZrr,    X86::VMOVQI2PQIZrm,        TB_NO_REVERSE },
  { X86::VMOVZPQILo2PQIrr,     X86::VMOVQI2PQIrm,         TB_NO_REVERSE },
  { X86::VPABSBYrr,            X86::VPABSBYrm,            0 },
  { X86::VPABSBZ128rr,         X86::VPABSBZ128rm,         0 },
  { X86::VPABSBZ256rr,         X86::VPABSBZ256rm,         0 },
  { X86::VPABSBZrr,            X86::VPABSBZrm,            0 },
  { X86::VPABSBrr,             X86::VPABSBrm,             0 },
  { X86::VPABSDYrr,            X86::VPABSDYrm,            0 },
  { X86::VPABSDZ128rr,         X86::VPABSDZ128rm,         0 },
  { X86::VPABSDZ256rr,         X86::VPABSDZ256rm,         0 },
  { X86::VPABSDZrr,            X86::VPABSDZrm,            0 },
  { X86::VPABSDrr,             X86::VPABSDrm,             0 },
  { X86::VPABSQZ128rr,         X86::VPABSQZ128rm,         0 },
  { X86::VPABSQZ256rr,         X86::VPABSQZ256rm,         0 },
  { X86::VPABSQZrr,            X86::VPABSQZrm,            0 },
  { X86::VPABSWYrr,            X86::VPABSWYrm,            0 },
  { X86::VPABSWZ128rr,         X86::VPABSWZ128rm,         0 },
  { X86::VPABSWZ256rr,         X86::VPABSWZ256rm,         0 },
  { X86::VPABSWZrr,            X86::VPABSWZrm,            0 },
  { X86::VPABSWrr,             X86::VPABSWrm,             0 },
  { X86::VPBROADCASTBYrr,      X86::VPBROADCASTBYrm,      TB_NO_REVERSE },
  { X86::VPBROADCASTBZ128r,    X86::VPBROADCASTBZ128m,    TB_NO_REVERSE },
  { X86::VPBROADCASTBZ256r,    X86::VPBROADCASTBZ256m,    TB_NO_REVERSE },
  { X86::VPBROADCASTBZr,       X86::VPBROADCASTBZm,       TB_NO_REVERSE },
  { X86::VPBROADCASTBrr,       X86::VPBROADCASTBrm,       TB_NO_REVERSE },
  { X86::VPBROADCASTDYrr,      X86::VPBROADCASTDYrm,      TB_NO_REVERSE },
  { X86::VPBROADCASTDZ128r,    X86::VPBROADCASTDZ128m,    TB_NO_REVERSE },
  { X86::VPBROADCASTDZ256r,    X86::VPBROADCASTDZ256m,    TB_NO_REVERSE },
  { X86::VPBROADCASTDZr,       X86::VPBROADCASTDZm,       TB_NO_REVERSE },
  { X86::VPBROADCASTDrr,       X86::VPBROADCASTDrm,       TB_NO_REVERSE },
  { X86::VPBROADCASTQYrr,      X86::VPBROADCASTQYrm,      TB_NO_REVERSE },
  { X86::VPBROADCASTQZ128r,    X86::VPBROADCASTQZ128m,    TB_NO_REVERSE },
  { X86::VPBROADCASTQZ256r,    X86::VPBROADCASTQZ256m,    TB_NO_REVERSE },
  { X86::VPBROADCASTQZr,       X86::VPBROADCASTQZm,       TB_NO_REVERSE },
  { X86::VPBROADCASTQrr,       X86::VPBROADCASTQrm,       TB_NO_REVERSE },
  { X86::VPBROADCASTWYrr,      X86::VPBROADCASTWYrm,      TB_NO_REVERSE },
  { X86::VPBROADCASTWZ128r,    X86::VPBROADCASTWZ128m,    TB_NO_REVERSE },
  { X86::VPBROADCASTWZ256r,    X86::VPBROADCASTWZ256m,    TB_NO_REVERSE },
  { X86::VPBROADCASTWZr,       X86::VPBROADCASTWZm,       TB_NO_REVERSE },
  { X86::VPBROADCASTWrr,       X86::VPBROADCASTWrm,       TB_NO_REVERSE },
  { X86::VPCMPESTRIrr,         X86::VPCMPESTRIrm,         0 },
  { X86::VPCMPESTRMrr,         X86::VPCMPESTRMrm,         0 },
  { X86::VPCMPISTRIrr,         X86::VPCMPISTRIrm,         0 },
  { X86::VPCMPISTRMrr,         X86::VPCMPISTRMrm,         0 },
  { X86::VPCONFLICTDZ128rr,    X86::VPCONFLICTDZ128rm,    0 },
  { X86::VPCONFLICTDZ256rr,    X86::VPCONFLICTDZ256rm,    0 },
  { X86::VPCONFLICTDZrr,       X86::VPCONFLICTDZrm,       0 },
  { X86::VPCONFLICTQZ128rr,    X86::VPCONFLICTQZ128rm,    0 },
  { X86::VPCONFLICTQZ256rr,    X86::VPCONFLICTQZ256rm,    0 },
  { X86::VPCONFLICTQZrr,       X86::VPCONFLICTQZrm,       0 },
  { X86::VPERMILPDYri,         X86::VPERMILPDYmi,         0 },
  { X86::VPERMILPDZ128ri,      X86::VPERMILPDZ128mi,      0 },
  { X86::VPERMILPDZ256ri,      X86::VPERMILPDZ256mi,      0 },
  { X86::VPERMILPDZri,         X86::VPERMILPDZmi,         0 },
  { X86::VPERMILPDri,          X86::VPERMILPDmi,          0 },
  { X86::VPERMILPSYri,         X86::VPERMILPSYmi,         0 },
  { X86::VPERMILPSZ128ri,      X86::VPERMILPSZ128mi,      0 },
  { X86::VPERMILPSZ256ri,      X86::VPERMILPSZ256mi,      0 },
  { X86::VPERMILPSZri,         X86::VPERMILPSZmi,         0 },
  { X86::VPERMILPSri,          X86::VPERMILPSmi,          0 },
  { X86::VPERMPDYri,           X86::VPERMPDYmi,           0 },
  { X86::VPERMPDZ256ri,        X86::VPERMPDZ256mi,        0 },
  { X86::VPERMPDZri,           X86::VPERMPDZmi,           0 },
  { X86::VPERMQYri,            X86::VPERMQYmi,            0 },
  { X86::VPERMQZ256ri,         X86::VPERMQZ256mi,         0 },
  { X86::VPERMQZri,            X86::VPERMQZmi,            0 },
  { X86::VPEXPANDBZ128rr,      X86::VPEXPANDBZ128rm,      TB_NO_REVERSE },
  { X86::VPEXPANDBZ256rr,      X86::VPEXPANDBZ256rm,      TB_NO_REVERSE },
  { X86::VPEXPANDBZrr,         X86::VPEXPANDBZrm,         TB_NO_REVERSE },
  { X86::VPEXPANDDZ128rr,      X86::VPEXPANDDZ128rm,      TB_NO_REVERSE },
  { X86::VPEXPANDDZ256rr,      X86::VPEXPANDDZ256rm,      TB_NO_REVERSE },
  { X86::VPEXPANDDZrr,         X86::VPEXPANDDZrm,         TB_NO_REVERSE },
  { X86::VPEXPANDQZ128rr,      X86::VPEXPANDQZ128rm,      TB_NO_REVERSE },
  { X86::VPEXPANDQZ256rr,      X86::VPEXPANDQZ256rm,      TB_NO_REVERSE },
  { X86::VPEXPANDQZrr,         X86::VPEXPANDQZrm,         TB_NO_REVERSE },
  { X86::VPEXPANDWZ128rr,      X86::VPEXPANDWZ128rm,      TB_NO_REVERSE },
  { X86::VPEXPANDWZ256rr,      X86::VPEXPANDWZ256rm,      TB_NO_REVERSE },
  { X86::VPEXPANDWZrr,         X86::VPEXPANDWZrm,         TB_NO_REVERSE },
  { X86::VPHADDBDrr,           X86::VPHADDBDrm,           0 },
  { X86::VPHADDBQrr,           X86::VPHADDBQrm,           0 },
  { X86::VPHADDBWrr,           X86::VPHADDBWrm,           0 },
  { X86::VPHADDDQrr,           X86::VPHADDDQrm,           0 },
  { X86::VPHADDUBDrr,          X86::VPHADDUBDrm,          0 },
  { X86::VPHADDUBQrr,          X86::VPHADDUBQrm,          0 },
  { X86::VPHADDUBWrr,          X86::VPHADDUBWrm,          0 },
  { X86::VPHADDUDQrr,          X86::VPHADDUDQrm,          0 },
  { X86::VPHADDUWDrr,          X86::VPHADDUWDrm,          0 },
  { X86::VPHADDUWQrr,          X86::VPHADDUWQrm,          0 },
  { X86::VPHADDWDrr,           X86::VPHADDWDrm,           0 },
  { X86::VPHADDWQrr,           X86::VPHADDWQrm,           0 },
  { X86::VPHMINPOSUWrr,        X86::VPHMINPOSUWrm,        0 },
  { X86::VPHSUBBWrr,           X86::VPHSUBBWrm,           0 },
  { X86::VPHSUBDQrr,           X86::VPHSUBDQrm,           0 },
  { X86::VPHSUBWDrr,           X86::VPHSUBWDrm,           0 },
  { X86::VPLZCNTDZ128rr,       X86::VPLZCNTDZ128rm,       0 },
  { X86::VPLZCNTDZ256rr,       X86::VPLZCNTDZ256rm,       0 },
  { X86::VPLZCNTDZrr,          X86::VPLZCNTDZrm,          0 },
  { X86::VPLZCNTQZ128rr,       X86::VPLZCNTQZ128rm,       0 },
  { X86::VPLZCNTQZ256rr,       X86::VPLZCNTQZ256rm,       0 },
  { X86::VPLZCNTQZrr,          X86::VPLZCNTQZrm,          0 },
  { X86::VPMOVSXBDYrr,         X86::VPMOVSXBDYrm,         TB_NO_REVERSE },
  { X86::VPMOVSXBDZ128rr,      X86::VPMOVSXBDZ128rm,      TB_NO_REVERSE },
  { X86::VPMOVSXBDZ256rr,      X86::VPMOVSXBDZ256rm,      TB_NO_REVERSE },
  { X86::VPMOVSXBDZrr,         X86::VPMOVSXBDZrm,         0 },
  { X86::VPMOVSXBDrr,          X86::VPMOVSXBDrm,          TB_NO_REVERSE },
  { X86::VPMOVSXBQYrr,         X86::VPMOVSXBQYrm,         TB_NO_REVERSE },
  { X86::VPMOVSXBQZ128rr,      X86::VPMOVSXBQZ128rm,      TB_NO_REVERSE },
  { X86::VPMOVSXBQZ256rr,      X86::VPMOVSXBQZ256rm,      TB_NO_REVERSE },
  { X86::VPMOVSXBQZrr,         X86::VPMOVSXBQZrm,         TB_NO_REVERSE },
  { X86::VPMOVSXBQrr,          X86::VPMOVSXBQrm,          TB_NO_REVERSE },
  { X86::VPMOVSXBWYrr,         X86::VPMOVSXBWYrm,         0 },
  { X86::VPMOVSXBWZ128rr,      X86::VPMOVSXBWZ128rm,      TB_NO_REVERSE },
  { X86::VPMOVSXBWZ256rr,      X86::VPMOVSXBWZ256rm,      0 },
  { X86::VPMOVSXBWZrr,         X86::VPMOVSXBWZrm,         0 },
  { X86::VPMOVSXBWrr,          X86::VPMOVSXBWrm,          TB_NO_REVERSE },
  { X86::VPMOVSXDQYrr,         X86::VPMOVSXDQYrm,         0 },
  { X86::VPMOVSXDQZ128rr,      X86::VPMOVSXDQZ128rm,      TB_NO_REVERSE },
  { X86::VPMOVSXDQZ256rr,      X86::VPMOVSXDQZ256rm,      0 },
  { X86::VPMOVSXDQZrr,         X86::VPMOVSXDQZrm,         0 },
  { X86::VPMOVSXDQrr,          X86::VPMOVSXDQrm,          TB_NO_REVERSE },
  { X86::VPMOVSXWDYrr,         X86::VPMOVSXWDYrm,         0 },
  { X86::VPMOVSXWDZ128rr,      X86::VPMOVSXWDZ128rm,      TB_NO_REVERSE },
  { X86::VPMOVSXWDZ256rr,      X86::VPMOVSXWDZ256rm,      0 },
  { X86::VPMOVSXWDZrr,         X86::VPMOVSXWDZrm,         0 },
  { X86::VPMOVSXWDrr,          X86::VPMOVSXWDrm,          TB_NO_REVERSE },
  { X86::VPMOVSXWQYrr,         X86::VPMOVSXWQYrm,         TB_NO_REVERSE },
  { X86::VPMOVSXWQZ128rr,      X86::VPMOVSXWQZ128rm,      TB_NO_REVERSE },
  { X86::VPMOVSXWQZ256rr,      X86::VPMOVSXWQZ256rm,      TB_NO_REVERSE },
  { X86::VPMOVSXWQZrr,         X86::VPMOVSXWQZrm,         0 },
  { X86::VPMOVSXWQrr,          X86::VPMOVSXWQrm,          TB_NO_REVERSE },
  { X86::VPMOVZXBDYrr,         X86::VPMOVZXBDYrm,         TB_NO_REVERSE },
  { X86::VPMOVZXBDZ128rr,      X86::VPMOVZXBDZ128rm,      TB_NO_REVERSE },
  { X86::VPMOVZXBDZ256rr,      X86::VPMOVZXBDZ256rm,      TB_NO_REVERSE },
  { X86::VPMOVZXBDZrr,         X86::VPMOVZXBDZrm,         0 },
  { X86::VPMOVZXBDrr,          X86::VPMOVZXBDrm,          TB_NO_REVERSE },
  { X86::VPMOVZXBQYrr,         X86::VPMOVZXBQYrm,         TB_NO_REVERSE },
  { X86::VPMOVZXBQZ128rr,      X86::VPMOVZXBQZ128rm,      TB_NO_REVERSE },
  { X86::VPMOVZXBQZ256rr,      X86::VPMOVZXBQZ256rm,      TB_NO_REVERSE },
  { X86::VPMOVZXBQZrr,         X86::VPMOVZXBQZrm,         TB_NO_REVERSE },
  { X86::VPMOVZXBQrr,          X86::VPMOVZXBQrm,          TB_NO_REVERSE },
  { X86::VPMOVZXBWYrr,         X86::VPMOVZXBWYrm,         0 },
  { X86::VPMOVZXBWZ128rr,      X86::VPMOVZXBWZ128rm,      TB_NO_REVERSE },
  { X86::VPMOVZXBWZ256rr,      X86::VPMOVZXBWZ256rm,      0 },
  { X86::VPMOVZXBWZrr,         X86::VPMOVZXBWZrm,         0 },
  { X86::VPMOVZXBWrr,          X86::VPMOVZXBWrm,          TB_NO_REVERSE },
  { X86::VPMOVZXDQYrr,         X86::VPMOVZXDQYrm,         0 },
  { X86::VPMOVZXDQZ128rr,      X86::VPMOVZXDQZ128rm,      TB_NO_REVERSE },
  { X86::VPMOVZXDQZ256rr,      X86::VPMOVZXDQZ256rm,      0 },
  { X86::VPMOVZXDQZrr,         X86::VPMOVZXDQZrm,         0 },
  { X86::VPMOVZXDQrr,          X86::VPMOVZXDQrm,          TB_NO_REVERSE },
  { X86::VPMOVZXWDYrr,         X86::VPMOVZXWDYrm,         0 },
  { X86::VPMOVZXWDZ128rr,      X86::VPMOVZXWDZ128rm,      TB_NO_REVERSE },
  { X86::VPMOVZXWDZ256rr,      X86::VPMOVZXWDZ256rm,      0 },
  { X86::VPMOVZXWDZrr,         X86::VPMOVZXWDZrm,         0 },
  { X86::VPMOVZXWDrr,          X86::VPMOVZXWDrm,          TB_NO_REVERSE },
  { X86::VPMOVZXWQYrr,         X86::VPMOVZXWQYrm,         TB_NO_REVERSE },
  { X86::VPMOVZXWQZ128rr,      X86::VPMOVZXWQZ128rm,      TB_NO_REVERSE },
  { X86::VPMOVZXWQZ256rr,      X86::VPMOVZXWQZ256rm,      TB_NO_REVERSE },
  { X86::VPMOVZXWQZrr,         X86::VPMOVZXWQZrm,         0 },
  { X86::VPMOVZXWQrr,          X86::VPMOVZXWQrm,          TB_NO_REVERSE },
  { X86::VPOPCNTBZ128rr,       X86::VPOPCNTBZ128rm,       0 },
  { X86::VPOPCNTBZ256rr,       X86::VPOPCNTBZ256rm,       0 },
  { X86::VPOPCNTBZrr,          X86::VPOPCNTBZrm,          0 },
  { X86::VPOPCNTDZ128rr,       X86::VPOPCNTDZ128rm,       0 },
  { X86::VPOPCNTDZ256rr,       X86::VPOPCNTDZ256rm,       0 },
  { X86::VPOPCNTDZrr,          X86::VPOPCNTDZrm,          0 },
  { X86::VPOPCNTQZ128rr,       X86::VPOPCNTQZ128rm,       0 },
  { X86::VPOPCNTQZ256rr,       X86::VPOPCNTQZ256rm,       0 },
  { X86::VPOPCNTQZrr,          X86::VPOPCNTQZrm,          0 },
  { X86::VPOPCNTWZ128rr,       X86::VPOPCNTWZ128rm,       0 },
  { X86::VPOPCNTWZ256rr,       X86::VPOPCNTWZ256rm,       0 },
  { X86::VPOPCNTWZrr,          X86::VPOPCNTWZrm,          0 },
  { X86::VPROLDZ128ri,         X86::VPROLDZ128mi,         0 },
  { X86::VPROLDZ256ri,         X86::VPROLDZ256mi,         0 },
  { X86::VPROLDZri,            X86::VPROLDZmi,            0 },
  { X86::VPROLQZ128ri,         X86::VPROLQZ128mi,         0 },
  { X86::VPROLQZ256ri,         X86::VPROLQZ256mi,         0 },
  { X86::VPROLQZri,            X86::VPROLQZmi,            0 },
  { X86::VPRORDZ128ri,         X86::VPRORDZ128mi,         0 },
  { X86::VPRORDZ256ri,         X86::VPRORDZ256mi,         0 },
  { X86::VPRORDZri,            X86::VPRORDZmi,            0 },
  { X86::VPRORQZ128ri,         X86::VPRORQZ128mi,         0 },
  { X86::VPRORQZ256ri,         X86::VPRORQZ256mi,         0 },
  { X86::VPRORQZri,            X86::VPRORQZmi,            0 },
  { X86::VPROTBri,             X86::VPROTBmi,             0 },
  { X86::VPROTBrr,             X86::VPROTBmr,             0 },
  { X86::VPROTDri,             X86::VPROTDmi,             0 },
  { X86::VPROTDrr,             X86::VPROTDmr,             0 },
  { X86::VPROTQri,             X86::VPROTQmi,             0 },
  { X86::VPROTQrr,             X86::VPROTQmr,             0 },
  { X86::VPROTWri,             X86::VPROTWmi,             0 },
  { X86::VPROTWrr,             X86::VPROTWmr,             0 },
  { X86::VPSHABrr,             X86::VPSHABmr,             0 },
  { X86::VPSHADrr,             X86::VPSHADmr,             0 },
  { X86::VPSHAQrr,             X86::VPSHAQmr,             0 },
  { X86::VPSHAWrr,             X86::VPSHAWmr,             0 },
  { X86::VPSHLBrr,             X86::VPSHLBmr,             0 },
  { X86::VPSHLDrr,             X86::VPSHLDmr,             0 },
  { X86::VPSHLQrr,             X86::VPSHLQmr,             0 },
  { X86::VPSHLWrr,             X86::VPSHLWmr,             0 },
  { X86::VPSHUFDYri,           X86::VPSHUFDYmi,           0 },
  { X86::VPSHUFDZ128ri,        X86::VPSHUFDZ128mi,        0 },
  { X86::VPSHUFDZ256ri,        X86::VPSHUFDZ256mi,        0 },
  { X86::VPSHUFDZri,           X86::VPSHUFDZmi,           0 },
  { X86::VPSHUFDri,            X86::VPSHUFDmi,            0 },
  { X86::VPSHUFHWYri,          X86::VPSHUFHWYmi,          0 },
  { X86::VPSHUFHWZ128ri,       X86::VPSHUFHWZ128mi,       0 },
  { X86::VPSHUFHWZ256ri,       X86::VPSHUFHWZ256mi,       0 },
  { X86::VPSHUFHWZri,          X86::VPSHUFHWZmi,          0 },
  { X86::VPSHUFHWri,           X86::VPSHUFHWmi,           0 },
  { X86::VPSHUFLWYri,          X86::VPSHUFLWYmi,          0 },
  { X86::VPSHUFLWZ128ri,       X86::VPSHUFLWZ128mi,       0 },
  { X86::VPSHUFLWZ256ri,       X86::VPSHUFLWZ256mi,       0 },
  { X86::VPSHUFLWZri,          X86::VPSHUFLWZmi,          0 },
  { X86::VPSHUFLWri,           X86::VPSHUFLWmi,           0 },
  { X86::VPSLLDQZ128rr,        X86::VPSLLDQZ128rm,        0 },
  { X86::VPSLLDQZ256rr,        X86::VPSLLDQZ256rm,        0 },
  { X86::VPSLLDQZrr,           X86::VPSLLDQZrm,           0 },
  { X86::VPSLLDZ128ri,         X86::VPSLLDZ128mi,         0 },
  { X86::VPSLLDZ256ri,         X86::VPSLLDZ256mi,         0 },
  { X86::VPSLLDZri,            X86::VPSLLDZmi,            0 },
  { X86::VPSLLQZ128ri,         X86::VPSLLQZ128mi,         0 },
  { X86::VPSLLQZ256ri,         X86::VPSLLQZ256mi,         0 },
  { X86::VPSLLQZri,            X86::VPSLLQZmi,            0 },
  { X86::VPSLLWZ128ri,         X86::VPSLLWZ128mi,         0 },
  { X86::VPSLLWZ256ri,         X86::VPSLLWZ256mi,         0 },
  { X86::VPSLLWZri,            X86::VPSLLWZmi,            0 },
  { X86::VPSRADZ128ri,         X86::VPSRADZ128mi,         0 },
  { X86::VPSRADZ256ri,         X86::VPSRADZ256mi,         0 },
  { X86::VPSRADZri,            X86::VPSRADZmi,            0 },
  { X86::VPSRAQZ128ri,         X86::VPSRAQZ128mi,         0 },
  { X86::VPSRAQZ256ri,         X86::VPSRAQZ256mi,         0 },
  { X86::VPSRAQZri,            X86::VPSRAQZmi,            0 },
  { X86::VPSRAWZ128ri,         X86::VPSRAWZ128mi,         0 },
  { X86::VPSRAWZ256ri,         X86::VPSRAWZ256mi,         0 },
  { X86::VPSRAWZri,            X86::VPSRAWZmi,            0 },
  { X86::VPSRLDQZ128rr,        X86::VPSRLDQZ128rm,        0 },
  { X86::VPSRLDQZ256rr,        X86::VPSRLDQZ256rm,        0 },
  { X86::VPSRLDQZrr,           X86::VPSRLDQZrm,           0 },
  { X86::VPSRLDZ128ri,         X86::VPSRLDZ128mi,         0 },
  { X86::VPSRLDZ256ri,         X86::VPSRLDZ256mi,         0 },
  { X86::VPSRLDZri,            X86::VPSRLDZmi,            0 },
  { X86::VPSRLQZ128ri,         X86::VPSRLQZ128mi,         0 },
  { X86::VPSRLQZ256ri,         X86::VPSRLQZ256mi,         0 },
  { X86::VPSRLQZri,            X86::VPSRLQZmi,            0 },
  { X86::VPSRLWZ128ri,         X86::VPSRLWZ128mi,         0 },
  { X86::VPSRLWZ256ri,         X86::VPSRLWZ256mi,         0 },
  { X86::VPSRLWZri,            X86::VPSRLWZmi,            0 },
  { X86::VPTESTYrr,            X86::VPTESTYrm,            0 },
  { X86::VPTESTrr,             X86::VPTESTrm,             0 },
  { X86::VRCP14PDZ128r,        X86::VRCP14PDZ128m,        0 },
  { X86::VRCP14PDZ256r,        X86::VRCP14PDZ256m,        0 },
  { X86::VRCP14PDZr,           X86::VRCP14PDZm,           0 },
  { X86::VRCP14PSZ128r,        X86::VRCP14PSZ128m,        0 },
  { X86::VRCP14PSZ256r,        X86::VRCP14PSZ256m,        0 },
  { X86::VRCP14PSZr,           X86::VRCP14PSZm,           0 },
  { X86::VRCP28PDZr,           X86::VRCP28PDZm,           0 },
  { X86::VRCP28PSZr,           X86::VRCP28PSZm,           0 },
  { X86::VRCPPSYr,             X86::VRCPPSYm,             0 },
  { X86::VRCPPSr,              X86::VRCPPSm,              0 },
  { X86::VREDUCEPDZ128rri,     X86::VREDUCEPDZ128rmi,     0 },
  { X86::VREDUCEPDZ256rri,     X86::VREDUCEPDZ256rmi,     0 },
  { X86::VREDUCEPDZrri,        X86::VREDUCEPDZrmi,        0 },
  { X86::VREDUCEPSZ128rri,     X86::VREDUCEPSZ128rmi,     0 },
  { X86::VREDUCEPSZ256rri,     X86::VREDUCEPSZ256rmi,     0 },
  { X86::VREDUCEPSZrri,        X86::VREDUCEPSZrmi,        0 },
  { X86::VRNDSCALEPDZ128rri,   X86::VRNDSCALEPDZ128rmi,   0 },
  { X86::VRNDSCALEPDZ256rri,   X86::VRNDSCALEPDZ256rmi,   0 },
  { X86::VRNDSCALEPDZrri,      X86::VRNDSCALEPDZrmi,      0 },
  { X86::VRNDSCALEPSZ128rri,   X86::VRNDSCALEPSZ128rmi,   0 },
  { X86::VRNDSCALEPSZ256rri,   X86::VRNDSCALEPSZ256rmi,   0 },
  { X86::VRNDSCALEPSZrri,      X86::VRNDSCALEPSZrmi,      0 },
  { X86::VROUNDPDYr,           X86::VROUNDPDYm,           0 },
  { X86::VROUNDPDr,            X86::VROUNDPDm,            0 },
  { X86::VROUNDPSYr,           X86::VROUNDPSYm,           0 },
  { X86::VROUNDPSr,            X86::VROUNDPSm,            0 },
  { X86::VRSQRT14PDZ128r,      X86::VRSQRT14PDZ128m,      0 },
  { X86::VRSQRT14PDZ256r,      X86::VRSQRT14PDZ256m,      0 },
  { X86::VRSQRT14PDZr,         X86::VRSQRT14PDZm,         0 },
  { X86::VRSQRT14PSZ128r,      X86::VRSQRT14PSZ128m,      0 },
  { X86::VRSQRT14PSZ256r,      X86::VRSQRT14PSZ256m,      0 },
  { X86::VRSQRT14PSZr,         X86::VRSQRT14PSZm,         0 },
  { X86::VRSQRT28PDZr,         X86::VRSQRT28PDZm,         0 },
  { X86::VRSQRT28PSZr,         X86::VRSQRT28PSZm,         0 },
  { X86::VRSQRTPSYr,           X86::VRSQRTPSYm,           0 },
  { X86::VRSQRTPSr,            X86::VRSQRTPSm,            0 },
  { X86::VSQRTPDYr,            X86::VSQRTPDYm,            0 },
  { X86::VSQRTPDZ128r,         X86::VSQRTPDZ128m,         0 },
  { X86::VSQRTPDZ256r,         X86::VSQRTPDZ256m,         0 },
  { X86::VSQRTPDZr,            X86::VSQRTPDZm,            0 },
  { X86::VSQRTPDr,             X86::VSQRTPDm,             0 },
  { X86::VSQRTPSYr,            X86::VSQRTPSYm,            0 },
  { X86::VSQRTPSZ128r,         X86::VSQRTPSZ128m,         0 },
  { X86::VSQRTPSZ256r,         X86::VSQRTPSZ256m,         0 },
  { X86::VSQRTPSZr,            X86::VSQRTPSZm,            0 },
  { X86::VSQRTPSr,             X86::VSQRTPSm,             0 },
  { X86::VTESTPDYrr,           X86::VTESTPDYrm,           0 },
  { X86::VTESTPDrr,            X86::VTESTPDrm,            0 },
  { X86::VTESTPSYrr,           X86::VTESTPSYrm,           0 },
  { X86::VTESTPSrr,            X86::VTESTPSrm,            0 },
  { X86::VUCOMISDZrr,          X86::VUCOMISDZrm,          0 },
  { X86::VUCOMISDZrr_Int,      X86::VUCOMISDZrm_Int,      TB_NO_REVERSE },
  { X86::VUCOMISDrr,           X86::VUCOMISDrm,           0 },
  { X86::VUCOMISDrr_Int,       X86::VUCOMISDrm_Int,       TB_NO_REVERSE },
  { X86::VUCOMISSZrr,          X86::VUCOMISSZrm,          0 },
  { X86::VUCOMISSZrr_Int,      X86::VUCOMISSZrm_Int,      TB_NO_REVERSE },
  { X86::VUCOMISSrr,           X86::VUCOMISSrm,           0 },
  { X86::VUCOMISSrr_Int,       X86::VUCOMISSrm_Int,       TB_NO_REVERSE },
};

static const X86MemoryFoldTableEntry MemoryFoldTable2[] = {
  { X86::ADD16rr_DB,               X86::ADD16rm,                  TB_NO_REVERSE },
  { X86::ADD32rr_DB,               X86::ADD32rm,                  TB_NO_REVERSE },
  { X86::ADD64rr_DB,               X86::ADD64rm,                  TB_NO_REVERSE },
  { X86::ADD8rr_DB,                X86::ADD8rm,                   TB_NO_REVERSE },
  { X86::ADC16rr,                  X86::ADC16rm,                  0 },
  { X86::ADC32rr,                  X86::ADC32rm,                  0 },
  { X86::ADC64rr,                  X86::ADC64rm,                  0 },
  { X86::ADC8rr,                   X86::ADC8rm,                   0 },
  { X86::ADCX32rr,                 X86::ADCX32rm,                 0 },
  { X86::ADCX64rr,                 X86::ADCX64rm,                 0 },
  { X86::ADD16rr,                  X86::ADD16rm,                  0 },
  { X86::ADD32rr,                  X86::ADD32rm,                  0 },
  { X86::ADD64rr,                  X86::ADD64rm,                  0 },
  { X86::ADD8rr,                   X86::ADD8rm,                   0 },
  { X86::ADDPDrr,                  X86::ADDPDrm,                  TB_ALIGN_16 },
  { X86::ADDPSrr,                  X86::ADDPSrm,                  TB_ALIGN_16 },
  { X86::ADDSDrr,                  X86::ADDSDrm,                  0 },
  { X86::ADDSDrr_Int,              X86::ADDSDrm_Int,              TB_NO_REVERSE },
  { X86::ADDSSrr,                  X86::ADDSSrm,                  0 },
  { X86::ADDSSrr_Int,              X86::ADDSSrm_Int,              TB_NO_REVERSE },
  { X86::ADDSUBPDrr,               X86::ADDSUBPDrm,               TB_ALIGN_16 },
  { X86::ADDSUBPSrr,               X86::ADDSUBPSrm,               TB_ALIGN_16 },
  { X86::ADOX32rr,                 X86::ADOX32rm,                 0 },
  { X86::ADOX64rr,                 X86::ADOX64rm,                 0 },
  { X86::AESDECLASTrr,             X86::AESDECLASTrm,             TB_ALIGN_16 },
  { X86::AESDECrr,                 X86::AESDECrm,                 TB_ALIGN_16 },
  { X86::AESENCLASTrr,             X86::AESENCLASTrm,             TB_ALIGN_16 },
  { X86::AESENCrr,                 X86::AESENCrm,                 TB_ALIGN_16 },
  { X86::AND16rr,                  X86::AND16rm,                  0 },
  { X86::AND32rr,                  X86::AND32rm,                  0 },
  { X86::AND64rr,                  X86::AND64rm,                  0 },
  { X86::AND8rr,                   X86::AND8rm,                   0 },
  { X86::ANDN32rr,                 X86::ANDN32rm,                 0 },
  { X86::ANDN64rr,                 X86::ANDN64rm,                 0 },
  { X86::ANDNPDrr,                 X86::ANDNPDrm,                 TB_ALIGN_16 },
  { X86::ANDNPSrr,                 X86::ANDNPSrm,                 TB_ALIGN_16 },
  { X86::ANDPDrr,                  X86::ANDPDrm,                  TB_ALIGN_16 },
  { X86::ANDPSrr,                  X86::ANDPSrm,                  TB_ALIGN_16 },
  { X86::BLENDPDrri,               X86::BLENDPDrmi,               TB_ALIGN_16 },
  { X86::BLENDPSrri,               X86::BLENDPSrmi,               TB_ALIGN_16 },
  { X86::BLENDVPDrr0,              X86::BLENDVPDrm0,              TB_ALIGN_16 },
  { X86::BLENDVPSrr0,              X86::BLENDVPSrm0,              TB_ALIGN_16 },
  { X86::CMOV16rr,                 X86::CMOV16rm,                 0 },
  { X86::CMOV32rr,                 X86::CMOV32rm,                 0 },
  { X86::CMOV64rr,                 X86::CMOV64rm,                 0 },
  { X86::CMPPDrri,                 X86::CMPPDrmi,                 TB_ALIGN_16 },
  { X86::CMPPSrri,                 X86::CMPPSrmi,                 TB_ALIGN_16 },
  { X86::CMPSDrr,                  X86::CMPSDrm,                  0 },
  { X86::CMPSDrr_Int,              X86::CMPSDrm_Int,              TB_NO_REVERSE },
  { X86::CMPSSrr,                  X86::CMPSSrm,                  0 },
  { X86::CMPSSrr_Int,              X86::CMPSSrm_Int,              TB_NO_REVERSE },
  { X86::CRC32r32r16,              X86::CRC32r32m16,              0 },
  { X86::CRC32r32r32,              X86::CRC32r32m32,              0 },
  { X86::CRC32r32r8,               X86::CRC32r32m8,               0 },
  { X86::CRC32r64r64,              X86::CRC32r64m64,              0 },
  { X86::CRC32r64r8,               X86::CRC32r64m8,               0 },
  { X86::CVTSD2SSrr_Int,           X86::CVTSD2SSrm_Int,           TB_NO_REVERSE },
  { X86::CVTSI2SDrr_Int,           X86::CVTSI2SDrm_Int,           0 },
  { X86::CVTSI2SSrr_Int,           X86::CVTSI2SSrm_Int,           0 },
  { X86::CVTSI642SDrr_Int,         X86::CVTSI642SDrm_Int,         0 },
  { X86::CVTSI642SSrr_Int,         X86::CVTSI642SSrm_Int,         0 },
  { X86::CVTSS2SDrr_Int,           X86::CVTSS2SDrm_Int,           TB_NO_REVERSE },
  { X86::DIVPDrr,                  X86::DIVPDrm,                  TB_ALIGN_16 },
  { X86::DIVPSrr,                  X86::DIVPSrm,                  TB_ALIGN_16 },
  { X86::DIVSDrr,                  X86::DIVSDrm,                  0 },
  { X86::DIVSDrr_Int,              X86::DIVSDrm_Int,              TB_NO_REVERSE },
  { X86::DIVSSrr,                  X86::DIVSSrm,                  0 },
  { X86::DIVSSrr_Int,              X86::DIVSSrm_Int,              TB_NO_REVERSE },
  { X86::DPPDrri,                  X86::DPPDrmi,                  TB_ALIGN_16 },
  { X86::DPPSrri,                  X86::DPPSrmi,                  TB_ALIGN_16 },
  { X86::GF2P8AFFINEINVQBrri,      X86::GF2P8AFFINEINVQBrmi,      TB_ALIGN_16 },
  { X86::GF2P8AFFINEQBrri,         X86::GF2P8AFFINEQBrmi,         TB_ALIGN_16 },
  { X86::GF2P8MULBrr,              X86::GF2P8MULBrm,              TB_ALIGN_16 },
  { X86::HADDPDrr,                 X86::HADDPDrm,                 TB_ALIGN_16 },
  { X86::HADDPSrr,                 X86::HADDPSrm,                 TB_ALIGN_16 },
  { X86::HSUBPDrr,                 X86::HSUBPDrm,                 TB_ALIGN_16 },
  { X86::HSUBPSrr,                 X86::HSUBPSrm,                 TB_ALIGN_16 },
  { X86::IMUL16rr,                 X86::IMUL16rm,                 0 },
  { X86::IMUL32rr,                 X86::IMUL32rm,                 0 },
  { X86::IMUL64rr,                 X86::IMUL64rm,                 0 },
  { X86::MAXCPDrr,                 X86::MAXCPDrm,                 TB_ALIGN_16 },
  { X86::MAXCPSrr,                 X86::MAXCPSrm,                 TB_ALIGN_16 },
  { X86::MAXCSDrr,                 X86::MAXCSDrm,                 0 },
  { X86::MAXCSSrr,                 X86::MAXCSSrm,                 0 },
  { X86::MAXPDrr,                  X86::MAXPDrm,                  TB_ALIGN_16 },
  { X86::MAXPSrr,                  X86::MAXPSrm,                  TB_ALIGN_16 },
  { X86::MAXSDrr,                  X86::MAXSDrm,                  0 },
  { X86::MAXSDrr_Int,              X86::MAXSDrm_Int,              TB_NO_REVERSE },
  { X86::MAXSSrr,                  X86::MAXSSrm,                  0 },
  { X86::MAXSSrr_Int,              X86::MAXSSrm_Int,              TB_NO_REVERSE },
  { X86::MINCPDrr,                 X86::MINCPDrm,                 TB_ALIGN_16 },
  { X86::MINCPSrr,                 X86::MINCPSrm,                 TB_ALIGN_16 },
  { X86::MINCSDrr,                 X86::MINCSDrm,                 0 },
  { X86::MINCSSrr,                 X86::MINCSSrm,                 0 },
  { X86::MINPDrr,                  X86::MINPDrm,                  TB_ALIGN_16 },
  { X86::MINPSrr,                  X86::MINPSrm,                  TB_ALIGN_16 },
  { X86::MINSDrr,                  X86::MINSDrm,                  0 },
  { X86::MINSDrr_Int,              X86::MINSDrm_Int,              TB_NO_REVERSE },
  { X86::MINSSrr,                  X86::MINSSrm,                  0 },
  { X86::MINSSrr_Int,              X86::MINSSrm_Int,              TB_NO_REVERSE },
  { X86::MMX_CVTPI2PSirr,          X86::MMX_CVTPI2PSirm,          0 },
  { X86::MMX_PACKSSDWirr,          X86::MMX_PACKSSDWirm,          0 },
  { X86::MMX_PACKSSWBirr,          X86::MMX_PACKSSWBirm,          0 },
  { X86::MMX_PACKUSWBirr,          X86::MMX_PACKUSWBirm,          0 },
  { X86::MMX_PADDBirr,             X86::MMX_PADDBirm,             0 },
  { X86::MMX_PADDDirr,             X86::MMX_PADDDirm,             0 },
  { X86::MMX_PADDQirr,             X86::MMX_PADDQirm,             0 },
  { X86::MMX_PADDSBirr,            X86::MMX_PADDSBirm,            0 },
  { X86::MMX_PADDSWirr,            X86::MMX_PADDSWirm,            0 },
  { X86::MMX_PADDUSBirr,           X86::MMX_PADDUSBirm,           0 },
  { X86::MMX_PADDUSWirr,           X86::MMX_PADDUSWirm,           0 },
  { X86::MMX_PADDWirr,             X86::MMX_PADDWirm,             0 },
  { X86::MMX_PALIGNRrri,           X86::MMX_PALIGNRrmi,           0 },
  { X86::MMX_PANDNirr,             X86::MMX_PANDNirm,             0 },
  { X86::MMX_PANDirr,              X86::MMX_PANDirm,              0 },
  { X86::MMX_PAVGBirr,             X86::MMX_PAVGBirm,             0 },
  { X86::MMX_PAVGWirr,             X86::MMX_PAVGWirm,             0 },
  { X86::MMX_PCMPEQBirr,           X86::MMX_PCMPEQBirm,           0 },
  { X86::MMX_PCMPEQDirr,           X86::MMX_PCMPEQDirm,           0 },
  { X86::MMX_PCMPEQWirr,           X86::MMX_PCMPEQWirm,           0 },
  { X86::MMX_PCMPGTBirr,           X86::MMX_PCMPGTBirm,           0 },
  { X86::MMX_PCMPGTDirr,           X86::MMX_PCMPGTDirm,           0 },
  { X86::MMX_PCMPGTWirr,           X86::MMX_PCMPGTWirm,           0 },
  { X86::MMX_PHADDDrr,             X86::MMX_PHADDDrm,             0 },
  { X86::MMX_PHADDSWrr,            X86::MMX_PHADDSWrm,            0 },
  { X86::MMX_PHADDWrr,             X86::MMX_PHADDWrm,             0 },
  { X86::MMX_PHSUBDrr,             X86::MMX_PHSUBDrm,             0 },
  { X86::MMX_PHSUBSWrr,            X86::MMX_PHSUBSWrm,            0 },
  { X86::MMX_PHSUBWrr,             X86::MMX_PHSUBWrm,             0 },
  { X86::MMX_PINSRWrr,             X86::MMX_PINSRWrm,             TB_NO_REVERSE },
  { X86::MMX_PMADDUBSWrr,          X86::MMX_PMADDUBSWrm,          0 },
  { X86::MMX_PMADDWDirr,           X86::MMX_PMADDWDirm,           0 },
  { X86::MMX_PMAXSWirr,            X86::MMX_PMAXSWirm,            0 },
  { X86::MMX_PMAXUBirr,            X86::MMX_PMAXUBirm,            0 },
  { X86::MMX_PMINSWirr,            X86::MMX_PMINSWirm,            0 },
  { X86::MMX_PMINUBirr,            X86::MMX_PMINUBirm,            0 },
  { X86::MMX_PMULHRSWrr,           X86::MMX_PMULHRSWrm,           0 },
  { X86::MMX_PMULHUWirr,           X86::MMX_PMULHUWirm,           0 },
  { X86::MMX_PMULHWirr,            X86::MMX_PMULHWirm,            0 },
  { X86::MMX_PMULLWirr,            X86::MMX_PMULLWirm,            0 },
  { X86::MMX_PMULUDQirr,           X86::MMX_PMULUDQirm,           0 },
  { X86::MMX_PORirr,               X86::MMX_PORirm,               0 },
  { X86::MMX_PSADBWirr,            X86::MMX_PSADBWirm,            0 },
  { X86::MMX_PSHUFBrr,             X86::MMX_PSHUFBrm,             0 },
  { X86::MMX_PSIGNBrr,             X86::MMX_PSIGNBrm,             0 },
  { X86::MMX_PSIGNDrr,             X86::MMX_PSIGNDrm,             0 },
  { X86::MMX_PSIGNWrr,             X86::MMX_PSIGNWrm,             0 },
  { X86::MMX_PSLLDrr,              X86::MMX_PSLLDrm,              0 },
  { X86::MMX_PSLLQrr,              X86::MMX_PSLLQrm,              0 },
  { X86::MMX_PSLLWrr,              X86::MMX_PSLLWrm,              0 },
  { X86::MMX_PSRADrr,              X86::MMX_PSRADrm,              0 },
  { X86::MMX_PSRAWrr,              X86::MMX_PSRAWrm,              0 },
  { X86::MMX_PSRLDrr,              X86::MMX_PSRLDrm,              0 },
  { X86::MMX_PSRLQrr,              X86::MMX_PSRLQrm,              0 },
  { X86::MMX_PSRLWrr,              X86::MMX_PSRLWrm,              0 },
  { X86::MMX_PSUBBirr,             X86::MMX_PSUBBirm,             0 },
  { X86::MMX_PSUBDirr,             X86::MMX_PSUBDirm,             0 },
  { X86::MMX_PSUBQirr,             X86::MMX_PSUBQirm,             0 },
  { X86::MMX_PSUBSBirr,            X86::MMX_PSUBSBirm,            0 },
  { X86::MMX_PSUBSWirr,            X86::MMX_PSUBSWirm,            0 },
  { X86::MMX_PSUBUSBirr,           X86::MMX_PSUBUSBirm,           0 },
  { X86::MMX_PSUBUSWirr,           X86::MMX_PSUBUSWirm,           0 },
  { X86::MMX_PSUBWirr,             X86::MMX_PSUBWirm,             0 },
  { X86::MMX_PUNPCKHBWirr,         X86::MMX_PUNPCKHBWirm,         0 },
  { X86::MMX_PUNPCKHDQirr,         X86::MMX_PUNPCKHDQirm,         0 },
  { X86::MMX_PUNPCKHWDirr,         X86::MMX_PUNPCKHWDirm,         0 },
  { X86::MMX_PUNPCKLBWirr,         X86::MMX_PUNPCKLBWirm,         TB_NO_REVERSE },
  { X86::MMX_PUNPCKLDQirr,         X86::MMX_PUNPCKLDQirm,         TB_NO_REVERSE },
  { X86::MMX_PUNPCKLWDirr,         X86::MMX_PUNPCKLWDirm,         TB_NO_REVERSE },
  { X86::MMX_PXORirr,              X86::MMX_PXORirm,              0 },
  { X86::MOVLHPSrr,                X86::MOVHPSrm,                 TB_NO_REVERSE },
  { X86::MOVSDrr,                  X86::MOVLPDrm,                 TB_NO_REVERSE },
  { X86::MPSADBWrri,               X86::MPSADBWrmi,               TB_ALIGN_16 },
  { X86::MULPDrr,                  X86::MULPDrm,                  TB_ALIGN_16 },
  { X86::MULPSrr,                  X86::MULPSrm,                  TB_ALIGN_16 },
  { X86::MULSDrr,                  X86::MULSDrm,                  0 },
  { X86::MULSDrr_Int,              X86::MULSDrm_Int,              TB_NO_REVERSE },
  { X86::MULSSrr,                  X86::MULSSrm,                  0 },
  { X86::MULSSrr_Int,              X86::MULSSrm_Int,              TB_NO_REVERSE },
  { X86::MULX32rr,                 X86::MULX32rm,                 0 },
  { X86::MULX64rr,                 X86::MULX64rm,                 0 },
  { X86::OR16rr,                   X86::OR16rm,                   0 },
  { X86::OR32rr,                   X86::OR32rm,                   0 },
  { X86::OR64rr,                   X86::OR64rm,                   0 },
  { X86::OR8rr,                    X86::OR8rm,                    0 },
  { X86::ORPDrr,                   X86::ORPDrm,                   TB_ALIGN_16 },
  { X86::ORPSrr,                   X86::ORPSrm,                   TB_ALIGN_16 },
  { X86::PACKSSDWrr,               X86::PACKSSDWrm,               TB_ALIGN_16 },
  { X86::PACKSSWBrr,               X86::PACKSSWBrm,               TB_ALIGN_16 },
  { X86::PACKUSDWrr,               X86::PACKUSDWrm,               TB_ALIGN_16 },
  { X86::PACKUSWBrr,               X86::PACKUSWBrm,               TB_ALIGN_16 },
  { X86::PADDBrr,                  X86::PADDBrm,                  TB_ALIGN_16 },
  { X86::PADDDrr,                  X86::PADDDrm,                  TB_ALIGN_16 },
  { X86::PADDQrr,                  X86::PADDQrm,                  TB_ALIGN_16 },
  { X86::PADDSBrr,                 X86::PADDSBrm,                 TB_ALIGN_16 },
  { X86::PADDSWrr,                 X86::PADDSWrm,                 TB_ALIGN_16 },
  { X86::PADDUSBrr,                X86::PADDUSBrm,                TB_ALIGN_16 },
  { X86::PADDUSWrr,                X86::PADDUSWrm,                TB_ALIGN_16 },
  { X86::PADDWrr,                  X86::PADDWrm,                  TB_ALIGN_16 },
  { X86::PALIGNRrri,               X86::PALIGNRrmi,               TB_ALIGN_16 },
  { X86::PANDNrr,                  X86::PANDNrm,                  TB_ALIGN_16 },
  { X86::PANDrr,                   X86::PANDrm,                   TB_ALIGN_16 },
  { X86::PAVGBrr,                  X86::PAVGBrm,                  TB_ALIGN_16 },
  { X86::PAVGUSBrr,                X86::PAVGUSBrm,                0 },
  { X86::PAVGWrr,                  X86::PAVGWrm,                  TB_ALIGN_16 },
  { X86::PBLENDVBrr0,              X86::PBLENDVBrm0,              TB_ALIGN_16 },
  { X86::PBLENDWrri,               X86::PBLENDWrmi,               TB_ALIGN_16 },
  { X86::PCLMULQDQrr,              X86::PCLMULQDQrm,              TB_ALIGN_16 },
  { X86::PCMPEQBrr,                X86::PCMPEQBrm,                TB_ALIGN_16 },
  { X86::PCMPEQDrr,                X86::PCMPEQDrm,                TB_ALIGN_16 },
  { X86::PCMPEQQrr,                X86::PCMPEQQrm,                TB_ALIGN_16 },
  { X86::PCMPEQWrr,                X86::PCMPEQWrm,                TB_ALIGN_16 },
  { X86::PCMPGTBrr,                X86::PCMPGTBrm,                TB_ALIGN_16 },
  { X86::PCMPGTDrr,                X86::PCMPGTDrm,                TB_ALIGN_16 },
  { X86::PCMPGTQrr,                X86::PCMPGTQrm,                TB_ALIGN_16 },
  { X86::PCMPGTWrr,                X86::PCMPGTWrm,                TB_ALIGN_16 },
  { X86::PDEP32rr,                 X86::PDEP32rm,                 0 },
  { X86::PDEP64rr,                 X86::PDEP64rm,                 0 },
  { X86::PEXT32rr,                 X86::PEXT32rm,                 0 },
  { X86::PEXT64rr,                 X86::PEXT64rm,                 0 },
  { X86::PFACCrr,                  X86::PFACCrm,                  0 },
  { X86::PFADDrr,                  X86::PFADDrm,                  0 },
  { X86::PFCMPEQrr,                X86::PFCMPEQrm,                0 },
  { X86::PFCMPGErr,                X86::PFCMPGErm,                0 },
  { X86::PFCMPGTrr,                X86::PFCMPGTrm,                0 },
  { X86::PFMAXrr,                  X86::PFMAXrm,                  0 },
  { X86::PFMINrr,                  X86::PFMINrm,                  0 },
  { X86::PFMULrr,                  X86::PFMULrm,                  0 },
  { X86::PFNACCrr,                 X86::PFNACCrm,                 0 },
  { X86::PFPNACCrr,                X86::PFPNACCrm,                0 },
  { X86::PFRCPIT1rr,               X86::PFRCPIT1rm,               0 },
  { X86::PFRCPIT2rr,               X86::PFRCPIT2rm,               0 },
  { X86::PFRSQIT1rr,               X86::PFRSQIT1rm,               0 },
  { X86::PFSUBRrr,                 X86::PFSUBRrm,                 0 },
  { X86::PFSUBrr,                  X86::PFSUBrm,                  0 },
  { X86::PHADDDrr,                 X86::PHADDDrm,                 TB_ALIGN_16 },
  { X86::PHADDSWrr,                X86::PHADDSWrm,                TB_ALIGN_16 },
  { X86::PHADDWrr,                 X86::PHADDWrm,                 TB_ALIGN_16 },
  { X86::PHSUBDrr,                 X86::PHSUBDrm,                 TB_ALIGN_16 },
  { X86::PHSUBSWrr,                X86::PHSUBSWrm,                TB_ALIGN_16 },
  { X86::PHSUBWrr,                 X86::PHSUBWrm,                 TB_ALIGN_16 },
  { X86::PINSRBrr,                 X86::PINSRBrm,                 TB_NO_REVERSE },
  { X86::PINSRDrr,                 X86::PINSRDrm,                 0 },
  { X86::PINSRQrr,                 X86::PINSRQrm,                 0 },
  { X86::PINSRWrr,                 X86::PINSRWrm,                 TB_NO_REVERSE },
  { X86::PMADDUBSWrr,              X86::PMADDUBSWrm,              TB_ALIGN_16 },
  { X86::PMADDWDrr,                X86::PMADDWDrm,                TB_ALIGN_16 },
  { X86::PMAXSBrr,                 X86::PMAXSBrm,                 TB_ALIGN_16 },
  { X86::PMAXSDrr,                 X86::PMAXSDrm,                 TB_ALIGN_16 },
  { X86::PMAXSWrr,                 X86::PMAXSWrm,                 TB_ALIGN_16 },
  { X86::PMAXUBrr,                 X86::PMAXUBrm,                 TB_ALIGN_16 },
  { X86::PMAXUDrr,                 X86::PMAXUDrm,                 TB_ALIGN_16 },
  { X86::PMAXUWrr,                 X86::PMAXUWrm,                 TB_ALIGN_16 },
  { X86::PMINSBrr,                 X86::PMINSBrm,                 TB_ALIGN_16 },
  { X86::PMINSDrr,                 X86::PMINSDrm,                 TB_ALIGN_16 },
  { X86::PMINSWrr,                 X86::PMINSWrm,                 TB_ALIGN_16 },
  { X86::PMINUBrr,                 X86::PMINUBrm,                 TB_ALIGN_16 },
  { X86::PMINUDrr,                 X86::PMINUDrm,                 TB_ALIGN_16 },
  { X86::PMINUWrr,                 X86::PMINUWrm,                 TB_ALIGN_16 },
  { X86::PMULDQrr,                 X86::PMULDQrm,                 TB_ALIGN_16 },
  { X86::PMULHRSWrr,               X86::PMULHRSWrm,               TB_ALIGN_16 },
  { X86::PMULHRWrr,                X86::PMULHRWrm,                0 },
  { X86::PMULHUWrr,                X86::PMULHUWrm,                TB_ALIGN_16 },
  { X86::PMULHWrr,                 X86::PMULHWrm,                 TB_ALIGN_16 },
  { X86::PMULLDrr,                 X86::PMULLDrm,                 TB_ALIGN_16 },
  { X86::PMULLWrr,                 X86::PMULLWrm,                 TB_ALIGN_16 },
  { X86::PMULUDQrr,                X86::PMULUDQrm,                TB_ALIGN_16 },
  { X86::PORrr,                    X86::PORrm,                    TB_ALIGN_16 },
  { X86::PSADBWrr,                 X86::PSADBWrm,                 TB_ALIGN_16 },
  { X86::PSHUFBrr,                 X86::PSHUFBrm,                 TB_ALIGN_16 },
  { X86::PSIGNBrr,                 X86::PSIGNBrm,                 TB_ALIGN_16 },
  { X86::PSIGNDrr,                 X86::PSIGNDrm,                 TB_ALIGN_16 },
  { X86::PSIGNWrr,                 X86::PSIGNWrm,                 TB_ALIGN_16 },
  { X86::PSLLDrr,                  X86::PSLLDrm,                  TB_ALIGN_16 },
  { X86::PSLLQrr,                  X86::PSLLQrm,                  TB_ALIGN_16 },
  { X86::PSLLWrr,                  X86::PSLLWrm,                  TB_ALIGN_16 },
  { X86::PSRADrr,                  X86::PSRADrm,                  TB_ALIGN_16 },
  { X86::PSRAWrr,                  X86::PSRAWrm,                  TB_ALIGN_16 },
  { X86::PSRLDrr,                  X86::PSRLDrm,                  TB_ALIGN_16 },
  { X86::PSRLQrr,                  X86::PSRLQrm,                  TB_ALIGN_16 },
  { X86::PSRLWrr,                  X86::PSRLWrm,                  TB_ALIGN_16 },
  { X86::PSUBBrr,                  X86::PSUBBrm,                  TB_ALIGN_16 },
  { X86::PSUBDrr,                  X86::PSUBDrm,                  TB_ALIGN_16 },
  { X86::PSUBQrr,                  X86::PSUBQrm,                  TB_ALIGN_16 },
  { X86::PSUBSBrr,                 X86::PSUBSBrm,                 TB_ALIGN_16 },
  { X86::PSUBSWrr,                 X86::PSUBSWrm,                 TB_ALIGN_16 },
  { X86::PSUBUSBrr,                X86::PSUBUSBrm,                TB_ALIGN_16 },
  { X86::PSUBUSWrr,                X86::PSUBUSWrm,                TB_ALIGN_16 },
  { X86::PSUBWrr,                  X86::PSUBWrm,                  TB_ALIGN_16 },
  { X86::PUNPCKHBWrr,              X86::PUNPCKHBWrm,              TB_ALIGN_16 },
  { X86::PUNPCKHDQrr,              X86::PUNPCKHDQrm,              TB_ALIGN_16 },
  { X86::PUNPCKHQDQrr,             X86::PUNPCKHQDQrm,             TB_ALIGN_16 },
  { X86::PUNPCKHWDrr,              X86::PUNPCKHWDrm,              TB_ALIGN_16 },
  { X86::PUNPCKLBWrr,              X86::PUNPCKLBWrm,              TB_ALIGN_16 },
  { X86::PUNPCKLDQrr,              X86::PUNPCKLDQrm,              TB_ALIGN_16 },
  { X86::PUNPCKLQDQrr,             X86::PUNPCKLQDQrm,             TB_ALIGN_16 },
  { X86::PUNPCKLWDrr,              X86::PUNPCKLWDrm,              TB_ALIGN_16 },
  { X86::PXORrr,                   X86::PXORrm,                   TB_ALIGN_16 },
  { X86::RCPSSr_Int,               X86::RCPSSm_Int,               TB_NO_REVERSE },
  { X86::ROUNDSDr_Int,             X86::ROUNDSDm_Int,             TB_NO_REVERSE },
  { X86::ROUNDSSr_Int,             X86::ROUNDSSm_Int,             TB_NO_REVERSE },
  { X86::RSQRTSSr_Int,             X86::RSQRTSSm_Int,             TB_NO_REVERSE },
  { X86::SBB16rr,                  X86::SBB16rm,                  0 },
  { X86::SBB32rr,                  X86::SBB32rm,                  0 },
  { X86::SBB64rr,                  X86::SBB64rm,                  0 },
  { X86::SBB8rr,                   X86::SBB8rm,                   0 },
  { X86::SHA1MSG1rr,               X86::SHA1MSG1rm,               TB_ALIGN_16 },
  { X86::SHA1MSG2rr,               X86::SHA1MSG2rm,               TB_ALIGN_16 },
  { X86::SHA1NEXTErr,              X86::SHA1NEXTErm,              TB_ALIGN_16 },
  { X86::SHA1RNDS4rri,             X86::SHA1RNDS4rmi,             TB_ALIGN_16 },
  { X86::SHA256MSG1rr,             X86::SHA256MSG1rm,             TB_ALIGN_16 },
  { X86::SHA256MSG2rr,             X86::SHA256MSG2rm,             TB_ALIGN_16 },
  { X86::SHA256RNDS2rr,            X86::SHA256RNDS2rm,            TB_ALIGN_16 },
  { X86::SHUFPDrri,                X86::SHUFPDrmi,                TB_ALIGN_16 },
  { X86::SHUFPSrri,                X86::SHUFPSrmi,                TB_ALIGN_16 },
  { X86::SQRTSDr_Int,              X86::SQRTSDm_Int,              TB_NO_REVERSE },
  { X86::SQRTSSr_Int,              X86::SQRTSSm_Int,              TB_NO_REVERSE },
  { X86::SUB16rr,                  X86::SUB16rm,                  0 },
  { X86::SUB32rr,                  X86::SUB32rm,                  0 },
  { X86::SUB64rr,                  X86::SUB64rm,                  0 },
  { X86::SUB8rr,                   X86::SUB8rm,                   0 },
  { X86::SUBPDrr,                  X86::SUBPDrm,                  TB_ALIGN_16 },
  { X86::SUBPSrr,                  X86::SUBPSrm,                  TB_ALIGN_16 },
  { X86::SUBSDrr,                  X86::SUBSDrm,                  0 },
  { X86::SUBSDrr_Int,              X86::SUBSDrm_Int,              TB_NO_REVERSE },
  { X86::SUBSSrr,                  X86::SUBSSrm,                  0 },
  { X86::SUBSSrr_Int,              X86::SUBSSrm_Int,              TB_NO_REVERSE },
  { X86::UNPCKHPDrr,               X86::UNPCKHPDrm,               TB_ALIGN_16 },
  { X86::UNPCKHPSrr,               X86::UNPCKHPSrm,               TB_ALIGN_16 },
  { X86::UNPCKLPDrr,               X86::UNPCKLPDrm,               TB_ALIGN_16 },
  { X86::UNPCKLPSrr,               X86::UNPCKLPSrm,               TB_ALIGN_16 },
  { X86::VADDPDYrr,                X86::VADDPDYrm,                0 },
  { X86::VADDPDZ128rr,             X86::VADDPDZ128rm,             0 },
  { X86::VADDPDZ256rr,             X86::VADDPDZ256rm,             0 },
  { X86::VADDPDZrr,                X86::VADDPDZrm,                0 },
  { X86::VADDPDrr,                 X86::VADDPDrm,                 0 },
  { X86::VADDPSYrr,                X86::VADDPSYrm,                0 },
  { X86::VADDPSZ128rr,             X86::VADDPSZ128rm,             0 },
  { X86::VADDPSZ256rr,             X86::VADDPSZ256rm,             0 },
  { X86::VADDPSZrr,                X86::VADDPSZrm,                0 },
  { X86::VADDPSrr,                 X86::VADDPSrm,                 0 },
  { X86::VADDSDZrr,                X86::VADDSDZrm,                0 },
  { X86::VADDSDZrr_Int,            X86::VADDSDZrm_Int,            TB_NO_REVERSE },
  { X86::VADDSDrr,                 X86::VADDSDrm,                 0 },
  { X86::VADDSDrr_Int,             X86::VADDSDrm_Int,             TB_NO_REVERSE },
  { X86::VADDSSZrr,                X86::VADDSSZrm,                0 },
  { X86::VADDSSZrr_Int,            X86::VADDSSZrm_Int,            TB_NO_REVERSE },
  { X86::VADDSSrr,                 X86::VADDSSrm,                 0 },
  { X86::VADDSSrr_Int,             X86::VADDSSrm_Int,             TB_NO_REVERSE },
  { X86::VADDSUBPDYrr,             X86::VADDSUBPDYrm,             0 },
  { X86::VADDSUBPDrr,              X86::VADDSUBPDrm,              0 },
  { X86::VADDSUBPSYrr,             X86::VADDSUBPSYrm,             0 },
  { X86::VADDSUBPSrr,              X86::VADDSUBPSrm,              0 },
  { X86::VAESDECLASTYrr,           X86::VAESDECLASTYrm,           0 },
  { X86::VAESDECLASTZ128rr,        X86::VAESDECLASTZ128rm,        0 },
  { X86::VAESDECLASTZ256rr,        X86::VAESDECLASTZ256rm,        0 },
  { X86::VAESDECLASTZrr,           X86::VAESDECLASTZrm,           0 },
  { X86::VAESDECLASTrr,            X86::VAESDECLASTrm,            0 },
  { X86::VAESDECYrr,               X86::VAESDECYrm,               0 },
  { X86::VAESDECZ128rr,            X86::VAESDECZ128rm,            0 },
  { X86::VAESDECZ256rr,            X86::VAESDECZ256rm,            0 },
  { X86::VAESDECZrr,               X86::VAESDECZrm,               0 },
  { X86::VAESDECrr,                X86::VAESDECrm,                0 },
  { X86::VAESENCLASTYrr,           X86::VAESENCLASTYrm,           0 },
  { X86::VAESENCLASTZ128rr,        X86::VAESENCLASTZ128rm,        0 },
  { X86::VAESENCLASTZ256rr,        X86::VAESENCLASTZ256rm,        0 },
  { X86::VAESENCLASTZrr,           X86::VAESENCLASTZrm,           0 },
  { X86::VAESENCLASTrr,            X86::VAESENCLASTrm,            0 },
  { X86::VAESENCYrr,               X86::VAESENCYrm,               0 },
  { X86::VAESENCZ128rr,            X86::VAESENCZ128rm,            0 },
  { X86::VAESENCZ256rr,            X86::VAESENCZ256rm,            0 },
  { X86::VAESENCZrr,               X86::VAESENCZrm,               0 },
  { X86::VAESENCrr,                X86::VAESENCrm,                0 },
  { X86::VALIGNDZ128rri,           X86::VALIGNDZ128rmi,           0 },
  { X86::VALIGNDZ256rri,           X86::VALIGNDZ256rmi,           0 },
  { X86::VALIGNDZrri,              X86::VALIGNDZrmi,              0 },
  { X86::VALIGNQZ128rri,           X86::VALIGNQZ128rmi,           0 },
  { X86::VALIGNQZ256rri,           X86::VALIGNQZ256rmi,           0 },
  { X86::VALIGNQZrri,              X86::VALIGNQZrmi,              0 },
  { X86::VANDNPDYrr,               X86::VANDNPDYrm,               0 },
  { X86::VANDNPDZ128rr,            X86::VANDNPDZ128rm,            0 },
  { X86::VANDNPDZ256rr,            X86::VANDNPDZ256rm,            0 },
  { X86::VANDNPDZrr,               X86::VANDNPDZrm,               0 },
  { X86::VANDNPDrr,                X86::VANDNPDrm,                0 },
  { X86::VANDNPSYrr,               X86::VANDNPSYrm,               0 },
  { X86::VANDNPSZ128rr,            X86::VANDNPSZ128rm,            0 },
  { X86::VANDNPSZ256rr,            X86::VANDNPSZ256rm,            0 },
  { X86::VANDNPSZrr,               X86::VANDNPSZrm,               0 },
  { X86::VANDNPSrr,                X86::VANDNPSrm,                0 },
  { X86::VANDPDYrr,                X86::VANDPDYrm,                0 },
  { X86::VANDPDZ128rr,             X86::VANDPDZ128rm,             0 },
  { X86::VANDPDZ256rr,             X86::VANDPDZ256rm,             0 },
  { X86::VANDPDZrr,                X86::VANDPDZrm,                0 },
  { X86::VANDPDrr,                 X86::VANDPDrm,                 0 },
  { X86::VANDPSYrr,                X86::VANDPSYrm,                0 },
  { X86::VANDPSZ128rr,             X86::VANDPSZ128rm,             0 },
  { X86::VANDPSZ256rr,             X86::VANDPSZ256rm,             0 },
  { X86::VANDPSZrr,                X86::VANDPSZrm,                0 },
  { X86::VANDPSrr,                 X86::VANDPSrm,                 0 },
  { X86::VBLENDMPDZ128rr,          X86::VBLENDMPDZ128rm,          0 },
  { X86::VBLENDMPDZ256rr,          X86::VBLENDMPDZ256rm,          0 },
  { X86::VBLENDMPDZrr,             X86::VBLENDMPDZrm,             0 },
  { X86::VBLENDMPSZ128rr,          X86::VBLENDMPSZ128rm,          0 },
  { X86::VBLENDMPSZ256rr,          X86::VBLENDMPSZ256rm,          0 },
  { X86::VBLENDMPSZrr,             X86::VBLENDMPSZrm,             0 },
  { X86::VBLENDPDYrri,             X86::VBLENDPDYrmi,             0 },
  { X86::VBLENDPDrri,              X86::VBLENDPDrmi,              0 },
  { X86::VBLENDPSYrri,             X86::VBLENDPSYrmi,             0 },
  { X86::VBLENDPSrri,              X86::VBLENDPSrmi,              0 },
  { X86::VBLENDVPDYrr,             X86::VBLENDVPDYrm,             0 },
  { X86::VBLENDVPDrr,              X86::VBLENDVPDrm,              0 },
  { X86::VBLENDVPSYrr,             X86::VBLENDVPSYrm,             0 },
  { X86::VBLENDVPSrr,              X86::VBLENDVPSrm,              0 },
  { X86::VBROADCASTF32X2Z256rkz,   X86::VBROADCASTF32X2Z256mkz,   TB_NO_REVERSE },
  { X86::VBROADCASTF32X2Zrkz,      X86::VBROADCASTF32X2Zmkz,      TB_NO_REVERSE },
  { X86::VBROADCASTI32X2Z128rkz,   X86::VBROADCASTI32X2Z128mkz,   TB_NO_REVERSE },
  { X86::VBROADCASTI32X2Z256rkz,   X86::VBROADCASTI32X2Z256mkz,   TB_NO_REVERSE },
  { X86::VBROADCASTI32X2Zrkz,      X86::VBROADCASTI32X2Zmkz,      TB_NO_REVERSE },
  { X86::VBROADCASTSDZ256rkz,      X86::VBROADCASTSDZ256mkz,      TB_NO_REVERSE },
  { X86::VBROADCASTSDZrkz,         X86::VBROADCASTSDZmkz,         TB_NO_REVERSE },
  { X86::VBROADCASTSSZ128rkz,      X86::VBROADCASTSSZ128mkz,      TB_NO_REVERSE },
  { X86::VBROADCASTSSZ256rkz,      X86::VBROADCASTSSZ256mkz,      TB_NO_REVERSE },
  { X86::VBROADCASTSSZrkz,         X86::VBROADCASTSSZmkz,         TB_NO_REVERSE },
  { X86::VCMPPDYrri,               X86::VCMPPDYrmi,               0 },
  { X86::VCMPPDZ128rri,            X86::VCMPPDZ128rmi,            0 },
  { X86::VCMPPDZ256rri,            X86::VCMPPDZ256rmi,            0 },
  { X86::VCMPPDZrri,               X86::VCMPPDZrmi,               0 },
  { X86::VCMPPDrri,                X86::VCMPPDrmi,                0 },
  { X86::VCMPPSYrri,               X86::VCMPPSYrmi,               0 },
  { X86::VCMPPSZ128rri,            X86::VCMPPSZ128rmi,            0 },
  { X86::VCMPPSZ256rri,            X86::VCMPPSZ256rmi,            0 },
  { X86::VCMPPSZrri,               X86::VCMPPSZrmi,               0 },
  { X86::VCMPPSrri,                X86::VCMPPSrmi,                0 },
  { X86::VCMPSDZrr,                X86::VCMPSDZrm,                0 },
  { X86::VCMPSDZrr_Int,            X86::VCMPSDZrm_Int,            TB_NO_REVERSE },
  { X86::VCMPSDrr,                 X86::VCMPSDrm,                 0 },
  { X86::VCMPSDrr_Int,             X86::VCMPSDrm_Int,             TB_NO_REVERSE },
  { X86::VCMPSSZrr,                X86::VCMPSSZrm,                0 },
  { X86::VCMPSSZrr_Int,            X86::VCMPSSZrm_Int,            TB_NO_REVERSE },
  { X86::VCMPSSrr,                 X86::VCMPSSrm,                 0 },
  { X86::VCMPSSrr_Int,             X86::VCMPSSrm_Int,             TB_NO_REVERSE },
  { X86::VCVTDQ2PDZ128rrkz,        X86::VCVTDQ2PDZ128rmkz,        TB_NO_REVERSE },
  { X86::VCVTDQ2PDZ256rrkz,        X86::VCVTDQ2PDZ256rmkz,        0 },
  { X86::VCVTDQ2PDZrrkz,           X86::VCVTDQ2PDZrmkz,           0 },
  { X86::VCVTDQ2PSZ128rrkz,        X86::VCVTDQ2PSZ128rmkz,        0 },
  { X86::VCVTDQ2PSZ256rrkz,        X86::VCVTDQ2PSZ256rmkz,        0 },
  { X86::VCVTDQ2PSZrrkz,           X86::VCVTDQ2PSZrmkz,           0 },
  { X86::VCVTNE2PS2BF16Z128rr,     X86::VCVTNE2PS2BF16Z128rm,     0 },
  { X86::VCVTNE2PS2BF16Z256rr,     X86::VCVTNE2PS2BF16Z256rm,     0 },
  { X86::VCVTNE2PS2BF16Zrr,        X86::VCVTNE2PS2BF16Zrm,        0 },
  { X86::VCVTNEPS2BF16Z128rrkz,    X86::VCVTNEPS2BF16Z128rmkz,    0 },
  { X86::VCVTNEPS2BF16Z256rrkz,    X86::VCVTNEPS2BF16Z256rmkz,    0 },
  { X86::VCVTNEPS2BF16Zrrkz,       X86::VCVTNEPS2BF16Zrmkz,       0 },
  { X86::VCVTPD2DQZ128rrkz,        X86::VCVTPD2DQZ128rmkz,        0 },
  { X86::VCVTPD2DQZ256rrkz,        X86::VCVTPD2DQZ256rmkz,        0 },
  { X86::VCVTPD2DQZrrkz,           X86::VCVTPD2DQZrmkz,           0 },
  { X86::VCVTPD2PSZ128rrkz,        X86::VCVTPD2PSZ128rmkz,        0 },
  { X86::VCVTPD2PSZ256rrkz,        X86::VCVTPD2PSZ256rmkz,        0 },
  { X86::VCVTPD2PSZrrkz,           X86::VCVTPD2PSZrmkz,           0 },
  { X86::VCVTPD2QQZ128rrkz,        X86::VCVTPD2QQZ128rmkz,        0 },
  { X86::VCVTPD2QQZ256rrkz,        X86::VCVTPD2QQZ256rmkz,        0 },
  { X86::VCVTPD2QQZrrkz,           X86::VCVTPD2QQZrmkz,           0 },
  { X86::VCVTPD2UDQZ128rrkz,       X86::VCVTPD2UDQZ128rmkz,       0 },
  { X86::VCVTPD2UDQZ256rrkz,       X86::VCVTPD2UDQZ256rmkz,       0 },
  { X86::VCVTPD2UDQZrrkz,          X86::VCVTPD2UDQZrmkz,          0 },
  { X86::VCVTPD2UQQZ128rrkz,       X86::VCVTPD2UQQZ128rmkz,       0 },
  { X86::VCVTPD2UQQZ256rrkz,       X86::VCVTPD2UQQZ256rmkz,       0 },
  { X86::VCVTPD2UQQZrrkz,          X86::VCVTPD2UQQZrmkz,          0 },
  { X86::VCVTPH2PSZ128rrkz,        X86::VCVTPH2PSZ128rmkz,        TB_NO_REVERSE },
  { X86::VCVTPH2PSZ256rrkz,        X86::VCVTPH2PSZ256rmkz,        0 },
  { X86::VCVTPH2PSZrrkz,           X86::VCVTPH2PSZrmkz,           0 },
  { X86::VCVTPS2DQZ128rrkz,        X86::VCVTPS2DQZ128rmkz,        0 },
  { X86::VCVTPS2DQZ256rrkz,        X86::VCVTPS2DQZ256rmkz,        0 },
  { X86::VCVTPS2DQZrrkz,           X86::VCVTPS2DQZrmkz,           0 },
  { X86::VCVTPS2PDZ128rrkz,        X86::VCVTPS2PDZ128rmkz,        TB_NO_REVERSE },
  { X86::VCVTPS2PDZ256rrkz,        X86::VCVTPS2PDZ256rmkz,        0 },
  { X86::VCVTPS2PDZrrkz,           X86::VCVTPS2PDZrmkz,           0 },
  { X86::VCVTPS2QQZ128rrkz,        X86::VCVTPS2QQZ128rmkz,        TB_NO_REVERSE },
  { X86::VCVTPS2QQZ256rrkz,        X86::VCVTPS2QQZ256rmkz,        0 },
  { X86::VCVTPS2QQZrrkz,           X86::VCVTPS2QQZrmkz,           0 },
  { X86::VCVTPS2UDQZ128rrkz,       X86::VCVTPS2UDQZ128rmkz,       0 },
  { X86::VCVTPS2UDQZ256rrkz,       X86::VCVTPS2UDQZ256rmkz,       0 },
  { X86::VCVTPS2UDQZrrkz,          X86::VCVTPS2UDQZrmkz,          0 },
  { X86::VCVTPS2UQQZ128rrkz,       X86::VCVTPS2UQQZ128rmkz,       TB_NO_REVERSE },
  { X86::VCVTPS2UQQZ256rrkz,       X86::VCVTPS2UQQZ256rmkz,       0 },
  { X86::VCVTPS2UQQZrrkz,          X86::VCVTPS2UQQZrmkz,          0 },
  { X86::VCVTQQ2PDZ128rrkz,        X86::VCVTQQ2PDZ128rmkz,        0 },
  { X86::VCVTQQ2PDZ256rrkz,        X86::VCVTQQ2PDZ256rmkz,        0 },
  { X86::VCVTQQ2PDZrrkz,           X86::VCVTQQ2PDZrmkz,           0 },
  { X86::VCVTQQ2PSZ128rrkz,        X86::VCVTQQ2PSZ128rmkz,        0 },
  { X86::VCVTQQ2PSZ256rrkz,        X86::VCVTQQ2PSZ256rmkz,        0 },
  { X86::VCVTQQ2PSZrrkz,           X86::VCVTQQ2PSZrmkz,           0 },
  { X86::VCVTSD2SSZrr,             X86::VCVTSD2SSZrm,             0 },
  { X86::VCVTSD2SSZrr_Int,         X86::VCVTSD2SSZrm_Int,         TB_NO_REVERSE },
  { X86::VCVTSD2SSrr,              X86::VCVTSD2SSrm,              0 },
  { X86::VCVTSD2SSrr_Int,          X86::VCVTSD2SSrm_Int,          TB_NO_REVERSE },
  { X86::VCVTSI2SDZrr,             X86::VCVTSI2SDZrm,             0 },
  { X86::VCVTSI2SDZrr_Int,         X86::VCVTSI2SDZrm_Int,         0 },
  { X86::VCVTSI2SDrr,              X86::VCVTSI2SDrm,              0 },
  { X86::VCVTSI2SDrr_Int,          X86::VCVTSI2SDrm_Int,          0 },
  { X86::VCVTSI2SSZrr,             X86::VCVTSI2SSZrm,             0 },
  { X86::VCVTSI2SSZrr_Int,         X86::VCVTSI2SSZrm_Int,         0 },
  { X86::VCVTSI2SSrr,              X86::VCVTSI2SSrm,              0 },
  { X86::VCVTSI2SSrr_Int,          X86::VCVTSI2SSrm_Int,          0 },
  { X86::VCVTSI642SDZrr,           X86::VCVTSI642SDZrm,           0 },
  { X86::VCVTSI642SDZrr_Int,       X86::VCVTSI642SDZrm_Int,       0 },
  { X86::VCVTSI642SDrr,            X86::VCVTSI642SDrm,            0 },
  { X86::VCVTSI642SDrr_Int,        X86::VCVTSI642SDrm_Int,        0 },
  { X86::VCVTSI642SSZrr,           X86::VCVTSI642SSZrm,           0 },
  { X86::VCVTSI642SSZrr_Int,       X86::VCVTSI642SSZrm_Int,       0 },
  { X86::VCVTSI642SSrr,            X86::VCVTSI642SSrm,            0 },
  { X86::VCVTSI642SSrr_Int,        X86::VCVTSI642SSrm_Int,        0 },
  { X86::VCVTSS2SDZrr,             X86::VCVTSS2SDZrm,             0 },
  { X86::VCVTSS2SDZrr_Int,         X86::VCVTSS2SDZrm_Int,         TB_NO_REVERSE },
  { X86::VCVTSS2SDrr,              X86::VCVTSS2SDrm,              0 },
  { X86::VCVTSS2SDrr_Int,          X86::VCVTSS2SDrm_Int,          TB_NO_REVERSE },
  { X86::VCVTTPD2DQZ128rrkz,       X86::VCVTTPD2DQZ128rmkz,       0 },
  { X86::VCVTTPD2DQZ256rrkz,       X86::VCVTTPD2DQZ256rmkz,       0 },
  { X86::VCVTTPD2DQZrrkz,          X86::VCVTTPD2DQZrmkz,          0 },
  { X86::VCVTTPD2QQZ128rrkz,       X86::VCVTTPD2QQZ128rmkz,       0 },
  { X86::VCVTTPD2QQZ256rrkz,       X86::VCVTTPD2QQZ256rmkz,       0 },
  { X86::VCVTTPD2QQZrrkz,          X86::VCVTTPD2QQZrmkz,          0 },
  { X86::VCVTTPD2UDQZ128rrkz,      X86::VCVTTPD2UDQZ128rmkz,      0 },
  { X86::VCVTTPD2UDQZ256rrkz,      X86::VCVTTPD2UDQZ256rmkz,      0 },
  { X86::VCVTTPD2UDQZrrkz,         X86::VCVTTPD2UDQZrmkz,         0 },
  { X86::VCVTTPD2UQQZ128rrkz,      X86::VCVTTPD2UQQZ128rmkz,      0 },
  { X86::VCVTTPD2UQQZ256rrkz,      X86::VCVTTPD2UQQZ256rmkz,      0 },
  { X86::VCVTTPD2UQQZrrkz,         X86::VCVTTPD2UQQZrmkz,         0 },
  { X86::VCVTTPS2DQZ128rrkz,       X86::VCVTTPS2DQZ128rmkz,       0 },
  { X86::VCVTTPS2DQZ256rrkz,       X86::VCVTTPS2DQZ256rmkz,       0 },
  { X86::VCVTTPS2DQZrrkz,          X86::VCVTTPS2DQZrmkz,          0 },
  { X86::VCVTTPS2QQZ128rrkz,       X86::VCVTTPS2QQZ128rmkz,       TB_NO_REVERSE },
  { X86::VCVTTPS2QQZ256rrkz,       X86::VCVTTPS2QQZ256rmkz,       0 },
  { X86::VCVTTPS2QQZrrkz,          X86::VCVTTPS2QQZrmkz,          0 },
  { X86::VCVTTPS2UDQZ128rrkz,      X86::VCVTTPS2UDQZ128rmkz,      0 },
  { X86::VCVTTPS2UDQZ256rrkz,      X86::VCVTTPS2UDQZ256rmkz,      0 },
  { X86::VCVTTPS2UDQZrrkz,         X86::VCVTTPS2UDQZrmkz,         0 },
  { X86::VCVTTPS2UQQZ128rrkz,      X86::VCVTTPS2UQQZ128rmkz,      TB_NO_REVERSE },
  { X86::VCVTTPS2UQQZ256rrkz,      X86::VCVTTPS2UQQZ256rmkz,      0 },
  { X86::VCVTTPS2UQQZrrkz,         X86::VCVTTPS2UQQZrmkz,         0 },
  { X86::VCVTUDQ2PDZ128rrkz,       X86::VCVTUDQ2PDZ128rmkz,       TB_NO_REVERSE },
  { X86::VCVTUDQ2PDZ256rrkz,       X86::VCVTUDQ2PDZ256rmkz,       0 },
  { X86::VCVTUDQ2PDZrrkz,          X86::VCVTUDQ2PDZrmkz,          0 },
  { X86::VCVTUDQ2PSZ128rrkz,       X86::VCVTUDQ2PSZ128rmkz,       0 },
  { X86::VCVTUDQ2PSZ256rrkz,       X86::VCVTUDQ2PSZ256rmkz,       0 },
  { X86::VCVTUDQ2PSZrrkz,          X86::VCVTUDQ2PSZrmkz,          0 },
  { X86::VCVTUQQ2PDZ128rrkz,       X86::VCVTUQQ2PDZ128rmkz,       0 },
  { X86::VCVTUQQ2PDZ256rrkz,       X86::VCVTUQQ2PDZ256rmkz,       0 },
  { X86::VCVTUQQ2PDZrrkz,          X86::VCVTUQQ2PDZrmkz,          0 },
  { X86::VCVTUQQ2PSZ128rrkz,       X86::VCVTUQQ2PSZ128rmkz,       0 },
  { X86::VCVTUQQ2PSZ256rrkz,       X86::VCVTUQQ2PSZ256rmkz,       0 },
  { X86::VCVTUQQ2PSZrrkz,          X86::VCVTUQQ2PSZrmkz,          0 },
  { X86::VCVTUSI2SDZrr,            X86::VCVTUSI2SDZrm,            0 },
  { X86::VCVTUSI2SDZrr_Int,        X86::VCVTUSI2SDZrm_Int,        0 },
  { X86::VCVTUSI2SSZrr,            X86::VCVTUSI2SSZrm,            0 },
  { X86::VCVTUSI2SSZrr_Int,        X86::VCVTUSI2SSZrm_Int,        0 },
  { X86::VCVTUSI642SDZrr,          X86::VCVTUSI642SDZrm,          0 },
  { X86::VCVTUSI642SDZrr_Int,      X86::VCVTUSI642SDZrm_Int,      0 },
  { X86::VCVTUSI642SSZrr,          X86::VCVTUSI642SSZrm,          0 },
  { X86::VCVTUSI642SSZrr_Int,      X86::VCVTUSI642SSZrm_Int,      0 },
  { X86::VDBPSADBWZ128rri,         X86::VDBPSADBWZ128rmi,         0 },
  { X86::VDBPSADBWZ256rri,         X86::VDBPSADBWZ256rmi,         0 },
  { X86::VDBPSADBWZrri,            X86::VDBPSADBWZrmi,            0 },
  { X86::VDIVPDYrr,                X86::VDIVPDYrm,                0 },
  { X86::VDIVPDZ128rr,             X86::VDIVPDZ128rm,             0 },
  { X86::VDIVPDZ256rr,             X86::VDIVPDZ256rm,             0 },
  { X86::VDIVPDZrr,                X86::VDIVPDZrm,                0 },
  { X86::VDIVPDrr,                 X86::VDIVPDrm,                 0 },
  { X86::VDIVPSYrr,                X86::VDIVPSYrm,                0 },
  { X86::VDIVPSZ128rr,             X86::VDIVPSZ128rm,             0 },
  { X86::VDIVPSZ256rr,             X86::VDIVPSZ256rm,             0 },
  { X86::VDIVPSZrr,                X86::VDIVPSZrm,                0 },
  { X86::VDIVPSrr,                 X86::VDIVPSrm,                 0 },
  { X86::VDIVSDZrr,                X86::VDIVSDZrm,                0 },
  { X86::VDIVSDZrr_Int,            X86::VDIVSDZrm_Int,            TB_NO_REVERSE },
  { X86::VDIVSDrr,                 X86::VDIVSDrm,                 0 },
  { X86::VDIVSDrr_Int,             X86::VDIVSDrm_Int,             TB_NO_REVERSE },
  { X86::VDIVSSZrr,                X86::VDIVSSZrm,                0 },
  { X86::VDIVSSZrr_Int,            X86::VDIVSSZrm_Int,            TB_NO_REVERSE },
  { X86::VDIVSSrr,                 X86::VDIVSSrm,                 0 },
  { X86::VDIVSSrr_Int,             X86::VDIVSSrm_Int,             TB_NO_REVERSE },
  { X86::VDPPDrri,                 X86::VDPPDrmi,                 0 },
  { X86::VDPPSYrri,                X86::VDPPSYrmi,                0 },
  { X86::VDPPSrri,                 X86::VDPPSrmi,                 0 },
  { X86::VEXP2PDZrkz,              X86::VEXP2PDZmkz,              0 },
  { X86::VEXP2PSZrkz,              X86::VEXP2PSZmkz,              0 },
  { X86::VEXPANDPDZ128rrkz,        X86::VEXPANDPDZ128rmkz,        TB_NO_REVERSE },
  { X86::VEXPANDPDZ256rrkz,        X86::VEXPANDPDZ256rmkz,        TB_NO_REVERSE },
  { X86::VEXPANDPDZrrkz,           X86::VEXPANDPDZrmkz,           TB_NO_REVERSE },
  { X86::VEXPANDPSZ128rrkz,        X86::VEXPANDPSZ128rmkz,        TB_NO_REVERSE },
  { X86::VEXPANDPSZ256rrkz,        X86::VEXPANDPSZ256rmkz,        TB_NO_REVERSE },
  { X86::VEXPANDPSZrrkz,           X86::VEXPANDPSZrmkz,           TB_NO_REVERSE },
  { X86::VFMADDPD4Yrr,             X86::VFMADDPD4Ymr,             0 },
  { X86::VFMADDPD4rr,              X86::VFMADDPD4mr,              0 },
  { X86::VFMADDPS4Yrr,             X86::VFMADDPS4Ymr,             0 },
  { X86::VFMADDPS4rr,              X86::VFMADDPS4mr,              0 },
  { X86::VFMADDSD4rr,              X86::VFMADDSD4mr,              0 },
  { X86::VFMADDSD4rr_Int,          X86::VFMADDSD4mr_Int,          TB_NO_REVERSE },
  { X86::VFMADDSS4rr,              X86::VFMADDSS4mr,              0 },
  { X86::VFMADDSS4rr_Int,          X86::VFMADDSS4mr_Int,          TB_NO_REVERSE },
  { X86::VFMADDSUBPD4Yrr,          X86::VFMADDSUBPD4Ymr,          0 },
  { X86::VFMADDSUBPD4rr,           X86::VFMADDSUBPD4mr,           0 },
  { X86::VFMADDSUBPS4Yrr,          X86::VFMADDSUBPS4Ymr,          0 },
  { X86::VFMADDSUBPS4rr,           X86::VFMADDSUBPS4mr,           0 },
  { X86::VFMSUBADDPD4Yrr,          X86::VFMSUBADDPD4Ymr,          0 },
  { X86::VFMSUBADDPD4rr,           X86::VFMSUBADDPD4mr,           0 },
  { X86::VFMSUBADDPS4Yrr,          X86::VFMSUBADDPS4Ymr,          0 },
  { X86::VFMSUBADDPS4rr,           X86::VFMSUBADDPS4mr,           0 },
  { X86::VFMSUBPD4Yrr,             X86::VFMSUBPD4Ymr,             0 },
  { X86::VFMSUBPD4rr,              X86::VFMSUBPD4mr,              0 },
  { X86::VFMSUBPS4Yrr,             X86::VFMSUBPS4Ymr,             0 },
  { X86::VFMSUBPS4rr,              X86::VFMSUBPS4mr,              0 },
  { X86::VFMSUBSD4rr,              X86::VFMSUBSD4mr,              0 },
  { X86::VFMSUBSD4rr_Int,          X86::VFMSUBSD4mr_Int,          TB_NO_REVERSE },
  { X86::VFMSUBSS4rr,              X86::VFMSUBSS4mr,              0 },
  { X86::VFMSUBSS4rr_Int,          X86::VFMSUBSS4mr_Int,          TB_NO_REVERSE },
  { X86::VFNMADDPD4Yrr,            X86::VFNMADDPD4Ymr,            0 },
  { X86::VFNMADDPD4rr,             X86::VFNMADDPD4mr,             0 },
  { X86::VFNMADDPS4Yrr,            X86::VFNMADDPS4Ymr,            0 },
  { X86::VFNMADDPS4rr,             X86::VFNMADDPS4mr,             0 },
  { X86::VFNMADDSD4rr,             X86::VFNMADDSD4mr,             0 },
  { X86::VFNMADDSD4rr_Int,         X86::VFNMADDSD4mr_Int,         TB_NO_REVERSE },
  { X86::VFNMADDSS4rr,             X86::VFNMADDSS4mr,             0 },
  { X86::VFNMADDSS4rr_Int,         X86::VFNMADDSS4mr_Int,         TB_NO_REVERSE },
  { X86::VFNMSUBPD4Yrr,            X86::VFNMSUBPD4Ymr,            0 },
  { X86::VFNMSUBPD4rr,             X86::VFNMSUBPD4mr,             0 },
  { X86::VFNMSUBPS4Yrr,            X86::VFNMSUBPS4Ymr,            0 },
  { X86::VFNMSUBPS4rr,             X86::VFNMSUBPS4mr,             0 },
  { X86::VFNMSUBSD4rr,             X86::VFNMSUBSD4mr,             0 },
  { X86::VFNMSUBSD4rr_Int,         X86::VFNMSUBSD4mr_Int,         TB_NO_REVERSE },
  { X86::VFNMSUBSS4rr,             X86::VFNMSUBSS4mr,             0 },
  { X86::VFNMSUBSS4rr_Int,         X86::VFNMSUBSS4mr_Int,         TB_NO_REVERSE },
  { X86::VFPCLASSPDZ128rrk,        X86::VFPCLASSPDZ128rmk,        0 },
  { X86::VFPCLASSPDZ256rrk,        X86::VFPCLASSPDZ256rmk,        0 },
  { X86::VFPCLASSPDZrrk,           X86::VFPCLASSPDZrmk,           0 },
  { X86::VFPCLASSPSZ128rrk,        X86::VFPCLASSPSZ128rmk,        0 },
  { X86::VFPCLASSPSZ256rrk,        X86::VFPCLASSPSZ256rmk,        0 },
  { X86::VFPCLASSPSZrrk,           X86::VFPCLASSPSZrmk,           0 },
  { X86::VFPCLASSSDZrrk,           X86::VFPCLASSSDZrmk,           TB_NO_REVERSE },
  { X86::VFPCLASSSSZrrk,           X86::VFPCLASSSSZrmk,           TB_NO_REVERSE },
  { X86::VGETEXPPDZ128rkz,         X86::VGETEXPPDZ128mkz,         0 },
  { X86::VGETEXPPDZ256rkz,         X86::VGETEXPPDZ256mkz,         0 },
  { X86::VGETEXPPDZrkz,            X86::VGETEXPPDZmkz,            0 },
  { X86::VGETEXPPSZ128rkz,         X86::VGETEXPPSZ128mkz,         0 },
  { X86::VGETEXPPSZ256rkz,         X86::VGETEXPPSZ256mkz,         0 },
  { X86::VGETEXPPSZrkz,            X86::VGETEXPPSZmkz,            0 },
  { X86::VGETEXPSDZr,              X86::VGETEXPSDZm,              TB_NO_REVERSE },
  { X86::VGETEXPSSZr,              X86::VGETEXPSSZm,              TB_NO_REVERSE },
  { X86::VGETMANTPDZ128rrikz,      X86::VGETMANTPDZ128rmikz,      0 },
  { X86::VGETMANTPDZ256rrikz,      X86::VGETMANTPDZ256rmikz,      0 },
  { X86::VGETMANTPDZrrikz,         X86::VGETMANTPDZrmikz,         0 },
  { X86::VGETMANTPSZ128rrikz,      X86::VGETMANTPSZ128rmikz,      0 },
  { X86::VGETMANTPSZ256rrikz,      X86::VGETMANTPSZ256rmikz,      0 },
  { X86::VGETMANTPSZrrikz,         X86::VGETMANTPSZrmikz,         0 },
  { X86::VGETMANTSDZrri,           X86::VGETMANTSDZrmi,           TB_NO_REVERSE },
  { X86::VGETMANTSSZrri,           X86::VGETMANTSSZrmi,           TB_NO_REVERSE },
  { X86::VGF2P8AFFINEINVQBYrri,    X86::VGF2P8AFFINEINVQBYrmi,    0 },
  { X86::VGF2P8AFFINEINVQBZ128rri, X86::VGF2P8AFFINEINVQBZ128rmi, 0 },
  { X86::VGF2P8AFFINEINVQBZ256rri, X86::VGF2P8AFFINEINVQBZ256rmi, 0 },
  { X86::VGF2P8AFFINEINVQBZrri,    X86::VGF2P8AFFINEINVQBZrmi,    0 },
  { X86::VGF2P8AFFINEINVQBrri,     X86::VGF2P8AFFINEINVQBrmi,     0 },
  { X86::VGF2P8AFFINEQBYrri,       X86::VGF2P8AFFINEQBYrmi,       0 },
  { X86::VGF2P8AFFINEQBZ128rri,    X86::VGF2P8AFFINEQBZ128rmi,    0 },
  { X86::VGF2P8AFFINEQBZ256rri,    X86::VGF2P8AFFINEQBZ256rmi,    0 },
  { X86::VGF2P8AFFINEQBZrri,       X86::VGF2P8AFFINEQBZrmi,       0 },
  { X86::VGF2P8AFFINEQBrri,        X86::VGF2P8AFFINEQBrmi,        0 },
  { X86::VGF2P8MULBYrr,            X86::VGF2P8MULBYrm,            0 },
  { X86::VGF2P8MULBZ128rr,         X86::VGF2P8MULBZ128rm,         0 },
  { X86::VGF2P8MULBZ256rr,         X86::VGF2P8MULBZ256rm,         0 },
  { X86::VGF2P8MULBZrr,            X86::VGF2P8MULBZrm,            0 },
  { X86::VGF2P8MULBrr,             X86::VGF2P8MULBrm,             0 },
  { X86::VHADDPDYrr,               X86::VHADDPDYrm,               0 },
  { X86::VHADDPDrr,                X86::VHADDPDrm,                0 },
  { X86::VHADDPSYrr,               X86::VHADDPSYrm,               0 },
  { X86::VHADDPSrr,                X86::VHADDPSrm,                0 },
  { X86::VHSUBPDYrr,               X86::VHSUBPDYrm,               0 },
  { X86::VHSUBPDrr,                X86::VHSUBPDrm,                0 },
  { X86::VHSUBPSYrr,               X86::VHSUBPSYrm,               0 },
  { X86::VHSUBPSrr,                X86::VHSUBPSrm,                0 },
  { X86::VINSERTF128rr,            X86::VINSERTF128rm,            0 },
  { X86::VINSERTF32x4Z256rr,       X86::VINSERTF32x4Z256rm,       0 },
  { X86::VINSERTF32x4Zrr,          X86::VINSERTF32x4Zrm,          0 },
  { X86::VINSERTF32x8Zrr,          X86::VINSERTF32x8Zrm,          0 },
  { X86::VINSERTF64x2Z256rr,       X86::VINSERTF64x2Z256rm,       0 },
  { X86::VINSERTF64x2Zrr,          X86::VINSERTF64x2Zrm,          0 },
  { X86::VINSERTF64x4Zrr,          X86::VINSERTF64x4Zrm,          0 },
  { X86::VINSERTI128rr,            X86::VINSERTI128rm,            0 },
  { X86::VINSERTI32x4Z256rr,       X86::VINSERTI32x4Z256rm,       0 },
  { X86::VINSERTI32x4Zrr,          X86::VINSERTI32x4Zrm,          0 },
  { X86::VINSERTI32x8Zrr,          X86::VINSERTI32x8Zrm,          0 },
  { X86::VINSERTI64x2Z256rr,       X86::VINSERTI64x2Z256rm,       0 },
  { X86::VINSERTI64x2Zrr,          X86::VINSERTI64x2Zrm,          0 },
  { X86::VINSERTI64x4Zrr,          X86::VINSERTI64x4Zrm,          0 },
  { X86::VMAXCPDYrr,               X86::VMAXCPDYrm,               0 },
  { X86::VMAXCPDZ128rr,            X86::VMAXCPDZ128rm,            0 },
  { X86::VMAXCPDZ256rr,            X86::VMAXCPDZ256rm,            0 },
  { X86::VMAXCPDZrr,               X86::VMAXCPDZrm,               0 },
  { X86::VMAXCPDrr,                X86::VMAXCPDrm,                0 },
  { X86::VMAXCPSYrr,               X86::VMAXCPSYrm,               0 },
  { X86::VMAXCPSZ128rr,            X86::VMAXCPSZ128rm,            0 },
  { X86::VMAXCPSZ256rr,            X86::VMAXCPSZ256rm,            0 },
  { X86::VMAXCPSZrr,               X86::VMAXCPSZrm,               0 },
  { X86::VMAXCPSrr,                X86::VMAXCPSrm,                0 },
  { X86::VMAXCSDZrr,               X86::VMAXCSDZrm,               0 },
  { X86::VMAXCSDrr,                X86::VMAXCSDrm,                0 },
  { X86::VMAXCSSZrr,               X86::VMAXCSSZrm,               0 },
  { X86::VMAXCSSrr,                X86::VMAXCSSrm,                0 },
  { X86::VMAXPDYrr,                X86::VMAXPDYrm,                0 },
  { X86::VMAXPDZ128rr,             X86::VMAXPDZ128rm,             0 },
  { X86::VMAXPDZ256rr,             X86::VMAXPDZ256rm,             0 },
  { X86::VMAXPDZrr,                X86::VMAXPDZrm,                0 },
  { X86::VMAXPDrr,                 X86::VMAXPDrm,                 0 },
  { X86::VMAXPSYrr,                X86::VMAXPSYrm,                0 },
  { X86::VMAXPSZ128rr,             X86::VMAXPSZ128rm,             0 },
  { X86::VMAXPSZ256rr,             X86::VMAXPSZ256rm,             0 },
  { X86::VMAXPSZrr,                X86::VMAXPSZrm,                0 },
  { X86::VMAXPSrr,                 X86::VMAXPSrm,                 0 },
  { X86::VMAXSDZrr,                X86::VMAXSDZrm,                0 },
  { X86::VMAXSDZrr_Int,            X86::VMAXSDZrm_Int,            TB_NO_REVERSE },
  { X86::VMAXSDrr,                 X86::VMAXSDrm,                 0 },
  { X86::VMAXSDrr_Int,             X86::VMAXSDrm_Int,             TB_NO_REVERSE },
  { X86::VMAXSSZrr,                X86::VMAXSSZrm,                0 },
  { X86::VMAXSSZrr_Int,            X86::VMAXSSZrm_Int,            TB_NO_REVERSE },
  { X86::VMAXSSrr,                 X86::VMAXSSrm,                 0 },
  { X86::VMAXSSrr_Int,             X86::VMAXSSrm_Int,             TB_NO_REVERSE },
  { X86::VMINCPDYrr,               X86::VMINCPDYrm,               0 },
  { X86::VMINCPDZ128rr,            X86::VMINCPDZ128rm,            0 },
  { X86::VMINCPDZ256rr,            X86::VMINCPDZ256rm,            0 },
  { X86::VMINCPDZrr,               X86::VMINCPDZrm,               0 },
  { X86::VMINCPDrr,                X86::VMINCPDrm,                0 },
  { X86::VMINCPSYrr,               X86::VMINCPSYrm,               0 },
  { X86::VMINCPSZ128rr,            X86::VMINCPSZ128rm,            0 },
  { X86::VMINCPSZ256rr,            X86::VMINCPSZ256rm,            0 },
  { X86::VMINCPSZrr,               X86::VMINCPSZrm,               0 },
  { X86::VMINCPSrr,                X86::VMINCPSrm,                0 },
  { X86::VMINCSDZrr,               X86::VMINCSDZrm,               0 },
  { X86::VMINCSDrr,                X86::VMINCSDrm,                0 },
  { X86::VMINCSSZrr,               X86::VMINCSSZrm,               0 },
  { X86::VMINCSSrr,                X86::VMINCSSrm,                0 },
  { X86::VMINPDYrr,                X86::VMINPDYrm,                0 },
  { X86::VMINPDZ128rr,             X86::VMINPDZ128rm,             0 },
  { X86::VMINPDZ256rr,             X86::VMINPDZ256rm,             0 },
  { X86::VMINPDZrr,                X86::VMINPDZrm,                0 },
  { X86::VMINPDrr,                 X86::VMINPDrm,                 0 },
  { X86::VMINPSYrr,                X86::VMINPSYrm,                0 },
  { X86::VMINPSZ128rr,             X86::VMINPSZ128rm,             0 },
  { X86::VMINPSZ256rr,             X86::VMINPSZ256rm,             0 },
  { X86::VMINPSZrr,                X86::VMINPSZrm,                0 },
  { X86::VMINPSrr,                 X86::VMINPSrm,                 0 },
  { X86::VMINSDZrr,                X86::VMINSDZrm,                0 },
  { X86::VMINSDZrr_Int,            X86::VMINSDZrm_Int,            TB_NO_REVERSE },
  { X86::VMINSDrr,                 X86::VMINSDrm,                 0 },
  { X86::VMINSDrr_Int,             X86::VMINSDrm_Int,             TB_NO_REVERSE },
  { X86::VMINSSZrr,                X86::VMINSSZrm,                0 },
  { X86::VMINSSZrr_Int,            X86::VMINSSZrm_Int,            TB_NO_REVERSE },
  { X86::VMINSSrr,                 X86::VMINSSrm,                 0 },
  { X86::VMINSSrr_Int,             X86::VMINSSrm_Int,             TB_NO_REVERSE },
  { X86::VMOVAPDZ128rrkz,          X86::VMOVAPDZ128rmkz,          TB_NO_REVERSE | TB_ALIGN_16 },
  { X86::VMOVAPDZ256rrkz,          X86::VMOVAPDZ256rmkz,          TB_NO_REVERSE | TB_ALIGN_32 },
  { X86::VMOVAPDZrrkz,             X86::VMOVAPDZrmkz,             TB_NO_REVERSE | TB_ALIGN_64 },
  { X86::VMOVAPSZ128rrkz,          X86::VMOVAPSZ128rmkz,          TB_NO_REVERSE | TB_ALIGN_16 },
  { X86::VMOVAPSZ256rrkz,          X86::VMOVAPSZ256rmkz,          TB_NO_REVERSE | TB_ALIGN_32 },
  { X86::VMOVAPSZrrkz,             X86::VMOVAPSZrmkz,             TB_NO_REVERSE | TB_ALIGN_64 },
  { X86::VMOVDDUPZ128rrkz,         X86::VMOVDDUPZ128rmkz,         TB_NO_REVERSE },
  { X86::VMOVDDUPZ256rrkz,         X86::VMOVDDUPZ256rmkz,         0 },
  { X86::VMOVDDUPZrrkz,            X86::VMOVDDUPZrmkz,            0 },
  { X86::VMOVDQA32Z128rrkz,        X86::VMOVDQA32Z128rmkz,        TB_NO_REVERSE | TB_ALIGN_16 },
  { X86::VMOVDQA32Z256rrkz,        X86::VMOVDQA32Z256rmkz,        TB_NO_REVERSE | TB_ALIGN_32 },
  { X86::VMOVDQA32Zrrkz,           X86::VMOVDQA32Zrmkz,           TB_NO_REVERSE | TB_ALIGN_64 },
  { X86::VMOVDQA64Z128rrkz,        X86::VMOVDQA64Z128rmkz,        TB_NO_REVERSE | TB_ALIGN_16 },
  { X86::VMOVDQA64Z256rrkz,        X86::VMOVDQA64Z256rmkz,        TB_NO_REVERSE | TB_ALIGN_32 },
  { X86::VMOVDQA64Zrrkz,           X86::VMOVDQA64Zrmkz,           TB_NO_REVERSE | TB_ALIGN_64 },
  { X86::VMOVDQU16Z128rrkz,        X86::VMOVDQU16Z128rmkz,        TB_NO_REVERSE },
  { X86::VMOVDQU16Z256rrkz,        X86::VMOVDQU16Z256rmkz,        TB_NO_REVERSE },
  { X86::VMOVDQU16Zrrkz,           X86::VMOVDQU16Zrmkz,           TB_NO_REVERSE },
  { X86::VMOVDQU32Z128rrkz,        X86::VMOVDQU32Z128rmkz,        TB_NO_REVERSE },
  { X86::VMOVDQU32Z256rrkz,        X86::VMOVDQU32Z256rmkz,        TB_NO_REVERSE },
  { X86::VMOVDQU32Zrrkz,           X86::VMOVDQU32Zrmkz,           TB_NO_REVERSE },
  { X86::VMOVDQU64Z128rrkz,        X86::VMOVDQU64Z128rmkz,        TB_NO_REVERSE },
  { X86::VMOVDQU64Z256rrkz,        X86::VMOVDQU64Z256rmkz,        TB_NO_REVERSE },
  { X86::VMOVDQU64Zrrkz,           X86::VMOVDQU64Zrmkz,           TB_NO_REVERSE },
  { X86::VMOVDQU8Z128rrkz,         X86::VMOVDQU8Z128rmkz,         TB_NO_REVERSE },
  { X86::VMOVDQU8Z256rrkz,         X86::VMOVDQU8Z256rmkz,         TB_NO_REVERSE },
  { X86::VMOVDQU8Zrrkz,            X86::VMOVDQU8Zrmkz,            TB_NO_REVERSE },
  { X86::VMOVLHPSZrr,              X86::VMOVHPSZ128rm,            TB_NO_REVERSE },
  { X86::VMOVLHPSrr,               X86::VMOVHPSrm,                TB_NO_REVERSE },
  { X86::VMOVSDZrr,                X86::VMOVLPDZ128rm,            TB_NO_REVERSE },
  { X86::VMOVSDrr,                 X86::VMOVLPDrm,                TB_NO_REVERSE },
  { X86::VMOVSHDUPZ128rrkz,        X86::VMOVSHDUPZ128rmkz,        0 },
  { X86::VMOVSHDUPZ256rrkz,        X86::VMOVSHDUPZ256rmkz,        0 },
  { X86::VMOVSHDUPZrrkz,           X86::VMOVSHDUPZrmkz,           0 },
  { X86::VMOVSLDUPZ128rrkz,        X86::VMOVSLDUPZ128rmkz,        0 },
  { X86::VMOVSLDUPZ256rrkz,        X86::VMOVSLDUPZ256rmkz,        0 },
  { X86::VMOVSLDUPZrrkz,           X86::VMOVSLDUPZrmkz,           0 },
  { X86::VMOVUPDZ128rrkz,          X86::VMOVUPDZ128rmkz,          TB_NO_REVERSE },
  { X86::VMOVUPDZ256rrkz,          X86::VMOVUPDZ256rmkz,          TB_NO_REVERSE },
  { X86::VMOVUPDZrrkz,             X86::VMOVUPDZrmkz,             TB_NO_REVERSE },
  { X86::VMOVUPSZ128rrkz,          X86::VMOVUPSZ128rmkz,          TB_NO_REVERSE },
  { X86::VMOVUPSZ256rrkz,          X86::VMOVUPSZ256rmkz,          TB_NO_REVERSE },
  { X86::VMOVUPSZrrkz,             X86::VMOVUPSZrmkz,             TB_NO_REVERSE },
  { X86::VMPSADBWYrri,             X86::VMPSADBWYrmi,             0 },
  { X86::VMPSADBWrri,              X86::VMPSADBWrmi,              0 },
  { X86::VMULPDYrr,                X86::VMULPDYrm,                0 },
  { X86::VMULPDZ128rr,             X86::VMULPDZ128rm,             0 },
  { X86::VMULPDZ256rr,             X86::VMULPDZ256rm,             0 },
  { X86::VMULPDZrr,                X86::VMULPDZrm,                0 },
  { X86::VMULPDrr,                 X86::VMULPDrm,                 0 },
  { X86::VMULPSYrr,                X86::VMULPSYrm,                0 },
  { X86::VMULPSZ128rr,             X86::VMULPSZ128rm,             0 },
  { X86::VMULPSZ256rr,             X86::VMULPSZ256rm,             0 },
  { X86::VMULPSZrr,                X86::VMULPSZrm,                0 },
  { X86::VMULPSrr,                 X86::VMULPSrm,                 0 },
  { X86::VMULSDZrr,                X86::VMULSDZrm,                0 },
  { X86::VMULSDZrr_Int,            X86::VMULSDZrm_Int,            TB_NO_REVERSE },
  { X86::VMULSDrr,                 X86::VMULSDrm,                 0 },
  { X86::VMULSDrr_Int,             X86::VMULSDrm_Int,             TB_NO_REVERSE },
  { X86::VMULSSZrr,                X86::VMULSSZrm,                0 },
  { X86::VMULSSZrr_Int,            X86::VMULSSZrm_Int,            TB_NO_REVERSE },
  { X86::VMULSSrr,                 X86::VMULSSrm,                 0 },
  { X86::VMULSSrr_Int,             X86::VMULSSrm_Int,             TB_NO_REVERSE },
  { X86::VORPDYrr,                 X86::VORPDYrm,                 0 },
  { X86::VORPDZ128rr,              X86::VORPDZ128rm,              0 },
  { X86::VORPDZ256rr,              X86::VORPDZ256rm,              0 },
  { X86::VORPDZrr,                 X86::VORPDZrm,                 0 },
  { X86::VORPDrr,                  X86::VORPDrm,                  0 },
  { X86::VORPSYrr,                 X86::VORPSYrm,                 0 },
  { X86::VORPSZ128rr,              X86::VORPSZ128rm,              0 },
  { X86::VORPSZ256rr,              X86::VORPSZ256rm,              0 },
  { X86::VORPSZrr,                 X86::VORPSZrm,                 0 },
  { X86::VORPSrr,                  X86::VORPSrm,                  0 },
  { X86::VP2INTERSECTDZ128rr,      X86::VP2INTERSECTDZ128rm,      0 },
  { X86::VP2INTERSECTDZ256rr,      X86::VP2INTERSECTDZ256rm,      0 },
  { X86::VP2INTERSECTDZrr,         X86::VP2INTERSECTDZrm,         0 },
  { X86::VP2INTERSECTQZ128rr,      X86::VP2INTERSECTQZ128rm,      0 },
  { X86::VP2INTERSECTQZ256rr,      X86::VP2INTERSECTQZ256rm,      0 },
  { X86::VP2INTERSECTQZrr,         X86::VP2INTERSECTQZrm,         0 },
  { X86::VPABSBZ128rrkz,           X86::VPABSBZ128rmkz,           0 },
  { X86::VPABSBZ256rrkz,           X86::VPABSBZ256rmkz,           0 },
  { X86::VPABSBZrrkz,              X86::VPABSBZrmkz,              0 },
  { X86::VPABSDZ128rrkz,           X86::VPABSDZ128rmkz,           0 },
  { X86::VPABSDZ256rrkz,           X86::VPABSDZ256rmkz,           0 },
  { X86::VPABSDZrrkz,              X86::VPABSDZrmkz,              0 },
  { X86::VPABSQZ128rrkz,           X86::VPABSQZ128rmkz,           0 },
  { X86::VPABSQZ256rrkz,           X86::VPABSQZ256rmkz,           0 },
  { X86::VPABSQZrrkz,              X86::VPABSQZrmkz,              0 },
  { X86::VPABSWZ128rrkz,           X86::VPABSWZ128rmkz,           0 },
  { X86::VPABSWZ256rrkz,           X86::VPABSWZ256rmkz,           0 },
  { X86::VPABSWZrrkz,              X86::VPABSWZrmkz,              0 },
  { X86::VPACKSSDWYrr,             X86::VPACKSSDWYrm,             0 },
  { X86::VPACKSSDWZ128rr,          X86::VPACKSSDWZ128rm,          0 },
  { X86::VPACKSSDWZ256rr,          X86::VPACKSSDWZ256rm,          0 },
  { X86::VPACKSSDWZrr,             X86::VPACKSSDWZrm,             0 },
  { X86::VPACKSSDWrr,              X86::VPACKSSDWrm,              0 },
  { X86::VPACKSSWBYrr,             X86::VPACKSSWBYrm,             0 },
  { X86::VPACKSSWBZ128rr,          X86::VPACKSSWBZ128rm,          0 },
  { X86::VPACKSSWBZ256rr,          X86::VPACKSSWBZ256rm,          0 },
  { X86::VPACKSSWBZrr,             X86::VPACKSSWBZrm,             0 },
  { X86::VPACKSSWBrr,              X86::VPACKSSWBrm,              0 },
  { X86::VPACKUSDWYrr,             X86::VPACKUSDWYrm,             0 },
  { X86::VPACKUSDWZ128rr,          X86::VPACKUSDWZ128rm,          0 },
  { X86::VPACKUSDWZ256rr,          X86::VPACKUSDWZ256rm,          0 },
  { X86::VPACKUSDWZrr,             X86::VPACKUSDWZrm,             0 },
  { X86::VPACKUSDWrr,              X86::VPACKUSDWrm,              0 },
  { X86::VPACKUSWBYrr,             X86::VPACKUSWBYrm,             0 },
  { X86::VPACKUSWBZ128rr,          X86::VPACKUSWBZ128rm,          0 },
  { X86::VPACKUSWBZ256rr,          X86::VPACKUSWBZ256rm,          0 },
  { X86::VPACKUSWBZrr,             X86::VPACKUSWBZrm,             0 },
  { X86::VPACKUSWBrr,              X86::VPACKUSWBrm,              0 },
  { X86::VPADDBYrr,                X86::VPADDBYrm,                0 },
  { X86::VPADDBZ128rr,             X86::VPADDBZ128rm,             0 },
  { X86::VPADDBZ256rr,             X86::VPADDBZ256rm,             0 },
  { X86::VPADDBZrr,                X86::VPADDBZrm,                0 },
  { X86::VPADDBrr,                 X86::VPADDBrm,                 0 },
  { X86::VPADDDYrr,                X86::VPADDDYrm,                0 },
  { X86::VPADDDZ128rr,             X86::VPADDDZ128rm,             0 },
  { X86::VPADDDZ256rr,             X86::VPADDDZ256rm,             0 },
  { X86::VPADDDZrr,                X86::VPADDDZrm,                0 },
  { X86::VPADDDrr,                 X86::VPADDDrm,                 0 },
  { X86::VPADDQYrr,                X86::VPADDQYrm,                0 },
  { X86::VPADDQZ128rr,             X86::VPADDQZ128rm,             0 },
  { X86::VPADDQZ256rr,             X86::VPADDQZ256rm,             0 },
  { X86::VPADDQZrr,                X86::VPADDQZrm,                0 },
  { X86::VPADDQrr,                 X86::VPADDQrm,                 0 },
  { X86::VPADDSBYrr,               X86::VPADDSBYrm,               0 },
  { X86::VPADDSBZ128rr,            X86::VPADDSBZ128rm,            0 },
  { X86::VPADDSBZ256rr,            X86::VPADDSBZ256rm,            0 },
  { X86::VPADDSBZrr,               X86::VPADDSBZrm,               0 },
  { X86::VPADDSBrr,                X86::VPADDSBrm,                0 },
  { X86::VPADDSWYrr,               X86::VPADDSWYrm,               0 },
  { X86::VPADDSWZ128rr,            X86::VPADDSWZ128rm,            0 },
  { X86::VPADDSWZ256rr,            X86::VPADDSWZ256rm,            0 },
  { X86::VPADDSWZrr,               X86::VPADDSWZrm,               0 },
  { X86::VPADDSWrr,                X86::VPADDSWrm,                0 },
  { X86::VPADDUSBYrr,              X86::VPADDUSBYrm,              0 },
  { X86::VPADDUSBZ128rr,           X86::VPADDUSBZ128rm,           0 },
  { X86::VPADDUSBZ256rr,           X86::VPADDUSBZ256rm,           0 },
  { X86::VPADDUSBZrr,              X86::VPADDUSBZrm,              0 },
  { X86::VPADDUSBrr,               X86::VPADDUSBrm,               0 },
  { X86::VPADDUSWYrr,              X86::VPADDUSWYrm,              0 },
  { X86::VPADDUSWZ128rr,           X86::VPADDUSWZ128rm,           0 },
  { X86::VPADDUSWZ256rr,           X86::VPADDUSWZ256rm,           0 },
  { X86::VPADDUSWZrr,              X86::VPADDUSWZrm,              0 },
  { X86::VPADDUSWrr,               X86::VPADDUSWrm,               0 },
  { X86::VPADDWYrr,                X86::VPADDWYrm,                0 },
  { X86::VPADDWZ128rr,             X86::VPADDWZ128rm,             0 },
  { X86::VPADDWZ256rr,             X86::VPADDWZ256rm,             0 },
  { X86::VPADDWZrr,                X86::VPADDWZrm,                0 },
  { X86::VPADDWrr,                 X86::VPADDWrm,                 0 },
  { X86::VPALIGNRYrri,             X86::VPALIGNRYrmi,             0 },
  { X86::VPALIGNRZ128rri,          X86::VPALIGNRZ128rmi,          0 },
  { X86::VPALIGNRZ256rri,          X86::VPALIGNRZ256rmi,          0 },
  { X86::VPALIGNRZrri,             X86::VPALIGNRZrmi,             0 },
  { X86::VPALIGNRrri,              X86::VPALIGNRrmi,              0 },
  { X86::VPANDDZ128rr,             X86::VPANDDZ128rm,             0 },
  { X86::VPANDDZ256rr,             X86::VPANDDZ256rm,             0 },
  { X86::VPANDDZrr,                X86::VPANDDZrm,                0 },
  { X86::VPANDNDZ128rr,            X86::VPANDNDZ128rm,            0 },
  { X86::VPANDNDZ256rr,            X86::VPANDNDZ256rm,            0 },
  { X86::VPANDNDZrr,               X86::VPANDNDZrm,               0 },
  { X86::VPANDNQZ128rr,            X86::VPANDNQZ128rm,            0 },
  { X86::VPANDNQZ256rr,            X86::VPANDNQZ256rm,            0 },
  { X86::VPANDNQZrr,               X86::VPANDNQZrm,               0 },
  { X86::VPANDNYrr,                X86::VPANDNYrm,                0 },
  { X86::VPANDNrr,                 X86::VPANDNrm,                 0 },
  { X86::VPANDQZ128rr,             X86::VPANDQZ128rm,             0 },
  { X86::VPANDQZ256rr,             X86::VPANDQZ256rm,             0 },
  { X86::VPANDQZrr,                X86::VPANDQZrm,                0 },
  { X86::VPANDYrr,                 X86::VPANDYrm,                 0 },
  { X86::VPANDrr,                  X86::VPANDrm,                  0 },
  { X86::VPAVGBYrr,                X86::VPAVGBYrm,                0 },
  { X86::VPAVGBZ128rr,             X86::VPAVGBZ128rm,             0 },
  { X86::VPAVGBZ256rr,             X86::VPAVGBZ256rm,             0 },
  { X86::VPAVGBZrr,                X86::VPAVGBZrm,                0 },
  { X86::VPAVGBrr,                 X86::VPAVGBrm,                 0 },
  { X86::VPAVGWYrr,                X86::VPAVGWYrm,                0 },
  { X86::VPAVGWZ128rr,             X86::VPAVGWZ128rm,             0 },
  { X86::VPAVGWZ256rr,             X86::VPAVGWZ256rm,             0 },
  { X86::VPAVGWZrr,                X86::VPAVGWZrm,                0 },
  { X86::VPAVGWrr,                 X86::VPAVGWrm,                 0 },
  { X86::VPBLENDDYrri,             X86::VPBLENDDYrmi,             0 },
  { X86::VPBLENDDrri,              X86::VPBLENDDrmi,              0 },
  { X86::VPBLENDMBZ128rr,          X86::VPBLENDMBZ128rm,          0 },
  { X86::VPBLENDMBZ256rr,          X86::VPBLENDMBZ256rm,          0 },
  { X86::VPBLENDMBZrr,             X86::VPBLENDMBZrm,             0 },
  { X86::VPBLENDMDZ128rr,          X86::VPBLENDMDZ128rm,          0 },
  { X86::VPBLENDMDZ256rr,          X86::VPBLENDMDZ256rm,          0 },
  { X86::VPBLENDMDZrr,             X86::VPBLENDMDZrm,             0 },
  { X86::VPBLENDMQZ128rr,          X86::VPBLENDMQZ128rm,          0 },
  { X86::VPBLENDMQZ256rr,          X86::VPBLENDMQZ256rm,          0 },
  { X86::VPBLENDMQZrr,             X86::VPBLENDMQZrm,             0 },
  { X86::VPBLENDMWZ128rr,          X86::VPBLENDMWZ128rm,          0 },
  { X86::VPBLENDMWZ256rr,          X86::VPBLENDMWZ256rm,          0 },
  { X86::VPBLENDMWZrr,             X86::VPBLENDMWZrm,             0 },
  { X86::VPBLENDVBYrr,             X86::VPBLENDVBYrm,             0 },
  { X86::VPBLENDVBrr,              X86::VPBLENDVBrm,              0 },
  { X86::VPBLENDWYrri,             X86::VPBLENDWYrmi,             0 },
  { X86::VPBLENDWrri,              X86::VPBLENDWrmi,              0 },
  { X86::VPBROADCASTBZ128rkz,      X86::VPBROADCASTBZ128mkz,      TB_NO_REVERSE },
  { X86::VPBROADCASTBZ256rkz,      X86::VPBROADCASTBZ256mkz,      TB_NO_REVERSE },
  { X86::VPBROADCASTBZrkz,         X86::VPBROADCASTBZmkz,         TB_NO_REVERSE },
  { X86::VPBROADCASTDZ128rkz,      X86::VPBROADCASTDZ128mkz,      TB_NO_REVERSE },
  { X86::VPBROADCASTDZ256rkz,      X86::VPBROADCASTDZ256mkz,      TB_NO_REVERSE },
  { X86::VPBROADCASTDZrkz,         X86::VPBROADCASTDZmkz,         TB_NO_REVERSE },
  { X86::VPBROADCASTQZ128rkz,      X86::VPBROADCASTQZ128mkz,      TB_NO_REVERSE },
  { X86::VPBROADCASTQZ256rkz,      X86::VPBROADCASTQZ256mkz,      TB_NO_REVERSE },
  { X86::VPBROADCASTQZrkz,         X86::VPBROADCASTQZmkz,         TB_NO_REVERSE },
  { X86::VPBROADCASTWZ128rkz,      X86::VPBROADCASTWZ128mkz,      TB_NO_REVERSE },
  { X86::VPBROADCASTWZ256rkz,      X86::VPBROADCASTWZ256mkz,      TB_NO_REVERSE },
  { X86::VPBROADCASTWZrkz,         X86::VPBROADCASTWZmkz,         TB_NO_REVERSE },
  { X86::VPCLMULQDQYrr,            X86::VPCLMULQDQYrm,            0 },
  { X86::VPCLMULQDQZ128rr,         X86::VPCLMULQDQZ128rm,         0 },
  { X86::VPCLMULQDQZ256rr,         X86::VPCLMULQDQZ256rm,         0 },
  { X86::VPCLMULQDQZrr,            X86::VPCLMULQDQZrm,            0 },
  { X86::VPCLMULQDQrr,             X86::VPCLMULQDQrm,             0 },
  { X86::VPCMOVYrrr,               X86::VPCMOVYrmr,               0 },
  { X86::VPCMOVrrr,                X86::VPCMOVrmr,                0 },
  { X86::VPCMPBZ128rri,            X86::VPCMPBZ128rmi,            0 },
  { X86::VPCMPBZ256rri,            X86::VPCMPBZ256rmi,            0 },
  { X86::VPCMPBZrri,               X86::VPCMPBZrmi,               0 },
  { X86::VPCMPDZ128rri,            X86::VPCMPDZ128rmi,            0 },
  { X86::VPCMPDZ256rri,            X86::VPCMPDZ256rmi,            0 },
  { X86::VPCMPDZrri,               X86::VPCMPDZrmi,               0 },
  { X86::VPCMPEQBYrr,              X86::VPCMPEQBYrm,              0 },
  { X86::VPCMPEQBZ128rr,           X86::VPCMPEQBZ128rm,           0 },
  { X86::VPCMPEQBZ256rr,           X86::VPCMPEQBZ256rm,           0 },
  { X86::VPCMPEQBZrr,              X86::VPCMPEQBZrm,              0 },
  { X86::VPCMPEQBrr,               X86::VPCMPEQBrm,               0 },
  { X86::VPCMPEQDYrr,              X86::VPCMPEQDYrm,              0 },
  { X86::VPCMPEQDZ128rr,           X86::VPCMPEQDZ128rm,           0 },
  { X86::VPCMPEQDZ256rr,           X86::VPCMPEQDZ256rm,           0 },
  { X86::VPCMPEQDZrr,              X86::VPCMPEQDZrm,              0 },
  { X86::VPCMPEQDrr,               X86::VPCMPEQDrm,               0 },
  { X86::VPCMPEQQYrr,              X86::VPCMPEQQYrm,              0 },
  { X86::VPCMPEQQZ128rr,           X86::VPCMPEQQZ128rm,           0 },
  { X86::VPCMPEQQZ256rr,           X86::VPCMPEQQZ256rm,           0 },
  { X86::VPCMPEQQZrr,              X86::VPCMPEQQZrm,              0 },
  { X86::VPCMPEQQrr,               X86::VPCMPEQQrm,               0 },
  { X86::VPCMPEQWYrr,              X86::VPCMPEQWYrm,              0 },
  { X86::VPCMPEQWZ128rr,           X86::VPCMPEQWZ128rm,           0 },
  { X86::VPCMPEQWZ256rr,           X86::VPCMPEQWZ256rm,           0 },
  { X86::VPCMPEQWZrr,              X86::VPCMPEQWZrm,              0 },
  { X86::VPCMPEQWrr,               X86::VPCMPEQWrm,               0 },
  { X86::VPCMPGTBYrr,              X86::VPCMPGTBYrm,              0 },
  { X86::VPCMPGTBZ128rr,           X86::VPCMPGTBZ128rm,           0 },
  { X86::VPCMPGTBZ256rr,           X86::VPCMPGTBZ256rm,           0 },
  { X86::VPCMPGTBZrr,              X86::VPCMPGTBZrm,              0 },
  { X86::VPCMPGTBrr,               X86::VPCMPGTBrm,               0 },
  { X86::VPCMPGTDYrr,              X86::VPCMPGTDYrm,              0 },
  { X86::VPCMPGTDZ128rr,           X86::VPCMPGTDZ128rm,           0 },
  { X86::VPCMPGTDZ256rr,           X86::VPCMPGTDZ256rm,           0 },
  { X86::VPCMPGTDZrr,              X86::VPCMPGTDZrm,              0 },
  { X86::VPCMPGTDrr,               X86::VPCMPGTDrm,               0 },
  { X86::VPCMPGTQYrr,              X86::VPCMPGTQYrm,              0 },
  { X86::VPCMPGTQZ128rr,           X86::VPCMPGTQZ128rm,           0 },
  { X86::VPCMPGTQZ256rr,           X86::VPCMPGTQZ256rm,           0 },
  { X86::VPCMPGTQZrr,              X86::VPCMPGTQZrm,              0 },
  { X86::VPCMPGTQrr,               X86::VPCMPGTQrm,               0 },
  { X86::VPCMPGTWYrr,              X86::VPCMPGTWYrm,              0 },
  { X86::VPCMPGTWZ128rr,           X86::VPCMPGTWZ128rm,           0 },
  { X86::VPCMPGTWZ256rr,           X86::VPCMPGTWZ256rm,           0 },
  { X86::VPCMPGTWZrr,              X86::VPCMPGTWZrm,              0 },
  { X86::VPCMPGTWrr,               X86::VPCMPGTWrm,               0 },
  { X86::VPCMPQZ128rri,            X86::VPCMPQZ128rmi,            0 },
  { X86::VPCMPQZ256rri,            X86::VPCMPQZ256rmi,            0 },
  { X86::VPCMPQZrri,               X86::VPCMPQZrmi,               0 },
  { X86::VPCMPUBZ128rri,           X86::VPCMPUBZ128rmi,           0 },
  { X86::VPCMPUBZ256rri,           X86::VPCMPUBZ256rmi,           0 },
  { X86::VPCMPUBZrri,              X86::VPCMPUBZrmi,              0 },
  { X86::VPCMPUDZ128rri,           X86::VPCMPUDZ128rmi,           0 },
  { X86::VPCMPUDZ256rri,           X86::VPCMPUDZ256rmi,           0 },
  { X86::VPCMPUDZrri,              X86::VPCMPUDZrmi,              0 },
  { X86::VPCMPUQZ128rri,           X86::VPCMPUQZ128rmi,           0 },
  { X86::VPCMPUQZ256rri,           X86::VPCMPUQZ256rmi,           0 },
  { X86::VPCMPUQZrri,              X86::VPCMPUQZrmi,              0 },
  { X86::VPCMPUWZ128rri,           X86::VPCMPUWZ128rmi,           0 },
  { X86::VPCMPUWZ256rri,           X86::VPCMPUWZ256rmi,           0 },
  { X86::VPCMPUWZrri,              X86::VPCMPUWZrmi,              0 },
  { X86::VPCMPWZ128rri,            X86::VPCMPWZ128rmi,            0 },
  { X86::VPCMPWZ256rri,            X86::VPCMPWZ256rmi,            0 },
  { X86::VPCMPWZrri,               X86::VPCMPWZrmi,               0 },
  { X86::VPCOMBri,                 X86::VPCOMBmi,                 0 },
  { X86::VPCOMDri,                 X86::VPCOMDmi,                 0 },
  { X86::VPCOMQri,                 X86::VPCOMQmi,                 0 },
  { X86::VPCOMUBri,                X86::VPCOMUBmi,                0 },
  { X86::VPCOMUDri,                X86::VPCOMUDmi,                0 },
  { X86::VPCOMUQri,                X86::VPCOMUQmi,                0 },
  { X86::VPCOMUWri,                X86::VPCOMUWmi,                0 },
  { X86::VPCOMWri,                 X86::VPCOMWmi,                 0 },
  { X86::VPCONFLICTDZ128rrkz,      X86::VPCONFLICTDZ128rmkz,      0 },
  { X86::VPCONFLICTDZ256rrkz,      X86::VPCONFLICTDZ256rmkz,      0 },
  { X86::VPCONFLICTDZrrkz,         X86::VPCONFLICTDZrmkz,         0 },
  { X86::VPCONFLICTQZ128rrkz,      X86::VPCONFLICTQZ128rmkz,      0 },
  { X86::VPCONFLICTQZ256rrkz,      X86::VPCONFLICTQZ256rmkz,      0 },
  { X86::VPCONFLICTQZrrkz,         X86::VPCONFLICTQZrmkz,         0 },
  { X86::VPERM2F128rr,             X86::VPERM2F128rm,             0 },
  { X86::VPERM2I128rr,             X86::VPERM2I128rm,             0 },
  { X86::VPERMBZ128rr,             X86::VPERMBZ128rm,             0 },
  { X86::VPERMBZ256rr,             X86::VPERMBZ256rm,             0 },
  { X86::VPERMBZrr,                X86::VPERMBZrm,                0 },
  { X86::VPERMDYrr,                X86::VPERMDYrm,                0 },
  { X86::VPERMDZ256rr,             X86::VPERMDZ256rm,             0 },
  { X86::VPERMDZrr,                X86::VPERMDZrm,                0 },
  { X86::VPERMIL2PDYrr,            X86::VPERMIL2PDYmr,            0 },
  { X86::VPERMIL2PDrr,             X86::VPERMIL2PDmr,             0 },
  { X86::VPERMIL2PSYrr,            X86::VPERMIL2PSYmr,            0 },
  { X86::VPERMIL2PSrr,             X86::VPERMIL2PSmr,             0 },
  { X86::VPERMILPDYrr,             X86::VPERMILPDYrm,             0 },
  { X86::VPERMILPDZ128rikz,        X86::VPERMILPDZ128mikz,        0 },
  { X86::VPERMILPDZ128rr,          X86::VPERMILPDZ128rm,          0 },
  { X86::VPERMILPDZ256rikz,        X86::VPERMILPDZ256mikz,        0 },
  { X86::VPERMILPDZ256rr,          X86::VPERMILPDZ256rm,          0 },
  { X86::VPERMILPDZrikz,           X86::VPERMILPDZmikz,           0 },
  { X86::VPERMILPDZrr,             X86::VPERMILPDZrm,             0 },
  { X86::VPERMILPDrr,              X86::VPERMILPDrm,              0 },
  { X86::VPERMILPSYrr,             X86::VPERMILPSYrm,             0 },
  { X86::VPERMILPSZ128rikz,        X86::VPERMILPSZ128mikz,        0 },
  { X86::VPERMILPSZ128rr,          X86::VPERMILPSZ128rm,          0 },
  { X86::VPERMILPSZ256rikz,        X86::VPERMILPSZ256mikz,        0 },
  { X86::VPERMILPSZ256rr,          X86::VPERMILPSZ256rm,          0 },
  { X86::VPERMILPSZrikz,           X86::VPERMILPSZmikz,           0 },
  { X86::VPERMILPSZrr,             X86::VPERMILPSZrm,             0 },
  { X86::VPERMILPSrr,              X86::VPERMILPSrm,              0 },
  { X86::VPERMPDZ256rikz,          X86::VPERMPDZ256mikz,          0 },
  { X86::VPERMPDZ256rr,            X86::VPERMPDZ256rm,            0 },
  { X86::VPERMPDZrikz,             X86::VPERMPDZmikz,             0 },
  { X86::VPERMPDZrr,               X86::VPERMPDZrm,               0 },
  { X86::VPERMPSYrr,               X86::VPERMPSYrm,               0 },
  { X86::VPERMPSZ256rr,            X86::VPERMPSZ256rm,            0 },
  { X86::VPERMPSZrr,               X86::VPERMPSZrm,               0 },
  { X86::VPERMQZ256rikz,           X86::VPERMQZ256mikz,           0 },
  { X86::VPERMQZ256rr,             X86::VPERMQZ256rm,             0 },
  { X86::VPERMQZrikz,              X86::VPERMQZmikz,              0 },
  { X86::VPERMQZrr,                X86::VPERMQZrm,                0 },
  { X86::VPERMWZ128rr,             X86::VPERMWZ128rm,             0 },
  { X86::VPERMWZ256rr,             X86::VPERMWZ256rm,             0 },
  { X86::VPERMWZrr,                X86::VPERMWZrm,                0 },
  { X86::VPEXPANDBZ128rrkz,        X86::VPEXPANDBZ128rmkz,        TB_NO_REVERSE },
  { X86::VPEXPANDBZ256rrkz,        X86::VPEXPANDBZ256rmkz,        TB_NO_REVERSE },
  { X86::VPEXPANDBZrrkz,           X86::VPEXPANDBZrmkz,           TB_NO_REVERSE },
  { X86::VPEXPANDDZ128rrkz,        X86::VPEXPANDDZ128rmkz,        TB_NO_REVERSE },
  { X86::VPEXPANDDZ256rrkz,        X86::VPEXPANDDZ256rmkz,        TB_NO_REVERSE },
  { X86::VPEXPANDDZrrkz,           X86::VPEXPANDDZrmkz,           TB_NO_REVERSE },
  { X86::VPEXPANDQZ128rrkz,        X86::VPEXPANDQZ128rmkz,        TB_NO_REVERSE },
  { X86::VPEXPANDQZ256rrkz,        X86::VPEXPANDQZ256rmkz,        TB_NO_REVERSE },
  { X86::VPEXPANDQZrrkz,           X86::VPEXPANDQZrmkz,           TB_NO_REVERSE },
  { X86::VPEXPANDWZ128rrkz,        X86::VPEXPANDWZ128rmkz,        TB_NO_REVERSE },
  { X86::VPEXPANDWZ256rrkz,        X86::VPEXPANDWZ256rmkz,        TB_NO_REVERSE },
  { X86::VPEXPANDWZrrkz,           X86::VPEXPANDWZrmkz,           TB_NO_REVERSE },
  { X86::VPHADDDYrr,               X86::VPHADDDYrm,               0 },
  { X86::VPHADDDrr,                X86::VPHADDDrm,                0 },
  { X86::VPHADDSWYrr,              X86::VPHADDSWYrm,              0 },
  { X86::VPHADDSWrr,               X86::VPHADDSWrm,               0 },
  { X86::VPHADDWYrr,               X86::VPHADDWYrm,               0 },
  { X86::VPHADDWrr,                X86::VPHADDWrm,                0 },
  { X86::VPHSUBDYrr,               X86::VPHSUBDYrm,               0 },
  { X86::VPHSUBDrr,                X86::VPHSUBDrm,                0 },
  { X86::VPHSUBSWYrr,              X86::VPHSUBSWYrm,              0 },
  { X86::VPHSUBSWrr,               X86::VPHSUBSWrm,               0 },
  { X86::VPHSUBWYrr,               X86::VPHSUBWYrm,               0 },
  { X86::VPHSUBWrr,                X86::VPHSUBWrm,                0 },
  { X86::VPINSRBZrr,               X86::VPINSRBZrm,               TB_NO_REVERSE },
  { X86::VPINSRBrr,                X86::VPINSRBrm,                TB_NO_REVERSE },
  { X86::VPINSRDZrr,               X86::VPINSRDZrm,               0 },
  { X86::VPINSRDrr,                X86::VPINSRDrm,                0 },
  { X86::VPINSRQZrr,               X86::VPINSRQZrm,               0 },
  { X86::VPINSRQrr,                X86::VPINSRQrm,                0 },
  { X86::VPINSRWZrr,               X86::VPINSRWZrm,               TB_NO_REVERSE },
  { X86::VPINSRWrr,                X86::VPINSRWrm,                TB_NO_REVERSE },
  { X86::VPLZCNTDZ128rrkz,         X86::VPLZCNTDZ128rmkz,         0 },
  { X86::VPLZCNTDZ256rrkz,         X86::VPLZCNTDZ256rmkz,         0 },
  { X86::VPLZCNTDZrrkz,            X86::VPLZCNTDZrmkz,            0 },
  { X86::VPLZCNTQZ128rrkz,         X86::VPLZCNTQZ128rmkz,         0 },
  { X86::VPLZCNTQZ256rrkz,         X86::VPLZCNTQZ256rmkz,         0 },
  { X86::VPLZCNTQZrrkz,            X86::VPLZCNTQZrmkz,            0 },
  { X86::VPMACSDDrr,               X86::VPMACSDDrm,               0 },
  { X86::VPMACSDQHrr,              X86::VPMACSDQHrm,              0 },
  { X86::VPMACSDQLrr,              X86::VPMACSDQLrm,              0 },
  { X86::VPMACSSDDrr,              X86::VPMACSSDDrm,              0 },
  { X86::VPMACSSDQHrr,             X86::VPMACSSDQHrm,             0 },
  { X86::VPMACSSDQLrr,             X86::VPMACSSDQLrm,             0 },
  { X86::VPMACSSWDrr,              X86::VPMACSSWDrm,              0 },
  { X86::VPMACSSWWrr,              X86::VPMACSSWWrm,              0 },
  { X86::VPMACSWDrr,               X86::VPMACSWDrm,               0 },
  { X86::VPMACSWWrr,               X86::VPMACSWWrm,               0 },
  { X86::VPMADCSSWDrr,             X86::VPMADCSSWDrm,             0 },
  { X86::VPMADCSWDrr,              X86::VPMADCSWDrm,              0 },
  { X86::VPMADDUBSWYrr,            X86::VPMADDUBSWYrm,            0 },
  { X86::VPMADDUBSWZ128rr,         X86::VPMADDUBSWZ128rm,         0 },
  { X86::VPMADDUBSWZ256rr,         X86::VPMADDUBSWZ256rm,         0 },
  { X86::VPMADDUBSWZrr,            X86::VPMADDUBSWZrm,            0 },
  { X86::VPMADDUBSWrr,             X86::VPMADDUBSWrm,             0 },
  { X86::VPMADDWDYrr,              X86::VPMADDWDYrm,              0 },
  { X86::VPMADDWDZ128rr,           X86::VPMADDWDZ128rm,           0 },
  { X86::VPMADDWDZ256rr,           X86::VPMADDWDZ256rm,           0 },
  { X86::VPMADDWDZrr,              X86::VPMADDWDZrm,              0 },
  { X86::VPMADDWDrr,               X86::VPMADDWDrm,               0 },
  { X86::VPMAXSBYrr,               X86::VPMAXSBYrm,               0 },
  { X86::VPMAXSBZ128rr,            X86::VPMAXSBZ128rm,            0 },
  { X86::VPMAXSBZ256rr,            X86::VPMAXSBZ256rm,            0 },
  { X86::VPMAXSBZrr,               X86::VPMAXSBZrm,               0 },
  { X86::VPMAXSBrr,                X86::VPMAXSBrm,                0 },
  { X86::VPMAXSDYrr,               X86::VPMAXSDYrm,               0 },
  { X86::VPMAXSDZ128rr,            X86::VPMAXSDZ128rm,            0 },
  { X86::VPMAXSDZ256rr,            X86::VPMAXSDZ256rm,            0 },
  { X86::VPMAXSDZrr,               X86::VPMAXSDZrm,               0 },
  { X86::VPMAXSDrr,                X86::VPMAXSDrm,                0 },
  { X86::VPMAXSQZ128rr,            X86::VPMAXSQZ128rm,            0 },
  { X86::VPMAXSQZ256rr,            X86::VPMAXSQZ256rm,            0 },
  { X86::VPMAXSQZrr,               X86::VPMAXSQZrm,               0 },
  { X86::VPMAXSWYrr,               X86::VPMAXSWYrm,               0 },
  { X86::VPMAXSWZ128rr,            X86::VPMAXSWZ128rm,            0 },
  { X86::VPMAXSWZ256rr,            X86::VPMAXSWZ256rm,            0 },
  { X86::VPMAXSWZrr,               X86::VPMAXSWZrm,               0 },
  { X86::VPMAXSWrr,                X86::VPMAXSWrm,                0 },
  { X86::VPMAXUBYrr,               X86::VPMAXUBYrm,               0 },
  { X86::VPMAXUBZ128rr,            X86::VPMAXUBZ128rm,            0 },
  { X86::VPMAXUBZ256rr,            X86::VPMAXUBZ256rm,            0 },
  { X86::VPMAXUBZrr,               X86::VPMAXUBZrm,               0 },
  { X86::VPMAXUBrr,                X86::VPMAXUBrm,                0 },
  { X86::VPMAXUDYrr,               X86::VPMAXUDYrm,               0 },
  { X86::VPMAXUDZ128rr,            X86::VPMAXUDZ128rm,            0 },
  { X86::VPMAXUDZ256rr,            X86::VPMAXUDZ256rm,            0 },
  { X86::VPMAXUDZrr,               X86::VPMAXUDZrm,               0 },
  { X86::VPMAXUDrr,                X86::VPMAXUDrm,                0 },
  { X86::VPMAXUQZ128rr,            X86::VPMAXUQZ128rm,            0 },
  { X86::VPMAXUQZ256rr,            X86::VPMAXUQZ256rm,            0 },
  { X86::VPMAXUQZrr,               X86::VPMAXUQZrm,               0 },
  { X86::VPMAXUWYrr,               X86::VPMAXUWYrm,               0 },
  { X86::VPMAXUWZ128rr,            X86::VPMAXUWZ128rm,            0 },
  { X86::VPMAXUWZ256rr,            X86::VPMAXUWZ256rm,            0 },
  { X86::VPMAXUWZrr,               X86::VPMAXUWZrm,               0 },
  { X86::VPMAXUWrr,                X86::VPMAXUWrm,                0 },
  { X86::VPMINSBYrr,               X86::VPMINSBYrm,               0 },
  { X86::VPMINSBZ128rr,            X86::VPMINSBZ128rm,            0 },
  { X86::VPMINSBZ256rr,            X86::VPMINSBZ256rm,            0 },
  { X86::VPMINSBZrr,               X86::VPMINSBZrm,               0 },
  { X86::VPMINSBrr,                X86::VPMINSBrm,                0 },
  { X86::VPMINSDYrr,               X86::VPMINSDYrm,               0 },
  { X86::VPMINSDZ128rr,            X86::VPMINSDZ128rm,            0 },
  { X86::VPMINSDZ256rr,            X86::VPMINSDZ256rm,            0 },
  { X86::VPMINSDZrr,               X86::VPMINSDZrm,               0 },
  { X86::VPMINSDrr,                X86::VPMINSDrm,                0 },
  { X86::VPMINSQZ128rr,            X86::VPMINSQZ128rm,            0 },
  { X86::VPMINSQZ256rr,            X86::VPMINSQZ256rm,            0 },
  { X86::VPMINSQZrr,               X86::VPMINSQZrm,               0 },
  { X86::VPMINSWYrr,               X86::VPMINSWYrm,               0 },
  { X86::VPMINSWZ128rr,            X86::VPMINSWZ128rm,            0 },
  { X86::VPMINSWZ256rr,            X86::VPMINSWZ256rm,            0 },
  { X86::VPMINSWZrr,               X86::VPMINSWZrm,               0 },
  { X86::VPMINSWrr,                X86::VPMINSWrm,                0 },
  { X86::VPMINUBYrr,               X86::VPMINUBYrm,               0 },
  { X86::VPMINUBZ128rr,            X86::VPMINUBZ128rm,            0 },
  { X86::VPMINUBZ256rr,            X86::VPMINUBZ256rm,            0 },
  { X86::VPMINUBZrr,               X86::VPMINUBZrm,               0 },
  { X86::VPMINUBrr,                X86::VPMINUBrm,                0 },
  { X86::VPMINUDYrr,               X86::VPMINUDYrm,               0 },
  { X86::VPMINUDZ128rr,            X86::VPMINUDZ128rm,            0 },
  { X86::VPMINUDZ256rr,            X86::VPMINUDZ256rm,            0 },
  { X86::VPMINUDZrr,               X86::VPMINUDZrm,               0 },
  { X86::VPMINUDrr,                X86::VPMINUDrm,                0 },
  { X86::VPMINUQZ128rr,            X86::VPMINUQZ128rm,            0 },
  { X86::VPMINUQZ256rr,            X86::VPMINUQZ256rm,            0 },
  { X86::VPMINUQZrr,               X86::VPMINUQZrm,               0 },
  { X86::VPMINUWYrr,               X86::VPMINUWYrm,               0 },
  { X86::VPMINUWZ128rr,            X86::VPMINUWZ128rm,            0 },
  { X86::VPMINUWZ256rr,            X86::VPMINUWZ256rm,            0 },
  { X86::VPMINUWZrr,               X86::VPMINUWZrm,               0 },
  { X86::VPMINUWrr,                X86::VPMINUWrm,                0 },
  { X86::VPMOVSXBDZ128rrkz,        X86::VPMOVSXBDZ128rmkz,        TB_NO_REVERSE },
  { X86::VPMOVSXBDZ256rrkz,        X86::VPMOVSXBDZ256rmkz,        TB_NO_REVERSE },
  { X86::VPMOVSXBDZrrkz,           X86::VPMOVSXBDZrmkz,           0 },
  { X86::VPMOVSXBQZ128rrkz,        X86::VPMOVSXBQZ128rmkz,        TB_NO_REVERSE },
  { X86::VPMOVSXBQZ256rrkz,        X86::VPMOVSXBQZ256rmkz,        TB_NO_REVERSE },
  { X86::VPMOVSXBQZrrkz,           X86::VPMOVSXBQZrmkz,           TB_NO_REVERSE },
  { X86::VPMOVSXBWZ128rrkz,        X86::VPMOVSXBWZ128rmkz,        TB_NO_REVERSE },
  { X86::VPMOVSXBWZ256rrkz,        X86::VPMOVSXBWZ256rmkz,        0 },
  { X86::VPMOVSXBWZrrkz,           X86::VPMOVSXBWZrmkz,           0 },
  { X86::VPMOVSXDQZ128rrkz,        X86::VPMOVSXDQZ128rmkz,        TB_NO_REVERSE },
  { X86::VPMOVSXDQZ256rrkz,        X86::VPMOVSXDQZ256rmkz,        0 },
  { X86::VPMOVSXDQZrrkz,           X86::VPMOVSXDQZrmkz,           0 },
  { X86::VPMOVSXWDZ128rrkz,        X86::VPMOVSXWDZ128rmkz,        TB_NO_REVERSE },
  { X86::VPMOVSXWDZ256rrkz,        X86::VPMOVSXWDZ256rmkz,        0 },
  { X86::VPMOVSXWDZrrkz,           X86::VPMOVSXWDZrmkz,           0 },
  { X86::VPMOVSXWQZ128rrkz,        X86::VPMOVSXWQZ128rmkz,        TB_NO_REVERSE },
  { X86::VPMOVSXWQZ256rrkz,        X86::VPMOVSXWQZ256rmkz,        TB_NO_REVERSE },
  { X86::VPMOVSXWQZrrkz,           X86::VPMOVSXWQZrmkz,           0 },
  { X86::VPMOVZXBDZ128rrkz,        X86::VPMOVZXBDZ128rmkz,        TB_NO_REVERSE },
  { X86::VPMOVZXBDZ256rrkz,        X86::VPMOVZXBDZ256rmkz,        TB_NO_REVERSE },
  { X86::VPMOVZXBDZrrkz,           X86::VPMOVZXBDZrmkz,           0 },
  { X86::VPMOVZXBQZ128rrkz,        X86::VPMOVZXBQZ128rmkz,        TB_NO_REVERSE },
  { X86::VPMOVZXBQZ256rrkz,        X86::VPMOVZXBQZ256rmkz,        TB_NO_REVERSE },
  { X86::VPMOVZXBQZrrkz,           X86::VPMOVZXBQZrmkz,           TB_NO_REVERSE },
  { X86::VPMOVZXBWZ128rrkz,        X86::VPMOVZXBWZ128rmkz,        TB_NO_REVERSE },
  { X86::VPMOVZXBWZ256rrkz,        X86::VPMOVZXBWZ256rmkz,        0 },
  { X86::VPMOVZXBWZrrkz,           X86::VPMOVZXBWZrmkz,           0 },
  { X86::VPMOVZXDQZ128rrkz,        X86::VPMOVZXDQZ128rmkz,        TB_NO_REVERSE },
  { X86::VPMOVZXDQZ256rrkz,        X86::VPMOVZXDQZ256rmkz,        0 },
  { X86::VPMOVZXDQZrrkz,           X86::VPMOVZXDQZrmkz,           0 },
  { X86::VPMOVZXWDZ128rrkz,        X86::VPMOVZXWDZ128rmkz,        TB_NO_REVERSE },
  { X86::VPMOVZXWDZ256rrkz,        X86::VPMOVZXWDZ256rmkz,        0 },
  { X86::VPMOVZXWDZrrkz,           X86::VPMOVZXWDZrmkz,           0 },
  { X86::VPMOVZXWQZ128rrkz,        X86::VPMOVZXWQZ128rmkz,        TB_NO_REVERSE },
  { X86::VPMOVZXWQZ256rrkz,        X86::VPMOVZXWQZ256rmkz,        TB_NO_REVERSE },
  { X86::VPMOVZXWQZrrkz,           X86::VPMOVZXWQZrmkz,           0 },
  { X86::VPMULDQYrr,               X86::VPMULDQYrm,               0 },
  { X86::VPMULDQZ128rr,            X86::VPMULDQZ128rm,            0 },
  { X86::VPMULDQZ256rr,            X86::VPMULDQZ256rm,            0 },
  { X86::VPMULDQZrr,               X86::VPMULDQZrm,               0 },
  { X86::VPMULDQrr,                X86::VPMULDQrm,                0 },
  { X86::VPMULHRSWYrr,             X86::VPMULHRSWYrm,             0 },
  { X86::VPMULHRSWZ128rr,          X86::VPMULHRSWZ128rm,          0 },
  { X86::VPMULHRSWZ256rr,          X86::VPMULHRSWZ256rm,          0 },
  { X86::VPMULHRSWZrr,             X86::VPMULHRSWZrm,             0 },
  { X86::VPMULHRSWrr,              X86::VPMULHRSWrm,              0 },
  { X86::VPMULHUWYrr,              X86::VPMULHUWYrm,              0 },
  { X86::VPMULHUWZ128rr,           X86::VPMULHUWZ128rm,           0 },
  { X86::VPMULHUWZ256rr,           X86::VPMULHUWZ256rm,           0 },
  { X86::VPMULHUWZrr,              X86::VPMULHUWZrm,              0 },
  { X86::VPMULHUWrr,               X86::VPMULHUWrm,               0 },
  { X86::VPMULHWYrr,               X86::VPMULHWYrm,               0 },
  { X86::VPMULHWZ128rr,            X86::VPMULHWZ128rm,            0 },
  { X86::VPMULHWZ256rr,            X86::VPMULHWZ256rm,            0 },
  { X86::VPMULHWZrr,               X86::VPMULHWZrm,               0 },
  { X86::VPMULHWrr,                X86::VPMULHWrm,                0 },
  { X86::VPMULLDYrr,               X86::VPMULLDYrm,               0 },
  { X86::VPMULLDZ128rr,            X86::VPMULLDZ128rm,            0 },
  { X86::VPMULLDZ256rr,            X86::VPMULLDZ256rm,            0 },
  { X86::VPMULLDZrr,               X86::VPMULLDZrm,               0 },
  { X86::VPMULLDrr,                X86::VPMULLDrm,                0 },
  { X86::VPMULLQZ128rr,            X86::VPMULLQZ128rm,            0 },
  { X86::VPMULLQZ256rr,            X86::VPMULLQZ256rm,            0 },
  { X86::VPMULLQZrr,               X86::VPMULLQZrm,               0 },
  { X86::VPMULLWYrr,               X86::VPMULLWYrm,               0 },
  { X86::VPMULLWZ128rr,            X86::VPMULLWZ128rm,            0 },
  { X86::VPMULLWZ256rr,            X86::VPMULLWZ256rm,            0 },
  { X86::VPMULLWZrr,               X86::VPMULLWZrm,               0 },
  { X86::VPMULLWrr,                X86::VPMULLWrm,                0 },
  { X86::VPMULTISHIFTQBZ128rr,     X86::VPMULTISHIFTQBZ128rm,     0 },
  { X86::VPMULTISHIFTQBZ256rr,     X86::VPMULTISHIFTQBZ256rm,     0 },
  { X86::VPMULTISHIFTQBZrr,        X86::VPMULTISHIFTQBZrm,        0 },
  { X86::VPMULUDQYrr,              X86::VPMULUDQYrm,              0 },
  { X86::VPMULUDQZ128rr,           X86::VPMULUDQZ128rm,           0 },
  { X86::VPMULUDQZ256rr,           X86::VPMULUDQZ256rm,           0 },
  { X86::VPMULUDQZrr,              X86::VPMULUDQZrm,              0 },
  { X86::VPMULUDQrr,               X86::VPMULUDQrm,               0 },
  { X86::VPOPCNTBZ128rrkz,         X86::VPOPCNTBZ128rmkz,         0 },
  { X86::VPOPCNTBZ256rrkz,         X86::VPOPCNTBZ256rmkz,         0 },
  { X86::VPOPCNTBZrrkz,            X86::VPOPCNTBZrmkz,            0 },
  { X86::VPOPCNTDZ128rrkz,         X86::VPOPCNTDZ128rmkz,         0 },
  { X86::VPOPCNTDZ256rrkz,         X86::VPOPCNTDZ256rmkz,         0 },
  { X86::VPOPCNTDZrrkz,            X86::VPOPCNTDZrmkz,            0 },
  { X86::VPOPCNTQZ128rrkz,         X86::VPOPCNTQZ128rmkz,         0 },
  { X86::VPOPCNTQZ256rrkz,         X86::VPOPCNTQZ256rmkz,         0 },
  { X86::VPOPCNTQZrrkz,            X86::VPOPCNTQZrmkz,            0 },
  { X86::VPOPCNTWZ128rrkz,         X86::VPOPCNTWZ128rmkz,         0 },
  { X86::VPOPCNTWZ256rrkz,         X86::VPOPCNTWZ256rmkz,         0 },
  { X86::VPOPCNTWZrrkz,            X86::VPOPCNTWZrmkz,            0 },
  { X86::VPORDZ128rr,              X86::VPORDZ128rm,              0 },
  { X86::VPORDZ256rr,              X86::VPORDZ256rm,              0 },
  { X86::VPORDZrr,                 X86::VPORDZrm,                 0 },
  { X86::VPORQZ128rr,              X86::VPORQZ128rm,              0 },
  { X86::VPORQZ256rr,              X86::VPORQZ256rm,              0 },
  { X86::VPORQZrr,                 X86::VPORQZrm,                 0 },
  { X86::VPORYrr,                  X86::VPORYrm,                  0 },
  { X86::VPORrr,                   X86::VPORrm,                   0 },
  { X86::VPPERMrrr,                X86::VPPERMrmr,                0 },
  { X86::VPROLDZ128rikz,           X86::VPROLDZ128mikz,           0 },
  { X86::VPROLDZ256rikz,           X86::VPROLDZ256mikz,           0 },
  { X86::VPROLDZrikz,              X86::VPROLDZmikz,              0 },
  { X86::VPROLQZ128rikz,           X86::VPROLQZ128mikz,           0 },
  { X86::VPROLQZ256rikz,           X86::VPROLQZ256mikz,           0 },
  { X86::VPROLQZrikz,              X86::VPROLQZmikz,              0 },
  { X86::VPROLVDZ128rr,            X86::VPROLVDZ128rm,            0 },
  { X86::VPROLVDZ256rr,            X86::VPROLVDZ256rm,            0 },
  { X86::VPROLVDZrr,               X86::VPROLVDZrm,               0 },
  { X86::VPROLVQZ128rr,            X86::VPROLVQZ128rm,            0 },
  { X86::VPROLVQZ256rr,            X86::VPROLVQZ256rm,            0 },
  { X86::VPROLVQZrr,               X86::VPROLVQZrm,               0 },
  { X86::VPRORDZ128rikz,           X86::VPRORDZ128mikz,           0 },
  { X86::VPRORDZ256rikz,           X86::VPRORDZ256mikz,           0 },
  { X86::VPRORDZrikz,              X86::VPRORDZmikz,              0 },
  { X86::VPRORQZ128rikz,           X86::VPRORQZ128mikz,           0 },
  { X86::VPRORQZ256rikz,           X86::VPRORQZ256mikz,           0 },
  { X86::VPRORQZrikz,              X86::VPRORQZmikz,              0 },
  { X86::VPRORVDZ128rr,            X86::VPRORVDZ128rm,            0 },
  { X86::VPRORVDZ256rr,            X86::VPRORVDZ256rm,            0 },
  { X86::VPRORVDZrr,               X86::VPRORVDZrm,               0 },
  { X86::VPRORVQZ128rr,            X86::VPRORVQZ128rm,            0 },
  { X86::VPRORVQZ256rr,            X86::VPRORVQZ256rm,            0 },
  { X86::VPRORVQZrr,               X86::VPRORVQZrm,               0 },
  { X86::VPROTBrr,                 X86::VPROTBrm,                 0 },
  { X86::VPROTDrr,                 X86::VPROTDrm,                 0 },
  { X86::VPROTQrr,                 X86::VPROTQrm,                 0 },
  { X86::VPROTWrr,                 X86::VPROTWrm,                 0 },
  { X86::VPSADBWYrr,               X86::VPSADBWYrm,               0 },
  { X86::VPSADBWZ128rr,            X86::VPSADBWZ128rm,            0 },
  { X86::VPSADBWZ256rr,            X86::VPSADBWZ256rm,            0 },
  { X86::VPSADBWZrr,               X86::VPSADBWZrm,               0 },
  { X86::VPSADBWrr,                X86::VPSADBWrm,                0 },
  { X86::VPSHABrr,                 X86::VPSHABrm,                 0 },
  { X86::VPSHADrr,                 X86::VPSHADrm,                 0 },
  { X86::VPSHAQrr,                 X86::VPSHAQrm,                 0 },
  { X86::VPSHAWrr,                 X86::VPSHAWrm,                 0 },
  { X86::VPSHLBrr,                 X86::VPSHLBrm,                 0 },
  { X86::VPSHLDDZ128rri,           X86::VPSHLDDZ128rmi,           0 },
  { X86::VPSHLDDZ256rri,           X86::VPSHLDDZ256rmi,           0 },
  { X86::VPSHLDDZrri,              X86::VPSHLDDZrmi,              0 },
  { X86::VPSHLDQZ128rri,           X86::VPSHLDQZ128rmi,           0 },
  { X86::VPSHLDQZ256rri,           X86::VPSHLDQZ256rmi,           0 },
  { X86::VPSHLDQZrri,              X86::VPSHLDQZrmi,              0 },
  { X86::VPSHLDWZ128rri,           X86::VPSHLDWZ128rmi,           0 },
  { X86::VPSHLDWZ256rri,           X86::VPSHLDWZ256rmi,           0 },
  { X86::VPSHLDWZrri,              X86::VPSHLDWZrmi,              0 },
  { X86::VPSHLDrr,                 X86::VPSHLDrm,                 0 },
  { X86::VPSHLQrr,                 X86::VPSHLQrm,                 0 },
  { X86::VPSHLWrr,                 X86::VPSHLWrm,                 0 },
  { X86::VPSHRDDZ128rri,           X86::VPSHRDDZ128rmi,           0 },
  { X86::VPSHRDDZ256rri,           X86::VPSHRDDZ256rmi,           0 },
  { X86::VPSHRDDZrri,              X86::VPSHRDDZrmi,              0 },
  { X86::VPSHRDQZ128rri,           X86::VPSHRDQZ128rmi,           0 },
  { X86::VPSHRDQZ256rri,           X86::VPSHRDQZ256rmi,           0 },
  { X86::VPSHRDQZrri,              X86::VPSHRDQZrmi,              0 },
  { X86::VPSHRDWZ128rri,           X86::VPSHRDWZ128rmi,           0 },
  { X86::VPSHRDWZ256rri,           X86::VPSHRDWZ256rmi,           0 },
  { X86::VPSHRDWZrri,              X86::VPSHRDWZrmi,              0 },
  { X86::VPSHUFBITQMBZ128rr,       X86::VPSHUFBITQMBZ128rm,       0 },
  { X86::VPSHUFBITQMBZ256rr,       X86::VPSHUFBITQMBZ256rm,       0 },
  { X86::VPSHUFBITQMBZrr,          X86::VPSHUFBITQMBZrm,          0 },
  { X86::VPSHUFBYrr,               X86::VPSHUFBYrm,               0 },
  { X86::VPSHUFBZ128rr,            X86::VPSHUFBZ128rm,            0 },
  { X86::VPSHUFBZ256rr,            X86::VPSHUFBZ256rm,            0 },
  { X86::VPSHUFBZrr,               X86::VPSHUFBZrm,               0 },
  { X86::VPSHUFBrr,                X86::VPSHUFBrm,                0 },
  { X86::VPSHUFDZ128rikz,          X86::VPSHUFDZ128mikz,          0 },
  { X86::VPSHUFDZ256rikz,          X86::VPSHUFDZ256mikz,          0 },
  { X86::VPSHUFDZrikz,             X86::VPSHUFDZmikz,             0 },
  { X86::VPSHUFHWZ128rikz,         X86::VPSHUFHWZ128mikz,         0 },
  { X86::VPSHUFHWZ256rikz,         X86::VPSHUFHWZ256mikz,         0 },
  { X86::VPSHUFHWZrikz,            X86::VPSHUFHWZmikz,            0 },
  { X86::VPSHUFLWZ128rikz,         X86::VPSHUFLWZ128mikz,         0 },
  { X86::VPSHUFLWZ256rikz,         X86::VPSHUFLWZ256mikz,         0 },
  { X86::VPSHUFLWZrikz,            X86::VPSHUFLWZmikz,            0 },
  { X86::VPSIGNBYrr,               X86::VPSIGNBYrm,               0 },
  { X86::VPSIGNBrr,                X86::VPSIGNBrm,                0 },
  { X86::VPSIGNDYrr,               X86::VPSIGNDYrm,               0 },
  { X86::VPSIGNDrr,                X86::VPSIGNDrm,                0 },
  { X86::VPSIGNWYrr,               X86::VPSIGNWYrm,               0 },
  { X86::VPSIGNWrr,                X86::VPSIGNWrm,                0 },
  { X86::VPSLLDYrr,                X86::VPSLLDYrm,                0 },
  { X86::VPSLLDZ128rikz,           X86::VPSLLDZ128mikz,           0 },
  { X86::VPSLLDZ128rr,             X86::VPSLLDZ128rm,             0 },
  { X86::VPSLLDZ256rikz,           X86::VPSLLDZ256mikz,           0 },
  { X86::VPSLLDZ256rr,             X86::VPSLLDZ256rm,             0 },
  { X86::VPSLLDZrikz,              X86::VPSLLDZmikz,              0 },
  { X86::VPSLLDZrr,                X86::VPSLLDZrm,                0 },
  { X86::VPSLLDrr,                 X86::VPSLLDrm,                 0 },
  { X86::VPSLLQYrr,                X86::VPSLLQYrm,                0 },
  { X86::VPSLLQZ128rikz,           X86::VPSLLQZ128mikz,           0 },
  { X86::VPSLLQZ128rr,             X86::VPSLLQZ128rm,             0 },
  { X86::VPSLLQZ256rikz,           X86::VPSLLQZ256mikz,           0 },
  { X86::VPSLLQZ256rr,             X86::VPSLLQZ256rm,             0 },
  { X86::VPSLLQZrikz,              X86::VPSLLQZmikz,              0 },
  { X86::VPSLLQZrr,                X86::VPSLLQZrm,                0 },
  { X86::VPSLLQrr,                 X86::VPSLLQrm,                 0 },
  { X86::VPSLLVDYrr,               X86::VPSLLVDYrm,               0 },
  { X86::VPSLLVDZ128rr,            X86::VPSLLVDZ128rm,            0 },
  { X86::VPSLLVDZ256rr,            X86::VPSLLVDZ256rm,            0 },
  { X86::VPSLLVDZrr,               X86::VPSLLVDZrm,               0 },
  { X86::VPSLLVDrr,                X86::VPSLLVDrm,                0 },
  { X86::VPSLLVQYrr,               X86::VPSLLVQYrm,               0 },
  { X86::VPSLLVQZ128rr,            X86::VPSLLVQZ128rm,            0 },
  { X86::VPSLLVQZ256rr,            X86::VPSLLVQZ256rm,            0 },
  { X86::VPSLLVQZrr,               X86::VPSLLVQZrm,               0 },
  { X86::VPSLLVQrr,                X86::VPSLLVQrm,                0 },
  { X86::VPSLLVWZ128rr,            X86::VPSLLVWZ128rm,            0 },
  { X86::VPSLLVWZ256rr,            X86::VPSLLVWZ256rm,            0 },
  { X86::VPSLLVWZrr,               X86::VPSLLVWZrm,               0 },
  { X86::VPSLLWYrr,                X86::VPSLLWYrm,                0 },
  { X86::VPSLLWZ128rikz,           X86::VPSLLWZ128mikz,           0 },
  { X86::VPSLLWZ128rr,             X86::VPSLLWZ128rm,             0 },
  { X86::VPSLLWZ256rikz,           X86::VPSLLWZ256mikz,           0 },
  { X86::VPSLLWZ256rr,             X86::VPSLLWZ256rm,             0 },
  { X86::VPSLLWZrikz,              X86::VPSLLWZmikz,              0 },
  { X86::VPSLLWZrr,                X86::VPSLLWZrm,                0 },
  { X86::VPSLLWrr,                 X86::VPSLLWrm,                 0 },
  { X86::VPSRADYrr,                X86::VPSRADYrm,                0 },
  { X86::VPSRADZ128rikz,           X86::VPSRADZ128mikz,           0 },
  { X86::VPSRADZ128rr,             X86::VPSRADZ128rm,             0 },
  { X86::VPSRADZ256rikz,           X86::VPSRADZ256mikz,           0 },
  { X86::VPSRADZ256rr,             X86::VPSRADZ256rm,             0 },
  { X86::VPSRADZrikz,              X86::VPSRADZmikz,              0 },
  { X86::VPSRADZrr,                X86::VPSRADZrm,                0 },
  { X86::VPSRADrr,                 X86::VPSRADrm,                 0 },
  { X86::VPSRAQZ128rikz,           X86::VPSRAQZ128mikz,           0 },
  { X86::VPSRAQZ128rr,             X86::VPSRAQZ128rm,             0 },
  { X86::VPSRAQZ256rikz,           X86::VPSRAQZ256mikz,           0 },
  { X86::VPSRAQZ256rr,             X86::VPSRAQZ256rm,             0 },
  { X86::VPSRAQZrikz,              X86::VPSRAQZmikz,              0 },
  { X86::VPSRAQZrr,                X86::VPSRAQZrm,                0 },
  { X86::VPSRAVDYrr,               X86::VPSRAVDYrm,               0 },
  { X86::VPSRAVDZ128rr,            X86::VPSRAVDZ128rm,            0 },
  { X86::VPSRAVDZ256rr,            X86::VPSRAVDZ256rm,            0 },
  { X86::VPSRAVDZrr,               X86::VPSRAVDZrm,               0 },
  { X86::VPSRAVDrr,                X86::VPSRAVDrm,                0 },
  { X86::VPSRAVQZ128rr,            X86::VPSRAVQZ128rm,            0 },
  { X86::VPSRAVQZ256rr,            X86::VPSRAVQZ256rm,            0 },
  { X86::VPSRAVQZrr,               X86::VPSRAVQZrm,               0 },
  { X86::VPSRAVWZ128rr,            X86::VPSRAVWZ128rm,            0 },
  { X86::VPSRAVWZ256rr,            X86::VPSRAVWZ256rm,            0 },
  { X86::VPSRAVWZrr,               X86::VPSRAVWZrm,               0 },
  { X86::VPSRAWYrr,                X86::VPSRAWYrm,                0 },
  { X86::VPSRAWZ128rikz,           X86::VPSRAWZ128mikz,           0 },
  { X86::VPSRAWZ128rr,             X86::VPSRAWZ128rm,             0 },
  { X86::VPSRAWZ256rikz,           X86::VPSRAWZ256mikz,           0 },
  { X86::VPSRAWZ256rr,             X86::VPSRAWZ256rm,             0 },
  { X86::VPSRAWZrikz,              X86::VPSRAWZmikz,              0 },
  { X86::VPSRAWZrr,                X86::VPSRAWZrm,                0 },
  { X86::VPSRAWrr,                 X86::VPSRAWrm,                 0 },
  { X86::VPSRLDYrr,                X86::VPSRLDYrm,                0 },
  { X86::VPSRLDZ128rikz,           X86::VPSRLDZ128mikz,           0 },
  { X86::VPSRLDZ128rr,             X86::VPSRLDZ128rm,             0 },
  { X86::VPSRLDZ256rikz,           X86::VPSRLDZ256mikz,           0 },
  { X86::VPSRLDZ256rr,             X86::VPSRLDZ256rm,             0 },
  { X86::VPSRLDZrikz,              X86::VPSRLDZmikz,              0 },
  { X86::VPSRLDZrr,                X86::VPSRLDZrm,                0 },
  { X86::VPSRLDrr,                 X86::VPSRLDrm,                 0 },
  { X86::VPSRLQYrr,                X86::VPSRLQYrm,                0 },
  { X86::VPSRLQZ128rikz,           X86::VPSRLQZ128mikz,           0 },
  { X86::VPSRLQZ128rr,             X86::VPSRLQZ128rm,             0 },
  { X86::VPSRLQZ256rikz,           X86::VPSRLQZ256mikz,           0 },
  { X86::VPSRLQZ256rr,             X86::VPSRLQZ256rm,             0 },
  { X86::VPSRLQZrikz,              X86::VPSRLQZmikz,              0 },
  { X86::VPSRLQZrr,                X86::VPSRLQZrm,                0 },
  { X86::VPSRLQrr,                 X86::VPSRLQrm,                 0 },
  { X86::VPSRLVDYrr,               X86::VPSRLVDYrm,               0 },
  { X86::VPSRLVDZ128rr,            X86::VPSRLVDZ128rm,            0 },
  { X86::VPSRLVDZ256rr,            X86::VPSRLVDZ256rm,            0 },
  { X86::VPSRLVDZrr,               X86::VPSRLVDZrm,               0 },
  { X86::VPSRLVDrr,                X86::VPSRLVDrm,                0 },
  { X86::VPSRLVQYrr,               X86::VPSRLVQYrm,               0 },
  { X86::VPSRLVQZ128rr,            X86::VPSRLVQZ128rm,            0 },
  { X86::VPSRLVQZ256rr,            X86::VPSRLVQZ256rm,            0 },
  { X86::VPSRLVQZrr,               X86::VPSRLVQZrm,               0 },
  { X86::VPSRLVQrr,                X86::VPSRLVQrm,                0 },
  { X86::VPSRLVWZ128rr,            X86::VPSRLVWZ128rm,            0 },
  { X86::VPSRLVWZ256rr,            X86::VPSRLVWZ256rm,            0 },
  { X86::VPSRLVWZrr,               X86::VPSRLVWZrm,               0 },
  { X86::VPSRLWYrr,                X86::VPSRLWYrm,                0 },
  { X86::VPSRLWZ128rikz,           X86::VPSRLWZ128mikz,           0 },
  { X86::VPSRLWZ128rr,             X86::VPSRLWZ128rm,             0 },
  { X86::VPSRLWZ256rikz,           X86::VPSRLWZ256mikz,           0 },
  { X86::VPSRLWZ256rr,             X86::VPSRLWZ256rm,             0 },
  { X86::VPSRLWZrikz,              X86::VPSRLWZmikz,              0 },
  { X86::VPSRLWZrr,                X86::VPSRLWZrm,                0 },
  { X86::VPSRLWrr,                 X86::VPSRLWrm,                 0 },
  { X86::VPSUBBYrr,                X86::VPSUBBYrm,                0 },
  { X86::VPSUBBZ128rr,             X86::VPSUBBZ128rm,             0 },
  { X86::VPSUBBZ256rr,             X86::VPSUBBZ256rm,             0 },
  { X86::VPSUBBZrr,                X86::VPSUBBZrm,                0 },
  { X86::VPSUBBrr,                 X86::VPSUBBrm,                 0 },
  { X86::VPSUBDYrr,                X86::VPSUBDYrm,                0 },
  { X86::VPSUBDZ128rr,             X86::VPSUBDZ128rm,             0 },
  { X86::VPSUBDZ256rr,             X86::VPSUBDZ256rm,             0 },
  { X86::VPSUBDZrr,                X86::VPSUBDZrm,                0 },
  { X86::VPSUBDrr,                 X86::VPSUBDrm,                 0 },
  { X86::VPSUBQYrr,                X86::VPSUBQYrm,                0 },
  { X86::VPSUBQZ128rr,             X86::VPSUBQZ128rm,             0 },
  { X86::VPSUBQZ256rr,             X86::VPSUBQZ256rm,             0 },
  { X86::VPSUBQZrr,                X86::VPSUBQZrm,                0 },
  { X86::VPSUBQrr,                 X86::VPSUBQrm,                 0 },
  { X86::VPSUBSBYrr,               X86::VPSUBSBYrm,               0 },
  { X86::VPSUBSBZ128rr,            X86::VPSUBSBZ128rm,            0 },
  { X86::VPSUBSBZ256rr,            X86::VPSUBSBZ256rm,            0 },
  { X86::VPSUBSBZrr,               X86::VPSUBSBZrm,               0 },
  { X86::VPSUBSBrr,                X86::VPSUBSBrm,                0 },
  { X86::VPSUBSWYrr,               X86::VPSUBSWYrm,               0 },
  { X86::VPSUBSWZ128rr,            X86::VPSUBSWZ128rm,            0 },
  { X86::VPSUBSWZ256rr,            X86::VPSUBSWZ256rm,            0 },
  { X86::VPSUBSWZrr,               X86::VPSUBSWZrm,               0 },
  { X86::VPSUBSWrr,                X86::VPSUBSWrm,                0 },
  { X86::VPSUBUSBYrr,              X86::VPSUBUSBYrm,              0 },
  { X86::VPSUBUSBZ128rr,           X86::VPSUBUSBZ128rm,           0 },
  { X86::VPSUBUSBZ256rr,           X86::VPSUBUSBZ256rm,           0 },
  { X86::VPSUBUSBZrr,              X86::VPSUBUSBZrm,              0 },
  { X86::VPSUBUSBrr,               X86::VPSUBUSBrm,               0 },
  { X86::VPSUBUSWYrr,              X86::VPSUBUSWYrm,              0 },
  { X86::VPSUBUSWZ128rr,           X86::VPSUBUSWZ128rm,           0 },
  { X86::VPSUBUSWZ256rr,           X86::VPSUBUSWZ256rm,           0 },
  { X86::VPSUBUSWZrr,              X86::VPSUBUSWZrm,              0 },
  { X86::VPSUBUSWrr,               X86::VPSUBUSWrm,               0 },
  { X86::VPSUBWYrr,                X86::VPSUBWYrm,                0 },
  { X86::VPSUBWZ128rr,             X86::VPSUBWZ128rm,             0 },
  { X86::VPSUBWZ256rr,             X86::VPSUBWZ256rm,             0 },
  { X86::VPSUBWZrr,                X86::VPSUBWZrm,                0 },
  { X86::VPSUBWrr,                 X86::VPSUBWrm,                 0 },
  { X86::VPTESTMBZ128rr,           X86::VPTESTMBZ128rm,           0 },
  { X86::VPTESTMBZ256rr,           X86::VPTESTMBZ256rm,           0 },
  { X86::VPTESTMBZrr,              X86::VPTESTMBZrm,              0 },
  { X86::VPTESTMDZ128rr,           X86::VPTESTMDZ128rm,           0 },
  { X86::VPTESTMDZ256rr,           X86::VPTESTMDZ256rm,           0 },
  { X86::VPTESTMDZrr,              X86::VPTESTMDZrm,              0 },
  { X86::VPTESTMQZ128rr,           X86::VPTESTMQZ128rm,           0 },
  { X86::VPTESTMQZ256rr,           X86::VPTESTMQZ256rm,           0 },
  { X86::VPTESTMQZrr,              X86::VPTESTMQZrm,              0 },
  { X86::VPTESTMWZ128rr,           X86::VPTESTMWZ128rm,           0 },
  { X86::VPTESTMWZ256rr,           X86::VPTESTMWZ256rm,           0 },
  { X86::VPTESTMWZrr,              X86::VPTESTMWZrm,              0 },
  { X86::VPTESTNMBZ128rr,          X86::VPTESTNMBZ128rm,          0 },
  { X86::VPTESTNMBZ256rr,          X86::VPTESTNMBZ256rm,          0 },
  { X86::VPTESTNMBZrr,             X86::VPTESTNMBZrm,             0 },
  { X86::VPTESTNMDZ128rr,          X86::VPTESTNMDZ128rm,          0 },
  { X86::VPTESTNMDZ256rr,          X86::VPTESTNMDZ256rm,          0 },
  { X86::VPTESTNMDZrr,             X86::VPTESTNMDZrm,             0 },
  { X86::VPTESTNMQZ128rr,          X86::VPTESTNMQZ128rm,          0 },
  { X86::VPTESTNMQZ256rr,          X86::VPTESTNMQZ256rm,          0 },
  { X86::VPTESTNMQZrr,             X86::VPTESTNMQZrm,             0 },
  { X86::VPTESTNMWZ128rr,          X86::VPTESTNMWZ128rm,          0 },
  { X86::VPTESTNMWZ256rr,          X86::VPTESTNMWZ256rm,          0 },
  { X86::VPTESTNMWZrr,             X86::VPTESTNMWZrm,             0 },
  { X86::VPUNPCKHBWYrr,            X86::VPUNPCKHBWYrm,            0 },
  { X86::VPUNPCKHBWZ128rr,         X86::VPUNPCKHBWZ128rm,         0 },
  { X86::VPUNPCKHBWZ256rr,         X86::VPUNPCKHBWZ256rm,         0 },
  { X86::VPUNPCKHBWZrr,            X86::VPUNPCKHBWZrm,            0 },
  { X86::VPUNPCKHBWrr,             X86::VPUNPCKHBWrm,             0 },
  { X86::VPUNPCKHDQYrr,            X86::VPUNPCKHDQYrm,            0 },
  { X86::VPUNPCKHDQZ128rr,         X86::VPUNPCKHDQZ128rm,         0 },
  { X86::VPUNPCKHDQZ256rr,         X86::VPUNPCKHDQZ256rm,         0 },
  { X86::VPUNPCKHDQZrr,            X86::VPUNPCKHDQZrm,            0 },
  { X86::VPUNPCKHDQrr,             X86::VPUNPCKHDQrm,             0 },
  { X86::VPUNPCKHQDQYrr,           X86::VPUNPCKHQDQYrm,           0 },
  { X86::VPUNPCKHQDQZ128rr,        X86::VPUNPCKHQDQZ128rm,        0 },
  { X86::VPUNPCKHQDQZ256rr,        X86::VPUNPCKHQDQZ256rm,        0 },
  { X86::VPUNPCKHQDQZrr,           X86::VPUNPCKHQDQZrm,           0 },
  { X86::VPUNPCKHQDQrr,            X86::VPUNPCKHQDQrm,            0 },
  { X86::VPUNPCKHWDYrr,            X86::VPUNPCKHWDYrm,            0 },
  { X86::VPUNPCKHWDZ128rr,         X86::VPUNPCKHWDZ128rm,         0 },
  { X86::VPUNPCKHWDZ256rr,         X86::VPUNPCKHWDZ256rm,         0 },
  { X86::VPUNPCKHWDZrr,            X86::VPUNPCKHWDZrm,            0 },
  { X86::VPUNPCKHWDrr,             X86::VPUNPCKHWDrm,             0 },
  { X86::VPUNPCKLBWYrr,            X86::VPUNPCKLBWYrm,            0 },
  { X86::VPUNPCKLBWZ128rr,         X86::VPUNPCKLBWZ128rm,         0 },
  { X86::VPUNPCKLBWZ256rr,         X86::VPUNPCKLBWZ256rm,         0 },
  { X86::VPUNPCKLBWZrr,            X86::VPUNPCKLBWZrm,            0 },
  { X86::VPUNPCKLBWrr,             X86::VPUNPCKLBWrm,             0 },
  { X86::VPUNPCKLDQYrr,            X86::VPUNPCKLDQYrm,            0 },
  { X86::VPUNPCKLDQZ128rr,         X86::VPUNPCKLDQZ128rm,         0 },
  { X86::VPUNPCKLDQZ256rr,         X86::VPUNPCKLDQZ256rm,         0 },
  { X86::VPUNPCKLDQZrr,            X86::VPUNPCKLDQZrm,            0 },
  { X86::VPUNPCKLDQrr,             X86::VPUNPCKLDQrm,             0 },
  { X86::VPUNPCKLQDQYrr,           X86::VPUNPCKLQDQYrm,           0 },
  { X86::VPUNPCKLQDQZ128rr,        X86::VPUNPCKLQDQZ128rm,        0 },
  { X86::VPUNPCKLQDQZ256rr,        X86::VPUNPCKLQDQZ256rm,        0 },
  { X86::VPUNPCKLQDQZrr,           X86::VPUNPCKLQDQZrm,           0 },
  { X86::VPUNPCKLQDQrr,            X86::VPUNPCKLQDQrm,            0 },
  { X86::VPUNPCKLWDYrr,            X86::VPUNPCKLWDYrm,            0 },
  { X86::VPUNPCKLWDZ128rr,         X86::VPUNPCKLWDZ128rm,         0 },
  { X86::VPUNPCKLWDZ256rr,         X86::VPUNPCKLWDZ256rm,         0 },
  { X86::VPUNPCKLWDZrr,            X86::VPUNPCKLWDZrm,            0 },
  { X86::VPUNPCKLWDrr,             X86::VPUNPCKLWDrm,             0 },
  { X86::VPXORDZ128rr,             X86::VPXORDZ128rm,             0 },
  { X86::VPXORDZ256rr,             X86::VPXORDZ256rm,             0 },
  { X86::VPXORDZrr,                X86::VPXORDZrm,                0 },
  { X86::VPXORQZ128rr,             X86::VPXORQZ128rm,             0 },
  { X86::VPXORQZ256rr,             X86::VPXORQZ256rm,             0 },
  { X86::VPXORQZrr,                X86::VPXORQZrm,                0 },
  { X86::VPXORYrr,                 X86::VPXORYrm,                 0 },
  { X86::VPXORrr,                  X86::VPXORrm,                  0 },
  { X86::VRANGEPDZ128rri,          X86::VRANGEPDZ128rmi,          0 },
  { X86::VRANGEPDZ256rri,          X86::VRANGEPDZ256rmi,          0 },
  { X86::VRANGEPDZrri,             X86::VRANGEPDZrmi,             0 },
  { X86::VRANGEPSZ128rri,          X86::VRANGEPSZ128rmi,          0 },
  { X86::VRANGEPSZ256rri,          X86::VRANGEPSZ256rmi,          0 },
  { X86::VRANGEPSZrri,             X86::VRANGEPSZrmi,             0 },
  { X86::VRANGESDZrri,             X86::VRANGESDZrmi,             TB_NO_REVERSE },
  { X86::VRANGESSZrri,             X86::VRANGESSZrmi,             TB_NO_REVERSE },
  { X86::VRCP14PDZ128rkz,          X86::VRCP14PDZ128mkz,          0 },
  { X86::VRCP14PDZ256rkz,          X86::VRCP14PDZ256mkz,          0 },
  { X86::VRCP14PDZrkz,             X86::VRCP14PDZmkz,             0 },
  { X86::VRCP14PSZ128rkz,          X86::VRCP14PSZ128mkz,          0 },
  { X86::VRCP14PSZ256rkz,          X86::VRCP14PSZ256mkz,          0 },
  { X86::VRCP14PSZrkz,             X86::VRCP14PSZmkz,             0 },
  { X86::VRCP14SDZrr,              X86::VRCP14SDZrm,              TB_NO_REVERSE },
  { X86::VRCP14SSZrr,              X86::VRCP14SSZrm,              TB_NO_REVERSE },
  { X86::VRCP28PDZrkz,             X86::VRCP28PDZmkz,             0 },
  { X86::VRCP28PSZrkz,             X86::VRCP28PSZmkz,             0 },
  { X86::VRCP28SDZr,               X86::VRCP28SDZm,               TB_NO_REVERSE },
  { X86::VRCP28SSZr,               X86::VRCP28SSZm,               TB_NO_REVERSE },
  { X86::VRCPSSr,                  X86::VRCPSSm,                  0 },
  { X86::VRCPSSr_Int,              X86::VRCPSSm_Int,              TB_NO_REVERSE },
  { X86::VREDUCEPDZ128rrikz,       X86::VREDUCEPDZ128rmikz,       0 },
  { X86::VREDUCEPDZ256rrikz,       X86::VREDUCEPDZ256rmikz,       0 },
  { X86::VREDUCEPDZrrikz,          X86::VREDUCEPDZrmikz,          0 },
  { X86::VREDUCEPSZ128rrikz,       X86::VREDUCEPSZ128rmikz,       0 },
  { X86::VREDUCEPSZ256rrikz,       X86::VREDUCEPSZ256rmikz,       0 },
  { X86::VREDUCEPSZrrikz,          X86::VREDUCEPSZrmikz,          0 },
  { X86::VREDUCESDZrri,            X86::VREDUCESDZrmi,            TB_NO_REVERSE },
  { X86::VREDUCESSZrri,            X86::VREDUCESSZrmi,            TB_NO_REVERSE },
  { X86::VRNDSCALEPDZ128rrikz,     X86::VRNDSCALEPDZ128rmikz,     0 },
  { X86::VRNDSCALEPDZ256rrikz,     X86::VRNDSCALEPDZ256rmikz,     0 },
  { X86::VRNDSCALEPDZrrikz,        X86::VRNDSCALEPDZrmikz,        0 },
  { X86::VRNDSCALEPSZ128rrikz,     X86::VRNDSCALEPSZ128rmikz,     0 },
  { X86::VRNDSCALEPSZ256rrikz,     X86::VRNDSCALEPSZ256rmikz,     0 },
  { X86::VRNDSCALEPSZrrikz,        X86::VRNDSCALEPSZrmikz,        0 },
  { X86::VRNDSCALESDZr,            X86::VRNDSCALESDZm,            0 },
  { X86::VRNDSCALESDZr_Int,        X86::VRNDSCALESDZm_Int,        TB_NO_REVERSE },
  { X86::VRNDSCALESSZr,            X86::VRNDSCALESSZm,            0 },
  { X86::VRNDSCALESSZr_Int,        X86::VRNDSCALESSZm_Int,        TB_NO_REVERSE },
  { X86::VROUNDSDr,                X86::VROUNDSDm,                0 },
  { X86::VROUNDSDr_Int,            X86::VROUNDSDm_Int,            TB_NO_REVERSE },
  { X86::VROUNDSSr,                X86::VROUNDSSm,                0 },
  { X86::VROUNDSSr_Int,            X86::VROUNDSSm_Int,            TB_NO_REVERSE },
  { X86::VRSQRT14PDZ128rkz,        X86::VRSQRT14PDZ128mkz,        0 },
  { X86::VRSQRT14PDZ256rkz,        X86::VRSQRT14PDZ256mkz,        0 },
  { X86::VRSQRT14PDZrkz,           X86::VRSQRT14PDZmkz,           0 },
  { X86::VRSQRT14PSZ128rkz,        X86::VRSQRT14PSZ128mkz,        0 },
  { X86::VRSQRT14PSZ256rkz,        X86::VRSQRT14PSZ256mkz,        0 },
  { X86::VRSQRT14PSZrkz,           X86::VRSQRT14PSZmkz,           0 },
  { X86::VRSQRT14SDZrr,            X86::VRSQRT14SDZrm,            TB_NO_REVERSE },
  { X86::VRSQRT14SSZrr,            X86::VRSQRT14SSZrm,            TB_NO_REVERSE },
  { X86::VRSQRT28PDZrkz,           X86::VRSQRT28PDZmkz,           0 },
  { X86::VRSQRT28PSZrkz,           X86::VRSQRT28PSZmkz,           0 },
  { X86::VRSQRT28SDZr,             X86::VRSQRT28SDZm,             TB_NO_REVERSE },
  { X86::VRSQRT28SSZr,             X86::VRSQRT28SSZm,             TB_NO_REVERSE },
  { X86::VRSQRTSSr,                X86::VRSQRTSSm,                0 },
  { X86::VRSQRTSSr_Int,            X86::VRSQRTSSm_Int,            TB_NO_REVERSE },
  { X86::VSCALEFPDZ128rr,          X86::VSCALEFPDZ128rm,          0 },
  { X86::VSCALEFPDZ256rr,          X86::VSCALEFPDZ256rm,          0 },
  { X86::VSCALEFPDZrr,             X86::VSCALEFPDZrm,             0 },
  { X86::VSCALEFPSZ128rr,          X86::VSCALEFPSZ128rm,          0 },
  { X86::VSCALEFPSZ256rr,          X86::VSCALEFPSZ256rm,          0 },
  { X86::VSCALEFPSZrr,             X86::VSCALEFPSZrm,             0 },
  { X86::VSCALEFSDZrr,             X86::VSCALEFSDZrm,             TB_NO_REVERSE },
  { X86::VSCALEFSSZrr,             X86::VSCALEFSSZrm,             TB_NO_REVERSE },
  { X86::VSHUFF32X4Z256rri,        X86::VSHUFF32X4Z256rmi,        0 },
  { X86::VSHUFF32X4Zrri,           X86::VSHUFF32X4Zrmi,           0 },
  { X86::VSHUFF64X2Z256rri,        X86::VSHUFF64X2Z256rmi,        0 },
  { X86::VSHUFF64X2Zrri,           X86::VSHUFF64X2Zrmi,           0 },
  { X86::VSHUFI32X4Z256rri,        X86::VSHUFI32X4Z256rmi,        0 },
  { X86::VSHUFI32X4Zrri,           X86::VSHUFI32X4Zrmi,           0 },
  { X86::VSHUFI64X2Z256rri,        X86::VSHUFI64X2Z256rmi,        0 },
  { X86::VSHUFI64X2Zrri,           X86::VSHUFI64X2Zrmi,           0 },
  { X86::VSHUFPDYrri,              X86::VSHUFPDYrmi,              0 },
  { X86::VSHUFPDZ128rri,           X86::VSHUFPDZ128rmi,           0 },
  { X86::VSHUFPDZ256rri,           X86::VSHUFPDZ256rmi,           0 },
  { X86::VSHUFPDZrri,              X86::VSHUFPDZrmi,              0 },
  { X86::VSHUFPDrri,               X86::VSHUFPDrmi,               0 },
  { X86::VSHUFPSYrri,              X86::VSHUFPSYrmi,              0 },
  { X86::VSHUFPSZ128rri,           X86::VSHUFPSZ128rmi,           0 },
  { X86::VSHUFPSZ256rri,           X86::VSHUFPSZ256rmi,           0 },
  { X86::VSHUFPSZrri,              X86::VSHUFPSZrmi,              0 },
  { X86::VSHUFPSrri,               X86::VSHUFPSrmi,               0 },
  { X86::VSQRTPDZ128rkz,           X86::VSQRTPDZ128mkz,           0 },
  { X86::VSQRTPDZ256rkz,           X86::VSQRTPDZ256mkz,           0 },
  { X86::VSQRTPDZrkz,              X86::VSQRTPDZmkz,              0 },
  { X86::VSQRTPSZ128rkz,           X86::VSQRTPSZ128mkz,           0 },
  { X86::VSQRTPSZ256rkz,           X86::VSQRTPSZ256mkz,           0 },
  { X86::VSQRTPSZrkz,              X86::VSQRTPSZmkz,              0 },
  { X86::VSQRTSDZr,                X86::VSQRTSDZm,                0 },
  { X86::VSQRTSDZr_Int,            X86::VSQRTSDZm_Int,            TB_NO_REVERSE },
  { X86::VSQRTSDr,                 X86::VSQRTSDm,                 0 },
  { X86::VSQRTSDr_Int,             X86::VSQRTSDm_Int,             TB_NO_REVERSE },
  { X86::VSQRTSSZr,                X86::VSQRTSSZm,                0 },
  { X86::VSQRTSSZr_Int,            X86::VSQRTSSZm_Int,            TB_NO_REVERSE },
  { X86::VSQRTSSr,                 X86::VSQRTSSm,                 0 },
  { X86::VSQRTSSr_Int,             X86::VSQRTSSm_Int,             TB_NO_REVERSE },
  { X86::VSUBPDYrr,                X86::VSUBPDYrm,                0 },
  { X86::VSUBPDZ128rr,             X86::VSUBPDZ128rm,             0 },
  { X86::VSUBPDZ256rr,             X86::VSUBPDZ256rm,             0 },
  { X86::VSUBPDZrr,                X86::VSUBPDZrm,                0 },
  { X86::VSUBPDrr,                 X86::VSUBPDrm,                 0 },
  { X86::VSUBPSYrr,                X86::VSUBPSYrm,                0 },
  { X86::VSUBPSZ128rr,             X86::VSUBPSZ128rm,             0 },
  { X86::VSUBPSZ256rr,             X86::VSUBPSZ256rm,             0 },
  { X86::VSUBPSZrr,                X86::VSUBPSZrm,                0 },
  { X86::VSUBPSrr,                 X86::VSUBPSrm,                 0 },
  { X86::VSUBSDZrr,                X86::VSUBSDZrm,                0 },
  { X86::VSUBSDZrr_Int,            X86::VSUBSDZrm_Int,            TB_NO_REVERSE },
  { X86::VSUBSDrr,                 X86::VSUBSDrm,                 0 },
  { X86::VSUBSDrr_Int,             X86::VSUBSDrm_Int,             TB_NO_REVERSE },
  { X86::VSUBSSZrr,                X86::VSUBSSZrm,                0 },
  { X86::VSUBSSZrr_Int,            X86::VSUBSSZrm_Int,            TB_NO_REVERSE },
  { X86::VSUBSSrr,                 X86::VSUBSSrm,                 0 },
  { X86::VSUBSSrr_Int,             X86::VSUBSSrm_Int,             TB_NO_REVERSE },
  { X86::VUNPCKHPDYrr,             X86::VUNPCKHPDYrm,             0 },
  { X86::VUNPCKHPDZ128rr,          X86::VUNPCKHPDZ128rm,          0 },
  { X86::VUNPCKHPDZ256rr,          X86::VUNPCKHPDZ256rm,          0 },
  { X86::VUNPCKHPDZrr,             X86::VUNPCKHPDZrm,             0 },
  { X86::VUNPCKHPDrr,              X86::VUNPCKHPDrm,              0 },
  { X86::VUNPCKHPSYrr,             X86::VUNPCKHPSYrm,             0 },
  { X86::VUNPCKHPSZ128rr,          X86::VUNPCKHPSZ128rm,          0 },
  { X86::VUNPCKHPSZ256rr,          X86::VUNPCKHPSZ256rm,          0 },
  { X86::VUNPCKHPSZrr,             X86::VUNPCKHPSZrm,             0 },
  { X86::VUNPCKHPSrr,              X86::VUNPCKHPSrm,              0 },
  { X86::VUNPCKLPDYrr,             X86::VUNPCKLPDYrm,             0 },
  { X86::VUNPCKLPDZ128rr,          X86::VUNPCKLPDZ128rm,          0 },
  { X86::VUNPCKLPDZ256rr,          X86::VUNPCKLPDZ256rm,          0 },
  { X86::VUNPCKLPDZrr,             X86::VUNPCKLPDZrm,             0 },
  { X86::VUNPCKLPDrr,              X86::VUNPCKLPDrm,              0 },
  { X86::VUNPCKLPSYrr,             X86::VUNPCKLPSYrm,             0 },
  { X86::VUNPCKLPSZ128rr,          X86::VUNPCKLPSZ128rm,          0 },
  { X86::VUNPCKLPSZ256rr,          X86::VUNPCKLPSZ256rm,          0 },
  { X86::VUNPCKLPSZrr,             X86::VUNPCKLPSZrm,             0 },
  { X86::VUNPCKLPSrr,              X86::VUNPCKLPSrm,              0 },
  { X86::VXORPDYrr,                X86::VXORPDYrm,                0 },
  { X86::VXORPDZ128rr,             X86::VXORPDZ128rm,             0 },
  { X86::VXORPDZ256rr,             X86::VXORPDZ256rm,             0 },
  { X86::VXORPDZrr,                X86::VXORPDZrm,                0 },
  { X86::VXORPDrr,                 X86::VXORPDrm,                 0 },
  { X86::VXORPSYrr,                X86::VXORPSYrm,                0 },
  { X86::VXORPSZ128rr,             X86::VXORPSZ128rm,             0 },
  { X86::VXORPSZ256rr,             X86::VXORPSZ256rm,             0 },
  { X86::VXORPSZrr,                X86::VXORPSZrm,                0 },
  { X86::VXORPSrr,                 X86::VXORPSrm,                 0 },
  { X86::XOR16rr,                  X86::XOR16rm,                  0 },
  { X86::XOR32rr,                  X86::XOR32rm,                  0 },
  { X86::XOR64rr,                  X86::XOR64rm,                  0 },
  { X86::XOR8rr,                   X86::XOR8rm,                   0 },
  { X86::XORPDrr,                  X86::XORPDrm,                  TB_ALIGN_16 },
  { X86::XORPSrr,                  X86::XORPSrm,                  TB_ALIGN_16 },
};

static const X86MemoryFoldTableEntry MemoryFoldTable3[] = {
  { X86::VADDPDZ128rrkz,             X86::VADDPDZ128rmkz,             0 },
  { X86::VADDPDZ256rrkz,             X86::VADDPDZ256rmkz,             0 },
  { X86::VADDPDZrrkz,                X86::VADDPDZrmkz,                0 },
  { X86::VADDPSZ128rrkz,             X86::VADDPSZ128rmkz,             0 },
  { X86::VADDPSZ256rrkz,             X86::VADDPSZ256rmkz,             0 },
  { X86::VADDPSZrrkz,                X86::VADDPSZrmkz,                0 },
  { X86::VADDSDZrr_Intkz,            X86::VADDSDZrm_Intkz,            TB_NO_REVERSE },
  { X86::VADDSSZrr_Intkz,            X86::VADDSSZrm_Intkz,            TB_NO_REVERSE },
  { X86::VALIGNDZ128rrikz,           X86::VALIGNDZ128rmikz,           0 },
  { X86::VALIGNDZ256rrikz,           X86::VALIGNDZ256rmikz,           0 },
  { X86::VALIGNDZrrikz,              X86::VALIGNDZrmikz,              0 },
  { X86::VALIGNQZ128rrikz,           X86::VALIGNQZ128rmikz,           0 },
  { X86::VALIGNQZ256rrikz,           X86::VALIGNQZ256rmikz,           0 },
  { X86::VALIGNQZrrikz,              X86::VALIGNQZrmikz,              0 },
  { X86::VANDNPDZ128rrkz,            X86::VANDNPDZ128rmkz,            0 },
  { X86::VANDNPDZ256rrkz,            X86::VANDNPDZ256rmkz,            0 },
  { X86::VANDNPDZrrkz,               X86::VANDNPDZrmkz,               0 },
  { X86::VANDNPSZ128rrkz,            X86::VANDNPSZ128rmkz,            0 },
  { X86::VANDNPSZ256rrkz,            X86::VANDNPSZ256rmkz,            0 },
  { X86::VANDNPSZrrkz,               X86::VANDNPSZrmkz,               0 },
  { X86::VANDPDZ128rrkz,             X86::VANDPDZ128rmkz,             0 },
  { X86::VANDPDZ256rrkz,             X86::VANDPDZ256rmkz,             0 },
  { X86::VANDPDZrrkz,                X86::VANDPDZrmkz,                0 },
  { X86::VANDPSZ128rrkz,             X86::VANDPSZ128rmkz,             0 },
  { X86::VANDPSZ256rrkz,             X86::VANDPSZ256rmkz,             0 },
  { X86::VANDPSZrrkz,                X86::VANDPSZrmkz,                0 },
  { X86::VBLENDMPDZ128rrk,           X86::VBLENDMPDZ128rmk,           0 },
  { X86::VBLENDMPDZ256rrk,           X86::VBLENDMPDZ256rmk,           0 },
  { X86::VBLENDMPDZrrk,              X86::VBLENDMPDZrmk,              0 },
  { X86::VBLENDMPSZ128rrk,           X86::VBLENDMPSZ128rmk,           0 },
  { X86::VBLENDMPSZ256rrk,           X86::VBLENDMPSZ256rmk,           0 },
  { X86::VBLENDMPSZrrk,              X86::VBLENDMPSZrmk,              0 },
  { X86::VBROADCASTF32X2Z256rk,      X86::VBROADCASTF32X2Z256mk,      TB_NO_REVERSE },
  { X86::VBROADCASTF32X2Zrk,         X86::VBROADCASTF32X2Zmk,         TB_NO_REVERSE },
  { X86::VBROADCASTI32X2Z128rk,      X86::VBROADCASTI32X2Z128mk,      TB_NO_REVERSE },
  { X86::VBROADCASTI32X2Z256rk,      X86::VBROADCASTI32X2Z256mk,      TB_NO_REVERSE },
  { X86::VBROADCASTI32X2Zrk,         X86::VBROADCASTI32X2Zmk,         TB_NO_REVERSE },
  { X86::VBROADCASTSDZ256rk,         X86::VBROADCASTSDZ256mk,         TB_NO_REVERSE },
  { X86::VBROADCASTSDZrk,            X86::VBROADCASTSDZmk,            TB_NO_REVERSE },
  { X86::VBROADCASTSSZ128rk,         X86::VBROADCASTSSZ128mk,         TB_NO_REVERSE },
  { X86::VBROADCASTSSZ256rk,         X86::VBROADCASTSSZ256mk,         TB_NO_REVERSE },
  { X86::VBROADCASTSSZrk,            X86::VBROADCASTSSZmk,            TB_NO_REVERSE },
  { X86::VCMPPDZ128rrik,             X86::VCMPPDZ128rmik,             0 },
  { X86::VCMPPDZ256rrik,             X86::VCMPPDZ256rmik,             0 },
  { X86::VCMPPDZrrik,                X86::VCMPPDZrmik,                0 },
  { X86::VCMPPSZ128rrik,             X86::VCMPPSZ128rmik,             0 },
  { X86::VCMPPSZ256rrik,             X86::VCMPPSZ256rmik,             0 },
  { X86::VCMPPSZrrik,                X86::VCMPPSZrmik,                0 },
  { X86::VCMPSDZrr_Intk,             X86::VCMPSDZrm_Intk,             TB_NO_REVERSE },
  { X86::VCMPSSZrr_Intk,             X86::VCMPSSZrm_Intk,             TB_NO_REVERSE },
  { X86::VCVTDQ2PDZ128rrk,           X86::VCVTDQ2PDZ128rmk,           TB_NO_REVERSE },
  { X86::VCVTDQ2PDZ256rrk,           X86::VCVTDQ2PDZ256rmk,           0 },
  { X86::VCVTDQ2PDZrrk,              X86::VCVTDQ2PDZrmk,              0 },
  { X86::VCVTDQ2PSZ128rrk,           X86::VCVTDQ2PSZ128rmk,           0 },
  { X86::VCVTDQ2PSZ256rrk,           X86::VCVTDQ2PSZ256rmk,           0 },
  { X86::VCVTDQ2PSZrrk,              X86::VCVTDQ2PSZrmk,              0 },
  { X86::VCVTNE2PS2BF16Z128rrkz,     X86::VCVTNE2PS2BF16Z128rmkz,     0 },
  { X86::VCVTNE2PS2BF16Z256rrkz,     X86::VCVTNE2PS2BF16Z256rmkz,     0 },
  { X86::VCVTNE2PS2BF16Zrrkz,        X86::VCVTNE2PS2BF16Zrmkz,        0 },
  { X86::VCVTNEPS2BF16Z128rrk,       X86::VCVTNEPS2BF16Z128rmk,       0 },
  { X86::VCVTNEPS2BF16Z256rrk,       X86::VCVTNEPS2BF16Z256rmk,       0 },
  { X86::VCVTNEPS2BF16Zrrk,          X86::VCVTNEPS2BF16Zrmk,          0 },
  { X86::VCVTPD2DQZ128rrk,           X86::VCVTPD2DQZ128rmk,           0 },
  { X86::VCVTPD2DQZ256rrk,           X86::VCVTPD2DQZ256rmk,           0 },
  { X86::VCVTPD2DQZrrk,              X86::VCVTPD2DQZrmk,              0 },
  { X86::VCVTPD2PSZ128rrk,           X86::VCVTPD2PSZ128rmk,           0 },
  { X86::VCVTPD2PSZ256rrk,           X86::VCVTPD2PSZ256rmk,           0 },
  { X86::VCVTPD2PSZrrk,              X86::VCVTPD2PSZrmk,              0 },
  { X86::VCVTPD2QQZ128rrk,           X86::VCVTPD2QQZ128rmk,           0 },
  { X86::VCVTPD2QQZ256rrk,           X86::VCVTPD2QQZ256rmk,           0 },
  { X86::VCVTPD2QQZrrk,              X86::VCVTPD2QQZrmk,              0 },
  { X86::VCVTPD2UDQZ128rrk,          X86::VCVTPD2UDQZ128rmk,          0 },
  { X86::VCVTPD2UDQZ256rrk,          X86::VCVTPD2UDQZ256rmk,          0 },
  { X86::VCVTPD2UDQZrrk,             X86::VCVTPD2UDQZrmk,             0 },
  { X86::VCVTPD2UQQZ128rrk,          X86::VCVTPD2UQQZ128rmk,          0 },
  { X86::VCVTPD2UQQZ256rrk,          X86::VCVTPD2UQQZ256rmk,          0 },
  { X86::VCVTPD2UQQZrrk,             X86::VCVTPD2UQQZrmk,             0 },
  { X86::VCVTPH2PSZ128rrk,           X86::VCVTPH2PSZ128rmk,           TB_NO_REVERSE },
  { X86::VCVTPH2PSZ256rrk,           X86::VCVTPH2PSZ256rmk,           0 },
  { X86::VCVTPH2PSZrrk,              X86::VCVTPH2PSZrmk,              0 },
  { X86::VCVTPS2DQZ128rrk,           X86::VCVTPS2DQZ128rmk,           0 },
  { X86::VCVTPS2DQZ256rrk,           X86::VCVTPS2DQZ256rmk,           0 },
  { X86::VCVTPS2DQZrrk,              X86::VCVTPS2DQZrmk,              0 },
  { X86::VCVTPS2PDZ128rrk,           X86::VCVTPS2PDZ128rmk,           TB_NO_REVERSE },
  { X86::VCVTPS2PDZ256rrk,           X86::VCVTPS2PDZ256rmk,           0 },
  { X86::VCVTPS2PDZrrk,              X86::VCVTPS2PDZrmk,              0 },
  { X86::VCVTPS2QQZ128rrk,           X86::VCVTPS2QQZ128rmk,           TB_NO_REVERSE },
  { X86::VCVTPS2QQZ256rrk,           X86::VCVTPS2QQZ256rmk,           0 },
  { X86::VCVTPS2QQZrrk,              X86::VCVTPS2QQZrmk,              0 },
  { X86::VCVTPS2UDQZ128rrk,          X86::VCVTPS2UDQZ128rmk,          0 },
  { X86::VCVTPS2UDQZ256rrk,          X86::VCVTPS2UDQZ256rmk,          0 },
  { X86::VCVTPS2UDQZrrk,             X86::VCVTPS2UDQZrmk,             0 },
  { X86::VCVTPS2UQQZ128rrk,          X86::VCVTPS2UQQZ128rmk,          TB_NO_REVERSE },
  { X86::VCVTPS2UQQZ256rrk,          X86::VCVTPS2UQQZ256rmk,          0 },
  { X86::VCVTPS2UQQZrrk,             X86::VCVTPS2UQQZrmk,             0 },
  { X86::VCVTQQ2PDZ128rrk,           X86::VCVTQQ2PDZ128rmk,           0 },
  { X86::VCVTQQ2PDZ256rrk,           X86::VCVTQQ2PDZ256rmk,           0 },
  { X86::VCVTQQ2PDZrrk,              X86::VCVTQQ2PDZrmk,              0 },
  { X86::VCVTQQ2PSZ128rrk,           X86::VCVTQQ2PSZ128rmk,           0 },
  { X86::VCVTQQ2PSZ256rrk,           X86::VCVTQQ2PSZ256rmk,           0 },
  { X86::VCVTQQ2PSZrrk,              X86::VCVTQQ2PSZrmk,              0 },
  { X86::VCVTSD2SSZrr_Intkz,         X86::VCVTSD2SSZrm_Intkz,         TB_NO_REVERSE },
  { X86::VCVTSS2SDZrr_Intkz,         X86::VCVTSS2SDZrm_Intkz,         TB_NO_REVERSE },
  { X86::VCVTTPD2DQZ128rrk,          X86::VCVTTPD2DQZ128rmk,          0 },
  { X86::VCVTTPD2DQZ256rrk,          X86::VCVTTPD2DQZ256rmk,          0 },
  { X86::VCVTTPD2DQZrrk,             X86::VCVTTPD2DQZrmk,             0 },
  { X86::VCVTTPD2QQZ128rrk,          X86::VCVTTPD2QQZ128rmk,          0 },
  { X86::VCVTTPD2QQZ256rrk,          X86::VCVTTPD2QQZ256rmk,          0 },
  { X86::VCVTTPD2QQZrrk,             X86::VCVTTPD2QQZrmk,             0 },
  { X86::VCVTTPD2UDQZ128rrk,         X86::VCVTTPD2UDQZ128rmk,         0 },
  { X86::VCVTTPD2UDQZ256rrk,         X86::VCVTTPD2UDQZ256rmk,         0 },
  { X86::VCVTTPD2UDQZrrk,            X86::VCVTTPD2UDQZrmk,            0 },
  { X86::VCVTTPD2UQQZ128rrk,         X86::VCVTTPD2UQQZ128rmk,         0 },
  { X86::VCVTTPD2UQQZ256rrk,         X86::VCVTTPD2UQQZ256rmk,         0 },
  { X86::VCVTTPD2UQQZrrk,            X86::VCVTTPD2UQQZrmk,            0 },
  { X86::VCVTTPS2DQZ128rrk,          X86::VCVTTPS2DQZ128rmk,          0 },
  { X86::VCVTTPS2DQZ256rrk,          X86::VCVTTPS2DQZ256rmk,          0 },
  { X86::VCVTTPS2DQZrrk,             X86::VCVTTPS2DQZrmk,             0 },
  { X86::VCVTTPS2QQZ128rrk,          X86::VCVTTPS2QQZ128rmk,          TB_NO_REVERSE },
  { X86::VCVTTPS2QQZ256rrk,          X86::VCVTTPS2QQZ256rmk,          0 },
  { X86::VCVTTPS2QQZrrk,             X86::VCVTTPS2QQZrmk,             0 },
  { X86::VCVTTPS2UDQZ128rrk,         X86::VCVTTPS2UDQZ128rmk,         0 },
  { X86::VCVTTPS2UDQZ256rrk,         X86::VCVTTPS2UDQZ256rmk,         0 },
  { X86::VCVTTPS2UDQZrrk,            X86::VCVTTPS2UDQZrmk,            0 },
  { X86::VCVTTPS2UQQZ128rrk,         X86::VCVTTPS2UQQZ128rmk,         TB_NO_REVERSE },
  { X86::VCVTTPS2UQQZ256rrk,         X86::VCVTTPS2UQQZ256rmk,         0 },
  { X86::VCVTTPS2UQQZrrk,            X86::VCVTTPS2UQQZrmk,            0 },
  { X86::VCVTUDQ2PDZ128rrk,          X86::VCVTUDQ2PDZ128rmk,          TB_NO_REVERSE },
  { X86::VCVTUDQ2PDZ256rrk,          X86::VCVTUDQ2PDZ256rmk,          0 },
  { X86::VCVTUDQ2PDZrrk,             X86::VCVTUDQ2PDZrmk,             0 },
  { X86::VCVTUDQ2PSZ128rrk,          X86::VCVTUDQ2PSZ128rmk,          0 },
  { X86::VCVTUDQ2PSZ256rrk,          X86::VCVTUDQ2PSZ256rmk,          0 },
  { X86::VCVTUDQ2PSZrrk,             X86::VCVTUDQ2PSZrmk,             0 },
  { X86::VCVTUQQ2PDZ128rrk,          X86::VCVTUQQ2PDZ128rmk,          0 },
  { X86::VCVTUQQ2PDZ256rrk,          X86::VCVTUQQ2PDZ256rmk,          0 },
  { X86::VCVTUQQ2PDZrrk,             X86::VCVTUQQ2PDZrmk,             0 },
  { X86::VCVTUQQ2PSZ128rrk,          X86::VCVTUQQ2PSZ128rmk,          0 },
  { X86::VCVTUQQ2PSZ256rrk,          X86::VCVTUQQ2PSZ256rmk,          0 },
  { X86::VCVTUQQ2PSZrrk,             X86::VCVTUQQ2PSZrmk,             0 },
  { X86::VDBPSADBWZ128rrikz,         X86::VDBPSADBWZ128rmikz,         0 },
  { X86::VDBPSADBWZ256rrikz,         X86::VDBPSADBWZ256rmikz,         0 },
  { X86::VDBPSADBWZrrikz,            X86::VDBPSADBWZrmikz,            0 },
  { X86::VDIVPDZ128rrkz,             X86::VDIVPDZ128rmkz,             0 },
  { X86::VDIVPDZ256rrkz,             X86::VDIVPDZ256rmkz,             0 },
  { X86::VDIVPDZrrkz,                X86::VDIVPDZrmkz,                0 },
  { X86::VDIVPSZ128rrkz,             X86::VDIVPSZ128rmkz,             0 },
  { X86::VDIVPSZ256rrkz,             X86::VDIVPSZ256rmkz,             0 },
  { X86::VDIVPSZrrkz,                X86::VDIVPSZrmkz,                0 },
  { X86::VDIVSDZrr_Intkz,            X86::VDIVSDZrm_Intkz,            TB_NO_REVERSE },
  { X86::VDIVSSZrr_Intkz,            X86::VDIVSSZrm_Intkz,            TB_NO_REVERSE },
  { X86::VDPBF16PSZ128r,             X86::VDPBF16PSZ128m,             0 },
  { X86::VDPBF16PSZ256r,             X86::VDPBF16PSZ256m,             0 },
  { X86::VDPBF16PSZr,                X86::VDPBF16PSZm,                0 },
  { X86::VEXP2PDZrk,                 X86::VEXP2PDZmk,                 0 },
  { X86::VEXP2PSZrk,                 X86::VEXP2PSZmk,                 0 },
  { X86::VEXPANDPDZ128rrk,           X86::VEXPANDPDZ128rmk,           TB_NO_REVERSE },
  { X86::VEXPANDPDZ256rrk,           X86::VEXPANDPDZ256rmk,           TB_NO_REVERSE },
  { X86::VEXPANDPDZrrk,              X86::VEXPANDPDZrmk,              TB_NO_REVERSE },
  { X86::VEXPANDPSZ128rrk,           X86::VEXPANDPSZ128rmk,           TB_NO_REVERSE },
  { X86::VEXPANDPSZ256rrk,           X86::VEXPANDPSZ256rmk,           TB_NO_REVERSE },
  { X86::VEXPANDPSZrrk,              X86::VEXPANDPSZrmk,              TB_NO_REVERSE },
  { X86::VFIXUPIMMPDZ128rri,         X86::VFIXUPIMMPDZ128rmi,         0 },
  { X86::VFIXUPIMMPDZ256rri,         X86::VFIXUPIMMPDZ256rmi,         0 },
  { X86::VFIXUPIMMPDZrri,            X86::VFIXUPIMMPDZrmi,            0 },
  { X86::VFIXUPIMMPSZ128rri,         X86::VFIXUPIMMPSZ128rmi,         0 },
  { X86::VFIXUPIMMPSZ256rri,         X86::VFIXUPIMMPSZ256rmi,         0 },
  { X86::VFIXUPIMMPSZrri,            X86::VFIXUPIMMPSZrmi,            0 },
  { X86::VFIXUPIMMSDZrri,            X86::VFIXUPIMMSDZrmi,            TB_NO_REVERSE },
  { X86::VFIXUPIMMSSZrri,            X86::VFIXUPIMMSSZrmi,            TB_NO_REVERSE },
  { X86::VFMADD132PDYr,              X86::VFMADD132PDYm,              0 },
  { X86::VFMADD132PDZ128r,           X86::VFMADD132PDZ128m,           0 },
  { X86::VFMADD132PDZ256r,           X86::VFMADD132PDZ256m,           0 },
  { X86::VFMADD132PDZr,              X86::VFMADD132PDZm,              0 },
  { X86::VFMADD132PDr,               X86::VFMADD132PDm,               0 },
  { X86::VFMADD132PSYr,              X86::VFMADD132PSYm,              0 },
  { X86::VFMADD132PSZ128r,           X86::VFMADD132PSZ128m,           0 },
  { X86::VFMADD132PSZ256r,           X86::VFMADD132PSZ256m,           0 },
  { X86::VFMADD132PSZr,              X86::VFMADD132PSZm,              0 },
  { X86::VFMADD132PSr,               X86::VFMADD132PSm,               0 },
  { X86::VFMADD132SDZr,              X86::VFMADD132SDZm,              0 },
  { X86::VFMADD132SDZr_Int,          X86::VFMADD132SDZm_Int,          TB_NO_REVERSE },
  { X86::VFMADD132SDr,               X86::VFMADD132SDm,               0 },
  { X86::VFMADD132SDr_Int,           X86::VFMADD132SDm_Int,           TB_NO_REVERSE },
  { X86::VFMADD132SSZr,              X86::VFMADD132SSZm,              0 },
  { X86::VFMADD132SSZr_Int,          X86::VFMADD132SSZm_Int,          TB_NO_REVERSE },
  { X86::VFMADD132SSr,               X86::VFMADD132SSm,               0 },
  { X86::VFMADD132SSr_Int,           X86::VFMADD132SSm_Int,           TB_NO_REVERSE },
  { X86::VFMADD213PDYr,              X86::VFMADD213PDYm,              0 },
  { X86::VFMADD213PDZ128r,           X86::VFMADD213PDZ128m,           0 },
  { X86::VFMADD213PDZ256r,           X86::VFMADD213PDZ256m,           0 },
  { X86::VFMADD213PDZr,              X86::VFMADD213PDZm,              0 },
  { X86::VFMADD213PDr,               X86::VFMADD213PDm,               0 },
  { X86::VFMADD213PSYr,              X86::VFMADD213PSYm,              0 },
  { X86::VFMADD213PSZ128r,           X86::VFMADD213PSZ128m,           0 },
  { X86::VFMADD213PSZ256r,           X86::VFMADD213PSZ256m,           0 },
  { X86::VFMADD213PSZr,              X86::VFMADD213PSZm,              0 },
  { X86::VFMADD213PSr,               X86::VFMADD213PSm,               0 },
  { X86::VFMADD213SDZr,              X86::VFMADD213SDZm,              0 },
  { X86::VFMADD213SDZr_Int,          X86::VFMADD213SDZm_Int,          TB_NO_REVERSE },
  { X86::VFMADD213SDr,               X86::VFMADD213SDm,               0 },
  { X86::VFMADD213SDr_Int,           X86::VFMADD213SDm_Int,           TB_NO_REVERSE },
  { X86::VFMADD213SSZr,              X86::VFMADD213SSZm,              0 },
  { X86::VFMADD213SSZr_Int,          X86::VFMADD213SSZm_Int,          TB_NO_REVERSE },
  { X86::VFMADD213SSr,               X86::VFMADD213SSm,               0 },
  { X86::VFMADD213SSr_Int,           X86::VFMADD213SSm_Int,           TB_NO_REVERSE },
  { X86::VFMADD231PDYr,              X86::VFMADD231PDYm,              0 },
  { X86::VFMADD231PDZ128r,           X86::VFMADD231PDZ128m,           0 },
  { X86::VFMADD231PDZ256r,           X86::VFMADD231PDZ256m,           0 },
  { X86::VFMADD231PDZr,              X86::VFMADD231PDZm,              0 },
  { X86::VFMADD231PDr,               X86::VFMADD231PDm,               0 },
  { X86::VFMADD231PSYr,              X86::VFMADD231PSYm,              0 },
  { X86::VFMADD231PSZ128r,           X86::VFMADD231PSZ128m,           0 },
  { X86::VFMADD231PSZ256r,           X86::VFMADD231PSZ256m,           0 },
  { X86::VFMADD231PSZr,              X86::VFMADD231PSZm,              0 },
  { X86::VFMADD231PSr,               X86::VFMADD231PSm,               0 },
  { X86::VFMADD231SDZr,              X86::VFMADD231SDZm,              0 },
  { X86::VFMADD231SDZr_Int,          X86::VFMADD231SDZm_Int,          TB_NO_REVERSE },
  { X86::VFMADD231SDr,               X86::VFMADD231SDm,               0 },
  { X86::VFMADD231SDr_Int,           X86::VFMADD231SDm_Int,           TB_NO_REVERSE },
  { X86::VFMADD231SSZr,              X86::VFMADD231SSZm,              0 },
  { X86::VFMADD231SSZr_Int,          X86::VFMADD231SSZm_Int,          TB_NO_REVERSE },
  { X86::VFMADD231SSr,               X86::VFMADD231SSm,               0 },
  { X86::VFMADD231SSr_Int,           X86::VFMADD231SSm_Int,           TB_NO_REVERSE },
  { X86::VFMADDPD4Yrr,               X86::VFMADDPD4Yrm,               0 },
  { X86::VFMADDPD4rr,                X86::VFMADDPD4rm,                0 },
  { X86::VFMADDPS4Yrr,               X86::VFMADDPS4Yrm,               0 },
  { X86::VFMADDPS4rr,                X86::VFMADDPS4rm,                0 },
  { X86::VFMADDSD4rr,                X86::VFMADDSD4rm,                0 },
  { X86::VFMADDSD4rr_Int,            X86::VFMADDSD4rm_Int,            TB_NO_REVERSE },
  { X86::VFMADDSS4rr,                X86::VFMADDSS4rm,                0 },
  { X86::VFMADDSS4rr_Int,            X86::VFMADDSS4rm_Int,            TB_NO_REVERSE },
  { X86::VFMADDSUB132PDYr,           X86::VFMADDSUB132PDYm,           0 },
  { X86::VFMADDSUB132PDZ128r,        X86::VFMADDSUB132PDZ128m,        0 },
  { X86::VFMADDSUB132PDZ256r,        X86::VFMADDSUB132PDZ256m,        0 },
  { X86::VFMADDSUB132PDZr,           X86::VFMADDSUB132PDZm,           0 },
  { X86::VFMADDSUB132PDr,            X86::VFMADDSUB132PDm,            0 },
  { X86::VFMADDSUB132PSYr,           X86::VFMADDSUB132PSYm,           0 },
  { X86::VFMADDSUB132PSZ128r,        X86::VFMADDSUB132PSZ128m,        0 },
  { X86::VFMADDSUB132PSZ256r,        X86::VFMADDSUB132PSZ256m,        0 },
  { X86::VFMADDSUB132PSZr,           X86::VFMADDSUB132PSZm,           0 },
  { X86::VFMADDSUB132PSr,            X86::VFMADDSUB132PSm,            0 },
  { X86::VFMADDSUB213PDYr,           X86::VFMADDSUB213PDYm,           0 },
  { X86::VFMADDSUB213PDZ128r,        X86::VFMADDSUB213PDZ128m,        0 },
  { X86::VFMADDSUB213PDZ256r,        X86::VFMADDSUB213PDZ256m,        0 },
  { X86::VFMADDSUB213PDZr,           X86::VFMADDSUB213PDZm,           0 },
  { X86::VFMADDSUB213PDr,            X86::VFMADDSUB213PDm,            0 },
  { X86::VFMADDSUB213PSYr,           X86::VFMADDSUB213PSYm,           0 },
  { X86::VFMADDSUB213PSZ128r,        X86::VFMADDSUB213PSZ128m,        0 },
  { X86::VFMADDSUB213PSZ256r,        X86::VFMADDSUB213PSZ256m,        0 },
  { X86::VFMADDSUB213PSZr,           X86::VFMADDSUB213PSZm,           0 },
  { X86::VFMADDSUB213PSr,            X86::VFMADDSUB213PSm,            0 },
  { X86::VFMADDSUB231PDYr,           X86::VFMADDSUB231PDYm,           0 },
  { X86::VFMADDSUB231PDZ128r,        X86::VFMADDSUB231PDZ128m,        0 },
  { X86::VFMADDSUB231PDZ256r,        X86::VFMADDSUB231PDZ256m,        0 },
  { X86::VFMADDSUB231PDZr,           X86::VFMADDSUB231PDZm,           0 },
  { X86::VFMADDSUB231PDr,            X86::VFMADDSUB231PDm,            0 },
  { X86::VFMADDSUB231PSYr,           X86::VFMADDSUB231PSYm,           0 },
  { X86::VFMADDSUB231PSZ128r,        X86::VFMADDSUB231PSZ128m,        0 },
  { X86::VFMADDSUB231PSZ256r,        X86::VFMADDSUB231PSZ256m,        0 },
  { X86::VFMADDSUB231PSZr,           X86::VFMADDSUB231PSZm,           0 },
  { X86::VFMADDSUB231PSr,            X86::VFMADDSUB231PSm,            0 },
  { X86::VFMADDSUBPD4Yrr,            X86::VFMADDSUBPD4Yrm,            0 },
  { X86::VFMADDSUBPD4rr,             X86::VFMADDSUBPD4rm,             0 },
  { X86::VFMADDSUBPS4Yrr,            X86::VFMADDSUBPS4Yrm,            0 },
  { X86::VFMADDSUBPS4rr,             X86::VFMADDSUBPS4rm,             0 },
  { X86::VFMSUB132PDYr,              X86::VFMSUB132PDYm,              0 },
  { X86::VFMSUB132PDZ128r,           X86::VFMSUB132PDZ128m,           0 },
  { X86::VFMSUB132PDZ256r,           X86::VFMSUB132PDZ256m,           0 },
  { X86::VFMSUB132PDZr,              X86::VFMSUB132PDZm,              0 },
  { X86::VFMSUB132PDr,               X86::VFMSUB132PDm,               0 },
  { X86::VFMSUB132PSYr,              X86::VFMSUB132PSYm,              0 },
  { X86::VFMSUB132PSZ128r,           X86::VFMSUB132PSZ128m,           0 },
  { X86::VFMSUB132PSZ256r,           X86::VFMSUB132PSZ256m,           0 },
  { X86::VFMSUB132PSZr,              X86::VFMSUB132PSZm,              0 },
  { X86::VFMSUB132PSr,               X86::VFMSUB132PSm,               0 },
  { X86::VFMSUB132SDZr,              X86::VFMSUB132SDZm,              0 },
  { X86::VFMSUB132SDZr_Int,          X86::VFMSUB132SDZm_Int,          TB_NO_REVERSE },
  { X86::VFMSUB132SDr,               X86::VFMSUB132SDm,               0 },
  { X86::VFMSUB132SDr_Int,           X86::VFMSUB132SDm_Int,           TB_NO_REVERSE },
  { X86::VFMSUB132SSZr,              X86::VFMSUB132SSZm,              0 },
  { X86::VFMSUB132SSZr_Int,          X86::VFMSUB132SSZm_Int,          TB_NO_REVERSE },
  { X86::VFMSUB132SSr,               X86::VFMSUB132SSm,               0 },
  { X86::VFMSUB132SSr_Int,           X86::VFMSUB132SSm_Int,           TB_NO_REVERSE },
  { X86::VFMSUB213PDYr,              X86::VFMSUB213PDYm,              0 },
  { X86::VFMSUB213PDZ128r,           X86::VFMSUB213PDZ128m,           0 },
  { X86::VFMSUB213PDZ256r,           X86::VFMSUB213PDZ256m,           0 },
  { X86::VFMSUB213PDZr,              X86::VFMSUB213PDZm,              0 },
  { X86::VFMSUB213PDr,               X86::VFMSUB213PDm,               0 },
  { X86::VFMSUB213PSYr,              X86::VFMSUB213PSYm,              0 },
  { X86::VFMSUB213PSZ128r,           X86::VFMSUB213PSZ128m,           0 },
  { X86::VFMSUB213PSZ256r,           X86::VFMSUB213PSZ256m,           0 },
  { X86::VFMSUB213PSZr,              X86::VFMSUB213PSZm,              0 },
  { X86::VFMSUB213PSr,               X86::VFMSUB213PSm,               0 },
  { X86::VFMSUB213SDZr,              X86::VFMSUB213SDZm,              0 },
  { X86::VFMSUB213SDZr_Int,          X86::VFMSUB213SDZm_Int,          TB_NO_REVERSE },
  { X86::VFMSUB213SDr,               X86::VFMSUB213SDm,               0 },
  { X86::VFMSUB213SDr_Int,           X86::VFMSUB213SDm_Int,           TB_NO_REVERSE },
  { X86::VFMSUB213SSZr,              X86::VFMSUB213SSZm,              0 },
  { X86::VFMSUB213SSZr_Int,          X86::VFMSUB213SSZm_Int,          TB_NO_REVERSE },
  { X86::VFMSUB213SSr,               X86::VFMSUB213SSm,               0 },
  { X86::VFMSUB213SSr_Int,           X86::VFMSUB213SSm_Int,           TB_NO_REVERSE },
  { X86::VFMSUB231PDYr,              X86::VFMSUB231PDYm,              0 },
  { X86::VFMSUB231PDZ128r,           X86::VFMSUB231PDZ128m,           0 },
  { X86::VFMSUB231PDZ256r,           X86::VFMSUB231PDZ256m,           0 },
  { X86::VFMSUB231PDZr,              X86::VFMSUB231PDZm,              0 },
  { X86::VFMSUB231PDr,               X86::VFMSUB231PDm,               0 },
  { X86::VFMSUB231PSYr,              X86::VFMSUB231PSYm,              0 },
  { X86::VFMSUB231PSZ128r,           X86::VFMSUB231PSZ128m,           0 },
  { X86::VFMSUB231PSZ256r,           X86::VFMSUB231PSZ256m,           0 },
  { X86::VFMSUB231PSZr,              X86::VFMSUB231PSZm,              0 },
  { X86::VFMSUB231PSr,               X86::VFMSUB231PSm,               0 },
  { X86::VFMSUB231SDZr,              X86::VFMSUB231SDZm,              0 },
  { X86::VFMSUB231SDZr_Int,          X86::VFMSUB231SDZm_Int,          TB_NO_REVERSE },
  { X86::VFMSUB231SDr,               X86::VFMSUB231SDm,               0 },
  { X86::VFMSUB231SDr_Int,           X86::VFMSUB231SDm_Int,           TB_NO_REVERSE },
  { X86::VFMSUB231SSZr,              X86::VFMSUB231SSZm,              0 },
  { X86::VFMSUB231SSZr_Int,          X86::VFMSUB231SSZm_Int,          TB_NO_REVERSE },
  { X86::VFMSUB231SSr,               X86::VFMSUB231SSm,               0 },
  { X86::VFMSUB231SSr_Int,           X86::VFMSUB231SSm_Int,           TB_NO_REVERSE },
  { X86::VFMSUBADD132PDYr,           X86::VFMSUBADD132PDYm,           0 },
  { X86::VFMSUBADD132PDZ128r,        X86::VFMSUBADD132PDZ128m,        0 },
  { X86::VFMSUBADD132PDZ256r,        X86::VFMSUBADD132PDZ256m,        0 },
  { X86::VFMSUBADD132PDZr,           X86::VFMSUBADD132PDZm,           0 },
  { X86::VFMSUBADD132PDr,            X86::VFMSUBADD132PDm,            0 },
  { X86::VFMSUBADD132PSYr,           X86::VFMSUBADD132PSYm,           0 },
  { X86::VFMSUBADD132PSZ128r,        X86::VFMSUBADD132PSZ128m,        0 },
  { X86::VFMSUBADD132PSZ256r,        X86::VFMSUBADD132PSZ256m,        0 },
  { X86::VFMSUBADD132PSZr,           X86::VFMSUBADD132PSZm,           0 },
  { X86::VFMSUBADD132PSr,            X86::VFMSUBADD132PSm,            0 },
  { X86::VFMSUBADD213PDYr,           X86::VFMSUBADD213PDYm,           0 },
  { X86::VFMSUBADD213PDZ128r,        X86::VFMSUBADD213PDZ128m,        0 },
  { X86::VFMSUBADD213PDZ256r,        X86::VFMSUBADD213PDZ256m,        0 },
  { X86::VFMSUBADD213PDZr,           X86::VFMSUBADD213PDZm,           0 },
  { X86::VFMSUBADD213PDr,            X86::VFMSUBADD213PDm,            0 },
  { X86::VFMSUBADD213PSYr,           X86::VFMSUBADD213PSYm,           0 },
  { X86::VFMSUBADD213PSZ128r,        X86::VFMSUBADD213PSZ128m,        0 },
  { X86::VFMSUBADD213PSZ256r,        X86::VFMSUBADD213PSZ256m,        0 },
  { X86::VFMSUBADD213PSZr,           X86::VFMSUBADD213PSZm,           0 },
  { X86::VFMSUBADD213PSr,            X86::VFMSUBADD213PSm,            0 },
  { X86::VFMSUBADD231PDYr,           X86::VFMSUBADD231PDYm,           0 },
  { X86::VFMSUBADD231PDZ128r,        X86::VFMSUBADD231PDZ128m,        0 },
  { X86::VFMSUBADD231PDZ256r,        X86::VFMSUBADD231PDZ256m,        0 },
  { X86::VFMSUBADD231PDZr,           X86::VFMSUBADD231PDZm,           0 },
  { X86::VFMSUBADD231PDr,            X86::VFMSUBADD231PDm,            0 },
  { X86::VFMSUBADD231PSYr,           X86::VFMSUBADD231PSYm,           0 },
  { X86::VFMSUBADD231PSZ128r,        X86::VFMSUBADD231PSZ128m,        0 },
  { X86::VFMSUBADD231PSZ256r,        X86::VFMSUBADD231PSZ256m,        0 },
  { X86::VFMSUBADD231PSZr,           X86::VFMSUBADD231PSZm,           0 },
  { X86::VFMSUBADD231PSr,            X86::VFMSUBADD231PSm,            0 },
  { X86::VFMSUBADDPD4Yrr,            X86::VFMSUBADDPD4Yrm,            0 },
  { X86::VFMSUBADDPD4rr,             X86::VFMSUBADDPD4rm,             0 },
  { X86::VFMSUBADDPS4Yrr,            X86::VFMSUBADDPS4Yrm,            0 },
  { X86::VFMSUBADDPS4rr,             X86::VFMSUBADDPS4rm,             0 },
  { X86::VFMSUBPD4Yrr,               X86::VFMSUBPD4Yrm,               0 },
  { X86::VFMSUBPD4rr,                X86::VFMSUBPD4rm,                0 },
  { X86::VFMSUBPS4Yrr,               X86::VFMSUBPS4Yrm,               0 },
  { X86::VFMSUBPS4rr,                X86::VFMSUBPS4rm,                0 },
  { X86::VFMSUBSD4rr,                X86::VFMSUBSD4rm,                0 },
  { X86::VFMSUBSD4rr_Int,            X86::VFMSUBSD4rm_Int,            TB_NO_REVERSE },
  { X86::VFMSUBSS4rr,                X86::VFMSUBSS4rm,                0 },
  { X86::VFMSUBSS4rr_Int,            X86::VFMSUBSS4rm_Int,            TB_NO_REVERSE },
  { X86::VFNMADD132PDYr,             X86::VFNMADD132PDYm,             0 },
  { X86::VFNMADD132PDZ128r,          X86::VFNMADD132PDZ128m,          0 },
  { X86::VFNMADD132PDZ256r,          X86::VFNMADD132PDZ256m,          0 },
  { X86::VFNMADD132PDZr,             X86::VFNMADD132PDZm,             0 },
  { X86::VFNMADD132PDr,              X86::VFNMADD132PDm,              0 },
  { X86::VFNMADD132PSYr,             X86::VFNMADD132PSYm,             0 },
  { X86::VFNMADD132PSZ128r,          X86::VFNMADD132PSZ128m,          0 },
  { X86::VFNMADD132PSZ256r,          X86::VFNMADD132PSZ256m,          0 },
  { X86::VFNMADD132PSZr,             X86::VFNMADD132PSZm,             0 },
  { X86::VFNMADD132PSr,              X86::VFNMADD132PSm,              0 },
  { X86::VFNMADD132SDZr,             X86::VFNMADD132SDZm,             0 },
  { X86::VFNMADD132SDZr_Int,         X86::VFNMADD132SDZm_Int,         TB_NO_REVERSE },
  { X86::VFNMADD132SDr,              X86::VFNMADD132SDm,              0 },
  { X86::VFNMADD132SDr_Int,          X86::VFNMADD132SDm_Int,          TB_NO_REVERSE },
  { X86::VFNMADD132SSZr,             X86::VFNMADD132SSZm,             0 },
  { X86::VFNMADD132SSZr_Int,         X86::VFNMADD132SSZm_Int,         TB_NO_REVERSE },
  { X86::VFNMADD132SSr,              X86::VFNMADD132SSm,              0 },
  { X86::VFNMADD132SSr_Int,          X86::VFNMADD132SSm_Int,          TB_NO_REVERSE },
  { X86::VFNMADD213PDYr,             X86::VFNMADD213PDYm,             0 },
  { X86::VFNMADD213PDZ128r,          X86::VFNMADD213PDZ128m,          0 },
  { X86::VFNMADD213PDZ256r,          X86::VFNMADD213PDZ256m,          0 },
  { X86::VFNMADD213PDZr,             X86::VFNMADD213PDZm,             0 },
  { X86::VFNMADD213PDr,              X86::VFNMADD213PDm,              0 },
  { X86::VFNMADD213PSYr,             X86::VFNMADD213PSYm,             0 },
  { X86::VFNMADD213PSZ128r,          X86::VFNMADD213PSZ128m,          0 },
  { X86::VFNMADD213PSZ256r,          X86::VFNMADD213PSZ256m,          0 },
  { X86::VFNMADD213PSZr,             X86::VFNMADD213PSZm,             0 },
  { X86::VFNMADD213PSr,              X86::VFNMADD213PSm,              0 },
  { X86::VFNMADD213SDZr,             X86::VFNMADD213SDZm,             0 },
  { X86::VFNMADD213SDZr_Int,         X86::VFNMADD213SDZm_Int,         TB_NO_REVERSE },
  { X86::VFNMADD213SDr,              X86::VFNMADD213SDm,              0 },
  { X86::VFNMADD213SDr_Int,          X86::VFNMADD213SDm_Int,          TB_NO_REVERSE },
  { X86::VFNMADD213SSZr,             X86::VFNMADD213SSZm,             0 },
  { X86::VFNMADD213SSZr_Int,         X86::VFNMADD213SSZm_Int,         TB_NO_REVERSE },
  { X86::VFNMADD213SSr,              X86::VFNMADD213SSm,              0 },
  { X86::VFNMADD213SSr_Int,          X86::VFNMADD213SSm_Int,          TB_NO_REVERSE },
  { X86::VFNMADD231PDYr,             X86::VFNMADD231PDYm,             0 },
  { X86::VFNMADD231PDZ128r,          X86::VFNMADD231PDZ128m,          0 },
  { X86::VFNMADD231PDZ256r,          X86::VFNMADD231PDZ256m,          0 },
  { X86::VFNMADD231PDZr,             X86::VFNMADD231PDZm,             0 },
  { X86::VFNMADD231PDr,              X86::VFNMADD231PDm,              0 },
  { X86::VFNMADD231PSYr,             X86::VFNMADD231PSYm,             0 },
  { X86::VFNMADD231PSZ128r,          X86::VFNMADD231PSZ128m,          0 },
  { X86::VFNMADD231PSZ256r,          X86::VFNMADD231PSZ256m,          0 },
  { X86::VFNMADD231PSZr,             X86::VFNMADD231PSZm,             0 },
  { X86::VFNMADD231PSr,              X86::VFNMADD231PSm,              0 },
  { X86::VFNMADD231SDZr,             X86::VFNMADD231SDZm,             0 },
  { X86::VFNMADD231SDZr_Int,         X86::VFNMADD231SDZm_Int,         TB_NO_REVERSE },
  { X86::VFNMADD231SDr,              X86::VFNMADD231SDm,              0 },
  { X86::VFNMADD231SDr_Int,          X86::VFNMADD231SDm_Int,          TB_NO_REVERSE },
  { X86::VFNMADD231SSZr,             X86::VFNMADD231SSZm,             0 },
  { X86::VFNMADD231SSZr_Int,         X86::VFNMADD231SSZm_Int,         TB_NO_REVERSE },
  { X86::VFNMADD231SSr,              X86::VFNMADD231SSm,              0 },
  { X86::VFNMADD231SSr_Int,          X86::VFNMADD231SSm_Int,          TB_NO_REVERSE },
  { X86::VFNMADDPD4Yrr,              X86::VFNMADDPD4Yrm,              0 },
  { X86::VFNMADDPD4rr,               X86::VFNMADDPD4rm,               0 },
  { X86::VFNMADDPS4Yrr,              X86::VFNMADDPS4Yrm,              0 },
  { X86::VFNMADDPS4rr,               X86::VFNMADDPS4rm,               0 },
  { X86::VFNMADDSD4rr,               X86::VFNMADDSD4rm,               0 },
  { X86::VFNMADDSD4rr_Int,           X86::VFNMADDSD4rm_Int,           TB_NO_REVERSE },
  { X86::VFNMADDSS4rr,               X86::VFNMADDSS4rm,               0 },
  { X86::VFNMADDSS4rr_Int,           X86::VFNMADDSS4rm_Int,           TB_NO_REVERSE },
  { X86::VFNMSUB132PDYr,             X86::VFNMSUB132PDYm,             0 },
  { X86::VFNMSUB132PDZ128r,          X86::VFNMSUB132PDZ128m,          0 },
  { X86::VFNMSUB132PDZ256r,          X86::VFNMSUB132PDZ256m,          0 },
  { X86::VFNMSUB132PDZr,             X86::VFNMSUB132PDZm,             0 },
  { X86::VFNMSUB132PDr,              X86::VFNMSUB132PDm,              0 },
  { X86::VFNMSUB132PSYr,             X86::VFNMSUB132PSYm,             0 },
  { X86::VFNMSUB132PSZ128r,          X86::VFNMSUB132PSZ128m,          0 },
  { X86::VFNMSUB132PSZ256r,          X86::VFNMSUB132PSZ256m,          0 },
  { X86::VFNMSUB132PSZr,             X86::VFNMSUB132PSZm,             0 },
  { X86::VFNMSUB132PSr,              X86::VFNMSUB132PSm,              0 },
  { X86::VFNMSUB132SDZr,             X86::VFNMSUB132SDZm,             0 },
  { X86::VFNMSUB132SDZr_Int,         X86::VFNMSUB132SDZm_Int,         TB_NO_REVERSE },
  { X86::VFNMSUB132SDr,              X86::VFNMSUB132SDm,              0 },
  { X86::VFNMSUB132SDr_Int,          X86::VFNMSUB132SDm_Int,          TB_NO_REVERSE },
  { X86::VFNMSUB132SSZr,             X86::VFNMSUB132SSZm,             0 },
  { X86::VFNMSUB132SSZr_Int,         X86::VFNMSUB132SSZm_Int,         TB_NO_REVERSE },
  { X86::VFNMSUB132SSr,              X86::VFNMSUB132SSm,              0 },
  { X86::VFNMSUB132SSr_Int,          X86::VFNMSUB132SSm_Int,          TB_NO_REVERSE },
  { X86::VFNMSUB213PDYr,             X86::VFNMSUB213PDYm,             0 },
  { X86::VFNMSUB213PDZ128r,          X86::VFNMSUB213PDZ128m,          0 },
  { X86::VFNMSUB213PDZ256r,          X86::VFNMSUB213PDZ256m,          0 },
  { X86::VFNMSUB213PDZr,             X86::VFNMSUB213PDZm,             0 },
  { X86::VFNMSUB213PDr,              X86::VFNMSUB213PDm,              0 },
  { X86::VFNMSUB213PSYr,             X86::VFNMSUB213PSYm,             0 },
  { X86::VFNMSUB213PSZ128r,          X86::VFNMSUB213PSZ128m,          0 },
  { X86::VFNMSUB213PSZ256r,          X86::VFNMSUB213PSZ256m,          0 },
  { X86::VFNMSUB213PSZr,             X86::VFNMSUB213PSZm,             0 },
  { X86::VFNMSUB213PSr,              X86::VFNMSUB213PSm,              0 },
  { X86::VFNMSUB213SDZr,             X86::VFNMSUB213SDZm,             0 },
  { X86::VFNMSUB213SDZr_Int,         X86::VFNMSUB213SDZm_Int,         TB_NO_REVERSE },
  { X86::VFNMSUB213SDr,              X86::VFNMSUB213SDm,              0 },
  { X86::VFNMSUB213SDr_Int,          X86::VFNMSUB213SDm_Int,          TB_NO_REVERSE },
  { X86::VFNMSUB213SSZr,             X86::VFNMSUB213SSZm,             0 },
  { X86::VFNMSUB213SSZr_Int,         X86::VFNMSUB213SSZm_Int,         TB_NO_REVERSE },
  { X86::VFNMSUB213SSr,              X86::VFNMSUB213SSm,              0 },
  { X86::VFNMSUB213SSr_Int,          X86::VFNMSUB213SSm_Int,          TB_NO_REVERSE },
  { X86::VFNMSUB231PDYr,             X86::VFNMSUB231PDYm,             0 },
  { X86::VFNMSUB231PDZ128r,          X86::VFNMSUB231PDZ128m,          0 },
  { X86::VFNMSUB231PDZ256r,          X86::VFNMSUB231PDZ256m,          0 },
  { X86::VFNMSUB231PDZr,             X86::VFNMSUB231PDZm,             0 },
  { X86::VFNMSUB231PDr,              X86::VFNMSUB231PDm,              0 },
  { X86::VFNMSUB231PSYr,             X86::VFNMSUB231PSYm,             0 },
  { X86::VFNMSUB231PSZ128r,          X86::VFNMSUB231PSZ128m,          0 },
  { X86::VFNMSUB231PSZ256r,          X86::VFNMSUB231PSZ256m,          0 },
  { X86::VFNMSUB231PSZr,             X86::VFNMSUB231PSZm,             0 },
  { X86::VFNMSUB231PSr,              X86::VFNMSUB231PSm,              0 },
  { X86::VFNMSUB231SDZr,             X86::VFNMSUB231SDZm,             0 },
  { X86::VFNMSUB231SDZr_Int,         X86::VFNMSUB231SDZm_Int,         TB_NO_REVERSE },
  { X86::VFNMSUB231SDr,              X86::VFNMSUB231SDm,              0 },
  { X86::VFNMSUB231SDr_Int,          X86::VFNMSUB231SDm_Int,          TB_NO_REVERSE },
  { X86::VFNMSUB231SSZr,             X86::VFNMSUB231SSZm,             0 },
  { X86::VFNMSUB231SSZr_Int,         X86::VFNMSUB231SSZm_Int,         TB_NO_REVERSE },
  { X86::VFNMSUB231SSr,              X86::VFNMSUB231SSm,              0 },
  { X86::VFNMSUB231SSr_Int,          X86::VFNMSUB231SSm_Int,          TB_NO_REVERSE },
  { X86::VFNMSUBPD4Yrr,              X86::VFNMSUBPD4Yrm,              0 },
  { X86::VFNMSUBPD4rr,               X86::VFNMSUBPD4rm,               0 },
  { X86::VFNMSUBPS4Yrr,              X86::VFNMSUBPS4Yrm,              0 },
  { X86::VFNMSUBPS4rr,               X86::VFNMSUBPS4rm,               0 },
  { X86::VFNMSUBSD4rr,               X86::VFNMSUBSD4rm,               0 },
  { X86::VFNMSUBSD4rr_Int,           X86::VFNMSUBSD4rm_Int,           TB_NO_REVERSE },
  { X86::VFNMSUBSS4rr,               X86::VFNMSUBSS4rm,               0 },
  { X86::VFNMSUBSS4rr_Int,           X86::VFNMSUBSS4rm_Int,           TB_NO_REVERSE },
  { X86::VGETEXPPDZ128rk,            X86::VGETEXPPDZ128mk,            0 },
  { X86::VGETEXPPDZ256rk,            X86::VGETEXPPDZ256mk,            0 },
  { X86::VGETEXPPDZrk,               X86::VGETEXPPDZmk,               0 },
  { X86::VGETEXPPSZ128rk,            X86::VGETEXPPSZ128mk,            0 },
  { X86::VGETEXPPSZ256rk,            X86::VGETEXPPSZ256mk,            0 },
  { X86::VGETEXPPSZrk,               X86::VGETEXPPSZmk,               0 },
  { X86::VGETEXPSDZrkz,              X86::VGETEXPSDZmkz,              TB_NO_REVERSE },
  { X86::VGETEXPSSZrkz,              X86::VGETEXPSSZmkz,              TB_NO_REVERSE },
  { X86::VGETMANTPDZ128rrik,         X86::VGETMANTPDZ128rmik,         0 },
  { X86::VGETMANTPDZ256rrik,         X86::VGETMANTPDZ256rmik,         0 },
  { X86::VGETMANTPDZrrik,            X86::VGETMANTPDZrmik,            0 },
  { X86::VGETMANTPSZ128rrik,         X86::VGETMANTPSZ128rmik,         0 },
  { X86::VGETMANTPSZ256rrik,         X86::VGETMANTPSZ256rmik,         0 },
  { X86::VGETMANTPSZrrik,            X86::VGETMANTPSZrmik,            0 },
  { X86::VGETMANTSDZrrikz,           X86::VGETMANTSDZrmikz,           TB_NO_REVERSE },
  { X86::VGETMANTSSZrrikz,           X86::VGETMANTSSZrmikz,           TB_NO_REVERSE },
  { X86::VGF2P8AFFINEINVQBZ128rrikz, X86::VGF2P8AFFINEINVQBZ128rmikz, 0 },
  { X86::VGF2P8AFFINEINVQBZ256rrikz, X86::VGF2P8AFFINEINVQBZ256rmikz, 0 },
  { X86::VGF2P8AFFINEINVQBZrrikz,    X86::VGF2P8AFFINEINVQBZrmikz,    0 },
  { X86::VGF2P8AFFINEQBZ128rrikz,    X86::VGF2P8AFFINEQBZ128rmikz,    0 },
  { X86::VGF2P8AFFINEQBZ256rrikz,    X86::VGF2P8AFFINEQBZ256rmikz,    0 },
  { X86::VGF2P8AFFINEQBZrrikz,       X86::VGF2P8AFFINEQBZrmikz,       0 },
  { X86::VGF2P8MULBZ128rrkz,         X86::VGF2P8MULBZ128rmkz,         0 },
  { X86::VGF2P8MULBZ256rrkz,         X86::VGF2P8MULBZ256rmkz,         0 },
  { X86::VGF2P8MULBZrrkz,            X86::VGF2P8MULBZrmkz,            0 },
  { X86::VINSERTF32x4Z256rrkz,       X86::VINSERTF32x4Z256rmkz,       0 },
  { X86::VINSERTF32x4Zrrkz,          X86::VINSERTF32x4Zrmkz,          0 },
  { X86::VINSERTF32x8Zrrkz,          X86::VINSERTF32x8Zrmkz,          0 },
  { X86::VINSERTF64x2Z256rrkz,       X86::VINSERTF64x2Z256rmkz,       0 },
  { X86::VINSERTF64x2Zrrkz,          X86::VINSERTF64x2Zrmkz,          0 },
  { X86::VINSERTF64x4Zrrkz,          X86::VINSERTF64x4Zrmkz,          0 },
  { X86::VINSERTI32x4Z256rrkz,       X86::VINSERTI32x4Z256rmkz,       0 },
  { X86::VINSERTI32x4Zrrkz,          X86::VINSERTI32x4Zrmkz,          0 },
  { X86::VINSERTI32x8Zrrkz,          X86::VINSERTI32x8Zrmkz,          0 },
  { X86::VINSERTI64x2Z256rrkz,       X86::VINSERTI64x2Z256rmkz,       0 },
  { X86::VINSERTI64x2Zrrkz,          X86::VINSERTI64x2Zrmkz,          0 },
  { X86::VINSERTI64x4Zrrkz,          X86::VINSERTI64x4Zrmkz,          0 },
  { X86::VMAXCPDZ128rrkz,            X86::VMAXCPDZ128rmkz,            0 },
  { X86::VMAXCPDZ256rrkz,            X86::VMAXCPDZ256rmkz,            0 },
  { X86::VMAXCPDZrrkz,               X86::VMAXCPDZrmkz,               0 },
  { X86::VMAXCPSZ128rrkz,            X86::VMAXCPSZ128rmkz,            0 },
  { X86::VMAXCPSZ256rrkz,            X86::VMAXCPSZ256rmkz,            0 },
  { X86::VMAXCPSZrrkz,               X86::VMAXCPSZrmkz,               0 },
  { X86::VMAXPDZ128rrkz,             X86::VMAXPDZ128rmkz,             0 },
  { X86::VMAXPDZ256rrkz,             X86::VMAXPDZ256rmkz,             0 },
  { X86::VMAXPDZrrkz,                X86::VMAXPDZrmkz,                0 },
  { X86::VMAXPSZ128rrkz,             X86::VMAXPSZ128rmkz,             0 },
  { X86::VMAXPSZ256rrkz,             X86::VMAXPSZ256rmkz,             0 },
  { X86::VMAXPSZrrkz,                X86::VMAXPSZrmkz,                0 },
  { X86::VMAXSDZrr_Intkz,            X86::VMAXSDZrm_Intkz,            TB_NO_REVERSE },
  { X86::VMAXSSZrr_Intkz,            X86::VMAXSSZrm_Intkz,            TB_NO_REVERSE },
  { X86::VMINCPDZ128rrkz,            X86::VMINCPDZ128rmkz,            0 },
  { X86::VMINCPDZ256rrkz,            X86::VMINCPDZ256rmkz,            0 },
  { X86::VMINCPDZrrkz,               X86::VMINCPDZrmkz,               0 },
  { X86::VMINCPSZ128rrkz,            X86::VMINCPSZ128rmkz,            0 },
  { X86::VMINCPSZ256rrkz,            X86::VMINCPSZ256rmkz,            0 },
  { X86::VMINCPSZrrkz,               X86::VMINCPSZrmkz,               0 },
  { X86::VMINPDZ128rrkz,             X86::VMINPDZ128rmkz,             0 },
  { X86::VMINPDZ256rrkz,             X86::VMINPDZ256rmkz,             0 },
  { X86::VMINPDZrrkz,                X86::VMINPDZrmkz,                0 },
  { X86::VMINPSZ128rrkz,             X86::VMINPSZ128rmkz,             0 },
  { X86::VMINPSZ256rrkz,             X86::VMINPSZ256rmkz,             0 },
  { X86::VMINPSZrrkz,                X86::VMINPSZrmkz,                0 },
  { X86::VMINSDZrr_Intkz,            X86::VMINSDZrm_Intkz,            TB_NO_REVERSE },
  { X86::VMINSSZrr_Intkz,            X86::VMINSSZrm_Intkz,            TB_NO_REVERSE },
  { X86::VMOVAPDZ128rrk,             X86::VMOVAPDZ128rmk,             TB_NO_REVERSE | TB_ALIGN_16 },
  { X86::VMOVAPDZ256rrk,             X86::VMOVAPDZ256rmk,             TB_NO_REVERSE | TB_ALIGN_32 },
  { X86::VMOVAPDZrrk,                X86::VMOVAPDZrmk,                TB_NO_REVERSE | TB_ALIGN_64 },
  { X86::VMOVAPSZ128rrk,             X86::VMOVAPSZ128rmk,             TB_NO_REVERSE | TB_ALIGN_16 },
  { X86::VMOVAPSZ256rrk,             X86::VMOVAPSZ256rmk,             TB_NO_REVERSE | TB_ALIGN_32 },
  { X86::VMOVAPSZrrk,                X86::VMOVAPSZrmk,                TB_NO_REVERSE | TB_ALIGN_64 },
  { X86::VMOVDDUPZ128rrk,            X86::VMOVDDUPZ128rmk,            TB_NO_REVERSE },
  { X86::VMOVDDUPZ256rrk,            X86::VMOVDDUPZ256rmk,            0 },
  { X86::VMOVDDUPZrrk,               X86::VMOVDDUPZrmk,               0 },
  { X86::VMOVDQA32Z128rrk,           X86::VMOVDQA32Z128rmk,           TB_NO_REVERSE | TB_ALIGN_16 },
  { X86::VMOVDQA32Z256rrk,           X86::VMOVDQA32Z256rmk,           TB_NO_REVERSE | TB_ALIGN_32 },
  { X86::VMOVDQA32Zrrk,              X86::VMOVDQA32Zrmk,              TB_NO_REVERSE | TB_ALIGN_64 },
  { X86::VMOVDQA64Z128rrk,           X86::VMOVDQA64Z128rmk,           TB_NO_REVERSE | TB_ALIGN_16 },
  { X86::VMOVDQA64Z256rrk,           X86::VMOVDQA64Z256rmk,           TB_NO_REVERSE | TB_ALIGN_32 },
  { X86::VMOVDQA64Zrrk,              X86::VMOVDQA64Zrmk,              TB_NO_REVERSE | TB_ALIGN_64 },
  { X86::VMOVDQU16Z128rrk,           X86::VMOVDQU16Z128rmk,           TB_NO_REVERSE },
  { X86::VMOVDQU16Z256rrk,           X86::VMOVDQU16Z256rmk,           TB_NO_REVERSE },
  { X86::VMOVDQU16Zrrk,              X86::VMOVDQU16Zrmk,              TB_NO_REVERSE },
  { X86::VMOVDQU32Z128rrk,           X86::VMOVDQU32Z128rmk,           TB_NO_REVERSE },
  { X86::VMOVDQU32Z256rrk,           X86::VMOVDQU32Z256rmk,           TB_NO_REVERSE },
  { X86::VMOVDQU32Zrrk,              X86::VMOVDQU32Zrmk,              TB_NO_REVERSE },
  { X86::VMOVDQU64Z128rrk,           X86::VMOVDQU64Z128rmk,           TB_NO_REVERSE },
  { X86::VMOVDQU64Z256rrk,           X86::VMOVDQU64Z256rmk,           TB_NO_REVERSE },
  { X86::VMOVDQU64Zrrk,              X86::VMOVDQU64Zrmk,              TB_NO_REVERSE },
  { X86::VMOVDQU8Z128rrk,            X86::VMOVDQU8Z128rmk,            TB_NO_REVERSE },
  { X86::VMOVDQU8Z256rrk,            X86::VMOVDQU8Z256rmk,            TB_NO_REVERSE },
  { X86::VMOVDQU8Zrrk,               X86::VMOVDQU8Zrmk,               TB_NO_REVERSE },
  { X86::VMOVSHDUPZ128rrk,           X86::VMOVSHDUPZ128rmk,           0 },
  { X86::VMOVSHDUPZ256rrk,           X86::VMOVSHDUPZ256rmk,           0 },
  { X86::VMOVSHDUPZrrk,              X86::VMOVSHDUPZrmk,              0 },
  { X86::VMOVSLDUPZ128rrk,           X86::VMOVSLDUPZ128rmk,           0 },
  { X86::VMOVSLDUPZ256rrk,           X86::VMOVSLDUPZ256rmk,           0 },
  { X86::VMOVSLDUPZrrk,              X86::VMOVSLDUPZrmk,              0 },
  { X86::VMOVUPDZ128rrk,             X86::VMOVUPDZ128rmk,             TB_NO_REVERSE },
  { X86::VMOVUPDZ256rrk,             X86::VMOVUPDZ256rmk,             TB_NO_REVERSE },
  { X86::VMOVUPDZrrk,                X86::VMOVUPDZrmk,                TB_NO_REVERSE },
  { X86::VMOVUPSZ128rrk,             X86::VMOVUPSZ128rmk,             TB_NO_REVERSE },
  { X86::VMOVUPSZ256rrk,             X86::VMOVUPSZ256rmk,             TB_NO_REVERSE },
  { X86::VMOVUPSZrrk,                X86::VMOVUPSZrmk,                TB_NO_REVERSE },
  { X86::VMULPDZ128rrkz,             X86::VMULPDZ128rmkz,             0 },
  { X86::VMULPDZ256rrkz,             X86::VMULPDZ256rmkz,             0 },
  { X86::VMULPDZrrkz,                X86::VMULPDZrmkz,                0 },
  { X86::VMULPSZ128rrkz,             X86::VMULPSZ128rmkz,             0 },
  { X86::VMULPSZ256rrkz,             X86::VMULPSZ256rmkz,             0 },
  { X86::VMULPSZrrkz,                X86::VMULPSZrmkz,                0 },
  { X86::VMULSDZrr_Intkz,            X86::VMULSDZrm_Intkz,            TB_NO_REVERSE },
  { X86::VMULSSZrr_Intkz,            X86::VMULSSZrm_Intkz,            TB_NO_REVERSE },
  { X86::VORPDZ128rrkz,              X86::VORPDZ128rmkz,              0 },
  { X86::VORPDZ256rrkz,              X86::VORPDZ256rmkz,              0 },
  { X86::VORPDZrrkz,                 X86::VORPDZrmkz,                 0 },
  { X86::VORPSZ128rrkz,              X86::VORPSZ128rmkz,              0 },
  { X86::VORPSZ256rrkz,              X86::VORPSZ256rmkz,              0 },
  { X86::VORPSZrrkz,                 X86::VORPSZrmkz,                 0 },
  { X86::VPABSBZ128rrk,              X86::VPABSBZ128rmk,              0 },
  { X86::VPABSBZ256rrk,              X86::VPABSBZ256rmk,              0 },
  { X86::VPABSBZrrk,                 X86::VPABSBZrmk,                 0 },
  { X86::VPABSDZ128rrk,              X86::VPABSDZ128rmk,              0 },
  { X86::VPABSDZ256rrk,              X86::VPABSDZ256rmk,              0 },
  { X86::VPABSDZrrk,                 X86::VPABSDZrmk,                 0 },
  { X86::VPABSQZ128rrk,              X86::VPABSQZ128rmk,              0 },
  { X86::VPABSQZ256rrk,              X86::VPABSQZ256rmk,              0 },
  { X86::VPABSQZrrk,                 X86::VPABSQZrmk,                 0 },
  { X86::VPABSWZ128rrk,              X86::VPABSWZ128rmk,              0 },
  { X86::VPABSWZ256rrk,              X86::VPABSWZ256rmk,              0 },
  { X86::VPABSWZrrk,                 X86::VPABSWZrmk,                 0 },
  { X86::VPACKSSDWZ128rrkz,          X86::VPACKSSDWZ128rmkz,          0 },
  { X86::VPACKSSDWZ256rrkz,          X86::VPACKSSDWZ256rmkz,          0 },
  { X86::VPACKSSDWZrrkz,             X86::VPACKSSDWZrmkz,             0 },
  { X86::VPACKSSWBZ128rrkz,          X86::VPACKSSWBZ128rmkz,          0 },
  { X86::VPACKSSWBZ256rrkz,          X86::VPACKSSWBZ256rmkz,          0 },
  { X86::VPACKSSWBZrrkz,             X86::VPACKSSWBZrmkz,             0 },
  { X86::VPACKUSDWZ128rrkz,          X86::VPACKUSDWZ128rmkz,          0 },
  { X86::VPACKUSDWZ256rrkz,          X86::VPACKUSDWZ256rmkz,          0 },
  { X86::VPACKUSDWZrrkz,             X86::VPACKUSDWZrmkz,             0 },
  { X86::VPACKUSWBZ128rrkz,          X86::VPACKUSWBZ128rmkz,          0 },
  { X86::VPACKUSWBZ256rrkz,          X86::VPACKUSWBZ256rmkz,          0 },
  { X86::VPACKUSWBZrrkz,             X86::VPACKUSWBZrmkz,             0 },
  { X86::VPADDBZ128rrkz,             X86::VPADDBZ128rmkz,             0 },
  { X86::VPADDBZ256rrkz,             X86::VPADDBZ256rmkz,             0 },
  { X86::VPADDBZrrkz,                X86::VPADDBZrmkz,                0 },
  { X86::VPADDDZ128rrkz,             X86::VPADDDZ128rmkz,             0 },
  { X86::VPADDDZ256rrkz,             X86::VPADDDZ256rmkz,             0 },
  { X86::VPADDDZrrkz,                X86::VPADDDZrmkz,                0 },
  { X86::VPADDQZ128rrkz,             X86::VPADDQZ128rmkz,             0 },
  { X86::VPADDQZ256rrkz,             X86::VPADDQZ256rmkz,             0 },
  { X86::VPADDQZrrkz,                X86::VPADDQZrmkz,                0 },
  { X86::VPADDSBZ128rrkz,            X86::VPADDSBZ128rmkz,            0 },
  { X86::VPADDSBZ256rrkz,            X86::VPADDSBZ256rmkz,            0 },
  { X86::VPADDSBZrrkz,               X86::VPADDSBZrmkz,               0 },
  { X86::VPADDSWZ128rrkz,            X86::VPADDSWZ128rmkz,            0 },
  { X86::VPADDSWZ256rrkz,            X86::VPADDSWZ256rmkz,            0 },
  { X86::VPADDSWZrrkz,               X86::VPADDSWZrmkz,               0 },
  { X86::VPADDUSBZ128rrkz,           X86::VPADDUSBZ128rmkz,           0 },
  { X86::VPADDUSBZ256rrkz,           X86::VPADDUSBZ256rmkz,           0 },
  { X86::VPADDUSBZrrkz,              X86::VPADDUSBZrmkz,              0 },
  { X86::VPADDUSWZ128rrkz,           X86::VPADDUSWZ128rmkz,           0 },
  { X86::VPADDUSWZ256rrkz,           X86::VPADDUSWZ256rmkz,           0 },
  { X86::VPADDUSWZrrkz,              X86::VPADDUSWZrmkz,              0 },
  { X86::VPADDWZ128rrkz,             X86::VPADDWZ128rmkz,             0 },
  { X86::VPADDWZ256rrkz,             X86::VPADDWZ256rmkz,             0 },
  { X86::VPADDWZrrkz,                X86::VPADDWZrmkz,                0 },
  { X86::VPALIGNRZ128rrikz,          X86::VPALIGNRZ128rmikz,          0 },
  { X86::VPALIGNRZ256rrikz,          X86::VPALIGNRZ256rmikz,          0 },
  { X86::VPALIGNRZrrikz,             X86::VPALIGNRZrmikz,             0 },
  { X86::VPANDDZ128rrkz,             X86::VPANDDZ128rmkz,             0 },
  { X86::VPANDDZ256rrkz,             X86::VPANDDZ256rmkz,             0 },
  { X86::VPANDDZrrkz,                X86::VPANDDZrmkz,                0 },
  { X86::VPANDNDZ128rrkz,            X86::VPANDNDZ128rmkz,            0 },
  { X86::VPANDNDZ256rrkz,            X86::VPANDNDZ256rmkz,            0 },
  { X86::VPANDNDZrrkz,               X86::VPANDNDZrmkz,               0 },
  { X86::VPANDNQZ128rrkz,            X86::VPANDNQZ128rmkz,            0 },
  { X86::VPANDNQZ256rrkz,            X86::VPANDNQZ256rmkz,            0 },
  { X86::VPANDNQZrrkz,               X86::VPANDNQZrmkz,               0 },
  { X86::VPANDQZ128rrkz,             X86::VPANDQZ128rmkz,             0 },
  { X86::VPANDQZ256rrkz,             X86::VPANDQZ256rmkz,             0 },
  { X86::VPANDQZrrkz,                X86::VPANDQZrmkz,                0 },
  { X86::VPAVGBZ128rrkz,             X86::VPAVGBZ128rmkz,             0 },
  { X86::VPAVGBZ256rrkz,             X86::VPAVGBZ256rmkz,             0 },
  { X86::VPAVGBZrrkz,                X86::VPAVGBZrmkz,                0 },
  { X86::VPAVGWZ128rrkz,             X86::VPAVGWZ128rmkz,             0 },
  { X86::VPAVGWZ256rrkz,             X86::VPAVGWZ256rmkz,             0 },
  { X86::VPAVGWZrrkz,                X86::VPAVGWZrmkz,                0 },
  { X86::VPBLENDMBZ128rrk,           X86::VPBLENDMBZ128rmk,           0 },
  { X86::VPBLENDMBZ256rrk,           X86::VPBLENDMBZ256rmk,           0 },
  { X86::VPBLENDMBZrrk,              X86::VPBLENDMBZrmk,              0 },
  { X86::VPBLENDMDZ128rrk,           X86::VPBLENDMDZ128rmk,           0 },
  { X86::VPBLENDMDZ256rrk,           X86::VPBLENDMDZ256rmk,           0 },
  { X86::VPBLENDMDZrrk,              X86::VPBLENDMDZrmk,              0 },
  { X86::VPBLENDMQZ128rrk,           X86::VPBLENDMQZ128rmk,           0 },
  { X86::VPBLENDMQZ256rrk,           X86::VPBLENDMQZ256rmk,           0 },
  { X86::VPBLENDMQZrrk,              X86::VPBLENDMQZrmk,              0 },
  { X86::VPBLENDMWZ128rrk,           X86::VPBLENDMWZ128rmk,           0 },
  { X86::VPBLENDMWZ256rrk,           X86::VPBLENDMWZ256rmk,           0 },
  { X86::VPBLENDMWZrrk,              X86::VPBLENDMWZrmk,              0 },
  { X86::VPBROADCASTBZ128rk,         X86::VPBROADCASTBZ128mk,         TB_NO_REVERSE },
  { X86::VPBROADCASTBZ256rk,         X86::VPBROADCASTBZ256mk,         TB_NO_REVERSE },
  { X86::VPBROADCASTBZrk,            X86::VPBROADCASTBZmk,            TB_NO_REVERSE },
  { X86::VPBROADCASTDZ128rk,         X86::VPBROADCASTDZ128mk,         TB_NO_REVERSE },
  { X86::VPBROADCASTDZ256rk,         X86::VPBROADCASTDZ256mk,         TB_NO_REVERSE },
  { X86::VPBROADCASTDZrk,            X86::VPBROADCASTDZmk,            TB_NO_REVERSE },
  { X86::VPBROADCASTQZ128rk,         X86::VPBROADCASTQZ128mk,         TB_NO_REVERSE },
  { X86::VPBROADCASTQZ256rk,         X86::VPBROADCASTQZ256mk,         TB_NO_REVERSE },
  { X86::VPBROADCASTQZrk,            X86::VPBROADCASTQZmk,            TB_NO_REVERSE },
  { X86::VPBROADCASTWZ128rk,         X86::VPBROADCASTWZ128mk,         TB_NO_REVERSE },
  { X86::VPBROADCASTWZ256rk,         X86::VPBROADCASTWZ256mk,         TB_NO_REVERSE },
  { X86::VPBROADCASTWZrk,            X86::VPBROADCASTWZmk,            TB_NO_REVERSE },
  { X86::VPCMOVYrrr,                 X86::VPCMOVYrrm,                 0 },
  { X86::VPCMOVrrr,                  X86::VPCMOVrrm,                  0 },
  { X86::VPCMPBZ128rrik,             X86::VPCMPBZ128rmik,             0 },
  { X86::VPCMPBZ256rrik,             X86::VPCMPBZ256rmik,             0 },
  { X86::VPCMPBZrrik,                X86::VPCMPBZrmik,                0 },
  { X86::VPCMPDZ128rrik,             X86::VPCMPDZ128rmik,             0 },
  { X86::VPCMPDZ256rrik,             X86::VPCMPDZ256rmik,             0 },
  { X86::VPCMPDZrrik,                X86::VPCMPDZrmik,                0 },
  { X86::VPCMPEQBZ128rrk,            X86::VPCMPEQBZ128rmk,            0 },
  { X86::VPCMPEQBZ256rrk,            X86::VPCMPEQBZ256rmk,            0 },
  { X86::VPCMPEQBZrrk,               X86::VPCMPEQBZrmk,               0 },
  { X86::VPCMPEQDZ128rrk,            X86::VPCMPEQDZ128rmk,            0 },
  { X86::VPCMPEQDZ256rrk,            X86::VPCMPEQDZ256rmk,            0 },
  { X86::VPCMPEQDZrrk,               X86::VPCMPEQDZrmk,               0 },
  { X86::VPCMPEQQZ128rrk,            X86::VPCMPEQQZ128rmk,            0 },
  { X86::VPCMPEQQZ256rrk,            X86::VPCMPEQQZ256rmk,            0 },
  { X86::VPCMPEQQZrrk,               X86::VPCMPEQQZrmk,               0 },
  { X86::VPCMPEQWZ128rrk,            X86::VPCMPEQWZ128rmk,            0 },
  { X86::VPCMPEQWZ256rrk,            X86::VPCMPEQWZ256rmk,            0 },
  { X86::VPCMPEQWZrrk,               X86::VPCMPEQWZrmk,               0 },
  { X86::VPCMPGTBZ128rrk,            X86::VPCMPGTBZ128rmk,            0 },
  { X86::VPCMPGTBZ256rrk,            X86::VPCMPGTBZ256rmk,            0 },
  { X86::VPCMPGTBZrrk,               X86::VPCMPGTBZrmk,               0 },
  { X86::VPCMPGTDZ128rrk,            X86::VPCMPGTDZ128rmk,            0 },
  { X86::VPCMPGTDZ256rrk,            X86::VPCMPGTDZ256rmk,            0 },
  { X86::VPCMPGTDZrrk,               X86::VPCMPGTDZrmk,               0 },
  { X86::VPCMPGTQZ128rrk,            X86::VPCMPGTQZ128rmk,            0 },
  { X86::VPCMPGTQZ256rrk,            X86::VPCMPGTQZ256rmk,            0 },
  { X86::VPCMPGTQZrrk,               X86::VPCMPGTQZrmk,               0 },
  { X86::VPCMPGTWZ128rrk,            X86::VPCMPGTWZ128rmk,            0 },
  { X86::VPCMPGTWZ256rrk,            X86::VPCMPGTWZ256rmk,            0 },
  { X86::VPCMPGTWZrrk,               X86::VPCMPGTWZrmk,               0 },
  { X86::VPCMPQZ128rrik,             X86::VPCMPQZ128rmik,             0 },
  { X86::VPCMPQZ256rrik,             X86::VPCMPQZ256rmik,             0 },
  { X86::VPCMPQZrrik,                X86::VPCMPQZrmik,                0 },
  { X86::VPCMPUBZ128rrik,            X86::VPCMPUBZ128rmik,            0 },
  { X86::VPCMPUBZ256rrik,            X86::VPCMPUBZ256rmik,            0 },
  { X86::VPCMPUBZrrik,               X86::VPCMPUBZrmik,               0 },
  { X86::VPCMPUDZ128rrik,            X86::VPCMPUDZ128rmik,            0 },
  { X86::VPCMPUDZ256rrik,            X86::VPCMPUDZ256rmik,            0 },
  { X86::VPCMPUDZrrik,               X86::VPCMPUDZrmik,               0 },
  { X86::VPCMPUQZ128rrik,            X86::VPCMPUQZ128rmik,            0 },
  { X86::VPCMPUQZ256rrik,            X86::VPCMPUQZ256rmik,            0 },
  { X86::VPCMPUQZrrik,               X86::VPCMPUQZrmik,               0 },
  { X86::VPCMPUWZ128rrik,            X86::VPCMPUWZ128rmik,            0 },
  { X86::VPCMPUWZ256rrik,            X86::VPCMPUWZ256rmik,            0 },
  { X86::VPCMPUWZrrik,               X86::VPCMPUWZrmik,               0 },
  { X86::VPCMPWZ128rrik,             X86::VPCMPWZ128rmik,             0 },
  { X86::VPCMPWZ256rrik,             X86::VPCMPWZ256rmik,             0 },
  { X86::VPCMPWZrrik,                X86::VPCMPWZrmik,                0 },
  { X86::VPCONFLICTDZ128rrk,         X86::VPCONFLICTDZ128rmk,         0 },
  { X86::VPCONFLICTDZ256rrk,         X86::VPCONFLICTDZ256rmk,         0 },
  { X86::VPCONFLICTDZrrk,            X86::VPCONFLICTDZrmk,            0 },
  { X86::VPCONFLICTQZ128rrk,         X86::VPCONFLICTQZ128rmk,         0 },
  { X86::VPCONFLICTQZ256rrk,         X86::VPCONFLICTQZ256rmk,         0 },
  { X86::VPCONFLICTQZrrk,            X86::VPCONFLICTQZrmk,            0 },
  { X86::VPDPBUSDSZ128r,             X86::VPDPBUSDSZ128m,             0 },
  { X86::VPDPBUSDSZ256r,             X86::VPDPBUSDSZ256m,             0 },
  { X86::VPDPBUSDSZr,                X86::VPDPBUSDSZm,                0 },
  { X86::VPDPBUSDZ128r,              X86::VPDPBUSDZ128m,              0 },
  { X86::VPDPBUSDZ256r,              X86::VPDPBUSDZ256m,              0 },
  { X86::VPDPBUSDZr,                 X86::VPDPBUSDZm,                 0 },
  { X86::VPDPWSSDSZ128r,             X86::VPDPWSSDSZ128m,             0 },
  { X86::VPDPWSSDSZ256r,             X86::VPDPWSSDSZ256m,             0 },
  { X86::VPDPWSSDSZr,                X86::VPDPWSSDSZm,                0 },
  { X86::VPDPWSSDZ128r,              X86::VPDPWSSDZ128m,              0 },
  { X86::VPDPWSSDZ256r,              X86::VPDPWSSDZ256m,              0 },
  { X86::VPDPWSSDZr,                 X86::VPDPWSSDZm,                 0 },
  { X86::VPERMBZ128rrkz,             X86::VPERMBZ128rmkz,             0 },
  { X86::VPERMBZ256rrkz,             X86::VPERMBZ256rmkz,             0 },
  { X86::VPERMBZrrkz,                X86::VPERMBZrmkz,                0 },
  { X86::VPERMDZ256rrkz,             X86::VPERMDZ256rmkz,             0 },
  { X86::VPERMDZrrkz,                X86::VPERMDZrmkz,                0 },
  { X86::VPERMI2B128rr,              X86::VPERMI2B128rm,              0 },
  { X86::VPERMI2B256rr,              X86::VPERMI2B256rm,              0 },
  { X86::VPERMI2Brr,                 X86::VPERMI2Brm,                 0 },
  { X86::VPERMI2D128rr,              X86::VPERMI2D128rm,              0 },
  { X86::VPERMI2D256rr,              X86::VPERMI2D256rm,              0 },
  { X86::VPERMI2Drr,                 X86::VPERMI2Drm,                 0 },
  { X86::VPERMI2PD128rr,             X86::VPERMI2PD128rm,             0 },
  { X86::VPERMI2PD256rr,             X86::VPERMI2PD256rm,             0 },
  { X86::VPERMI2PDrr,                X86::VPERMI2PDrm,                0 },
  { X86::VPERMI2PS128rr,             X86::VPERMI2PS128rm,             0 },
  { X86::VPERMI2PS256rr,             X86::VPERMI2PS256rm,             0 },
  { X86::VPERMI2PSrr,                X86::VPERMI2PSrm,                0 },
  { X86::VPERMI2Q128rr,              X86::VPERMI2Q128rm,              0 },
  { X86::VPERMI2Q256rr,              X86::VPERMI2Q256rm,              0 },
  { X86::VPERMI2Qrr,                 X86::VPERMI2Qrm,                 0 },
  { X86::VPERMI2W128rr,              X86::VPERMI2W128rm,              0 },
  { X86::VPERMI2W256rr,              X86::VPERMI2W256rm,              0 },
  { X86::VPERMI2Wrr,                 X86::VPERMI2Wrm,                 0 },
  { X86::VPERMIL2PDYrr,              X86::VPERMIL2PDYrm,              0 },
  { X86::VPERMIL2PDrr,               X86::VPERMIL2PDrm,               0 },
  { X86::VPERMIL2PSYrr,              X86::VPERMIL2PSYrm,              0 },
  { X86::VPERMIL2PSrr,               X86::VPERMIL2PSrm,               0 },
  { X86::VPERMILPDZ128rik,           X86::VPERMILPDZ128mik,           0 },
  { X86::VPERMILPDZ128rrkz,          X86::VPERMILPDZ128rmkz,          0 },
  { X86::VPERMILPDZ256rik,           X86::VPERMILPDZ256mik,           0 },
  { X86::VPERMILPDZ256rrkz,          X86::VPERMILPDZ256rmkz,          0 },
  { X86::VPERMILPDZrik,              X86::VPERMILPDZmik,              0 },
  { X86::VPERMILPDZrrkz,             X86::VPERMILPDZrmkz,             0 },
  { X86::VPERMILPSZ128rik,           X86::VPERMILPSZ128mik,           0 },
  { X86::VPERMILPSZ128rrkz,          X86::VPERMILPSZ128rmkz,          0 },
  { X86::VPERMILPSZ256rik,           X86::VPERMILPSZ256mik,           0 },
  { X86::VPERMILPSZ256rrkz,          X86::VPERMILPSZ256rmkz,          0 },
  { X86::VPERMILPSZrik,              X86::VPERMILPSZmik,              0 },
  { X86::VPERMILPSZrrkz,             X86::VPERMILPSZrmkz,             0 },
  { X86::VPERMPDZ256rik,             X86::VPERMPDZ256mik,             0 },
  { X86::VPERMPDZ256rrkz,            X86::VPERMPDZ256rmkz,            0 },
  { X86::VPERMPDZrik,                X86::VPERMPDZmik,                0 },
  { X86::VPERMPDZrrkz,               X86::VPERMPDZrmkz,               0 },
  { X86::VPERMPSZ256rrkz,            X86::VPERMPSZ256rmkz,            0 },
  { X86::VPERMPSZrrkz,               X86::VPERMPSZrmkz,               0 },
  { X86::VPERMQZ256rik,              X86::VPERMQZ256mik,              0 },
  { X86::VPERMQZ256rrkz,             X86::VPERMQZ256rmkz,             0 },
  { X86::VPERMQZrik,                 X86::VPERMQZmik,                 0 },
  { X86::VPERMQZrrkz,                X86::VPERMQZrmkz,                0 },
  { X86::VPERMT2B128rr,              X86::VPERMT2B128rm,              0 },
  { X86::VPERMT2B256rr,              X86::VPERMT2B256rm,              0 },
  { X86::VPERMT2Brr,                 X86::VPERMT2Brm,                 0 },
  { X86::VPERMT2D128rr,              X86::VPERMT2D128rm,              0 },
  { X86::VPERMT2D256rr,              X86::VPERMT2D256rm,              0 },
  { X86::VPERMT2Drr,                 X86::VPERMT2Drm,                 0 },
  { X86::VPERMT2PD128rr,             X86::VPERMT2PD128rm,             0 },
  { X86::VPERMT2PD256rr,             X86::VPERMT2PD256rm,             0 },
  { X86::VPERMT2PDrr,                X86::VPERMT2PDrm,                0 },
  { X86::VPERMT2PS128rr,             X86::VPERMT2PS128rm,             0 },
  { X86::VPERMT2PS256rr,             X86::VPERMT2PS256rm,             0 },
  { X86::VPERMT2PSrr,                X86::VPERMT2PSrm,                0 },
  { X86::VPERMT2Q128rr,              X86::VPERMT2Q128rm,              0 },
  { X86::VPERMT2Q256rr,              X86::VPERMT2Q256rm,              0 },
  { X86::VPERMT2Qrr,                 X86::VPERMT2Qrm,                 0 },
  { X86::VPERMT2W128rr,              X86::VPERMT2W128rm,              0 },
  { X86::VPERMT2W256rr,              X86::VPERMT2W256rm,              0 },
  { X86::VPERMT2Wrr,                 X86::VPERMT2Wrm,                 0 },
  { X86::VPERMWZ128rrkz,             X86::VPERMWZ128rmkz,             0 },
  { X86::VPERMWZ256rrkz,             X86::VPERMWZ256rmkz,             0 },
  { X86::VPERMWZrrkz,                X86::VPERMWZrmkz,                0 },
  { X86::VPEXPANDBZ128rrk,           X86::VPEXPANDBZ128rmk,           TB_NO_REVERSE },
  { X86::VPEXPANDBZ256rrk,           X86::VPEXPANDBZ256rmk,           TB_NO_REVERSE },
  { X86::VPEXPANDBZrrk,              X86::VPEXPANDBZrmk,              TB_NO_REVERSE },
  { X86::VPEXPANDDZ128rrk,           X86::VPEXPANDDZ128rmk,           TB_NO_REVERSE },
  { X86::VPEXPANDDZ256rrk,           X86::VPEXPANDDZ256rmk,           TB_NO_REVERSE },
  { X86::VPEXPANDDZrrk,              X86::VPEXPANDDZrmk,              TB_NO_REVERSE },
  { X86::VPEXPANDQZ128rrk,           X86::VPEXPANDQZ128rmk,           TB_NO_REVERSE },
  { X86::VPEXPANDQZ256rrk,           X86::VPEXPANDQZ256rmk,           TB_NO_REVERSE },
  { X86::VPEXPANDQZrrk,              X86::VPEXPANDQZrmk,              TB_NO_REVERSE },
  { X86::VPEXPANDWZ128rrk,           X86::VPEXPANDWZ128rmk,           TB_NO_REVERSE },
  { X86::VPEXPANDWZ256rrk,           X86::VPEXPANDWZ256rmk,           TB_NO_REVERSE },
  { X86::VPEXPANDWZrrk,              X86::VPEXPANDWZrmk,              TB_NO_REVERSE },
  { X86::VPLZCNTDZ128rrk,            X86::VPLZCNTDZ128rmk,            0 },
  { X86::VPLZCNTDZ256rrk,            X86::VPLZCNTDZ256rmk,            0 },
  { X86::VPLZCNTDZrrk,               X86::VPLZCNTDZrmk,               0 },
  { X86::VPLZCNTQZ128rrk,            X86::VPLZCNTQZ128rmk,            0 },
  { X86::VPLZCNTQZ256rrk,            X86::VPLZCNTQZ256rmk,            0 },
  { X86::VPLZCNTQZrrk,               X86::VPLZCNTQZrmk,               0 },
  { X86::VPMADD52HUQZ128r,           X86::VPMADD52HUQZ128m,           0 },
  { X86::VPMADD52HUQZ256r,           X86::VPMADD52HUQZ256m,           0 },
  { X86::VPMADD52HUQZr,              X86::VPMADD52HUQZm,              0 },
  { X86::VPMADD52LUQZ128r,           X86::VPMADD52LUQZ128m,           0 },
  { X86::VPMADD52LUQZ256r,           X86::VPMADD52LUQZ256m,           0 },
  { X86::VPMADD52LUQZr,              X86::VPMADD52LUQZm,              0 },
  { X86::VPMADDUBSWZ128rrkz,         X86::VPMADDUBSWZ128rmkz,         0 },
  { X86::VPMADDUBSWZ256rrkz,         X86::VPMADDUBSWZ256rmkz,         0 },
  { X86::VPMADDUBSWZrrkz,            X86::VPMADDUBSWZrmkz,            0 },
  { X86::VPMADDWDZ128rrkz,           X86::VPMADDWDZ128rmkz,           0 },
  { X86::VPMADDWDZ256rrkz,           X86::VPMADDWDZ256rmkz,           0 },
  { X86::VPMADDWDZrrkz,              X86::VPMADDWDZrmkz,              0 },
  { X86::VPMAXSBZ128rrkz,            X86::VPMAXSBZ128rmkz,            0 },
  { X86::VPMAXSBZ256rrkz,            X86::VPMAXSBZ256rmkz,            0 },
  { X86::VPMAXSBZrrkz,               X86::VPMAXSBZrmkz,               0 },
  { X86::VPMAXSDZ128rrkz,            X86::VPMAXSDZ128rmkz,            0 },
  { X86::VPMAXSDZ256rrkz,            X86::VPMAXSDZ256rmkz,            0 },
  { X86::VPMAXSDZrrkz,               X86::VPMAXSDZrmkz,               0 },
  { X86::VPMAXSQZ128rrkz,            X86::VPMAXSQZ128rmkz,            0 },
  { X86::VPMAXSQZ256rrkz,            X86::VPMAXSQZ256rmkz,            0 },
  { X86::VPMAXSQZrrkz,               X86::VPMAXSQZrmkz,               0 },
  { X86::VPMAXSWZ128rrkz,            X86::VPMAXSWZ128rmkz,            0 },
  { X86::VPMAXSWZ256rrkz,            X86::VPMAXSWZ256rmkz,            0 },
  { X86::VPMAXSWZrrkz,               X86::VPMAXSWZrmkz,               0 },
  { X86::VPMAXUBZ128rrkz,            X86::VPMAXUBZ128rmkz,            0 },
  { X86::VPMAXUBZ256rrkz,            X86::VPMAXUBZ256rmkz,            0 },
  { X86::VPMAXUBZrrkz,               X86::VPMAXUBZrmkz,               0 },
  { X86::VPMAXUDZ128rrkz,            X86::VPMAXUDZ128rmkz,            0 },
  { X86::VPMAXUDZ256rrkz,            X86::VPMAXUDZ256rmkz,            0 },
  { X86::VPMAXUDZrrkz,               X86::VPMAXUDZrmkz,               0 },
  { X86::VPMAXUQZ128rrkz,            X86::VPMAXUQZ128rmkz,            0 },
  { X86::VPMAXUQZ256rrkz,            X86::VPMAXUQZ256rmkz,            0 },
  { X86::VPMAXUQZrrkz,               X86::VPMAXUQZrmkz,               0 },
  { X86::VPMAXUWZ128rrkz,            X86::VPMAXUWZ128rmkz,            0 },
  { X86::VPMAXUWZ256rrkz,            X86::VPMAXUWZ256rmkz,            0 },
  { X86::VPMAXUWZrrkz,               X86::VPMAXUWZrmkz,               0 },
  { X86::VPMINSBZ128rrkz,            X86::VPMINSBZ128rmkz,            0 },
  { X86::VPMINSBZ256rrkz,            X86::VPMINSBZ256rmkz,            0 },
  { X86::VPMINSBZrrkz,               X86::VPMINSBZrmkz,               0 },
  { X86::VPMINSDZ128rrkz,            X86::VPMINSDZ128rmkz,            0 },
  { X86::VPMINSDZ256rrkz,            X86::VPMINSDZ256rmkz,            0 },
  { X86::VPMINSDZrrkz,               X86::VPMINSDZrmkz,               0 },
  { X86::VPMINSQZ128rrkz,            X86::VPMINSQZ128rmkz,            0 },
  { X86::VPMINSQZ256rrkz,            X86::VPMINSQZ256rmkz,            0 },
  { X86::VPMINSQZrrkz,               X86::VPMINSQZrmkz,               0 },
  { X86::VPMINSWZ128rrkz,            X86::VPMINSWZ128rmkz,            0 },
  { X86::VPMINSWZ256rrkz,            X86::VPMINSWZ256rmkz,            0 },
  { X86::VPMINSWZrrkz,               X86::VPMINSWZrmkz,               0 },
  { X86::VPMINUBZ128rrkz,            X86::VPMINUBZ128rmkz,            0 },
  { X86::VPMINUBZ256rrkz,            X86::VPMINUBZ256rmkz,            0 },
  { X86::VPMINUBZrrkz,               X86::VPMINUBZrmkz,               0 },
  { X86::VPMINUDZ128rrkz,            X86::VPMINUDZ128rmkz,            0 },
  { X86::VPMINUDZ256rrkz,            X86::VPMINUDZ256rmkz,            0 },
  { X86::VPMINUDZrrkz,               X86::VPMINUDZrmkz,               0 },
  { X86::VPMINUQZ128rrkz,            X86::VPMINUQZ128rmkz,            0 },
  { X86::VPMINUQZ256rrkz,            X86::VPMINUQZ256rmkz,            0 },
  { X86::VPMINUQZrrkz,               X86::VPMINUQZrmkz,               0 },
  { X86::VPMINUWZ128rrkz,            X86::VPMINUWZ128rmkz,            0 },
  { X86::VPMINUWZ256rrkz,            X86::VPMINUWZ256rmkz,            0 },
  { X86::VPMINUWZrrkz,               X86::VPMINUWZrmkz,               0 },
  { X86::VPMOVSXBDZ128rrk,           X86::VPMOVSXBDZ128rmk,           TB_NO_REVERSE },
  { X86::VPMOVSXBDZ256rrk,           X86::VPMOVSXBDZ256rmk,           TB_NO_REVERSE },
  { X86::VPMOVSXBDZrrk,              X86::VPMOVSXBDZrmk,              0 },
  { X86::VPMOVSXBQZ128rrk,           X86::VPMOVSXBQZ128rmk,           TB_NO_REVERSE },
  { X86::VPMOVSXBQZ256rrk,           X86::VPMOVSXBQZ256rmk,           TB_NO_REVERSE },
  { X86::VPMOVSXBQZrrk,              X86::VPMOVSXBQZrmk,              TB_NO_REVERSE },
  { X86::VPMOVSXBWZ128rrk,           X86::VPMOVSXBWZ128rmk,           TB_NO_REVERSE },
  { X86::VPMOVSXBWZ256rrk,           X86::VPMOVSXBWZ256rmk,           0 },
  { X86::VPMOVSXBWZrrk,              X86::VPMOVSXBWZrmk,              0 },
  { X86::VPMOVSXDQZ128rrk,           X86::VPMOVSXDQZ128rmk,           TB_NO_REVERSE },
  { X86::VPMOVSXDQZ256rrk,           X86::VPMOVSXDQZ256rmk,           0 },
  { X86::VPMOVSXDQZrrk,              X86::VPMOVSXDQZrmk,              0 },
  { X86::VPMOVSXWDZ128rrk,           X86::VPMOVSXWDZ128rmk,           TB_NO_REVERSE },
  { X86::VPMOVSXWDZ256rrk,           X86::VPMOVSXWDZ256rmk,           0 },
  { X86::VPMOVSXWDZrrk,              X86::VPMOVSXWDZrmk,              0 },
  { X86::VPMOVSXWQZ128rrk,           X86::VPMOVSXWQZ128rmk,           TB_NO_REVERSE },
  { X86::VPMOVSXWQZ256rrk,           X86::VPMOVSXWQZ256rmk,           TB_NO_REVERSE },
  { X86::VPMOVSXWQZrrk,              X86::VPMOVSXWQZrmk,              0 },
  { X86::VPMOVZXBDZ128rrk,           X86::VPMOVZXBDZ128rmk,           TB_NO_REVERSE },
  { X86::VPMOVZXBDZ256rrk,           X86::VPMOVZXBDZ256rmk,           TB_NO_REVERSE },
  { X86::VPMOVZXBDZrrk,              X86::VPMOVZXBDZrmk,              0 },
  { X86::VPMOVZXBQZ128rrk,           X86::VPMOVZXBQZ128rmk,           TB_NO_REVERSE },
  { X86::VPMOVZXBQZ256rrk,           X86::VPMOVZXBQZ256rmk,           TB_NO_REVERSE },
  { X86::VPMOVZXBQZrrk,              X86::VPMOVZXBQZrmk,              TB_NO_REVERSE },
  { X86::VPMOVZXBWZ128rrk,           X86::VPMOVZXBWZ128rmk,           TB_NO_REVERSE },
  { X86::VPMOVZXBWZ256rrk,           X86::VPMOVZXBWZ256rmk,           0 },
  { X86::VPMOVZXBWZrrk,              X86::VPMOVZXBWZrmk,              0 },
  { X86::VPMOVZXDQZ128rrk,           X86::VPMOVZXDQZ128rmk,           TB_NO_REVERSE },
  { X86::VPMOVZXDQZ256rrk,           X86::VPMOVZXDQZ256rmk,           0 },
  { X86::VPMOVZXDQZrrk,              X86::VPMOVZXDQZrmk,              0 },
  { X86::VPMOVZXWDZ128rrk,           X86::VPMOVZXWDZ128rmk,           TB_NO_REVERSE },
  { X86::VPMOVZXWDZ256rrk,           X86::VPMOVZXWDZ256rmk,           0 },
  { X86::VPMOVZXWDZrrk,              X86::VPMOVZXWDZrmk,              0 },
  { X86::VPMOVZXWQZ128rrk,           X86::VPMOVZXWQZ128rmk,           TB_NO_REVERSE },
  { X86::VPMOVZXWQZ256rrk,           X86::VPMOVZXWQZ256rmk,           TB_NO_REVERSE },
  { X86::VPMOVZXWQZrrk,              X86::VPMOVZXWQZrmk,              0 },
  { X86::VPMULDQZ128rrkz,            X86::VPMULDQZ128rmkz,            0 },
  { X86::VPMULDQZ256rrkz,            X86::VPMULDQZ256rmkz,            0 },
  { X86::VPMULDQZrrkz,               X86::VPMULDQZrmkz,               0 },
  { X86::VPMULHRSWZ128rrkz,          X86::VPMULHRSWZ128rmkz,          0 },
  { X86::VPMULHRSWZ256rrkz,          X86::VPMULHRSWZ256rmkz,          0 },
  { X86::VPMULHRSWZrrkz,             X86::VPMULHRSWZrmkz,             0 },
  { X86::VPMULHUWZ128rrkz,           X86::VPMULHUWZ128rmkz,           0 },
  { X86::VPMULHUWZ256rrkz,           X86::VPMULHUWZ256rmkz,           0 },
  { X86::VPMULHUWZrrkz,              X86::VPMULHUWZrmkz,              0 },
  { X86::VPMULHWZ128rrkz,            X86::VPMULHWZ128rmkz,            0 },
  { X86::VPMULHWZ256rrkz,            X86::VPMULHWZ256rmkz,            0 },
  { X86::VPMULHWZrrkz,               X86::VPMULHWZrmkz,               0 },
  { X86::VPMULLDZ128rrkz,            X86::VPMULLDZ128rmkz,            0 },
  { X86::VPMULLDZ256rrkz,            X86::VPMULLDZ256rmkz,            0 },
  { X86::VPMULLDZrrkz,               X86::VPMULLDZrmkz,               0 },
  { X86::VPMULLQZ128rrkz,            X86::VPMULLQZ128rmkz,            0 },
  { X86::VPMULLQZ256rrkz,            X86::VPMULLQZ256rmkz,            0 },
  { X86::VPMULLQZrrkz,               X86::VPMULLQZrmkz,               0 },
  { X86::VPMULLWZ128rrkz,            X86::VPMULLWZ128rmkz,            0 },
  { X86::VPMULLWZ256rrkz,            X86::VPMULLWZ256rmkz,            0 },
  { X86::VPMULLWZrrkz,               X86::VPMULLWZrmkz,               0 },
  { X86::VPMULTISHIFTQBZ128rrkz,     X86::VPMULTISHIFTQBZ128rmkz,     0 },
  { X86::VPMULTISHIFTQBZ256rrkz,     X86::VPMULTISHIFTQBZ256rmkz,     0 },
  { X86::VPMULTISHIFTQBZrrkz,        X86::VPMULTISHIFTQBZrmkz,        0 },
  { X86::VPMULUDQZ128rrkz,           X86::VPMULUDQZ128rmkz,           0 },
  { X86::VPMULUDQZ256rrkz,           X86::VPMULUDQZ256rmkz,           0 },
  { X86::VPMULUDQZrrkz,              X86::VPMULUDQZrmkz,              0 },
  { X86::VPOPCNTBZ128rrk,            X86::VPOPCNTBZ128rmk,            0 },
  { X86::VPOPCNTBZ256rrk,            X86::VPOPCNTBZ256rmk,            0 },
  { X86::VPOPCNTBZrrk,               X86::VPOPCNTBZrmk,               0 },
  { X86::VPOPCNTDZ128rrk,            X86::VPOPCNTDZ128rmk,            0 },
  { X86::VPOPCNTDZ256rrk,            X86::VPOPCNTDZ256rmk,            0 },
  { X86::VPOPCNTDZrrk,               X86::VPOPCNTDZrmk,               0 },
  { X86::VPOPCNTQZ128rrk,            X86::VPOPCNTQZ128rmk,            0 },
  { X86::VPOPCNTQZ256rrk,            X86::VPOPCNTQZ256rmk,            0 },
  { X86::VPOPCNTQZrrk,               X86::VPOPCNTQZrmk,               0 },
  { X86::VPOPCNTWZ128rrk,            X86::VPOPCNTWZ128rmk,            0 },
  { X86::VPOPCNTWZ256rrk,            X86::VPOPCNTWZ256rmk,            0 },
  { X86::VPOPCNTWZrrk,               X86::VPOPCNTWZrmk,               0 },
  { X86::VPORDZ128rrkz,              X86::VPORDZ128rmkz,              0 },
  { X86::VPORDZ256rrkz,              X86::VPORDZ256rmkz,              0 },
  { X86::VPORDZrrkz,                 X86::VPORDZrmkz,                 0 },
  { X86::VPORQZ128rrkz,              X86::VPORQZ128rmkz,              0 },
  { X86::VPORQZ256rrkz,              X86::VPORQZ256rmkz,              0 },
  { X86::VPORQZrrkz,                 X86::VPORQZrmkz,                 0 },
  { X86::VPPERMrrr,                  X86::VPPERMrrm,                  0 },
  { X86::VPROLDZ128rik,              X86::VPROLDZ128mik,              0 },
  { X86::VPROLDZ256rik,              X86::VPROLDZ256mik,              0 },
  { X86::VPROLDZrik,                 X86::VPROLDZmik,                 0 },
  { X86::VPROLQZ128rik,              X86::VPROLQZ128mik,              0 },
  { X86::VPROLQZ256rik,              X86::VPROLQZ256mik,              0 },
  { X86::VPROLQZrik,                 X86::VPROLQZmik,                 0 },
  { X86::VPROLVDZ128rrkz,            X86::VPROLVDZ128rmkz,            0 },
  { X86::VPROLVDZ256rrkz,            X86::VPROLVDZ256rmkz,            0 },
  { X86::VPROLVDZrrkz,               X86::VPROLVDZrmkz,               0 },
  { X86::VPROLVQZ128rrkz,            X86::VPROLVQZ128rmkz,            0 },
  { X86::VPROLVQZ256rrkz,            X86::VPROLVQZ256rmkz,            0 },
  { X86::VPROLVQZrrkz,               X86::VPROLVQZrmkz,               0 },
  { X86::VPRORDZ128rik,              X86::VPRORDZ128mik,              0 },
  { X86::VPRORDZ256rik,              X86::VPRORDZ256mik,              0 },
  { X86::VPRORDZrik,                 X86::VPRORDZmik,                 0 },
  { X86::VPRORQZ128rik,              X86::VPRORQZ128mik,              0 },
  { X86::VPRORQZ256rik,              X86::VPRORQZ256mik,              0 },
  { X86::VPRORQZrik,                 X86::VPRORQZmik,                 0 },
  { X86::VPRORVDZ128rrkz,            X86::VPRORVDZ128rmkz,            0 },
  { X86::VPRORVDZ256rrkz,            X86::VPRORVDZ256rmkz,            0 },
  { X86::VPRORVDZrrkz,               X86::VPRORVDZrmkz,               0 },
  { X86::VPRORVQZ128rrkz,            X86::VPRORVQZ128rmkz,            0 },
  { X86::VPRORVQZ256rrkz,            X86::VPRORVQZ256rmkz,            0 },
  { X86::VPRORVQZrrkz,               X86::VPRORVQZrmkz,               0 },
  { X86::VPSHLDDZ128rrikz,           X86::VPSHLDDZ128rmikz,           0 },
  { X86::VPSHLDDZ256rrikz,           X86::VPSHLDDZ256rmikz,           0 },
  { X86::VPSHLDDZrrikz,              X86::VPSHLDDZrmikz,              0 },
  { X86::VPSHLDQZ128rrikz,           X86::VPSHLDQZ128rmikz,           0 },
  { X86::VPSHLDQZ256rrikz,           X86::VPSHLDQZ256rmikz,           0 },
  { X86::VPSHLDQZrrikz,              X86::VPSHLDQZrmikz,              0 },
  { X86::VPSHLDVDZ128r,              X86::VPSHLDVDZ128m,              0 },
  { X86::VPSHLDVDZ256r,              X86::VPSHLDVDZ256m,              0 },
  { X86::VPSHLDVDZr,                 X86::VPSHLDVDZm,                 0 },
  { X86::VPSHLDVQZ128r,              X86::VPSHLDVQZ128m,              0 },
  { X86::VPSHLDVQZ256r,              X86::VPSHLDVQZ256m,              0 },
  { X86::VPSHLDVQZr,                 X86::VPSHLDVQZm,                 0 },
  { X86::VPSHLDVWZ128r,              X86::VPSHLDVWZ128m,              0 },
  { X86::VPSHLDVWZ256r,              X86::VPSHLDVWZ256m,              0 },
  { X86::VPSHLDVWZr,                 X86::VPSHLDVWZm,                 0 },
  { X86::VPSHLDWZ128rrikz,           X86::VPSHLDWZ128rmikz,           0 },
  { X86::VPSHLDWZ256rrikz,           X86::VPSHLDWZ256rmikz,           0 },
  { X86::VPSHLDWZrrikz,              X86::VPSHLDWZrmikz,              0 },
  { X86::VPSHRDDZ128rrikz,           X86::VPSHRDDZ128rmikz,           0 },
  { X86::VPSHRDDZ256rrikz,           X86::VPSHRDDZ256rmikz,           0 },
  { X86::VPSHRDDZrrikz,              X86::VPSHRDDZrmikz,              0 },
  { X86::VPSHRDQZ128rrikz,           X86::VPSHRDQZ128rmikz,           0 },
  { X86::VPSHRDQZ256rrikz,           X86::VPSHRDQZ256rmikz,           0 },
  { X86::VPSHRDQZrrikz,              X86::VPSHRDQZrmikz,              0 },
  { X86::VPSHRDVDZ128r,              X86::VPSHRDVDZ128m,              0 },
  { X86::VPSHRDVDZ256r,              X86::VPSHRDVDZ256m,              0 },
  { X86::VPSHRDVDZr,                 X86::VPSHRDVDZm,                 0 },
  { X86::VPSHRDVQZ128r,              X86::VPSHRDVQZ128m,              0 },
  { X86::VPSHRDVQZ256r,              X86::VPSHRDVQZ256m,              0 },
  { X86::VPSHRDVQZr,                 X86::VPSHRDVQZm,                 0 },
  { X86::VPSHRDVWZ128r,              X86::VPSHRDVWZ128m,              0 },
  { X86::VPSHRDVWZ256r,              X86::VPSHRDVWZ256m,              0 },
  { X86::VPSHRDVWZr,                 X86::VPSHRDVWZm,                 0 },
  { X86::VPSHRDWZ128rrikz,           X86::VPSHRDWZ128rmikz,           0 },
  { X86::VPSHRDWZ256rrikz,           X86::VPSHRDWZ256rmikz,           0 },
  { X86::VPSHRDWZrrikz,              X86::VPSHRDWZrmikz,              0 },
  { X86::VPSHUFBITQMBZ128rrk,        X86::VPSHUFBITQMBZ128rmk,        0 },
  { X86::VPSHUFBITQMBZ256rrk,        X86::VPSHUFBITQMBZ256rmk,        0 },
  { X86::VPSHUFBITQMBZrrk,           X86::VPSHUFBITQMBZrmk,           0 },
  { X86::VPSHUFBZ128rrkz,            X86::VPSHUFBZ128rmkz,            0 },
  { X86::VPSHUFBZ256rrkz,            X86::VPSHUFBZ256rmkz,            0 },
  { X86::VPSHUFBZrrkz,               X86::VPSHUFBZrmkz,               0 },
  { X86::VPSHUFDZ128rik,             X86::VPSHUFDZ128mik,             0 },
  { X86::VPSHUFDZ256rik,             X86::VPSHUFDZ256mik,             0 },
  { X86::VPSHUFDZrik,                X86::VPSHUFDZmik,                0 },
  { X86::VPSHUFHWZ128rik,            X86::VPSHUFHWZ128mik,            0 },
  { X86::VPSHUFHWZ256rik,            X86::VPSHUFHWZ256mik,            0 },
  { X86::VPSHUFHWZrik,               X86::VPSHUFHWZmik,               0 },
  { X86::VPSHUFLWZ128rik,            X86::VPSHUFLWZ128mik,            0 },
  { X86::VPSHUFLWZ256rik,            X86::VPSHUFLWZ256mik,            0 },
  { X86::VPSHUFLWZrik,               X86::VPSHUFLWZmik,               0 },
  { X86::VPSLLDZ128rik,              X86::VPSLLDZ128mik,              0 },
  { X86::VPSLLDZ128rrkz,             X86::VPSLLDZ128rmkz,             0 },
  { X86::VPSLLDZ256rik,              X86::VPSLLDZ256mik,              0 },
  { X86::VPSLLDZ256rrkz,             X86::VPSLLDZ256rmkz,             0 },
  { X86::VPSLLDZrik,                 X86::VPSLLDZmik,                 0 },
  { X86::VPSLLDZrrkz,                X86::VPSLLDZrmkz,                0 },
  { X86::VPSLLQZ128rik,              X86::VPSLLQZ128mik,              0 },
  { X86::VPSLLQZ128rrkz,             X86::VPSLLQZ128rmkz,             0 },
  { X86::VPSLLQZ256rik,              X86::VPSLLQZ256mik,              0 },
  { X86::VPSLLQZ256rrkz,             X86::VPSLLQZ256rmkz,             0 },
  { X86::VPSLLQZrik,                 X86::VPSLLQZmik,                 0 },
  { X86::VPSLLQZrrkz,                X86::VPSLLQZrmkz,                0 },
  { X86::VPSLLVDZ128rrkz,            X86::VPSLLVDZ128rmkz,            0 },
  { X86::VPSLLVDZ256rrkz,            X86::VPSLLVDZ256rmkz,            0 },
  { X86::VPSLLVDZrrkz,               X86::VPSLLVDZrmkz,               0 },
  { X86::VPSLLVQZ128rrkz,            X86::VPSLLVQZ128rmkz,            0 },
  { X86::VPSLLVQZ256rrkz,            X86::VPSLLVQZ256rmkz,            0 },
  { X86::VPSLLVQZrrkz,               X86::VPSLLVQZrmkz,               0 },
  { X86::VPSLLVWZ128rrkz,            X86::VPSLLVWZ128rmkz,            0 },
  { X86::VPSLLVWZ256rrkz,            X86::VPSLLVWZ256rmkz,            0 },
  { X86::VPSLLVWZrrkz,               X86::VPSLLVWZrmkz,               0 },
  { X86::VPSLLWZ128rik,              X86::VPSLLWZ128mik,              0 },
  { X86::VPSLLWZ128rrkz,             X86::VPSLLWZ128rmkz,             0 },
  { X86::VPSLLWZ256rik,              X86::VPSLLWZ256mik,              0 },
  { X86::VPSLLWZ256rrkz,             X86::VPSLLWZ256rmkz,             0 },
  { X86::VPSLLWZrik,                 X86::VPSLLWZmik,                 0 },
  { X86::VPSLLWZrrkz,                X86::VPSLLWZrmkz,                0 },
  { X86::VPSRADZ128rik,              X86::VPSRADZ128mik,              0 },
  { X86::VPSRADZ128rrkz,             X86::VPSRADZ128rmkz,             0 },
  { X86::VPSRADZ256rik,              X86::VPSRADZ256mik,              0 },
  { X86::VPSRADZ256rrkz,             X86::VPSRADZ256rmkz,             0 },
  { X86::VPSRADZrik,                 X86::VPSRADZmik,                 0 },
  { X86::VPSRADZrrkz,                X86::VPSRADZrmkz,                0 },
  { X86::VPSRAQZ128rik,              X86::VPSRAQZ128mik,              0 },
  { X86::VPSRAQZ128rrkz,             X86::VPSRAQZ128rmkz,             0 },
  { X86::VPSRAQZ256rik,              X86::VPSRAQZ256mik,              0 },
  { X86::VPSRAQZ256rrkz,             X86::VPSRAQZ256rmkz,             0 },
  { X86::VPSRAQZrik,                 X86::VPSRAQZmik,                 0 },
  { X86::VPSRAQZrrkz,                X86::VPSRAQZrmkz,                0 },
  { X86::VPSRAVDZ128rrkz,            X86::VPSRAVDZ128rmkz,            0 },
  { X86::VPSRAVDZ256rrkz,            X86::VPSRAVDZ256rmkz,            0 },
  { X86::VPSRAVDZrrkz,               X86::VPSRAVDZrmkz,               0 },
  { X86::VPSRAVQZ128rrkz,            X86::VPSRAVQZ128rmkz,            0 },
  { X86::VPSRAVQZ256rrkz,            X86::VPSRAVQZ256rmkz,            0 },
  { X86::VPSRAVQZrrkz,               X86::VPSRAVQZrmkz,               0 },
  { X86::VPSRAVWZ128rrkz,            X86::VPSRAVWZ128rmkz,            0 },
  { X86::VPSRAVWZ256rrkz,            X86::VPSRAVWZ256rmkz,            0 },
  { X86::VPSRAVWZrrkz,               X86::VPSRAVWZrmkz,               0 },
  { X86::VPSRAWZ128rik,              X86::VPSRAWZ128mik,              0 },
  { X86::VPSRAWZ128rrkz,             X86::VPSRAWZ128rmkz,             0 },
  { X86::VPSRAWZ256rik,              X86::VPSRAWZ256mik,              0 },
  { X86::VPSRAWZ256rrkz,             X86::VPSRAWZ256rmkz,             0 },
  { X86::VPSRAWZrik,                 X86::VPSRAWZmik,                 0 },
  { X86::VPSRAWZrrkz,                X86::VPSRAWZrmkz,                0 },
  { X86::VPSRLDZ128rik,              X86::VPSRLDZ128mik,              0 },
  { X86::VPSRLDZ128rrkz,             X86::VPSRLDZ128rmkz,             0 },
  { X86::VPSRLDZ256rik,              X86::VPSRLDZ256mik,              0 },
  { X86::VPSRLDZ256rrkz,             X86::VPSRLDZ256rmkz,             0 },
  { X86::VPSRLDZrik,                 X86::VPSRLDZmik,                 0 },
  { X86::VPSRLDZrrkz,                X86::VPSRLDZrmkz,                0 },
  { X86::VPSRLQZ128rik,              X86::VPSRLQZ128mik,              0 },
  { X86::VPSRLQZ128rrkz,             X86::VPSRLQZ128rmkz,             0 },
  { X86::VPSRLQZ256rik,              X86::VPSRLQZ256mik,              0 },
  { X86::VPSRLQZ256rrkz,             X86::VPSRLQZ256rmkz,             0 },
  { X86::VPSRLQZrik,                 X86::VPSRLQZmik,                 0 },
  { X86::VPSRLQZrrkz,                X86::VPSRLQZrmkz,                0 },
  { X86::VPSRLVDZ128rrkz,            X86::VPSRLVDZ128rmkz,            0 },
  { X86::VPSRLVDZ256rrkz,            X86::VPSRLVDZ256rmkz,            0 },
  { X86::VPSRLVDZrrkz,               X86::VPSRLVDZrmkz,               0 },
  { X86::VPSRLVQZ128rrkz,            X86::VPSRLVQZ128rmkz,            0 },
  { X86::VPSRLVQZ256rrkz,            X86::VPSRLVQZ256rmkz,            0 },
  { X86::VPSRLVQZrrkz,               X86::VPSRLVQZrmkz,               0 },
  { X86::VPSRLVWZ128rrkz,            X86::VPSRLVWZ128rmkz,            0 },
  { X86::VPSRLVWZ256rrkz,            X86::VPSRLVWZ256rmkz,            0 },
  { X86::VPSRLVWZrrkz,               X86::VPSRLVWZrmkz,               0 },
  { X86::VPSRLWZ128rik,              X86::VPSRLWZ128mik,              0 },
  { X86::VPSRLWZ128rrkz,             X86::VPSRLWZ128rmkz,             0 },
  { X86::VPSRLWZ256rik,              X86::VPSRLWZ256mik,              0 },
  { X86::VPSRLWZ256rrkz,             X86::VPSRLWZ256rmkz,             0 },
  { X86::VPSRLWZrik,                 X86::VPSRLWZmik,                 0 },
  { X86::VPSRLWZrrkz,                X86::VPSRLWZrmkz,                0 },
  { X86::VPSUBBZ128rrkz,             X86::VPSUBBZ128rmkz,             0 },
  { X86::VPSUBBZ256rrkz,             X86::VPSUBBZ256rmkz,             0 },
  { X86::VPSUBBZrrkz,                X86::VPSUBBZrmkz,                0 },
  { X86::VPSUBDZ128rrkz,             X86::VPSUBDZ128rmkz,             0 },
  { X86::VPSUBDZ256rrkz,             X86::VPSUBDZ256rmkz,             0 },
  { X86::VPSUBDZrrkz,                X86::VPSUBDZrmkz,                0 },
  { X86::VPSUBQZ128rrkz,             X86::VPSUBQZ128rmkz,             0 },
  { X86::VPSUBQZ256rrkz,             X86::VPSUBQZ256rmkz,             0 },
  { X86::VPSUBQZrrkz,                X86::VPSUBQZrmkz,                0 },
  { X86::VPSUBSBZ128rrkz,            X86::VPSUBSBZ128rmkz,            0 },
  { X86::VPSUBSBZ256rrkz,            X86::VPSUBSBZ256rmkz,            0 },
  { X86::VPSUBSBZrrkz,               X86::VPSUBSBZrmkz,               0 },
  { X86::VPSUBSWZ128rrkz,            X86::VPSUBSWZ128rmkz,            0 },
  { X86::VPSUBSWZ256rrkz,            X86::VPSUBSWZ256rmkz,            0 },
  { X86::VPSUBSWZrrkz,               X86::VPSUBSWZrmkz,               0 },
  { X86::VPSUBUSBZ128rrkz,           X86::VPSUBUSBZ128rmkz,           0 },
  { X86::VPSUBUSBZ256rrkz,           X86::VPSUBUSBZ256rmkz,           0 },
  { X86::VPSUBUSBZrrkz,              X86::VPSUBUSBZrmkz,              0 },
  { X86::VPSUBUSWZ128rrkz,           X86::VPSUBUSWZ128rmkz,           0 },
  { X86::VPSUBUSWZ256rrkz,           X86::VPSUBUSWZ256rmkz,           0 },
  { X86::VPSUBUSWZrrkz,              X86::VPSUBUSWZrmkz,              0 },
  { X86::VPSUBWZ128rrkz,             X86::VPSUBWZ128rmkz,             0 },
  { X86::VPSUBWZ256rrkz,             X86::VPSUBWZ256rmkz,             0 },
  { X86::VPSUBWZrrkz,                X86::VPSUBWZrmkz,                0 },
  { X86::VPTERNLOGDZ128rri,          X86::VPTERNLOGDZ128rmi,          0 },
  { X86::VPTERNLOGDZ256rri,          X86::VPTERNLOGDZ256rmi,          0 },
  { X86::VPTERNLOGDZrri,             X86::VPTERNLOGDZrmi,             0 },
  { X86::VPTERNLOGQZ128rri,          X86::VPTERNLOGQZ128rmi,          0 },
  { X86::VPTERNLOGQZ256rri,          X86::VPTERNLOGQZ256rmi,          0 },
  { X86::VPTERNLOGQZrri,             X86::VPTERNLOGQZrmi,             0 },
  { X86::VPTESTMBZ128rrk,            X86::VPTESTMBZ128rmk,            0 },
  { X86::VPTESTMBZ256rrk,            X86::VPTESTMBZ256rmk,            0 },
  { X86::VPTESTMBZrrk,               X86::VPTESTMBZrmk,               0 },
  { X86::VPTESTMDZ128rrk,            X86::VPTESTMDZ128rmk,            0 },
  { X86::VPTESTMDZ256rrk,            X86::VPTESTMDZ256rmk,            0 },
  { X86::VPTESTMDZrrk,               X86::VPTESTMDZrmk,               0 },
  { X86::VPTESTMQZ128rrk,            X86::VPTESTMQZ128rmk,            0 },
  { X86::VPTESTMQZ256rrk,            X86::VPTESTMQZ256rmk,            0 },
  { X86::VPTESTMQZrrk,               X86::VPTESTMQZrmk,               0 },
  { X86::VPTESTMWZ128rrk,            X86::VPTESTMWZ128rmk,            0 },
  { X86::VPTESTMWZ256rrk,            X86::VPTESTMWZ256rmk,            0 },
  { X86::VPTESTMWZrrk,               X86::VPTESTMWZrmk,               0 },
  { X86::VPTESTNMBZ128rrk,           X86::VPTESTNMBZ128rmk,           0 },
  { X86::VPTESTNMBZ256rrk,           X86::VPTESTNMBZ256rmk,           0 },
  { X86::VPTESTNMBZrrk,              X86::VPTESTNMBZrmk,              0 },
  { X86::VPTESTNMDZ128rrk,           X86::VPTESTNMDZ128rmk,           0 },
  { X86::VPTESTNMDZ256rrk,           X86::VPTESTNMDZ256rmk,           0 },
  { X86::VPTESTNMDZrrk,              X86::VPTESTNMDZrmk,              0 },
  { X86::VPTESTNMQZ128rrk,           X86::VPTESTNMQZ128rmk,           0 },
  { X86::VPTESTNMQZ256rrk,           X86::VPTESTNMQZ256rmk,           0 },
  { X86::VPTESTNMQZrrk,              X86::VPTESTNMQZrmk,              0 },
  { X86::VPTESTNMWZ128rrk,           X86::VPTESTNMWZ128rmk,           0 },
  { X86::VPTESTNMWZ256rrk,           X86::VPTESTNMWZ256rmk,           0 },
  { X86::VPTESTNMWZrrk,              X86::VPTESTNMWZrmk,              0 },
  { X86::VPUNPCKHBWZ128rrkz,         X86::VPUNPCKHBWZ128rmkz,         0 },
  { X86::VPUNPCKHBWZ256rrkz,         X86::VPUNPCKHBWZ256rmkz,         0 },
  { X86::VPUNPCKHBWZrrkz,            X86::VPUNPCKHBWZrmkz,            0 },
  { X86::VPUNPCKHDQZ128rrkz,         X86::VPUNPCKHDQZ128rmkz,         0 },
  { X86::VPUNPCKHDQZ256rrkz,         X86::VPUNPCKHDQZ256rmkz,         0 },
  { X86::VPUNPCKHDQZrrkz,            X86::VPUNPCKHDQZrmkz,            0 },
  { X86::VPUNPCKHQDQZ128rrkz,        X86::VPUNPCKHQDQZ128rmkz,        0 },
  { X86::VPUNPCKHQDQZ256rrkz,        X86::VPUNPCKHQDQZ256rmkz,        0 },
  { X86::VPUNPCKHQDQZrrkz,           X86::VPUNPCKHQDQZrmkz,           0 },
  { X86::VPUNPCKHWDZ128rrkz,         X86::VPUNPCKHWDZ128rmkz,         0 },
  { X86::VPUNPCKHWDZ256rrkz,         X86::VPUNPCKHWDZ256rmkz,         0 },
  { X86::VPUNPCKHWDZrrkz,            X86::VPUNPCKHWDZrmkz,            0 },
  { X86::VPUNPCKLBWZ128rrkz,         X86::VPUNPCKLBWZ128rmkz,         0 },
  { X86::VPUNPCKLBWZ256rrkz,         X86::VPUNPCKLBWZ256rmkz,         0 },
  { X86::VPUNPCKLBWZrrkz,            X86::VPUNPCKLBWZrmkz,            0 },
  { X86::VPUNPCKLDQZ128rrkz,         X86::VPUNPCKLDQZ128rmkz,         0 },
  { X86::VPUNPCKLDQZ256rrkz,         X86::VPUNPCKLDQZ256rmkz,         0 },
  { X86::VPUNPCKLDQZrrkz,            X86::VPUNPCKLDQZrmkz,            0 },
  { X86::VPUNPCKLQDQZ128rrkz,        X86::VPUNPCKLQDQZ128rmkz,        0 },
  { X86::VPUNPCKLQDQZ256rrkz,        X86::VPUNPCKLQDQZ256rmkz,        0 },
  { X86::VPUNPCKLQDQZrrkz,           X86::VPUNPCKLQDQZrmkz,           0 },
  { X86::VPUNPCKLWDZ128rrkz,         X86::VPUNPCKLWDZ128rmkz,         0 },
  { X86::VPUNPCKLWDZ256rrkz,         X86::VPUNPCKLWDZ256rmkz,         0 },
  { X86::VPUNPCKLWDZrrkz,            X86::VPUNPCKLWDZrmkz,            0 },
  { X86::VPXORDZ128rrkz,             X86::VPXORDZ128rmkz,             0 },
  { X86::VPXORDZ256rrkz,             X86::VPXORDZ256rmkz,             0 },
  { X86::VPXORDZrrkz,                X86::VPXORDZrmkz,                0 },
  { X86::VPXORQZ128rrkz,             X86::VPXORQZ128rmkz,             0 },
  { X86::VPXORQZ256rrkz,             X86::VPXORQZ256rmkz,             0 },
  { X86::VPXORQZrrkz,                X86::VPXORQZrmkz,                0 },
  { X86::VRANGEPDZ128rrikz,          X86::VRANGEPDZ128rmikz,          0 },
  { X86::VRANGEPDZ256rrikz,          X86::VRANGEPDZ256rmikz,          0 },
  { X86::VRANGEPDZrrikz,             X86::VRANGEPDZrmikz,             0 },
  { X86::VRANGEPSZ128rrikz,          X86::VRANGEPSZ128rmikz,          0 },
  { X86::VRANGEPSZ256rrikz,          X86::VRANGEPSZ256rmikz,          0 },
  { X86::VRANGEPSZrrikz,             X86::VRANGEPSZrmikz,             0 },
  { X86::VRANGESDZrrikz,             X86::VRANGESDZrmikz,             TB_NO_REVERSE },
  { X86::VRANGESSZrrikz,             X86::VRANGESSZrmikz,             TB_NO_REVERSE },
  { X86::VRCP14PDZ128rk,             X86::VRCP14PDZ128mk,             0 },
  { X86::VRCP14PDZ256rk,             X86::VRCP14PDZ256mk,             0 },
  { X86::VRCP14PDZrk,                X86::VRCP14PDZmk,                0 },
  { X86::VRCP14PSZ128rk,             X86::VRCP14PSZ128mk,             0 },
  { X86::VRCP14PSZ256rk,             X86::VRCP14PSZ256mk,             0 },
  { X86::VRCP14PSZrk,                X86::VRCP14PSZmk,                0 },
  { X86::VRCP14SDZrrkz,              X86::VRCP14SDZrmkz,              TB_NO_REVERSE },
  { X86::VRCP14SSZrrkz,              X86::VRCP14SSZrmkz,              TB_NO_REVERSE },
  { X86::VRCP28PDZrk,                X86::VRCP28PDZmk,                0 },
  { X86::VRCP28PSZrk,                X86::VRCP28PSZmk,                0 },
  { X86::VRCP28SDZrkz,               X86::VRCP28SDZmkz,               TB_NO_REVERSE },
  { X86::VRCP28SSZrkz,               X86::VRCP28SSZmkz,               TB_NO_REVERSE },
  { X86::VREDUCEPDZ128rrik,          X86::VREDUCEPDZ128rmik,          0 },
  { X86::VREDUCEPDZ256rrik,          X86::VREDUCEPDZ256rmik,          0 },
  { X86::VREDUCEPDZrrik,             X86::VREDUCEPDZrmik,             0 },
  { X86::VREDUCEPSZ128rrik,          X86::VREDUCEPSZ128rmik,          0 },
  { X86::VREDUCEPSZ256rrik,          X86::VREDUCEPSZ256rmik,          0 },
  { X86::VREDUCEPSZrrik,             X86::VREDUCEPSZrmik,             0 },
  { X86::VREDUCESDZrrikz,            X86::VREDUCESDZrmikz,            TB_NO_REVERSE },
  { X86::VREDUCESSZrrikz,            X86::VREDUCESSZrmikz,            TB_NO_REVERSE },
  { X86::VRNDSCALEPDZ128rrik,        X86::VRNDSCALEPDZ128rmik,        0 },
  { X86::VRNDSCALEPDZ256rrik,        X86::VRNDSCALEPDZ256rmik,        0 },
  { X86::VRNDSCALEPDZrrik,           X86::VRNDSCALEPDZrmik,           0 },
  { X86::VRNDSCALEPSZ128rrik,        X86::VRNDSCALEPSZ128rmik,        0 },
  { X86::VRNDSCALEPSZ256rrik,        X86::VRNDSCALEPSZ256rmik,        0 },
  { X86::VRNDSCALEPSZrrik,           X86::VRNDSCALEPSZrmik,           0 },
  { X86::VRNDSCALESDZr_Intkz,        X86::VRNDSCALESDZm_Intkz,        TB_NO_REVERSE },
  { X86::VRNDSCALESSZr_Intkz,        X86::VRNDSCALESSZm_Intkz,        TB_NO_REVERSE },
  { X86::VRSQRT14PDZ128rk,           X86::VRSQRT14PDZ128mk,           0 },
  { X86::VRSQRT14PDZ256rk,           X86::VRSQRT14PDZ256mk,           0 },
  { X86::VRSQRT14PDZrk,              X86::VRSQRT14PDZmk,              0 },
  { X86::VRSQRT14PSZ128rk,           X86::VRSQRT14PSZ128mk,           0 },
  { X86::VRSQRT14PSZ256rk,           X86::VRSQRT14PSZ256mk,           0 },
  { X86::VRSQRT14PSZrk,              X86::VRSQRT14PSZmk,              0 },
  { X86::VRSQRT14SDZrrkz,            X86::VRSQRT14SDZrmkz,            TB_NO_REVERSE },
  { X86::VRSQRT14SSZrrkz,            X86::VRSQRT14SSZrmkz,            TB_NO_REVERSE },
  { X86::VRSQRT28PDZrk,              X86::VRSQRT28PDZmk,              0 },
  { X86::VRSQRT28PSZrk,              X86::VRSQRT28PSZmk,              0 },
  { X86::VRSQRT28SDZrkz,             X86::VRSQRT28SDZmkz,             TB_NO_REVERSE },
  { X86::VRSQRT28SSZrkz,             X86::VRSQRT28SSZmkz,             TB_NO_REVERSE },
  { X86::VSCALEFPDZ128rrkz,          X86::VSCALEFPDZ128rmkz,          0 },
  { X86::VSCALEFPDZ256rrkz,          X86::VSCALEFPDZ256rmkz,          0 },
  { X86::VSCALEFPDZrrkz,             X86::VSCALEFPDZrmkz,             0 },
  { X86::VSCALEFPSZ128rrkz,          X86::VSCALEFPSZ128rmkz,          0 },
  { X86::VSCALEFPSZ256rrkz,          X86::VSCALEFPSZ256rmkz,          0 },
  { X86::VSCALEFPSZrrkz,             X86::VSCALEFPSZrmkz,             0 },
  { X86::VSCALEFSDZrrkz,             X86::VSCALEFSDZrmkz,             TB_NO_REVERSE },
  { X86::VSCALEFSSZrrkz,             X86::VSCALEFSSZrmkz,             TB_NO_REVERSE },
  { X86::VSHUFF32X4Z256rrikz,        X86::VSHUFF32X4Z256rmikz,        0 },
  { X86::VSHUFF32X4Zrrikz,           X86::VSHUFF32X4Zrmikz,           0 },
  { X86::VSHUFF64X2Z256rrikz,        X86::VSHUFF64X2Z256rmikz,        0 },
  { X86::VSHUFF64X2Zrrikz,           X86::VSHUFF64X2Zrmikz,           0 },
  { X86::VSHUFI32X4Z256rrikz,        X86::VSHUFI32X4Z256rmikz,        0 },
  { X86::VSHUFI32X4Zrrikz,           X86::VSHUFI32X4Zrmikz,           0 },
  { X86::VSHUFI64X2Z256rrikz,        X86::VSHUFI64X2Z256rmikz,        0 },
  { X86::VSHUFI64X2Zrrikz,           X86::VSHUFI64X2Zrmikz,           0 },
  { X86::VSHUFPDZ128rrikz,           X86::VSHUFPDZ128rmikz,           0 },
  { X86::VSHUFPDZ256rrikz,           X86::VSHUFPDZ256rmikz,           0 },
  { X86::VSHUFPDZrrikz,              X86::VSHUFPDZrmikz,              0 },
  { X86::VSHUFPSZ128rrikz,           X86::VSHUFPSZ128rmikz,           0 },
  { X86::VSHUFPSZ256rrikz,           X86::VSHUFPSZ256rmikz,           0 },
  { X86::VSHUFPSZrrikz,              X86::VSHUFPSZrmikz,              0 },
  { X86::VSQRTPDZ128rk,              X86::VSQRTPDZ128mk,              0 },
  { X86::VSQRTPDZ256rk,              X86::VSQRTPDZ256mk,              0 },
  { X86::VSQRTPDZrk,                 X86::VSQRTPDZmk,                 0 },
  { X86::VSQRTPSZ128rk,              X86::VSQRTPSZ128mk,              0 },
  { X86::VSQRTPSZ256rk,              X86::VSQRTPSZ256mk,              0 },
  { X86::VSQRTPSZrk,                 X86::VSQRTPSZmk,                 0 },
  { X86::VSQRTSDZr_Intkz,            X86::VSQRTSDZm_Intkz,            TB_NO_REVERSE },
  { X86::VSQRTSSZr_Intkz,            X86::VSQRTSSZm_Intkz,            TB_NO_REVERSE },
  { X86::VSUBPDZ128rrkz,             X86::VSUBPDZ128rmkz,             0 },
  { X86::VSUBPDZ256rrkz,             X86::VSUBPDZ256rmkz,             0 },
  { X86::VSUBPDZrrkz,                X86::VSUBPDZrmkz,                0 },
  { X86::VSUBPSZ128rrkz,             X86::VSUBPSZ128rmkz,             0 },
  { X86::VSUBPSZ256rrkz,             X86::VSUBPSZ256rmkz,             0 },
  { X86::VSUBPSZrrkz,                X86::VSUBPSZrmkz,                0 },
  { X86::VSUBSDZrr_Intkz,            X86::VSUBSDZrm_Intkz,            TB_NO_REVERSE },
  { X86::VSUBSSZrr_Intkz,            X86::VSUBSSZrm_Intkz,            TB_NO_REVERSE },
  { X86::VUNPCKHPDZ128rrkz,          X86::VUNPCKHPDZ128rmkz,          0 },
  { X86::VUNPCKHPDZ256rrkz,          X86::VUNPCKHPDZ256rmkz,          0 },
  { X86::VUNPCKHPDZrrkz,             X86::VUNPCKHPDZrmkz,             0 },
  { X86::VUNPCKHPSZ128rrkz,          X86::VUNPCKHPSZ128rmkz,          0 },
  { X86::VUNPCKHPSZ256rrkz,          X86::VUNPCKHPSZ256rmkz,          0 },
  { X86::VUNPCKHPSZrrkz,             X86::VUNPCKHPSZrmkz,             0 },
  { X86::VUNPCKLPDZ128rrkz,          X86::VUNPCKLPDZ128rmkz,          0 },
  { X86::VUNPCKLPDZ256rrkz,          X86::VUNPCKLPDZ256rmkz,          0 },
  { X86::VUNPCKLPDZrrkz,             X86::VUNPCKLPDZrmkz,             0 },
  { X86::VUNPCKLPSZ128rrkz,          X86::VUNPCKLPSZ128rmkz,          0 },
  { X86::VUNPCKLPSZ256rrkz,          X86::VUNPCKLPSZ256rmkz,          0 },
  { X86::VUNPCKLPSZrrkz,             X86::VUNPCKLPSZrmkz,             0 },
  { X86::VXORPDZ128rrkz,             X86::VXORPDZ128rmkz,             0 },
  { X86::VXORPDZ256rrkz,             X86::VXORPDZ256rmkz,             0 },
  { X86::VXORPDZrrkz,                X86::VXORPDZrmkz,                0 },
  { X86::VXORPSZ128rrkz,             X86::VXORPSZ128rmkz,             0 },
  { X86::VXORPSZ256rrkz,             X86::VXORPSZ256rmkz,             0 },
  { X86::VXORPSZrrkz,                X86::VXORPSZrmkz,                0 },
};

static const X86MemoryFoldTableEntry MemoryFoldTable4[] = {
  { X86::VADDPDZ128rrk,             X86::VADDPDZ128rmk,             0 },
  { X86::VADDPDZ256rrk,             X86::VADDPDZ256rmk,             0 },
  { X86::VADDPDZrrk,                X86::VADDPDZrmk,                0 },
  { X86::VADDPSZ128rrk,             X86::VADDPSZ128rmk,             0 },
  { X86::VADDPSZ256rrk,             X86::VADDPSZ256rmk,             0 },
  { X86::VADDPSZrrk,                X86::VADDPSZrmk,                0 },
  { X86::VADDSDZrr_Intk,            X86::VADDSDZrm_Intk,            TB_NO_REVERSE },
  { X86::VADDSSZrr_Intk,            X86::VADDSSZrm_Intk,            TB_NO_REVERSE },
  { X86::VALIGNDZ128rrik,           X86::VALIGNDZ128rmik,           0 },
  { X86::VALIGNDZ256rrik,           X86::VALIGNDZ256rmik,           0 },
  { X86::VALIGNDZrrik,              X86::VALIGNDZrmik,              0 },
  { X86::VALIGNQZ128rrik,           X86::VALIGNQZ128rmik,           0 },
  { X86::VALIGNQZ256rrik,           X86::VALIGNQZ256rmik,           0 },
  { X86::VALIGNQZrrik,              X86::VALIGNQZrmik,              0 },
  { X86::VANDNPDZ128rrk,            X86::VANDNPDZ128rmk,            0 },
  { X86::VANDNPDZ256rrk,            X86::VANDNPDZ256rmk,            0 },
  { X86::VANDNPDZrrk,               X86::VANDNPDZrmk,               0 },
  { X86::VANDNPSZ128rrk,            X86::VANDNPSZ128rmk,            0 },
  { X86::VANDNPSZ256rrk,            X86::VANDNPSZ256rmk,            0 },
  { X86::VANDNPSZrrk,               X86::VANDNPSZrmk,               0 },
  { X86::VANDPDZ128rrk,             X86::VANDPDZ128rmk,             0 },
  { X86::VANDPDZ256rrk,             X86::VANDPDZ256rmk,             0 },
  { X86::VANDPDZrrk,                X86::VANDPDZrmk,                0 },
  { X86::VANDPSZ128rrk,             X86::VANDPSZ128rmk,             0 },
  { X86::VANDPSZ256rrk,             X86::VANDPSZ256rmk,             0 },
  { X86::VANDPSZrrk,                X86::VANDPSZrmk,                0 },
  { X86::VCVTNE2PS2BF16Z128rrk,     X86::VCVTNE2PS2BF16Z128rmk,     0 },
  { X86::VCVTNE2PS2BF16Z256rrk,     X86::VCVTNE2PS2BF16Z256rmk,     0 },
  { X86::VCVTNE2PS2BF16Zrrk,        X86::VCVTNE2PS2BF16Zrmk,        0 },
  { X86::VCVTSD2SSZrr_Intk,         X86::VCVTSD2SSZrm_Intk,         TB_NO_REVERSE },
  { X86::VCVTSS2SDZrr_Intk,         X86::VCVTSS2SDZrm_Intk,         TB_NO_REVERSE },
  { X86::VDBPSADBWZ128rrik,         X86::VDBPSADBWZ128rmik,         0 },
  { X86::VDBPSADBWZ256rrik,         X86::VDBPSADBWZ256rmik,         0 },
  { X86::VDBPSADBWZrrik,            X86::VDBPSADBWZrmik,            0 },
  { X86::VDIVPDZ128rrk,             X86::VDIVPDZ128rmk,             0 },
  { X86::VDIVPDZ256rrk,             X86::VDIVPDZ256rmk,             0 },
  { X86::VDIVPDZrrk,                X86::VDIVPDZrmk,                0 },
  { X86::VDIVPSZ128rrk,             X86::VDIVPSZ128rmk,             0 },
  { X86::VDIVPSZ256rrk,             X86::VDIVPSZ256rmk,             0 },
  { X86::VDIVPSZrrk,                X86::VDIVPSZrmk,                0 },
  { X86::VDIVSDZrr_Intk,            X86::VDIVSDZrm_Intk,            TB_NO_REVERSE },
  { X86::VDIVSSZrr_Intk,            X86::VDIVSSZrm_Intk,            TB_NO_REVERSE },
  { X86::VDPBF16PSZ128rk,           X86::VDPBF16PSZ128mk,           0 },
  { X86::VDPBF16PSZ128rkz,          X86::VDPBF16PSZ128mkz,          0 },
  { X86::VDPBF16PSZ256rk,           X86::VDPBF16PSZ256mk,           0 },
  { X86::VDPBF16PSZ256rkz,          X86::VDPBF16PSZ256mkz,          0 },
  { X86::VDPBF16PSZrk,              X86::VDPBF16PSZmk,              0 },
  { X86::VDPBF16PSZrkz,             X86::VDPBF16PSZmkz,             0 },
  { X86::VFIXUPIMMPDZ128rrik,       X86::VFIXUPIMMPDZ128rmik,       0 },
  { X86::VFIXUPIMMPDZ128rrikz,      X86::VFIXUPIMMPDZ128rmikz,      0 },
  { X86::VFIXUPIMMPDZ256rrik,       X86::VFIXUPIMMPDZ256rmik,       0 },
  { X86::VFIXUPIMMPDZ256rrikz,      X86::VFIXUPIMMPDZ256rmikz,      0 },
  { X86::VFIXUPIMMPDZrrik,          X86::VFIXUPIMMPDZrmik,          0 },
  { X86::VFIXUPIMMPDZrrikz,         X86::VFIXUPIMMPDZrmikz,         0 },
  { X86::VFIXUPIMMPSZ128rrik,       X86::VFIXUPIMMPSZ128rmik,       0 },
  { X86::VFIXUPIMMPSZ128rrikz,      X86::VFIXUPIMMPSZ128rmikz,      0 },
  { X86::VFIXUPIMMPSZ256rrik,       X86::VFIXUPIMMPSZ256rmik,       0 },
  { X86::VFIXUPIMMPSZ256rrikz,      X86::VFIXUPIMMPSZ256rmikz,      0 },
  { X86::VFIXUPIMMPSZrrik,          X86::VFIXUPIMMPSZrmik,          0 },
  { X86::VFIXUPIMMPSZrrikz,         X86::VFIXUPIMMPSZrmikz,         0 },
  { X86::VFIXUPIMMSDZrrik,          X86::VFIXUPIMMSDZrmik,          TB_NO_REVERSE },
  { X86::VFIXUPIMMSDZrrikz,         X86::VFIXUPIMMSDZrmikz,         TB_NO_REVERSE },
  { X86::VFIXUPIMMSSZrrik,          X86::VFIXUPIMMSSZrmik,          TB_NO_REVERSE },
  { X86::VFIXUPIMMSSZrrikz,         X86::VFIXUPIMMSSZrmikz,         TB_NO_REVERSE },
  { X86::VFMADD132PDZ128rk,         X86::VFMADD132PDZ128mk,         0 },
  { X86::VFMADD132PDZ128rkz,        X86::VFMADD132PDZ128mkz,        0 },
  { X86::VFMADD132PDZ256rk,         X86::VFMADD132PDZ256mk,         0 },
  { X86::VFMADD132PDZ256rkz,        X86::VFMADD132PDZ256mkz,        0 },
  { X86::VFMADD132PDZrk,            X86::VFMADD132PDZmk,            0 },
  { X86::VFMADD132PDZrkz,           X86::VFMADD132PDZmkz,           0 },
  { X86::VFMADD132PSZ128rk,         X86::VFMADD132PSZ128mk,         0 },
  { X86::VFMADD132PSZ128rkz,        X86::VFMADD132PSZ128mkz,        0 },
  { X86::VFMADD132PSZ256rk,         X86::VFMADD132PSZ256mk,         0 },
  { X86::VFMADD132PSZ256rkz,        X86::VFMADD132PSZ256mkz,        0 },
  { X86::VFMADD132PSZrk,            X86::VFMADD132PSZmk,            0 },
  { X86::VFMADD132PSZrkz,           X86::VFMADD132PSZmkz,           0 },
  { X86::VFMADD132SDZr_Intk,        X86::VFMADD132SDZm_Intk,        TB_NO_REVERSE },
  { X86::VFMADD132SDZr_Intkz,       X86::VFMADD132SDZm_Intkz,       TB_NO_REVERSE },
  { X86::VFMADD132SSZr_Intk,        X86::VFMADD132SSZm_Intk,        TB_NO_REVERSE },
  { X86::VFMADD132SSZr_Intkz,       X86::VFMADD132SSZm_Intkz,       TB_NO_REVERSE },
  { X86::VFMADD213PDZ128rk,         X86::VFMADD213PDZ128mk,         0 },
  { X86::VFMADD213PDZ128rkz,        X86::VFMADD213PDZ128mkz,        0 },
  { X86::VFMADD213PDZ256rk,         X86::VFMADD213PDZ256mk,         0 },
  { X86::VFMADD213PDZ256rkz,        X86::VFMADD213PDZ256mkz,        0 },
  { X86::VFMADD213PDZrk,            X86::VFMADD213PDZmk,            0 },
  { X86::VFMADD213PDZrkz,           X86::VFMADD213PDZmkz,           0 },
  { X86::VFMADD213PSZ128rk,         X86::VFMADD213PSZ128mk,         0 },
  { X86::VFMADD213PSZ128rkz,        X86::VFMADD213PSZ128mkz,        0 },
  { X86::VFMADD213PSZ256rk,         X86::VFMADD213PSZ256mk,         0 },
  { X86::VFMADD213PSZ256rkz,        X86::VFMADD213PSZ256mkz,        0 },
  { X86::VFMADD213PSZrk,            X86::VFMADD213PSZmk,            0 },
  { X86::VFMADD213PSZrkz,           X86::VFMADD213PSZmkz,           0 },
  { X86::VFMADD213SDZr_Intk,        X86::VFMADD213SDZm_Intk,        TB_NO_REVERSE },
  { X86::VFMADD213SDZr_Intkz,       X86::VFMADD213SDZm_Intkz,       TB_NO_REVERSE },
  { X86::VFMADD213SSZr_Intk,        X86::VFMADD213SSZm_Intk,        TB_NO_REVERSE },
  { X86::VFMADD213SSZr_Intkz,       X86::VFMADD213SSZm_Intkz,       TB_NO_REVERSE },
  { X86::VFMADD231PDZ128rk,         X86::VFMADD231PDZ128mk,         0 },
  { X86::VFMADD231PDZ128rkz,        X86::VFMADD231PDZ128mkz,        0 },
  { X86::VFMADD231PDZ256rk,         X86::VFMADD231PDZ256mk,         0 },
  { X86::VFMADD231PDZ256rkz,        X86::VFMADD231PDZ256mkz,        0 },
  { X86::VFMADD231PDZrk,            X86::VFMADD231PDZmk,            0 },
  { X86::VFMADD231PDZrkz,           X86::VFMADD231PDZmkz,           0 },
  { X86::VFMADD231PSZ128rk,         X86::VFMADD231PSZ128mk,         0 },
  { X86::VFMADD231PSZ128rkz,        X86::VFMADD231PSZ128mkz,        0 },
  { X86::VFMADD231PSZ256rk,         X86::VFMADD231PSZ256mk,         0 },
  { X86::VFMADD231PSZ256rkz,        X86::VFMADD231PSZ256mkz,        0 },
  { X86::VFMADD231PSZrk,            X86::VFMADD231PSZmk,            0 },
  { X86::VFMADD231PSZrkz,           X86::VFMADD231PSZmkz,           0 },
  { X86::VFMADD231SDZr_Intk,        X86::VFMADD231SDZm_Intk,        TB_NO_REVERSE },
  { X86::VFMADD231SDZr_Intkz,       X86::VFMADD231SDZm_Intkz,       TB_NO_REVERSE },
  { X86::VFMADD231SSZr_Intk,        X86::VFMADD231SSZm_Intk,        TB_NO_REVERSE },
  { X86::VFMADD231SSZr_Intkz,       X86::VFMADD231SSZm_Intkz,       TB_NO_REVERSE },
  { X86::VFMADDSUB132PDZ128rk,      X86::VFMADDSUB132PDZ128mk,      0 },
  { X86::VFMADDSUB132PDZ128rkz,     X86::VFMADDSUB132PDZ128mkz,     0 },
  { X86::VFMADDSUB132PDZ256rk,      X86::VFMADDSUB132PDZ256mk,      0 },
  { X86::VFMADDSUB132PDZ256rkz,     X86::VFMADDSUB132PDZ256mkz,     0 },
  { X86::VFMADDSUB132PDZrk,         X86::VFMADDSUB132PDZmk,         0 },
  { X86::VFMADDSUB132PDZrkz,        X86::VFMADDSUB132PDZmkz,        0 },
  { X86::VFMADDSUB132PSZ128rk,      X86::VFMADDSUB132PSZ128mk,      0 },
  { X86::VFMADDSUB132PSZ128rkz,     X86::VFMADDSUB132PSZ128mkz,     0 },
  { X86::VFMADDSUB132PSZ256rk,      X86::VFMADDSUB132PSZ256mk,      0 },
  { X86::VFMADDSUB132PSZ256rkz,     X86::VFMADDSUB132PSZ256mkz,     0 },
  { X86::VFMADDSUB132PSZrk,         X86::VFMADDSUB132PSZmk,         0 },
  { X86::VFMADDSUB132PSZrkz,        X86::VFMADDSUB132PSZmkz,        0 },
  { X86::VFMADDSUB213PDZ128rk,      X86::VFMADDSUB213PDZ128mk,      0 },
  { X86::VFMADDSUB213PDZ128rkz,     X86::VFMADDSUB213PDZ128mkz,     0 },
  { X86::VFMADDSUB213PDZ256rk,      X86::VFMADDSUB213PDZ256mk,      0 },
  { X86::VFMADDSUB213PDZ256rkz,     X86::VFMADDSUB213PDZ256mkz,     0 },
  { X86::VFMADDSUB213PDZrk,         X86::VFMADDSUB213PDZmk,         0 },
  { X86::VFMADDSUB213PDZrkz,        X86::VFMADDSUB213PDZmkz,        0 },
  { X86::VFMADDSUB213PSZ128rk,      X86::VFMADDSUB213PSZ128mk,      0 },
  { X86::VFMADDSUB213PSZ128rkz,     X86::VFMADDSUB213PSZ128mkz,     0 },
  { X86::VFMADDSUB213PSZ256rk,      X86::VFMADDSUB213PSZ256mk,      0 },
  { X86::VFMADDSUB213PSZ256rkz,     X86::VFMADDSUB213PSZ256mkz,     0 },
  { X86::VFMADDSUB213PSZrk,         X86::VFMADDSUB213PSZmk,         0 },
  { X86::VFMADDSUB213PSZrkz,        X86::VFMADDSUB213PSZmkz,        0 },
  { X86::VFMADDSUB231PDZ128rk,      X86::VFMADDSUB231PDZ128mk,      0 },
  { X86::VFMADDSUB231PDZ128rkz,     X86::VFMADDSUB231PDZ128mkz,     0 },
  { X86::VFMADDSUB231PDZ256rk,      X86::VFMADDSUB231PDZ256mk,      0 },
  { X86::VFMADDSUB231PDZ256rkz,     X86::VFMADDSUB231PDZ256mkz,     0 },
  { X86::VFMADDSUB231PDZrk,         X86::VFMADDSUB231PDZmk,         0 },
  { X86::VFMADDSUB231PDZrkz,        X86::VFMADDSUB231PDZmkz,        0 },
  { X86::VFMADDSUB231PSZ128rk,      X86::VFMADDSUB231PSZ128mk,      0 },
  { X86::VFMADDSUB231PSZ128rkz,     X86::VFMADDSUB231PSZ128mkz,     0 },
  { X86::VFMADDSUB231PSZ256rk,      X86::VFMADDSUB231PSZ256mk,      0 },
  { X86::VFMADDSUB231PSZ256rkz,     X86::VFMADDSUB231PSZ256mkz,     0 },
  { X86::VFMADDSUB231PSZrk,         X86::VFMADDSUB231PSZmk,         0 },
  { X86::VFMADDSUB231PSZrkz,        X86::VFMADDSUB231PSZmkz,        0 },
  { X86::VFMSUB132PDZ128rk,         X86::VFMSUB132PDZ128mk,         0 },
  { X86::VFMSUB132PDZ128rkz,        X86::VFMSUB132PDZ128mkz,        0 },
  { X86::VFMSUB132PDZ256rk,         X86::VFMSUB132PDZ256mk,         0 },
  { X86::VFMSUB132PDZ256rkz,        X86::VFMSUB132PDZ256mkz,        0 },
  { X86::VFMSUB132PDZrk,            X86::VFMSUB132PDZmk,            0 },
  { X86::VFMSUB132PDZrkz,           X86::VFMSUB132PDZmkz,           0 },
  { X86::VFMSUB132PSZ128rk,         X86::VFMSUB132PSZ128mk,         0 },
  { X86::VFMSUB132PSZ128rkz,        X86::VFMSUB132PSZ128mkz,        0 },
  { X86::VFMSUB132PSZ256rk,         X86::VFMSUB132PSZ256mk,         0 },
  { X86::VFMSUB132PSZ256rkz,        X86::VFMSUB132PSZ256mkz,        0 },
  { X86::VFMSUB132PSZrk,            X86::VFMSUB132PSZmk,            0 },
  { X86::VFMSUB132PSZrkz,           X86::VFMSUB132PSZmkz,           0 },
  { X86::VFMSUB132SDZr_Intk,        X86::VFMSUB132SDZm_Intk,        TB_NO_REVERSE },
  { X86::VFMSUB132SDZr_Intkz,       X86::VFMSUB132SDZm_Intkz,       TB_NO_REVERSE },
  { X86::VFMSUB132SSZr_Intk,        X86::VFMSUB132SSZm_Intk,        TB_NO_REVERSE },
  { X86::VFMSUB132SSZr_Intkz,       X86::VFMSUB132SSZm_Intkz,       TB_NO_REVERSE },
  { X86::VFMSUB213PDZ128rk,         X86::VFMSUB213PDZ128mk,         0 },
  { X86::VFMSUB213PDZ128rkz,        X86::VFMSUB213PDZ128mkz,        0 },
  { X86::VFMSUB213PDZ256rk,         X86::VFMSUB213PDZ256mk,         0 },
  { X86::VFMSUB213PDZ256rkz,        X86::VFMSUB213PDZ256mkz,        0 },
  { X86::VFMSUB213PDZrk,            X86::VFMSUB213PDZmk,            0 },
  { X86::VFMSUB213PDZrkz,           X86::VFMSUB213PDZmkz,           0 },
  { X86::VFMSUB213PSZ128rk,         X86::VFMSUB213PSZ128mk,         0 },
  { X86::VFMSUB213PSZ128rkz,        X86::VFMSUB213PSZ128mkz,        0 },
  { X86::VFMSUB213PSZ256rk,         X86::VFMSUB213PSZ256mk,         0 },
  { X86::VFMSUB213PSZ256rkz,        X86::VFMSUB213PSZ256mkz,        0 },
  { X86::VFMSUB213PSZrk,            X86::VFMSUB213PSZmk,            0 },
  { X86::VFMSUB213PSZrkz,           X86::VFMSUB213PSZmkz,           0 },
  { X86::VFMSUB213SDZr_Intk,        X86::VFMSUB213SDZm_Intk,        TB_NO_REVERSE },
  { X86::VFMSUB213SDZr_Intkz,       X86::VFMSUB213SDZm_Intkz,       TB_NO_REVERSE },
  { X86::VFMSUB213SSZr_Intk,        X86::VFMSUB213SSZm_Intk,        TB_NO_REVERSE },
  { X86::VFMSUB213SSZr_Intkz,       X86::VFMSUB213SSZm_Intkz,       TB_NO_REVERSE },
  { X86::VFMSUB231PDZ128rk,         X86::VFMSUB231PDZ128mk,         0 },
  { X86::VFMSUB231PDZ128rkz,        X86::VFMSUB231PDZ128mkz,        0 },
  { X86::VFMSUB231PDZ256rk,         X86::VFMSUB231PDZ256mk,         0 },
  { X86::VFMSUB231PDZ256rkz,        X86::VFMSUB231PDZ256mkz,        0 },
  { X86::VFMSUB231PDZrk,            X86::VFMSUB231PDZmk,            0 },
  { X86::VFMSUB231PDZrkz,           X86::VFMSUB231PDZmkz,           0 },
  { X86::VFMSUB231PSZ128rk,         X86::VFMSUB231PSZ128mk,         0 },
  { X86::VFMSUB231PSZ128rkz,        X86::VFMSUB231PSZ128mkz,        0 },
  { X86::VFMSUB231PSZ256rk,         X86::VFMSUB231PSZ256mk,         0 },
  { X86::VFMSUB231PSZ256rkz,        X86::VFMSUB231PSZ256mkz,        0 },
  { X86::VFMSUB231PSZrk,            X86::VFMSUB231PSZmk,            0 },
  { X86::VFMSUB231PSZrkz,           X86::VFMSUB231PSZmkz,           0 },
  { X86::VFMSUB231SDZr_Intk,        X86::VFMSUB231SDZm_Intk,        TB_NO_REVERSE },
  { X86::VFMSUB231SDZr_Intkz,       X86::VFMSUB231SDZm_Intkz,       TB_NO_REVERSE },
  { X86::VFMSUB231SSZr_Intk,        X86::VFMSUB231SSZm_Intk,        TB_NO_REVERSE },
  { X86::VFMSUB231SSZr_Intkz,       X86::VFMSUB231SSZm_Intkz,       TB_NO_REVERSE },
  { X86::VFMSUBADD132PDZ128rk,      X86::VFMSUBADD132PDZ128mk,      0 },
  { X86::VFMSUBADD132PDZ128rkz,     X86::VFMSUBADD132PDZ128mkz,     0 },
  { X86::VFMSUBADD132PDZ256rk,      X86::VFMSUBADD132PDZ256mk,      0 },
  { X86::VFMSUBADD132PDZ256rkz,     X86::VFMSUBADD132PDZ256mkz,     0 },
  { X86::VFMSUBADD132PDZrk,         X86::VFMSUBADD132PDZmk,         0 },
  { X86::VFMSUBADD132PDZrkz,        X86::VFMSUBADD132PDZmkz,        0 },
  { X86::VFMSUBADD132PSZ128rk,      X86::VFMSUBADD132PSZ128mk,      0 },
  { X86::VFMSUBADD132PSZ128rkz,     X86::VFMSUBADD132PSZ128mkz,     0 },
  { X86::VFMSUBADD132PSZ256rk,      X86::VFMSUBADD132PSZ256mk,      0 },
  { X86::VFMSUBADD132PSZ256rkz,     X86::VFMSUBADD132PSZ256mkz,     0 },
  { X86::VFMSUBADD132PSZrk,         X86::VFMSUBADD132PSZmk,         0 },
  { X86::VFMSUBADD132PSZrkz,        X86::VFMSUBADD132PSZmkz,        0 },
  { X86::VFMSUBADD213PDZ128rk,      X86::VFMSUBADD213PDZ128mk,      0 },
  { X86::VFMSUBADD213PDZ128rkz,     X86::VFMSUBADD213PDZ128mkz,     0 },
  { X86::VFMSUBADD213PDZ256rk,      X86::VFMSUBADD213PDZ256mk,      0 },
  { X86::VFMSUBADD213PDZ256rkz,     X86::VFMSUBADD213PDZ256mkz,     0 },
  { X86::VFMSUBADD213PDZrk,         X86::VFMSUBADD213PDZmk,         0 },
  { X86::VFMSUBADD213PDZrkz,        X86::VFMSUBADD213PDZmkz,        0 },
  { X86::VFMSUBADD213PSZ128rk,      X86::VFMSUBADD213PSZ128mk,      0 },
  { X86::VFMSUBADD213PSZ128rkz,     X86::VFMSUBADD213PSZ128mkz,     0 },
  { X86::VFMSUBADD213PSZ256rk,      X86::VFMSUBADD213PSZ256mk,      0 },
  { X86::VFMSUBADD213PSZ256rkz,     X86::VFMSUBADD213PSZ256mkz,     0 },
  { X86::VFMSUBADD213PSZrk,         X86::VFMSUBADD213PSZmk,         0 },
  { X86::VFMSUBADD213PSZrkz,        X86::VFMSUBADD213PSZmkz,        0 },
  { X86::VFMSUBADD231PDZ128rk,      X86::VFMSUBADD231PDZ128mk,      0 },
  { X86::VFMSUBADD231PDZ128rkz,     X86::VFMSUBADD231PDZ128mkz,     0 },
  { X86::VFMSUBADD231PDZ256rk,      X86::VFMSUBADD231PDZ256mk,      0 },
  { X86::VFMSUBADD231PDZ256rkz,     X86::VFMSUBADD231PDZ256mkz,     0 },
  { X86::VFMSUBADD231PDZrk,         X86::VFMSUBADD231PDZmk,         0 },
  { X86::VFMSUBADD231PDZrkz,        X86::VFMSUBADD231PDZmkz,        0 },
  { X86::VFMSUBADD231PSZ128rk,      X86::VFMSUBADD231PSZ128mk,      0 },
  { X86::VFMSUBADD231PSZ128rkz,     X86::VFMSUBADD231PSZ128mkz,     0 },
  { X86::VFMSUBADD231PSZ256rk,      X86::VFMSUBADD231PSZ256mk,      0 },
  { X86::VFMSUBADD231PSZ256rkz,     X86::VFMSUBADD231PSZ256mkz,     0 },
  { X86::VFMSUBADD231PSZrk,         X86::VFMSUBADD231PSZmk,         0 },
  { X86::VFMSUBADD231PSZrkz,        X86::VFMSUBADD231PSZmkz,        0 },
  { X86::VFNMADD132PDZ128rk,        X86::VFNMADD132PDZ128mk,        0 },
  { X86::VFNMADD132PDZ128rkz,       X86::VFNMADD132PDZ128mkz,       0 },
  { X86::VFNMADD132PDZ256rk,        X86::VFNMADD132PDZ256mk,        0 },
  { X86::VFNMADD132PDZ256rkz,       X86::VFNMADD132PDZ256mkz,       0 },
  { X86::VFNMADD132PDZrk,           X86::VFNMADD132PDZmk,           0 },
  { X86::VFNMADD132PDZrkz,          X86::VFNMADD132PDZmkz,          0 },
  { X86::VFNMADD132PSZ128rk,        X86::VFNMADD132PSZ128mk,        0 },
  { X86::VFNMADD132PSZ128rkz,       X86::VFNMADD132PSZ128mkz,       0 },
  { X86::VFNMADD132PSZ256rk,        X86::VFNMADD132PSZ256mk,        0 },
  { X86::VFNMADD132PSZ256rkz,       X86::VFNMADD132PSZ256mkz,       0 },
  { X86::VFNMADD132PSZrk,           X86::VFNMADD132PSZmk,           0 },
  { X86::VFNMADD132PSZrkz,          X86::VFNMADD132PSZmkz,          0 },
  { X86::VFNMADD132SDZr_Intk,       X86::VFNMADD132SDZm_Intk,       TB_NO_REVERSE },
  { X86::VFNMADD132SDZr_Intkz,      X86::VFNMADD132SDZm_Intkz,      TB_NO_REVERSE },
  { X86::VFNMADD132SSZr_Intk,       X86::VFNMADD132SSZm_Intk,       TB_NO_REVERSE },
  { X86::VFNMADD132SSZr_Intkz,      X86::VFNMADD132SSZm_Intkz,      TB_NO_REVERSE },
  { X86::VFNMADD213PDZ128rk,        X86::VFNMADD213PDZ128mk,        0 },
  { X86::VFNMADD213PDZ128rkz,       X86::VFNMADD213PDZ128mkz,       0 },
  { X86::VFNMADD213PDZ256rk,        X86::VFNMADD213PDZ256mk,        0 },
  { X86::VFNMADD213PDZ256rkz,       X86::VFNMADD213PDZ256mkz,       0 },
  { X86::VFNMADD213PDZrk,           X86::VFNMADD213PDZmk,           0 },
  { X86::VFNMADD213PDZrkz,          X86::VFNMADD213PDZmkz,          0 },
  { X86::VFNMADD213PSZ128rk,        X86::VFNMADD213PSZ128mk,        0 },
  { X86::VFNMADD213PSZ128rkz,       X86::VFNMADD213PSZ128mkz,       0 },
  { X86::VFNMADD213PSZ256rk,        X86::VFNMADD213PSZ256mk,        0 },
  { X86::VFNMADD213PSZ256rkz,       X86::VFNMADD213PSZ256mkz,       0 },
  { X86::VFNMADD213PSZrk,           X86::VFNMADD213PSZmk,           0 },
  { X86::VFNMADD213PSZrkz,          X86::VFNMADD213PSZmkz,          0 },
  { X86::VFNMADD213SDZr_Intk,       X86::VFNMADD213SDZm_Intk,       TB_NO_REVERSE },
  { X86::VFNMADD213SDZr_Intkz,      X86::VFNMADD213SDZm_Intkz,      TB_NO_REVERSE },
  { X86::VFNMADD213SSZr_Intk,       X86::VFNMADD213SSZm_Intk,       TB_NO_REVERSE },
  { X86::VFNMADD213SSZr_Intkz,      X86::VFNMADD213SSZm_Intkz,      TB_NO_REVERSE },
  { X86::VFNMADD231PDZ128rk,        X86::VFNMADD231PDZ128mk,        0 },
  { X86::VFNMADD231PDZ128rkz,       X86::VFNMADD231PDZ128mkz,       0 },
  { X86::VFNMADD231PDZ256rk,        X86::VFNMADD231PDZ256mk,        0 },
  { X86::VFNMADD231PDZ256rkz,       X86::VFNMADD231PDZ256mkz,       0 },
  { X86::VFNMADD231PDZrk,           X86::VFNMADD231PDZmk,           0 },
  { X86::VFNMADD231PDZrkz,          X86::VFNMADD231PDZmkz,          0 },
  { X86::VFNMADD231PSZ128rk,        X86::VFNMADD231PSZ128mk,        0 },
  { X86::VFNMADD231PSZ128rkz,       X86::VFNMADD231PSZ128mkz,       0 },
  { X86::VFNMADD231PSZ256rk,        X86::VFNMADD231PSZ256mk,        0 },
  { X86::VFNMADD231PSZ256rkz,       X86::VFNMADD231PSZ256mkz,       0 },
  { X86::VFNMADD231PSZrk,           X86::VFNMADD231PSZmk,           0 },
  { X86::VFNMADD231PSZrkz,          X86::VFNMADD231PSZmkz,          0 },
  { X86::VFNMADD231SDZr_Intk,       X86::VFNMADD231SDZm_Intk,       TB_NO_REVERSE },
  { X86::VFNMADD231SDZr_Intkz,      X86::VFNMADD231SDZm_Intkz,      TB_NO_REVERSE },
  { X86::VFNMADD231SSZr_Intk,       X86::VFNMADD231SSZm_Intk,       TB_NO_REVERSE },
  { X86::VFNMADD231SSZr_Intkz,      X86::VFNMADD231SSZm_Intkz,      TB_NO_REVERSE },
  { X86::VFNMSUB132PDZ128rk,        X86::VFNMSUB132PDZ128mk,        0 },
  { X86::VFNMSUB132PDZ128rkz,       X86::VFNMSUB132PDZ128mkz,       0 },
  { X86::VFNMSUB132PDZ256rk,        X86::VFNMSUB132PDZ256mk,        0 },
  { X86::VFNMSUB132PDZ256rkz,       X86::VFNMSUB132PDZ256mkz,       0 },
  { X86::VFNMSUB132PDZrk,           X86::VFNMSUB132PDZmk,           0 },
  { X86::VFNMSUB132PDZrkz,          X86::VFNMSUB132PDZmkz,          0 },
  { X86::VFNMSUB132PSZ128rk,        X86::VFNMSUB132PSZ128mk,        0 },
  { X86::VFNMSUB132PSZ128rkz,       X86::VFNMSUB132PSZ128mkz,       0 },
  { X86::VFNMSUB132PSZ256rk,        X86::VFNMSUB132PSZ256mk,        0 },
  { X86::VFNMSUB132PSZ256rkz,       X86::VFNMSUB132PSZ256mkz,       0 },
  { X86::VFNMSUB132PSZrk,           X86::VFNMSUB132PSZmk,           0 },
  { X86::VFNMSUB132PSZrkz,          X86::VFNMSUB132PSZmkz,          0 },
  { X86::VFNMSUB132SDZr_Intk,       X86::VFNMSUB132SDZm_Intk,       TB_NO_REVERSE },
  { X86::VFNMSUB132SDZr_Intkz,      X86::VFNMSUB132SDZm_Intkz,      TB_NO_REVERSE },
  { X86::VFNMSUB132SSZr_Intk,       X86::VFNMSUB132SSZm_Intk,       TB_NO_REVERSE },
  { X86::VFNMSUB132SSZr_Intkz,      X86::VFNMSUB132SSZm_Intkz,      TB_NO_REVERSE },
  { X86::VFNMSUB213PDZ128rk,        X86::VFNMSUB213PDZ128mk,        0 },
  { X86::VFNMSUB213PDZ128rkz,       X86::VFNMSUB213PDZ128mkz,       0 },
  { X86::VFNMSUB213PDZ256rk,        X86::VFNMSUB213PDZ256mk,        0 },
  { X86::VFNMSUB213PDZ256rkz,       X86::VFNMSUB213PDZ256mkz,       0 },
  { X86::VFNMSUB213PDZrk,           X86::VFNMSUB213PDZmk,           0 },
  { X86::VFNMSUB213PDZrkz,          X86::VFNMSUB213PDZmkz,          0 },
  { X86::VFNMSUB213PSZ128rk,        X86::VFNMSUB213PSZ128mk,        0 },
  { X86::VFNMSUB213PSZ128rkz,       X86::VFNMSUB213PSZ128mkz,       0 },
  { X86::VFNMSUB213PSZ256rk,        X86::VFNMSUB213PSZ256mk,        0 },
  { X86::VFNMSUB213PSZ256rkz,       X86::VFNMSUB213PSZ256mkz,       0 },
  { X86::VFNMSUB213PSZrk,           X86::VFNMSUB213PSZmk,           0 },
  { X86::VFNMSUB213PSZrkz,          X86::VFNMSUB213PSZmkz,          0 },
  { X86::VFNMSUB213SDZr_Intk,       X86::VFNMSUB213SDZm_Intk,       TB_NO_REVERSE },
  { X86::VFNMSUB213SDZr_Intkz,      X86::VFNMSUB213SDZm_Intkz,      TB_NO_REVERSE },
  { X86::VFNMSUB213SSZr_Intk,       X86::VFNMSUB213SSZm_Intk,       TB_NO_REVERSE },
  { X86::VFNMSUB213SSZr_Intkz,      X86::VFNMSUB213SSZm_Intkz,      TB_NO_REVERSE },
  { X86::VFNMSUB231PDZ128rk,        X86::VFNMSUB231PDZ128mk,        0 },
  { X86::VFNMSUB231PDZ128rkz,       X86::VFNMSUB231PDZ128mkz,       0 },
  { X86::VFNMSUB231PDZ256rk,        X86::VFNMSUB231PDZ256mk,        0 },
  { X86::VFNMSUB231PDZ256rkz,       X86::VFNMSUB231PDZ256mkz,       0 },
  { X86::VFNMSUB231PDZrk,           X86::VFNMSUB231PDZmk,           0 },
  { X86::VFNMSUB231PDZrkz,          X86::VFNMSUB231PDZmkz,          0 },
  { X86::VFNMSUB231PSZ128rk,        X86::VFNMSUB231PSZ128mk,        0 },
  { X86::VFNMSUB231PSZ128rkz,       X86::VFNMSUB231PSZ128mkz,       0 },
  { X86::VFNMSUB231PSZ256rk,        X86::VFNMSUB231PSZ256mk,        0 },
  { X86::VFNMSUB231PSZ256rkz,       X86::VFNMSUB231PSZ256mkz,       0 },
  { X86::VFNMSUB231PSZrk,           X86::VFNMSUB231PSZmk,           0 },
  { X86::VFNMSUB231PSZrkz,          X86::VFNMSUB231PSZmkz,          0 },
  { X86::VFNMSUB231SDZr_Intk,       X86::VFNMSUB231SDZm_Intk,       TB_NO_REVERSE },
  { X86::VFNMSUB231SDZr_Intkz,      X86::VFNMSUB231SDZm_Intkz,      TB_NO_REVERSE },
  { X86::VFNMSUB231SSZr_Intk,       X86::VFNMSUB231SSZm_Intk,       TB_NO_REVERSE },
  { X86::VFNMSUB231SSZr_Intkz,      X86::VFNMSUB231SSZm_Intkz,      TB_NO_REVERSE },
  { X86::VGETEXPSDZrk,              X86::VGETEXPSDZmk,              TB_NO_REVERSE },
  { X86::VGETEXPSSZrk,              X86::VGETEXPSSZmk,              TB_NO_REVERSE },
  { X86::VGETMANTSDZrrik,           X86::VGETMANTSDZrmik,           TB_NO_REVERSE },
  { X86::VGETMANTSSZrrik,           X86::VGETMANTSSZrmik,           TB_NO_REVERSE },
  { X86::VGF2P8AFFINEINVQBZ128rrik, X86::VGF2P8AFFINEINVQBZ128rmik, 0 },
  { X86::VGF2P8AFFINEINVQBZ256rrik, X86::VGF2P8AFFINEINVQBZ256rmik, 0 },
  { X86::VGF2P8AFFINEINVQBZrrik,    X86::VGF2P8AFFINEINVQBZrmik,    0 },
  { X86::VGF2P8AFFINEQBZ128rrik,    X86::VGF2P8AFFINEQBZ128rmik,    0 },
  { X86::VGF2P8AFFINEQBZ256rrik,    X86::VGF2P8AFFINEQBZ256rmik,    0 },
  { X86::VGF2P8AFFINEQBZrrik,       X86::VGF2P8AFFINEQBZrmik,       0 },
  { X86::VGF2P8MULBZ128rrk,         X86::VGF2P8MULBZ128rmk,         0 },
  { X86::VGF2P8MULBZ256rrk,         X86::VGF2P8MULBZ256rmk,         0 },
  { X86::VGF2P8MULBZrrk,            X86::VGF2P8MULBZrmk,            0 },
  { X86::VINSERTF32x4Z256rrk,       X86::VINSERTF32x4Z256rmk,       0 },
  { X86::VINSERTF32x4Zrrk,          X86::VINSERTF32x4Zrmk,          0 },
  { X86::VINSERTF32x8Zrrk,          X86::VINSERTF32x8Zrmk,          0 },
  { X86::VINSERTF64x2Z256rrk,       X86::VINSERTF64x2Z256rmk,       0 },
  { X86::VINSERTF64x2Zrrk,          X86::VINSERTF64x2Zrmk,          0 },
  { X86::VINSERTF64x4Zrrk,          X86::VINSERTF64x4Zrmk,          0 },
  { X86::VINSERTI32x4Z256rrk,       X86::VINSERTI32x4Z256rmk,       0 },
  { X86::VINSERTI32x4Zrrk,          X86::VINSERTI32x4Zrmk,          0 },
  { X86::VINSERTI32x8Zrrk,          X86::VINSERTI32x8Zrmk,          0 },
  { X86::VINSERTI64x2Z256rrk,       X86::VINSERTI64x2Z256rmk,       0 },
  { X86::VINSERTI64x2Zrrk,          X86::VINSERTI64x2Zrmk,          0 },
  { X86::VINSERTI64x4Zrrk,          X86::VINSERTI64x4Zrmk,          0 },
  { X86::VMAXCPDZ128rrk,            X86::VMAXCPDZ128rmk,            0 },
  { X86::VMAXCPDZ256rrk,            X86::VMAXCPDZ256rmk,            0 },
  { X86::VMAXCPDZrrk,               X86::VMAXCPDZrmk,               0 },
  { X86::VMAXCPSZ128rrk,            X86::VMAXCPSZ128rmk,            0 },
  { X86::VMAXCPSZ256rrk,            X86::VMAXCPSZ256rmk,            0 },
  { X86::VMAXCPSZrrk,               X86::VMAXCPSZrmk,               0 },
  { X86::VMAXPDZ128rrk,             X86::VMAXPDZ128rmk,             0 },
  { X86::VMAXPDZ256rrk,             X86::VMAXPDZ256rmk,             0 },
  { X86::VMAXPDZrrk,                X86::VMAXPDZrmk,                0 },
  { X86::VMAXPSZ128rrk,             X86::VMAXPSZ128rmk,             0 },
  { X86::VMAXPSZ256rrk,             X86::VMAXPSZ256rmk,             0 },
  { X86::VMAXPSZrrk,                X86::VMAXPSZrmk,                0 },
  { X86::VMAXSDZrr_Intk,            X86::VMAXSDZrm_Intk,            TB_NO_REVERSE },
  { X86::VMAXSSZrr_Intk,            X86::VMAXSSZrm_Intk,            TB_NO_REVERSE },
  { X86::VMINCPDZ128rrk,            X86::VMINCPDZ128rmk,            0 },
  { X86::VMINCPDZ256rrk,            X86::VMINCPDZ256rmk,            0 },
  { X86::VMINCPDZrrk,               X86::VMINCPDZrmk,               0 },
  { X86::VMINCPSZ128rrk,            X86::VMINCPSZ128rmk,            0 },
  { X86::VMINCPSZ256rrk,            X86::VMINCPSZ256rmk,            0 },
  { X86::VMINCPSZrrk,               X86::VMINCPSZrmk,               0 },
  { X86::VMINPDZ128rrk,             X86::VMINPDZ128rmk,             0 },
  { X86::VMINPDZ256rrk,             X86::VMINPDZ256rmk,             0 },
  { X86::VMINPDZrrk,                X86::VMINPDZrmk,                0 },
  { X86::VMINPSZ128rrk,             X86::VMINPSZ128rmk,             0 },
  { X86::VMINPSZ256rrk,             X86::VMINPSZ256rmk,             0 },
  { X86::VMINPSZrrk,                X86::VMINPSZrmk,                0 },
  { X86::VMINSDZrr_Intk,            X86::VMINSDZrm_Intk,            TB_NO_REVERSE },
  { X86::VMINSSZrr_Intk,            X86::VMINSSZrm_Intk,            TB_NO_REVERSE },
  { X86::VMULPDZ128rrk,             X86::VMULPDZ128rmk,             0 },
  { X86::VMULPDZ256rrk,             X86::VMULPDZ256rmk,             0 },
  { X86::VMULPDZrrk,                X86::VMULPDZrmk,                0 },
  { X86::VMULPSZ128rrk,             X86::VMULPSZ128rmk,             0 },
  { X86::VMULPSZ256rrk,             X86::VMULPSZ256rmk,             0 },
  { X86::VMULPSZrrk,                X86::VMULPSZrmk,                0 },
  { X86::VMULSDZrr_Intk,            X86::VMULSDZrm_Intk,            TB_NO_REVERSE },
  { X86::VMULSSZrr_Intk,            X86::VMULSSZrm_Intk,            TB_NO_REVERSE },
  { X86::VORPDZ128rrk,              X86::VORPDZ128rmk,              0 },
  { X86::VORPDZ256rrk,              X86::VORPDZ256rmk,              0 },
  { X86::VORPDZrrk,                 X86::VORPDZrmk,                 0 },
  { X86::VORPSZ128rrk,              X86::VORPSZ128rmk,              0 },
  { X86::VORPSZ256rrk,              X86::VORPSZ256rmk,              0 },
  { X86::VORPSZrrk,                 X86::VORPSZrmk,                 0 },
  { X86::VPACKSSDWZ128rrk,          X86::VPACKSSDWZ128rmk,          0 },
  { X86::VPACKSSDWZ256rrk,          X86::VPACKSSDWZ256rmk,          0 },
  { X86::VPACKSSDWZrrk,             X86::VPACKSSDWZrmk,             0 },
  { X86::VPACKSSWBZ128rrk,          X86::VPACKSSWBZ128rmk,          0 },
  { X86::VPACKSSWBZ256rrk,          X86::VPACKSSWBZ256rmk,          0 },
  { X86::VPACKSSWBZrrk,             X86::VPACKSSWBZrmk,             0 },
  { X86::VPACKUSDWZ128rrk,          X86::VPACKUSDWZ128rmk,          0 },
  { X86::VPACKUSDWZ256rrk,          X86::VPACKUSDWZ256rmk,          0 },
  { X86::VPACKUSDWZrrk,             X86::VPACKUSDWZrmk,             0 },
  { X86::VPACKUSWBZ128rrk,          X86::VPACKUSWBZ128rmk,          0 },
  { X86::VPACKUSWBZ256rrk,          X86::VPACKUSWBZ256rmk,          0 },
  { X86::VPACKUSWBZrrk,             X86::VPACKUSWBZrmk,             0 },
  { X86::VPADDBZ128rrk,             X86::VPADDBZ128rmk,             0 },
  { X86::VPADDBZ256rrk,             X86::VPADDBZ256rmk,             0 },
  { X86::VPADDBZrrk,                X86::VPADDBZrmk,                0 },
  { X86::VPADDDZ128rrk,             X86::VPADDDZ128rmk,             0 },
  { X86::VPADDDZ256rrk,             X86::VPADDDZ256rmk,             0 },
  { X86::VPADDDZrrk,                X86::VPADDDZrmk,                0 },
  { X86::VPADDQZ128rrk,             X86::VPADDQZ128rmk,             0 },
  { X86::VPADDQZ256rrk,             X86::VPADDQZ256rmk,             0 },
  { X86::VPADDQZrrk,                X86::VPADDQZrmk,                0 },
  { X86::VPADDSBZ128rrk,            X86::VPADDSBZ128rmk,            0 },
  { X86::VPADDSBZ256rrk,            X86::VPADDSBZ256rmk,            0 },
  { X86::VPADDSBZrrk,               X86::VPADDSBZrmk,               0 },
  { X86::VPADDSWZ128rrk,            X86::VPADDSWZ128rmk,            0 },
  { X86::VPADDSWZ256rrk,            X86::VPADDSWZ256rmk,            0 },
  { X86::VPADDSWZrrk,               X86::VPADDSWZrmk,               0 },
  { X86::VPADDUSBZ128rrk,           X86::VPADDUSBZ128rmk,           0 },
  { X86::VPADDUSBZ256rrk,           X86::VPADDUSBZ256rmk,           0 },
  { X86::VPADDUSBZrrk,              X86::VPADDUSBZrmk,              0 },
  { X86::VPADDUSWZ128rrk,           X86::VPADDUSWZ128rmk,           0 },
  { X86::VPADDUSWZ256rrk,           X86::VPADDUSWZ256rmk,           0 },
  { X86::VPADDUSWZrrk,              X86::VPADDUSWZrmk,              0 },
  { X86::VPADDWZ128rrk,             X86::VPADDWZ128rmk,             0 },
  { X86::VPADDWZ256rrk,             X86::VPADDWZ256rmk,             0 },
  { X86::VPADDWZrrk,                X86::VPADDWZrmk,                0 },
  { X86::VPALIGNRZ128rrik,          X86::VPALIGNRZ128rmik,          0 },
  { X86::VPALIGNRZ256rrik,          X86::VPALIGNRZ256rmik,          0 },
  { X86::VPALIGNRZrrik,             X86::VPALIGNRZrmik,             0 },
  { X86::VPANDDZ128rrk,             X86::VPANDDZ128rmk,             0 },
  { X86::VPANDDZ256rrk,             X86::VPANDDZ256rmk,             0 },
  { X86::VPANDDZrrk,                X86::VPANDDZrmk,                0 },
  { X86::VPANDNDZ128rrk,            X86::VPANDNDZ128rmk,            0 },
  { X86::VPANDNDZ256rrk,            X86::VPANDNDZ256rmk,            0 },
  { X86::VPANDNDZrrk,               X86::VPANDNDZrmk,               0 },
  { X86::VPANDNQZ128rrk,            X86::VPANDNQZ128rmk,            0 },
  { X86::VPANDNQZ256rrk,            X86::VPANDNQZ256rmk,            0 },
  { X86::VPANDNQZrrk,               X86::VPANDNQZrmk,               0 },
  { X86::VPANDQZ128rrk,             X86::VPANDQZ128rmk,             0 },
  { X86::VPANDQZ256rrk,             X86::VPANDQZ256rmk,             0 },
  { X86::VPANDQZrrk,                X86::VPANDQZrmk,                0 },
  { X86::VPAVGBZ128rrk,             X86::VPAVGBZ128rmk,             0 },
  { X86::VPAVGBZ256rrk,             X86::VPAVGBZ256rmk,             0 },
  { X86::VPAVGBZrrk,                X86::VPAVGBZrmk,                0 },
  { X86::VPAVGWZ128rrk,             X86::VPAVGWZ128rmk,             0 },
  { X86::VPAVGWZ256rrk,             X86::VPAVGWZ256rmk,             0 },
  { X86::VPAVGWZrrk,                X86::VPAVGWZrmk,                0 },
  { X86::VPDPBUSDSZ128rk,           X86::VPDPBUSDSZ128mk,           0 },
  { X86::VPDPBUSDSZ128rkz,          X86::VPDPBUSDSZ128mkz,          0 },
  { X86::VPDPBUSDSZ256rk,           X86::VPDPBUSDSZ256mk,           0 },
  { X86::VPDPBUSDSZ256rkz,          X86::VPDPBUSDSZ256mkz,          0 },
  { X86::VPDPBUSDSZrk,              X86::VPDPBUSDSZmk,              0 },
  { X86::VPDPBUSDSZrkz,             X86::VPDPBUSDSZmkz,             0 },
  { X86::VPDPBUSDZ128rk,            X86::VPDPBUSDZ128mk,            0 },
  { X86::VPDPBUSDZ128rkz,           X86::VPDPBUSDZ128mkz,           0 },
  { X86::VPDPBUSDZ256rk,            X86::VPDPBUSDZ256mk,            0 },
  { X86::VPDPBUSDZ256rkz,           X86::VPDPBUSDZ256mkz,           0 },
  { X86::VPDPBUSDZrk,               X86::VPDPBUSDZmk,               0 },
  { X86::VPDPBUSDZrkz,              X86::VPDPBUSDZmkz,              0 },
  { X86::VPDPWSSDSZ128rk,           X86::VPDPWSSDSZ128mk,           0 },
  { X86::VPDPWSSDSZ128rkz,          X86::VPDPWSSDSZ128mkz,          0 },
  { X86::VPDPWSSDSZ256rk,           X86::VPDPWSSDSZ256mk,           0 },
  { X86::VPDPWSSDSZ256rkz,          X86::VPDPWSSDSZ256mkz,          0 },
  { X86::VPDPWSSDSZrk,              X86::VPDPWSSDSZmk,              0 },
  { X86::VPDPWSSDSZrkz,             X86::VPDPWSSDSZmkz,             0 },
  { X86::VPDPWSSDZ128rk,            X86::VPDPWSSDZ128mk,            0 },
  { X86::VPDPWSSDZ128rkz,           X86::VPDPWSSDZ128mkz,           0 },
  { X86::VPDPWSSDZ256rk,            X86::VPDPWSSDZ256mk,            0 },
  { X86::VPDPWSSDZ256rkz,           X86::VPDPWSSDZ256mkz,           0 },
  { X86::VPDPWSSDZrk,               X86::VPDPWSSDZmk,               0 },
  { X86::VPDPWSSDZrkz,              X86::VPDPWSSDZmkz,              0 },
  { X86::VPERMBZ128rrk,             X86::VPERMBZ128rmk,             0 },
  { X86::VPERMBZ256rrk,             X86::VPERMBZ256rmk,             0 },
  { X86::VPERMBZrrk,                X86::VPERMBZrmk,                0 },
  { X86::VPERMDZ256rrk,             X86::VPERMDZ256rmk,             0 },
  { X86::VPERMDZrrk,                X86::VPERMDZrmk,                0 },
  { X86::VPERMI2B128rrk,            X86::VPERMI2B128rmk,            0 },
  { X86::VPERMI2B128rrkz,           X86::VPERMI2B128rmkz,           0 },
  { X86::VPERMI2B256rrk,            X86::VPERMI2B256rmk,            0 },
  { X86::VPERMI2B256rrkz,           X86::VPERMI2B256rmkz,           0 },
  { X86::VPERMI2Brrk,               X86::VPERMI2Brmk,               0 },
  { X86::VPERMI2Brrkz,              X86::VPERMI2Brmkz,              0 },
  { X86::VPERMI2D128rrk,            X86::VPERMI2D128rmk,            0 },
  { X86::VPERMI2D128rrkz,           X86::VPERMI2D128rmkz,           0 },
  { X86::VPERMI2D256rrk,            X86::VPERMI2D256rmk,            0 },
  { X86::VPERMI2D256rrkz,           X86::VPERMI2D256rmkz,           0 },
  { X86::VPERMI2Drrk,               X86::VPERMI2Drmk,               0 },
  { X86::VPERMI2Drrkz,              X86::VPERMI2Drmkz,              0 },
  { X86::VPERMI2PD128rrk,           X86::VPERMI2PD128rmk,           0 },
  { X86::VPERMI2PD128rrkz,          X86::VPERMI2PD128rmkz,          0 },
  { X86::VPERMI2PD256rrk,           X86::VPERMI2PD256rmk,           0 },
  { X86::VPERMI2PD256rrkz,          X86::VPERMI2PD256rmkz,          0 },
  { X86::VPERMI2PDrrk,              X86::VPERMI2PDrmk,              0 },
  { X86::VPERMI2PDrrkz,             X86::VPERMI2PDrmkz,             0 },
  { X86::VPERMI2PS128rrk,           X86::VPERMI2PS128rmk,           0 },
  { X86::VPERMI2PS128rrkz,          X86::VPERMI2PS128rmkz,          0 },
  { X86::VPERMI2PS256rrk,           X86::VPERMI2PS256rmk,           0 },
  { X86::VPERMI2PS256rrkz,          X86::VPERMI2PS256rmkz,          0 },
  { X86::VPERMI2PSrrk,              X86::VPERMI2PSrmk,              0 },
  { X86::VPERMI2PSrrkz,             X86::VPERMI2PSrmkz,             0 },
  { X86::VPERMI2Q128rrk,            X86::VPERMI2Q128rmk,            0 },
  { X86::VPERMI2Q128rrkz,           X86::VPERMI2Q128rmkz,           0 },
  { X86::VPERMI2Q256rrk,            X86::VPERMI2Q256rmk,            0 },
  { X86::VPERMI2Q256rrkz,           X86::VPERMI2Q256rmkz,           0 },
  { X86::VPERMI2Qrrk,               X86::VPERMI2Qrmk,               0 },
  { X86::VPERMI2Qrrkz,              X86::VPERMI2Qrmkz,              0 },
  { X86::VPERMI2W128rrk,            X86::VPERMI2W128rmk,            0 },
  { X86::VPERMI2W128rrkz,           X86::VPERMI2W128rmkz,           0 },
  { X86::VPERMI2W256rrk,            X86::VPERMI2W256rmk,            0 },
  { X86::VPERMI2W256rrkz,           X86::VPERMI2W256rmkz,           0 },
  { X86::VPERMI2Wrrk,               X86::VPERMI2Wrmk,               0 },
  { X86::VPERMI2Wrrkz,              X86::VPERMI2Wrmkz,              0 },
  { X86::VPERMILPDZ128rrk,          X86::VPERMILPDZ128rmk,          0 },
  { X86::VPERMILPDZ256rrk,          X86::VPERMILPDZ256rmk,          0 },
  { X86::VPERMILPDZrrk,             X86::VPERMILPDZrmk,             0 },
  { X86::VPERMILPSZ128rrk,          X86::VPERMILPSZ128rmk,          0 },
  { X86::VPERMILPSZ256rrk,          X86::VPERMILPSZ256rmk,          0 },
  { X86::VPERMILPSZrrk,             X86::VPERMILPSZrmk,             0 },
  { X86::VPERMPDZ256rrk,            X86::VPERMPDZ256rmk,            0 },
  { X86::VPERMPDZrrk,               X86::VPERMPDZrmk,               0 },
  { X86::VPERMPSZ256rrk,            X86::VPERMPSZ256rmk,            0 },
  { X86::VPERMPSZrrk,               X86::VPERMPSZrmk,               0 },
  { X86::VPERMQZ256rrk,             X86::VPERMQZ256rmk,             0 },
  { X86::VPERMQZrrk,                X86::VPERMQZrmk,                0 },
  { X86::VPERMT2B128rrk,            X86::VPERMT2B128rmk,            0 },
  { X86::VPERMT2B128rrkz,           X86::VPERMT2B128rmkz,           0 },
  { X86::VPERMT2B256rrk,            X86::VPERMT2B256rmk,            0 },
  { X86::VPERMT2B256rrkz,           X86::VPERMT2B256rmkz,           0 },
  { X86::VPERMT2Brrk,               X86::VPERMT2Brmk,               0 },
  { X86::VPERMT2Brrkz,              X86::VPERMT2Brmkz,              0 },
  { X86::VPERMT2D128rrk,            X86::VPERMT2D128rmk,            0 },
  { X86::VPERMT2D128rrkz,           X86::VPERMT2D128rmkz,           0 },
  { X86::VPERMT2D256rrk,            X86::VPERMT2D256rmk,            0 },
  { X86::VPERMT2D256rrkz,           X86::VPERMT2D256rmkz,           0 },
  { X86::VPERMT2Drrk,               X86::VPERMT2Drmk,               0 },
  { X86::VPERMT2Drrkz,              X86::VPERMT2Drmkz,              0 },
  { X86::VPERMT2PD128rrk,           X86::VPERMT2PD128rmk,           0 },
  { X86::VPERMT2PD128rrkz,          X86::VPERMT2PD128rmkz,          0 },
  { X86::VPERMT2PD256rrk,           X86::VPERMT2PD256rmk,           0 },
  { X86::VPERMT2PD256rrkz,          X86::VPERMT2PD256rmkz,          0 },
  { X86::VPERMT2PDrrk,              X86::VPERMT2PDrmk,              0 },
  { X86::VPERMT2PDrrkz,             X86::VPERMT2PDrmkz,             0 },
  { X86::VPERMT2PS128rrk,           X86::VPERMT2PS128rmk,           0 },
  { X86::VPERMT2PS128rrkz,          X86::VPERMT2PS128rmkz,          0 },
  { X86::VPERMT2PS256rrk,           X86::VPERMT2PS256rmk,           0 },
  { X86::VPERMT2PS256rrkz,          X86::VPERMT2PS256rmkz,          0 },
  { X86::VPERMT2PSrrk,              X86::VPERMT2PSrmk,              0 },
  { X86::VPERMT2PSrrkz,             X86::VPERMT2PSrmkz,             0 },
  { X86::VPERMT2Q128rrk,            X86::VPERMT2Q128rmk,            0 },
  { X86::VPERMT2Q128rrkz,           X86::VPERMT2Q128rmkz,           0 },
  { X86::VPERMT2Q256rrk,            X86::VPERMT2Q256rmk,            0 },
  { X86::VPERMT2Q256rrkz,           X86::VPERMT2Q256rmkz,           0 },
  { X86::VPERMT2Qrrk,               X86::VPERMT2Qrmk,               0 },
  { X86::VPERMT2Qrrkz,              X86::VPERMT2Qrmkz,              0 },
  { X86::VPERMT2W128rrk,            X86::VPERMT2W128rmk,            0 },
  { X86::VPERMT2W128rrkz,           X86::VPERMT2W128rmkz,           0 },
  { X86::VPERMT2W256rrk,            X86::VPERMT2W256rmk,            0 },
  { X86::VPERMT2W256rrkz,           X86::VPERMT2W256rmkz,           0 },
  { X86::VPERMT2Wrrk,               X86::VPERMT2Wrmk,               0 },
  { X86::VPERMT2Wrrkz,              X86::VPERMT2Wrmkz,              0 },
  { X86::VPERMWZ128rrk,             X86::VPERMWZ128rmk,             0 },
  { X86::VPERMWZ256rrk,             X86::VPERMWZ256rmk,             0 },
  { X86::VPERMWZrrk,                X86::VPERMWZrmk,                0 },
  { X86::VPMADD52HUQZ128rk,         X86::VPMADD52HUQZ128mk,         0 },
  { X86::VPMADD52HUQZ128rkz,        X86::VPMADD52HUQZ128mkz,        0 },
  { X86::VPMADD52HUQZ256rk,         X86::VPMADD52HUQZ256mk,         0 },
  { X86::VPMADD52HUQZ256rkz,        X86::VPMADD52HUQZ256mkz,        0 },
  { X86::VPMADD52HUQZrk,            X86::VPMADD52HUQZmk,            0 },
  { X86::VPMADD52HUQZrkz,           X86::VPMADD52HUQZmkz,           0 },
  { X86::VPMADD52LUQZ128rk,         X86::VPMADD52LUQZ128mk,         0 },
  { X86::VPMADD52LUQZ128rkz,        X86::VPMADD52LUQZ128mkz,        0 },
  { X86::VPMADD52LUQZ256rk,         X86::VPMADD52LUQZ256mk,         0 },
  { X86::VPMADD52LUQZ256rkz,        X86::VPMADD52LUQZ256mkz,        0 },
  { X86::VPMADD52LUQZrk,            X86::VPMADD52LUQZmk,            0 },
  { X86::VPMADD52LUQZrkz,           X86::VPMADD52LUQZmkz,           0 },
  { X86::VPMADDUBSWZ128rrk,         X86::VPMADDUBSWZ128rmk,         0 },
  { X86::VPMADDUBSWZ256rrk,         X86::VPMADDUBSWZ256rmk,         0 },
  { X86::VPMADDUBSWZrrk,            X86::VPMADDUBSWZrmk,            0 },
  { X86::VPMADDWDZ128rrk,           X86::VPMADDWDZ128rmk,           0 },
  { X86::VPMADDWDZ256rrk,           X86::VPMADDWDZ256rmk,           0 },
  { X86::VPMADDWDZrrk,              X86::VPMADDWDZrmk,              0 },
  { X86::VPMAXSBZ128rrk,            X86::VPMAXSBZ128rmk,            0 },
  { X86::VPMAXSBZ256rrk,            X86::VPMAXSBZ256rmk,            0 },
  { X86::VPMAXSBZrrk,               X86::VPMAXSBZrmk,               0 },
  { X86::VPMAXSDZ128rrk,            X86::VPMAXSDZ128rmk,            0 },
  { X86::VPMAXSDZ256rrk,            X86::VPMAXSDZ256rmk,            0 },
  { X86::VPMAXSDZrrk,               X86::VPMAXSDZrmk,               0 },
  { X86::VPMAXSQZ128rrk,            X86::VPMAXSQZ128rmk,            0 },
  { X86::VPMAXSQZ256rrk,            X86::VPMAXSQZ256rmk,            0 },
  { X86::VPMAXSQZrrk,               X86::VPMAXSQZrmk,               0 },
  { X86::VPMAXSWZ128rrk,            X86::VPMAXSWZ128rmk,            0 },
  { X86::VPMAXSWZ256rrk,            X86::VPMAXSWZ256rmk,            0 },
  { X86::VPMAXSWZrrk,               X86::VPMAXSWZrmk,               0 },
  { X86::VPMAXUBZ128rrk,            X86::VPMAXUBZ128rmk,            0 },
  { X86::VPMAXUBZ256rrk,            X86::VPMAXUBZ256rmk,            0 },
  { X86::VPMAXUBZrrk,               X86::VPMAXUBZrmk,               0 },
  { X86::VPMAXUDZ128rrk,            X86::VPMAXUDZ128rmk,            0 },
  { X86::VPMAXUDZ256rrk,            X86::VPMAXUDZ256rmk,            0 },
  { X86::VPMAXUDZrrk,               X86::VPMAXUDZrmk,               0 },
  { X86::VPMAXUQZ128rrk,            X86::VPMAXUQZ128rmk,            0 },
  { X86::VPMAXUQZ256rrk,            X86::VPMAXUQZ256rmk,            0 },
  { X86::VPMAXUQZrrk,               X86::VPMAXUQZrmk,               0 },
  { X86::VPMAXUWZ128rrk,            X86::VPMAXUWZ128rmk,            0 },
  { X86::VPMAXUWZ256rrk,            X86::VPMAXUWZ256rmk,            0 },
  { X86::VPMAXUWZrrk,               X86::VPMAXUWZrmk,               0 },
  { X86::VPMINSBZ128rrk,            X86::VPMINSBZ128rmk,            0 },
  { X86::VPMINSBZ256rrk,            X86::VPMINSBZ256rmk,            0 },
  { X86::VPMINSBZrrk,               X86::VPMINSBZrmk,               0 },
  { X86::VPMINSDZ128rrk,            X86::VPMINSDZ128rmk,            0 },
  { X86::VPMINSDZ256rrk,            X86::VPMINSDZ256rmk,            0 },
  { X86::VPMINSDZrrk,               X86::VPMINSDZrmk,               0 },
  { X86::VPMINSQZ128rrk,            X86::VPMINSQZ128rmk,            0 },
  { X86::VPMINSQZ256rrk,            X86::VPMINSQZ256rmk,            0 },
  { X86::VPMINSQZrrk,               X86::VPMINSQZrmk,               0 },
  { X86::VPMINSWZ128rrk,            X86::VPMINSWZ128rmk,            0 },
  { X86::VPMINSWZ256rrk,            X86::VPMINSWZ256rmk,            0 },
  { X86::VPMINSWZrrk,               X86::VPMINSWZrmk,               0 },
  { X86::VPMINUBZ128rrk,            X86::VPMINUBZ128rmk,            0 },
  { X86::VPMINUBZ256rrk,            X86::VPMINUBZ256rmk,            0 },
  { X86::VPMINUBZrrk,               X86::VPMINUBZrmk,               0 },
  { X86::VPMINUDZ128rrk,            X86::VPMINUDZ128rmk,            0 },
  { X86::VPMINUDZ256rrk,            X86::VPMINUDZ256rmk,            0 },
  { X86::VPMINUDZrrk,               X86::VPMINUDZrmk,               0 },
  { X86::VPMINUQZ128rrk,            X86::VPMINUQZ128rmk,            0 },
  { X86::VPMINUQZ256rrk,            X86::VPMINUQZ256rmk,            0 },
  { X86::VPMINUQZrrk,               X86::VPMINUQZrmk,               0 },
  { X86::VPMINUWZ128rrk,            X86::VPMINUWZ128rmk,            0 },
  { X86::VPMINUWZ256rrk,            X86::VPMINUWZ256rmk,            0 },
  { X86::VPMINUWZrrk,               X86::VPMINUWZrmk,               0 },
  { X86::VPMULDQZ128rrk,            X86::VPMULDQZ128rmk,            0 },
  { X86::VPMULDQZ256rrk,            X86::VPMULDQZ256rmk,            0 },
  { X86::VPMULDQZrrk,               X86::VPMULDQZrmk,               0 },
  { X86::VPMULHRSWZ128rrk,          X86::VPMULHRSWZ128rmk,          0 },
  { X86::VPMULHRSWZ256rrk,          X86::VPMULHRSWZ256rmk,          0 },
  { X86::VPMULHRSWZrrk,             X86::VPMULHRSWZrmk,             0 },
  { X86::VPMULHUWZ128rrk,           X86::VPMULHUWZ128rmk,           0 },
  { X86::VPMULHUWZ256rrk,           X86::VPMULHUWZ256rmk,           0 },
  { X86::VPMULHUWZrrk,              X86::VPMULHUWZrmk,              0 },
  { X86::VPMULHWZ128rrk,            X86::VPMULHWZ128rmk,            0 },
  { X86::VPMULHWZ256rrk,            X86::VPMULHWZ256rmk,            0 },
  { X86::VPMULHWZrrk,               X86::VPMULHWZrmk,               0 },
  { X86::VPMULLDZ128rrk,            X86::VPMULLDZ128rmk,            0 },
  { X86::VPMULLDZ256rrk,            X86::VPMULLDZ256rmk,            0 },
  { X86::VPMULLDZrrk,               X86::VPMULLDZrmk,               0 },
  { X86::VPMULLQZ128rrk,            X86::VPMULLQZ128rmk,            0 },
  { X86::VPMULLQZ256rrk,            X86::VPMULLQZ256rmk,            0 },
  { X86::VPMULLQZrrk,               X86::VPMULLQZrmk,               0 },
  { X86::VPMULLWZ128rrk,            X86::VPMULLWZ128rmk,            0 },
  { X86::VPMULLWZ256rrk,            X86::VPMULLWZ256rmk,            0 },
  { X86::VPMULLWZrrk,               X86::VPMULLWZrmk,               0 },
  { X86::VPMULTISHIFTQBZ128rrk,     X86::VPMULTISHIFTQBZ128rmk,     0 },
  { X86::VPMULTISHIFTQBZ256rrk,     X86::VPMULTISHIFTQBZ256rmk,     0 },
  { X86::VPMULTISHIFTQBZrrk,        X86::VPMULTISHIFTQBZrmk,        0 },
  { X86::VPMULUDQZ128rrk,           X86::VPMULUDQZ128rmk,           0 },
  { X86::VPMULUDQZ256rrk,           X86::VPMULUDQZ256rmk,           0 },
  { X86::VPMULUDQZrrk,              X86::VPMULUDQZrmk,              0 },
  { X86::VPORDZ128rrk,              X86::VPORDZ128rmk,              0 },
  { X86::VPORDZ256rrk,              X86::VPORDZ256rmk,              0 },
  { X86::VPORDZrrk,                 X86::VPORDZrmk,                 0 },
  { X86::VPORQZ128rrk,              X86::VPORQZ128rmk,              0 },
  { X86::VPORQZ256rrk,              X86::VPORQZ256rmk,              0 },
  { X86::VPORQZrrk,                 X86::VPORQZrmk,                 0 },
  { X86::VPROLVDZ128rrk,            X86::VPROLVDZ128rmk,            0 },
  { X86::VPROLVDZ256rrk,            X86::VPROLVDZ256rmk,            0 },
  { X86::VPROLVDZrrk,               X86::VPROLVDZrmk,               0 },
  { X86::VPROLVQZ128rrk,            X86::VPROLVQZ128rmk,            0 },
  { X86::VPROLVQZ256rrk,            X86::VPROLVQZ256rmk,            0 },
  { X86::VPROLVQZrrk,               X86::VPROLVQZrmk,               0 },
  { X86::VPRORVDZ128rrk,            X86::VPRORVDZ128rmk,            0 },
  { X86::VPRORVDZ256rrk,            X86::VPRORVDZ256rmk,            0 },
  { X86::VPRORVDZrrk,               X86::VPRORVDZrmk,               0 },
  { X86::VPRORVQZ128rrk,            X86::VPRORVQZ128rmk,            0 },
  { X86::VPRORVQZ256rrk,            X86::VPRORVQZ256rmk,            0 },
  { X86::VPRORVQZrrk,               X86::VPRORVQZrmk,               0 },
  { X86::VPSHLDDZ128rrik,           X86::VPSHLDDZ128rmik,           0 },
  { X86::VPSHLDDZ256rrik,           X86::VPSHLDDZ256rmik,           0 },
  { X86::VPSHLDDZrrik,              X86::VPSHLDDZrmik,              0 },
  { X86::VPSHLDQZ128rrik,           X86::VPSHLDQZ128rmik,           0 },
  { X86::VPSHLDQZ256rrik,           X86::VPSHLDQZ256rmik,           0 },
  { X86::VPSHLDQZrrik,              X86::VPSHLDQZrmik,              0 },
  { X86::VPSHLDVDZ128rk,            X86::VPSHLDVDZ128mk,            0 },
  { X86::VPSHLDVDZ128rkz,           X86::VPSHLDVDZ128mkz,           0 },
  { X86::VPSHLDVDZ256rk,            X86::VPSHLDVDZ256mk,            0 },
  { X86::VPSHLDVDZ256rkz,           X86::VPSHLDVDZ256mkz,           0 },
  { X86::VPSHLDVDZrk,               X86::VPSHLDVDZmk,               0 },
  { X86::VPSHLDVDZrkz,              X86::VPSHLDVDZmkz,              0 },
  { X86::VPSHLDVQZ128rk,            X86::VPSHLDVQZ128mk,            0 },
  { X86::VPSHLDVQZ128rkz,           X86::VPSHLDVQZ128mkz,           0 },
  { X86::VPSHLDVQZ256rk,            X86::VPSHLDVQZ256mk,            0 },
  { X86::VPSHLDVQZ256rkz,           X86::VPSHLDVQZ256mkz,           0 },
  { X86::VPSHLDVQZrk,               X86::VPSHLDVQZmk,               0 },
  { X86::VPSHLDVQZrkz,              X86::VPSHLDVQZmkz,              0 },
  { X86::VPSHLDVWZ128rk,            X86::VPSHLDVWZ128mk,            0 },
  { X86::VPSHLDVWZ128rkz,           X86::VPSHLDVWZ128mkz,           0 },
  { X86::VPSHLDVWZ256rk,            X86::VPSHLDVWZ256mk,            0 },
  { X86::VPSHLDVWZ256rkz,           X86::VPSHLDVWZ256mkz,           0 },
  { X86::VPSHLDVWZrk,               X86::VPSHLDVWZmk,               0 },
  { X86::VPSHLDVWZrkz,              X86::VPSHLDVWZmkz,              0 },
  { X86::VPSHLDWZ128rrik,           X86::VPSHLDWZ128rmik,           0 },
  { X86::VPSHLDWZ256rrik,           X86::VPSHLDWZ256rmik,           0 },
  { X86::VPSHLDWZrrik,              X86::VPSHLDWZrmik,              0 },
  { X86::VPSHRDDZ128rrik,           X86::VPSHRDDZ128rmik,           0 },
  { X86::VPSHRDDZ256rrik,           X86::VPSHRDDZ256rmik,           0 },
  { X86::VPSHRDDZrrik,              X86::VPSHRDDZrmik,              0 },
  { X86::VPSHRDQZ128rrik,           X86::VPSHRDQZ128rmik,           0 },
  { X86::VPSHRDQZ256rrik,           X86::VPSHRDQZ256rmik,           0 },
  { X86::VPSHRDQZrrik,              X86::VPSHRDQZrmik,              0 },
  { X86::VPSHRDVDZ128rk,            X86::VPSHRDVDZ128mk,            0 },
  { X86::VPSHRDVDZ128rkz,           X86::VPSHRDVDZ128mkz,           0 },
  { X86::VPSHRDVDZ256rk,            X86::VPSHRDVDZ256mk,            0 },
  { X86::VPSHRDVDZ256rkz,           X86::VPSHRDVDZ256mkz,           0 },
  { X86::VPSHRDVDZrk,               X86::VPSHRDVDZmk,               0 },
  { X86::VPSHRDVDZrkz,              X86::VPSHRDVDZmkz,              0 },
  { X86::VPSHRDVQZ128rk,            X86::VPSHRDVQZ128mk,            0 },
  { X86::VPSHRDVQZ128rkz,           X86::VPSHRDVQZ128mkz,           0 },
  { X86::VPSHRDVQZ256rk,            X86::VPSHRDVQZ256mk,            0 },
  { X86::VPSHRDVQZ256rkz,           X86::VPSHRDVQZ256mkz,           0 },
  { X86::VPSHRDVQZrk,               X86::VPSHRDVQZmk,               0 },
  { X86::VPSHRDVQZrkz,              X86::VPSHRDVQZmkz,              0 },
  { X86::VPSHRDVWZ128rk,            X86::VPSHRDVWZ128mk,            0 },
  { X86::VPSHRDVWZ128rkz,           X86::VPSHRDVWZ128mkz,           0 },
  { X86::VPSHRDVWZ256rk,            X86::VPSHRDVWZ256mk,            0 },
  { X86::VPSHRDVWZ256rkz,           X86::VPSHRDVWZ256mkz,           0 },
  { X86::VPSHRDVWZrk,               X86::VPSHRDVWZmk,               0 },
  { X86::VPSHRDVWZrkz,              X86::VPSHRDVWZmkz,              0 },
  { X86::VPSHRDWZ128rrik,           X86::VPSHRDWZ128rmik,           0 },
  { X86::VPSHRDWZ256rrik,           X86::VPSHRDWZ256rmik,           0 },
  { X86::VPSHRDWZrrik,              X86::VPSHRDWZrmik,              0 },
  { X86::VPSHUFBZ128rrk,            X86::VPSHUFBZ128rmk,            0 },
  { X86::VPSHUFBZ256rrk,            X86::VPSHUFBZ256rmk,            0 },
  { X86::VPSHUFBZrrk,               X86::VPSHUFBZrmk,               0 },
  { X86::VPSLLDZ128rrk,             X86::VPSLLDZ128rmk,             0 },
  { X86::VPSLLDZ256rrk,             X86::VPSLLDZ256rmk,             0 },
  { X86::VPSLLDZrrk,                X86::VPSLLDZrmk,                0 },
  { X86::VPSLLQZ128rrk,             X86::VPSLLQZ128rmk,             0 },
  { X86::VPSLLQZ256rrk,             X86::VPSLLQZ256rmk,             0 },
  { X86::VPSLLQZrrk,                X86::VPSLLQZrmk,                0 },
  { X86::VPSLLVDZ128rrk,            X86::VPSLLVDZ128rmk,            0 },
  { X86::VPSLLVDZ256rrk,            X86::VPSLLVDZ256rmk,            0 },
  { X86::VPSLLVDZrrk,               X86::VPSLLVDZrmk,               0 },
  { X86::VPSLLVQZ128rrk,            X86::VPSLLVQZ128rmk,            0 },
  { X86::VPSLLVQZ256rrk,            X86::VPSLLVQZ256rmk,            0 },
  { X86::VPSLLVQZrrk,               X86::VPSLLVQZrmk,               0 },
  { X86::VPSLLVWZ128rrk,            X86::VPSLLVWZ128rmk,            0 },
  { X86::VPSLLVWZ256rrk,            X86::VPSLLVWZ256rmk,            0 },
  { X86::VPSLLVWZrrk,               X86::VPSLLVWZrmk,               0 },
  { X86::VPSLLWZ128rrk,             X86::VPSLLWZ128rmk,             0 },
  { X86::VPSLLWZ256rrk,             X86::VPSLLWZ256rmk,             0 },
  { X86::VPSLLWZrrk,                X86::VPSLLWZrmk,                0 },
  { X86::VPSRADZ128rrk,             X86::VPSRADZ128rmk,             0 },
  { X86::VPSRADZ256rrk,             X86::VPSRADZ256rmk,             0 },
  { X86::VPSRADZrrk,                X86::VPSRADZrmk,                0 },
  { X86::VPSRAQZ128rrk,             X86::VPSRAQZ128rmk,             0 },
  { X86::VPSRAQZ256rrk,             X86::VPSRAQZ256rmk,             0 },
  { X86::VPSRAQZrrk,                X86::VPSRAQZrmk,                0 },
  { X86::VPSRAVDZ128rrk,            X86::VPSRAVDZ128rmk,            0 },
  { X86::VPSRAVDZ256rrk,            X86::VPSRAVDZ256rmk,            0 },
  { X86::VPSRAVDZrrk,               X86::VPSRAVDZrmk,               0 },
  { X86::VPSRAVQZ128rrk,            X86::VPSRAVQZ128rmk,            0 },
  { X86::VPSRAVQZ256rrk,            X86::VPSRAVQZ256rmk,            0 },
  { X86::VPSRAVQZrrk,               X86::VPSRAVQZrmk,               0 },
  { X86::VPSRAVWZ128rrk,            X86::VPSRAVWZ128rmk,            0 },
  { X86::VPSRAVWZ256rrk,            X86::VPSRAVWZ256rmk,            0 },
  { X86::VPSRAVWZrrk,               X86::VPSRAVWZrmk,               0 },
  { X86::VPSRAWZ128rrk,             X86::VPSRAWZ128rmk,             0 },
  { X86::VPSRAWZ256rrk,             X86::VPSRAWZ256rmk,             0 },
  { X86::VPSRAWZrrk,                X86::VPSRAWZrmk,                0 },
  { X86::VPSRLDZ128rrk,             X86::VPSRLDZ128rmk,             0 },
  { X86::VPSRLDZ256rrk,             X86::VPSRLDZ256rmk,             0 },
  { X86::VPSRLDZrrk,                X86::VPSRLDZrmk,                0 },
  { X86::VPSRLQZ128rrk,             X86::VPSRLQZ128rmk,             0 },
  { X86::VPSRLQZ256rrk,             X86::VPSRLQZ256rmk,             0 },
  { X86::VPSRLQZrrk,                X86::VPSRLQZrmk,                0 },
  { X86::VPSRLVDZ128rrk,            X86::VPSRLVDZ128rmk,            0 },
  { X86::VPSRLVDZ256rrk,            X86::VPSRLVDZ256rmk,            0 },
  { X86::VPSRLVDZrrk,               X86::VPSRLVDZrmk,               0 },
  { X86::VPSRLVQZ128rrk,            X86::VPSRLVQZ128rmk,            0 },
  { X86::VPSRLVQZ256rrk,            X86::VPSRLVQZ256rmk,            0 },
  { X86::VPSRLVQZrrk,               X86::VPSRLVQZrmk,               0 },
  { X86::VPSRLVWZ128rrk,            X86::VPSRLVWZ128rmk,            0 },
  { X86::VPSRLVWZ256rrk,            X86::VPSRLVWZ256rmk,            0 },
  { X86::VPSRLVWZrrk,               X86::VPSRLVWZrmk,               0 },
  { X86::VPSRLWZ128rrk,             X86::VPSRLWZ128rmk,             0 },
  { X86::VPSRLWZ256rrk,             X86::VPSRLWZ256rmk,             0 },
  { X86::VPSRLWZrrk,                X86::VPSRLWZrmk,                0 },
  { X86::VPSUBBZ128rrk,             X86::VPSUBBZ128rmk,             0 },
  { X86::VPSUBBZ256rrk,             X86::VPSUBBZ256rmk,             0 },
  { X86::VPSUBBZrrk,                X86::VPSUBBZrmk,                0 },
  { X86::VPSUBDZ128rrk,             X86::VPSUBDZ128rmk,             0 },
  { X86::VPSUBDZ256rrk,             X86::VPSUBDZ256rmk,             0 },
  { X86::VPSUBDZrrk,                X86::VPSUBDZrmk,                0 },
  { X86::VPSUBQZ128rrk,             X86::VPSUBQZ128rmk,             0 },
  { X86::VPSUBQZ256rrk,             X86::VPSUBQZ256rmk,             0 },
  { X86::VPSUBQZrrk,                X86::VPSUBQZrmk,                0 },
  { X86::VPSUBSBZ128rrk,            X86::VPSUBSBZ128rmk,            0 },
  { X86::VPSUBSBZ256rrk,            X86::VPSUBSBZ256rmk,            0 },
  { X86::VPSUBSBZrrk,               X86::VPSUBSBZrmk,               0 },
  { X86::VPSUBSWZ128rrk,            X86::VPSUBSWZ128rmk,            0 },
  { X86::VPSUBSWZ256rrk,            X86::VPSUBSWZ256rmk,            0 },
  { X86::VPSUBSWZrrk,               X86::VPSUBSWZrmk,               0 },
  { X86::VPSUBUSBZ128rrk,           X86::VPSUBUSBZ128rmk,           0 },
  { X86::VPSUBUSBZ256rrk,           X86::VPSUBUSBZ256rmk,           0 },
  { X86::VPSUBUSBZrrk,              X86::VPSUBUSBZrmk,              0 },
  { X86::VPSUBUSWZ128rrk,           X86::VPSUBUSWZ128rmk,           0 },
  { X86::VPSUBUSWZ256rrk,           X86::VPSUBUSWZ256rmk,           0 },
  { X86::VPSUBUSWZrrk,              X86::VPSUBUSWZrmk,              0 },
  { X86::VPSUBWZ128rrk,             X86::VPSUBWZ128rmk,             0 },
  { X86::VPSUBWZ256rrk,             X86::VPSUBWZ256rmk,             0 },
  { X86::VPSUBWZrrk,                X86::VPSUBWZrmk,                0 },
  { X86::VPTERNLOGDZ128rrik,        X86::VPTERNLOGDZ128rmik,        0 },
  { X86::VPTERNLOGDZ128rrikz,       X86::VPTERNLOGDZ128rmikz,       0 },
  { X86::VPTERNLOGDZ256rrik,        X86::VPTERNLOGDZ256rmik,        0 },
  { X86::VPTERNLOGDZ256rrikz,       X86::VPTERNLOGDZ256rmikz,       0 },
  { X86::VPTERNLOGDZrrik,           X86::VPTERNLOGDZrmik,           0 },
  { X86::VPTERNLOGDZrrikz,          X86::VPTERNLOGDZrmikz,          0 },
  { X86::VPTERNLOGQZ128rrik,        X86::VPTERNLOGQZ128rmik,        0 },
  { X86::VPTERNLOGQZ128rrikz,       X86::VPTERNLOGQZ128rmikz,       0 },
  { X86::VPTERNLOGQZ256rrik,        X86::VPTERNLOGQZ256rmik,        0 },
  { X86::VPTERNLOGQZ256rrikz,       X86::VPTERNLOGQZ256rmikz,       0 },
  { X86::VPTERNLOGQZrrik,           X86::VPTERNLOGQZrmik,           0 },
  { X86::VPTERNLOGQZrrikz,          X86::VPTERNLOGQZrmikz,          0 },
  { X86::VPUNPCKHBWZ128rrk,         X86::VPUNPCKHBWZ128rmk,         0 },
  { X86::VPUNPCKHBWZ256rrk,         X86::VPUNPCKHBWZ256rmk,         0 },
  { X86::VPUNPCKHBWZrrk,            X86::VPUNPCKHBWZrmk,            0 },
  { X86::VPUNPCKHDQZ128rrk,         X86::VPUNPCKHDQZ128rmk,         0 },
  { X86::VPUNPCKHDQZ256rrk,         X86::VPUNPCKHDQZ256rmk,         0 },
  { X86::VPUNPCKHDQZrrk,            X86::VPUNPCKHDQZrmk,            0 },
  { X86::VPUNPCKHQDQZ128rrk,        X86::VPUNPCKHQDQZ128rmk,        0 },
  { X86::VPUNPCKHQDQZ256rrk,        X86::VPUNPCKHQDQZ256rmk,        0 },
  { X86::VPUNPCKHQDQZrrk,           X86::VPUNPCKHQDQZrmk,           0 },
  { X86::VPUNPCKHWDZ128rrk,         X86::VPUNPCKHWDZ128rmk,         0 },
  { X86::VPUNPCKHWDZ256rrk,         X86::VPUNPCKHWDZ256rmk,         0 },
  { X86::VPUNPCKHWDZrrk,            X86::VPUNPCKHWDZrmk,            0 },
  { X86::VPUNPCKLBWZ128rrk,         X86::VPUNPCKLBWZ128rmk,         0 },
  { X86::VPUNPCKLBWZ256rrk,         X86::VPUNPCKLBWZ256rmk,         0 },
  { X86::VPUNPCKLBWZrrk,            X86::VPUNPCKLBWZrmk,            0 },
  { X86::VPUNPCKLDQZ128rrk,         X86::VPUNPCKLDQZ128rmk,         0 },
  { X86::VPUNPCKLDQZ256rrk,         X86::VPUNPCKLDQZ256rmk,         0 },
  { X86::VPUNPCKLDQZrrk,            X86::VPUNPCKLDQZrmk,            0 },
  { X86::VPUNPCKLQDQZ128rrk,        X86::VPUNPCKLQDQZ128rmk,        0 },
  { X86::VPUNPCKLQDQZ256rrk,        X86::VPUNPCKLQDQZ256rmk,        0 },
  { X86::VPUNPCKLQDQZrrk,           X86::VPUNPCKLQDQZrmk,           0 },
  { X86::VPUNPCKLWDZ128rrk,         X86::VPUNPCKLWDZ128rmk,         0 },
  { X86::VPUNPCKLWDZ256rrk,         X86::VPUNPCKLWDZ256rmk,         0 },
  { X86::VPUNPCKLWDZrrk,            X86::VPUNPCKLWDZrmk,            0 },
  { X86::VPXORDZ128rrk,             X86::VPXORDZ128rmk,             0 },
  { X86::VPXORDZ256rrk,             X86::VPXORDZ256rmk,             0 },
  { X86::VPXORDZrrk,                X86::VPXORDZrmk,                0 },
  { X86::VPXORQZ128rrk,             X86::VPXORQZ128rmk,             0 },
  { X86::VPXORQZ256rrk,             X86::VPXORQZ256rmk,             0 },
  { X86::VPXORQZrrk,                X86::VPXORQZrmk,                0 },
  { X86::VRANGEPDZ128rrik,          X86::VRANGEPDZ128rmik,          0 },
  { X86::VRANGEPDZ256rrik,          X86::VRANGEPDZ256rmik,          0 },
  { X86::VRANGEPDZrrik,             X86::VRANGEPDZrmik,             0 },
  { X86::VRANGEPSZ128rrik,          X86::VRANGEPSZ128rmik,          0 },
  { X86::VRANGEPSZ256rrik,          X86::VRANGEPSZ256rmik,          0 },
  { X86::VRANGEPSZrrik,             X86::VRANGEPSZrmik,             0 },
  { X86::VRANGESDZrrik,             X86::VRANGESDZrmik,             TB_NO_REVERSE },
  { X86::VRANGESSZrrik,             X86::VRANGESSZrmik,             TB_NO_REVERSE },
  { X86::VRCP14SDZrrk,              X86::VRCP14SDZrmk,              TB_NO_REVERSE },
  { X86::VRCP14SSZrrk,              X86::VRCP14SSZrmk,              TB_NO_REVERSE },
  { X86::VRCP28SDZrk,               X86::VRCP28SDZmk,               TB_NO_REVERSE },
  { X86::VRCP28SSZrk,               X86::VRCP28SSZmk,               TB_NO_REVERSE },
  { X86::VREDUCESDZrrik,            X86::VREDUCESDZrmik,            TB_NO_REVERSE },
  { X86::VREDUCESSZrrik,            X86::VREDUCESSZrmik,            TB_NO_REVERSE },
  { X86::VRNDSCALESDZr_Intk,        X86::VRNDSCALESDZm_Intk,        TB_NO_REVERSE },
  { X86::VRNDSCALESSZr_Intk,        X86::VRNDSCALESSZm_Intk,        TB_NO_REVERSE },
  { X86::VRSQRT14SDZrrk,            X86::VRSQRT14SDZrmk,            TB_NO_REVERSE },
  { X86::VRSQRT14SSZrrk,            X86::VRSQRT14SSZrmk,            TB_NO_REVERSE },
  { X86::VRSQRT28SDZrk,             X86::VRSQRT28SDZmk,             TB_NO_REVERSE },
  { X86::VRSQRT28SSZrk,             X86::VRSQRT28SSZmk,             TB_NO_REVERSE },
  { X86::VSCALEFPDZ128rrk,          X86::VSCALEFPDZ128rmk,          0 },
  { X86::VSCALEFPDZ256rrk,          X86::VSCALEFPDZ256rmk,          0 },
  { X86::VSCALEFPDZrrk,             X86::VSCALEFPDZrmk,             0 },
  { X86::VSCALEFPSZ128rrk,          X86::VSCALEFPSZ128rmk,          0 },
  { X86::VSCALEFPSZ256rrk,          X86::VSCALEFPSZ256rmk,          0 },
  { X86::VSCALEFPSZrrk,             X86::VSCALEFPSZrmk,             0 },
  { X86::VSCALEFSDZrrk,             X86::VSCALEFSDZrmk,             TB_NO_REVERSE },
  { X86::VSCALEFSSZrrk,             X86::VSCALEFSSZrmk,             TB_NO_REVERSE },
  { X86::VSHUFF32X4Z256rrik,        X86::VSHUFF32X4Z256rmik,        0 },
  { X86::VSHUFF32X4Zrrik,           X86::VSHUFF32X4Zrmik,           0 },
  { X86::VSHUFF64X2Z256rrik,        X86::VSHUFF64X2Z256rmik,        0 },
  { X86::VSHUFF64X2Zrrik,           X86::VSHUFF64X2Zrmik,           0 },
  { X86::VSHUFI32X4Z256rrik,        X86::VSHUFI32X4Z256rmik,        0 },
  { X86::VSHUFI32X4Zrrik,           X86::VSHUFI32X4Zrmik,           0 },
  { X86::VSHUFI64X2Z256rrik,        X86::VSHUFI64X2Z256rmik,        0 },
  { X86::VSHUFI64X2Zrrik,           X86::VSHUFI64X2Zrmik,           0 },
  { X86::VSHUFPDZ128rrik,           X86::VSHUFPDZ128rmik,           0 },
  { X86::VSHUFPDZ256rrik,           X86::VSHUFPDZ256rmik,           0 },
  { X86::VSHUFPDZrrik,              X86::VSHUFPDZrmik,              0 },
  { X86::VSHUFPSZ128rrik,           X86::VSHUFPSZ128rmik,           0 },
  { X86::VSHUFPSZ256rrik,           X86::VSHUFPSZ256rmik,           0 },
  { X86::VSHUFPSZrrik,              X86::VSHUFPSZrmik,              0 },
  { X86::VSQRTSDZr_Intk,            X86::VSQRTSDZm_Intk,            TB_NO_REVERSE },
  { X86::VSQRTSSZr_Intk,            X86::VSQRTSSZm_Intk,            TB_NO_REVERSE },
  { X86::VSUBPDZ128rrk,             X86::VSUBPDZ128rmk,             0 },
  { X86::VSUBPDZ256rrk,             X86::VSUBPDZ256rmk,             0 },
  { X86::VSUBPDZrrk,                X86::VSUBPDZrmk,                0 },
  { X86::VSUBPSZ128rrk,             X86::VSUBPSZ128rmk,             0 },
  { X86::VSUBPSZ256rrk,             X86::VSUBPSZ256rmk,             0 },
  { X86::VSUBPSZrrk,                X86::VSUBPSZrmk,                0 },
  { X86::VSUBSDZrr_Intk,            X86::VSUBSDZrm_Intk,            TB_NO_REVERSE },
  { X86::VSUBSSZrr_Intk,            X86::VSUBSSZrm_Intk,            TB_NO_REVERSE },
  { X86::VUNPCKHPDZ128rrk,          X86::VUNPCKHPDZ128rmk,          0 },
  { X86::VUNPCKHPDZ256rrk,          X86::VUNPCKHPDZ256rmk,          0 },
  { X86::VUNPCKHPDZrrk,             X86::VUNPCKHPDZrmk,             0 },
  { X86::VUNPCKHPSZ128rrk,          X86::VUNPCKHPSZ128rmk,          0 },
  { X86::VUNPCKHPSZ256rrk,          X86::VUNPCKHPSZ256rmk,          0 },
  { X86::VUNPCKHPSZrrk,             X86::VUNPCKHPSZrmk,             0 },
  { X86::VUNPCKLPDZ128rrk,          X86::VUNPCKLPDZ128rmk,          0 },
  { X86::VUNPCKLPDZ256rrk,          X86::VUNPCKLPDZ256rmk,          0 },
  { X86::VUNPCKLPDZrrk,             X86::VUNPCKLPDZrmk,             0 },
  { X86::VUNPCKLPSZ128rrk,          X86::VUNPCKLPSZ128rmk,          0 },
  { X86::VUNPCKLPSZ256rrk,          X86::VUNPCKLPSZ256rmk,          0 },
  { X86::VUNPCKLPSZrrk,             X86::VUNPCKLPSZrmk,             0 },
  { X86::VXORPDZ128rrk,             X86::VXORPDZ128rmk,             0 },
  { X86::VXORPDZ256rrk,             X86::VXORPDZ256rmk,             0 },
  { X86::VXORPDZrrk,                X86::VXORPDZrmk,                0 },
  { X86::VXORPSZ128rrk,             X86::VXORPSZ128rmk,             0 },
  { X86::VXORPSZ256rrk,             X86::VXORPSZ256rmk,             0 },
  { X86::VXORPSZrrk,                X86::VXORPSZrmk,                0 },
};

static const X86MemoryFoldTableEntry *
lookupFoldTableImpl(ArrayRef<X86MemoryFoldTableEntry> Table, unsigned RegOp) {
#ifndef NDEBUG
  // Make sure the tables are sorted.
  static std::atomic<bool> FoldTablesChecked(false);
  if (!FoldTablesChecked.load(std::memory_order_relaxed)) {
    assert(std::is_sorted(std::begin(MemoryFoldTable2Addr),
                          std::end(MemoryFoldTable2Addr)) &&
           std::adjacent_find(std::begin(MemoryFoldTable2Addr),
                              std::end(MemoryFoldTable2Addr)) ==
           std::end(MemoryFoldTable2Addr) &&
           "MemoryFoldTable2Addr is not sorted and unique!");
    assert(std::is_sorted(std::begin(MemoryFoldTable0),
                          std::end(MemoryFoldTable0)) &&
           std::adjacent_find(std::begin(MemoryFoldTable0),
                              std::end(MemoryFoldTable0)) ==
           std::end(MemoryFoldTable0) &&
           "MemoryFoldTable0 is not sorted and unique!");
    assert(std::is_sorted(std::begin(MemoryFoldTable1),
                          std::end(MemoryFoldTable1)) &&
           std::adjacent_find(std::begin(MemoryFoldTable1),
                              std::end(MemoryFoldTable1)) ==
           std::end(MemoryFoldTable1) &&
           "MemoryFoldTable1 is not sorted and unique!");
    assert(std::is_sorted(std::begin(MemoryFoldTable2),
                          std::end(MemoryFoldTable2)) &&
           std::adjacent_find(std::begin(MemoryFoldTable2),
                              std::end(MemoryFoldTable2)) ==
           std::end(MemoryFoldTable2) &&
           "MemoryFoldTable2 is not sorted and unique!");
    assert(std::is_sorted(std::begin(MemoryFoldTable3),
                          std::end(MemoryFoldTable3)) &&
           std::adjacent_find(std::begin(MemoryFoldTable3),
                              std::end(MemoryFoldTable3)) ==
           std::end(MemoryFoldTable3) &&
           "MemoryFoldTable3 is not sorted and unique!");
    assert(std::is_sorted(std::begin(MemoryFoldTable4),
                          std::end(MemoryFoldTable4)) &&
           std::adjacent_find(std::begin(MemoryFoldTable4),
                              std::end(MemoryFoldTable4)) ==
           std::end(MemoryFoldTable4) &&
           "MemoryFoldTable4 is not sorted and unique!");
    FoldTablesChecked.store(true, std::memory_order_relaxed);
  }
#endif

  const X86MemoryFoldTableEntry *Data = llvm::lower_bound(Table, RegOp);
  if (Data != Table.end() && Data->KeyOp == RegOp &&
      !(Data->Flags & TB_NO_FORWARD))
    return Data;
  return nullptr;
}

const X86MemoryFoldTableEntry *
llvm::lookupTwoAddrFoldTable(unsigned RegOp) {
  return lookupFoldTableImpl(MemoryFoldTable2Addr, RegOp);
}

const X86MemoryFoldTableEntry *
llvm::lookupFoldTable(unsigned RegOp, unsigned OpNum) {
  ArrayRef<X86MemoryFoldTableEntry> FoldTable;
  if (OpNum == 0)
    FoldTable = makeArrayRef(MemoryFoldTable0);
  else if (OpNum == 1)
    FoldTable = makeArrayRef(MemoryFoldTable1);
  else if (OpNum == 2)
    FoldTable = makeArrayRef(MemoryFoldTable2);
  else if (OpNum == 3)
    FoldTable = makeArrayRef(MemoryFoldTable3);
  else if (OpNum == 4)
    FoldTable = makeArrayRef(MemoryFoldTable4);
  else
    return nullptr;

  return lookupFoldTableImpl(FoldTable, RegOp);
}

namespace {

// This class stores the memory unfolding tables. It is instantiated as a
// ManagedStatic to lazily init the unfolding table.
struct X86MemUnfoldTable {
  // Stores memory unfolding tables entries sorted by opcode.
  std::vector<X86MemoryFoldTableEntry> Table;

  X86MemUnfoldTable() {
    for (const X86MemoryFoldTableEntry &Entry : MemoryFoldTable2Addr)
      // Index 0, folded load and store, no alignment requirement.
      addTableEntry(Entry, TB_INDEX_0 | TB_FOLDED_LOAD | TB_FOLDED_STORE);

    for (const X86MemoryFoldTableEntry &Entry : MemoryFoldTable0)
      // Index 0, mix of loads and stores.
      addTableEntry(Entry, TB_INDEX_0);

    for (const X86MemoryFoldTableEntry &Entry : MemoryFoldTable1)
      // Index 1, folded load
      addTableEntry(Entry, TB_INDEX_1 | TB_FOLDED_LOAD);

    for (const X86MemoryFoldTableEntry &Entry : MemoryFoldTable2)
      // Index 2, folded load
      addTableEntry(Entry, TB_INDEX_2 | TB_FOLDED_LOAD);

    for (const X86MemoryFoldTableEntry &Entry : MemoryFoldTable3)
      // Index 3, folded load
      addTableEntry(Entry, TB_INDEX_3 | TB_FOLDED_LOAD);

    for (const X86MemoryFoldTableEntry &Entry : MemoryFoldTable4)
      // Index 4, folded load
      addTableEntry(Entry, TB_INDEX_4 | TB_FOLDED_LOAD);

    // Sort the memory->reg unfold table.
    array_pod_sort(Table.begin(), Table.end());

    // Now that it's sorted, ensure its unique.
    assert(std::adjacent_find(Table.begin(), Table.end()) == Table.end() &&
           "Memory unfolding table is not unique!");
  }

  void addTableEntry(const X86MemoryFoldTableEntry &Entry,
                     uint16_t ExtraFlags) {
    // NOTE: This swaps the KeyOp and DstOp in the table so we can sort it.
    if ((Entry.Flags & TB_NO_REVERSE) == 0)
      Table.push_back({Entry.DstOp, Entry.KeyOp,
                      static_cast<uint16_t>(Entry.Flags | ExtraFlags) });
  }
};
}

static ManagedStatic<X86MemUnfoldTable> MemUnfoldTable;

const X86MemoryFoldTableEntry *
llvm::lookupUnfoldTable(unsigned MemOp) {
  auto &Table = MemUnfoldTable->Table;
  auto I = llvm::lower_bound(Table, MemOp);
  if (I != Table.end() && I->KeyOp == MemOp)
    return &*I;
  return nullptr;
}

